//================================================================
// EQX game engine
//
// - Preferences
//
// AUTHOR:  Gabor Nagy
// DATE:    2019-Jun-23 02:08:44
//
// Copyright (C) 1994-2021 by Gabor Nagy. All rights reserved.
//================================================================
#include <stdio.h>
#include <stdarg.h>
#include <time.h>

#include <EMalloc.h>

#include "common.h"

#include <E3D/UI/Prefs.h>

//----------------------------------------
// Panel/Pick.c
//----------------------------------------
extern int  E3dp_PickBufferSize;


// Application settings to be saved with a Scene. This is how EQUINOX-3D is able to save user GUI settings, etc. per Scene.
// We don't have any, but we could...
// Either way, libE3D's 3D file I/O plugins look for this symbol.
//
EResource  E3dApp_SceneSettings[]=
{
	{ "PointSize",                 EResINT32,           &E3dp_Prefs.PointSize },
	{ "SelectedPointSize",         EResINT32,           &E3dp_Prefs.SelectedPointSize },
	{ "LineWidth",                 EResFLOAT32,         &E3dp_Prefs.LineWidth },
	{ "LineWidthSelected",         EResFLOAT32,         &E3dp_Prefs.LineWidthSelected },
	{ "LineWidthSplines",          EResFLOAT32,         &E3dp_Prefs.LineWidthSplines },
	{ "LineWidthSelectedSplines",  EResFLOAT32,         &E3dp_Prefs.LineWidthSelectedSplines },
	{ "ManipulatorSize",           EResFLOAT32,         &E3dp_Prefs.ManipulatorSize },
	{ "RotatorSize",               EResFLOAT32,         &E3dp_Prefs.ManipulatorRotSize },

	{ "PolyPaintOnSelectionOnly",  EResBOOLEAN,         &E3dp_Prefs.PolyPaintOnSelectionOnly },
	{ "MaxUndoSteps",              EResINT32,           &E3dp_Prefs.MaxUndoSteps },

// Camera control
//
	{ "OrbitStepLon",              E3dCOORDINATE_TYPE,  &E3dp_Prefs.OrbitStepLongitude },
	{ "OrbitStepLat",              E3dCOORDINATE_TYPE,  &E3dp_Prefs.OrbitStepLatitude },
	{ "DollyStep",                 E3dCOORDINATE_TYPE,  &E3dp_Prefs.DollyStep },
	{ "TrackStep",                 E3dCOORDINATE_TYPE,  &E3dp_Prefs.TrackStep },
	{ "BankStep",                  E3dCOORDINATE_TYPE,  &E3dp_Prefs.BankStep },
	{ "FrameToTime",               E3dCOORDINATE_TYPE,  &E3dp_Prefs.FrameToTime },
	{ "FlyToOrtho",                EResBOOLEAN,         &E3dp_Prefs.CameraFlyToVMOrtho },
	{ "FlyToSchematics",           EResBOOLEAN,         &E3dp_Prefs.CameraFlyToVMSchematics },
	{ "FlyToAnimation",            EResBOOLEAN,         &E3dp_Prefs.CameraFlyToVMAnimation },
//{ "MouseWheelZoomAndDolly",    EResBOOLEAN,         &E3dp_Prefs.MouseWheelZoomAndDolly },
//{ "MouseWheelForwardZoomIn",   EResBOOLEAN,         &E3dp_Prefs.MouseWheelForwardZoomIn },

// Display control flags
//
	{ "ShowCenters",               EResBOOLEAN,         &E3dp_Prefs.ShowNodeCenters },
	{ "ShowPolygonCenters",        EResBOOLEAN,         &E3dp_Prefs.ShowPolygonCenters },
	{ "ShowPolygonNormals",        EResBOOLEAN,         &E3dp_Prefs.ShowPolygonNormals },
	{ "ShowVertexNormals",         EResBOOLEAN,         &E3dp_Prefs.ShowVertexNormals },
	{ "ShowVertexTangents",        EResBOOLEAN,         &E3dp_Prefs.ShowTangents },
	{ "ShowTags",                  EResBOOLEAN,         &E3dp_Prefs.ShowTags },
	{ "ShowSelectedPolygons",      EResBOOLEAN,         &E3dp_Prefs.ShowSelectedPolygons },
	{ "ShowWireOfHiddenObjects",   EResBOOLEAN,         &E3dp_Prefs.ShowWireframeOfHiddenObjects },
	{ "ShowTextureProjectors",     EResBOOLEAN,         &E3dp_Prefs.ShowTextureProjectors },
	{ "ShowGridsInShadedMode",     EResBOOLEAN,         &E3dp_Prefs.ShowGridsInShadedMode },
	{ "ShowLampsInWireMode",       EResBOOLEAN,         &E3dp_Prefs.ShowLampsInWireMode },
	{ "ShowLampsInSolidMode",      EResBOOLEAN,         &E3dp_Prefs.ShowLampsInSolidMode },

	{ "GL_antialiasing",           EResINT16,           &E3dp_Prefs.AntiAliasing },

	{ "SchematicsOrientation",     EResINT8,            &E3dp_Prefs.Schematics_Orientation },
	{ "SchematicsLinkStyle",       EResINT8,            &E3dp_Prefs.Schematics_LinkStyle },
	{ "SchematicsShowObjects",     EResBOOLEAN,         &E3dp_Prefs.Schematics_ShowObjects },
	{ "SchematicsShowPolyGroups",  EResBOOLEAN,         &E3dp_Prefs.Schematics_ShowPolyGroups },

	{ "MaxUndoSteps",              EResINT32,           &E3dp_Prefs.MaxUndoSteps },

// Animation control
//
	{ "CurrentFrame",              EResFLOAT64,         &E3dp_Prefs.CurrentFrame },
	{ "StartFrame",                EResFLOAT64,         &E3dp_Prefs.StartFrame },
	{ "EndFrame",                  EResFLOAT64,         &E3dp_Prefs.EndFrame },
	{ "FramesPerSec",              EResFLOAT32,         &E3dp_Prefs.FramesPerSecond },

	{ "DeleteSelection_DeleteSelectedPoints",  EResBOOLEAN,  &E3dp_Prefs.DeleteSelection_DeleteSelectedPoints },
	{ "DeleteSelection_DeleteSelectedEdges",   EResBOOLEAN,  &E3dp_Prefs.DeleteSelection_DeleteSelectedEdges  },
	{ "DeleteSelection_DeleteUnusedVertices",  EResBOOLEAN,  &E3dp_Prefs.DeleteSelection_DeleteUnusedVertices },
};


//================================================================================================
// Logging callback function, used by libE3D to report errors (usuall logged into a file)
// We'll just print the error here (in an app with a GUI, you could pop up a dialog box, etc.)
//================================================================================================
void E3d_Log(const int PType, const char* PPrefix, const char* PMessage, ...)
{
	char  LTmpStr[E3dMAX_PRINTF_BUFFER_SIZE];

	va_list  LAL;va_start(LAL, PMessage);
		vsnprintf(LTmpStr, sizeof(LTmpStr), PMessage, LAL);
	va_end(LAL);

	char  LTimeStr[64];
	time_t  LTime = time(NULL);if(LTime != -1)
	{
		struct tm*  LTimeS = localtime(&LTime);
		sprintf(LTimeStr, "%d-%02d-%02d %02d:%02d:%02d", LTimeS->tm_year + 1900, LTimeS->tm_mon + 1, LTimeS->tm_mday, LTimeS->tm_hour, LTimeS->tm_min, LTimeS->tm_sec);
	}

	switch(PType & (~ELogSILENT))
	{
		case ELogERROR:
			printf("%s  %sERROR%s    %s > %s\n", LTimeStr, ERedCode, ENormalCode, PPrefix, LTmpStr);fflush(stdout);
		break;

		case ELogWARNING:
			printf("%s  %sWARNING%s  %s > %s\n", LTimeStr, EYellowCode, ENormalCode, PPrefix, LTmpStr);fflush(stdout);
		break;

		case ELogMESSAGE:
			printf("%s  %sMESSAGE%s  %s > %s\n", LTimeStr, EGreenCode, ENormalCode, PPrefix, LTmpStr);fflush(stdout);
		break;
	}
}





E3dpPrefs  E3dp_Prefs;

//========================================
// Initialize 3D-panel preferences
//========================================
void E3dp_PreferencesInit(const int PArgC, const char* const* PArgv)
{
	E3dpPrefs*  LPrefs = &E3dp_Prefs;

	LPrefs->GLValueType = E3dGLSINGLE;

	LPrefs->SchematicsFontName = NULL;

	LPrefs->VMenuIconSize         = 28;
	LPrefs->ToolBarIconSize       = 32;
	LPrefs->DropDownPanelIconSize = 48;
	LPrefs->WindowHeaderIconSize  = 16;

	LPrefs->MenuBarOn = TRUE;
	LPrefs->MenuBarPlacement = E3dpPLACE_LEFT;

	LPrefs->MaxUndoSteps = 50;

	LPrefs->AntiAliasing = 1;  // 2x2X

	LPrefs->PointSize = 5;LPrefs->SelectedPointSize = 7;
	LPrefs->RectangleSelectThreshold = 5;
	LPrefs->PickPointSize            = 9;

	LPrefs->LineWidth                = 1.0;
	LPrefs->LineWidthSelected        = 1.5;
	LPrefs->LineWidthSplines         = 2.0;
	LPrefs->LineWidthSelectedSplines = 3.0;

	LPrefs->ManipulatorSize          = 120.0;
	LPrefs->ManipulatorRotSize       = 160.0;
	LPrefs->MenuBarCascadeButtonType = E3dpMBT_ICON;
	LPrefs->PolyPaintOnSelectionOnly = FALSE;

	LPrefs->HeadersOn        = TRUE;
	LPrefs->TransformPanelOn = TRUE;
	LPrefs->TimePanelOn      = TRUE;LPrefs->TimeExtPanelOn = TRUE;

	LPrefs->StereoEnabled = TRUE;

	LPrefs->ToolBarOn           = TRUE;
	LPrefs->CreateFolderOpen    = TRUE;
	LPrefs->CloneFolderOpen     = TRUE;
	LPrefs->InfoFolderOpen      = TRUE;
	LPrefs->DisplayFolderOpen   = TRUE;
	LPrefs->CameraFolderOpen    = TRUE;
	LPrefs->LightFolderOpen     = TRUE;
	LPrefs->TransformFolderOpen = FALSE;
	LPrefs->HierarchyFolderOpen = TRUE;
	LPrefs->GeometryFolderOpen  = TRUE;
	LPrefs->SplineFolderOpen    = TRUE;
	LPrefs->MaterialFolderOpen  = TRUE;
	LPrefs->SurfaceFolderOpen   = TRUE;
	LPrefs->DeleteFolderOpen    = TRUE;
	LPrefs->EffectFolderOpen    = TRUE;
	LPrefs->RenderFolderOpen    = TRUE;
	LPrefs->SelectFolderOpen    = TRUE;
	LPrefs->HistoryFolderOpen   = TRUE;
	LPrefs->AnimationFolderOpen = TRUE;

	LPrefs->HighlightActive3DWindow = FALSE;

	LPrefs->StatusPanelOn = TRUE;
	LPrefs->StatusPanelTextAntiAliasing = 4;  // 2x2

// Colors
//
	Ec_SetRGBAi(&LPrefs->BackgroundRGBAiColor,    0x9000, 0x9000, 0x9000, 0x0000);

	E3dp_UpdateGridColors(&LPrefs->BackgroundRGBAiColor);


// Wireframe color
//
	Ec_SetRGBAi(&LPrefs->WireRGBAiColor,          0x2000, 0x2000, 0x5000, 0xFFFF);
//Ec_SetRGBAi(&LPrefs->WireRGBAiColor,          0x4800, 0x4800, 0x4800, 0xFFFF);

// Selected Model wireframe color
//
	Ec_SetRGBAi(&LPrefs->HWireRGBAiColor,         0xFF00, 0xFF00, 0xFF00, 0xFFFF);

// Selected Geometry wireframe color
//
	Ec_SetRGBAi(&LPrefs->SelGeoWireRGBAiColor,    0x4800, 0xE800, 0xD800, 0xFFFF);
//	Ec_SetRGBAi(&LPrefs->SelGeoWireRGBAiColor,  0x2400, 0x7400, 0x6C00, 0xFFFF);
//	Ec_SetRGBAi(&LPrefs->SelGeoWireRGBAiColor,  0x2400, 0x7400, 0x6C00, 0xFFFF);
//	Ec_SetRGBAi(&LPrefs->SelGeoWireRGBAiColor,  0x2000, 0x2000, 0x5000, 0xFFFF);

// Selected SubGeometry (PolyGroup, Spline segment) wireframe color
//
	Ec_SetRGBAi(&LPrefs->SelSubGeoWireRGBAiColor, 0xF000, 0xF000, 0x1000, 0xFFFF);



	Ec_SetRGBAi(&LPrefs->LockedItemRGBAiColor, 0x7000,    0x0000, 0x7000, 0xFFFF);
//	Ec_SetRGBAi(&LPrefs->TextureProjectorRGBAiColor,      0x0000, 0xE000, 0xE000, 0xFFFF);
	Ec_SetRGBAi(&LPrefs->TextureProjectorRGBAiColor,      0x0000, 0x6000, 0xA000, 0xFFFF);
	Ec_SetRGBAi(&LPrefs->VertexRGBAiColor,                0x0000, 0x0000, 0xFF00, 0xFFFF);
	Ec_SetRGBAi(&LPrefs->Line1RGBAiColor,                 0x0000, 0x7000, 0x7000, 0xFFFF);
	Ec_SetRGBAi(&LPrefs->Line2RGBAiColor,                 0x8000, 0x2000, 0x8000, 0xFFFF);
	Ec_SetRGBAi(&LPrefs->HVertexRGBAiColor,               0x0000, 0xEE00, 0x0000, 0xFFFF);
	Ec_SetRGBAi(&LPrefs->ActivePointRGBAiColor,           0xF000, 0xF000, 0x0000, 0xFFFF);
	Ec_SetRGBAi(&LPrefs->SelectedPointRGBAiColor,         0xB000, 0x0000, 0x0000, 0xFFFF);

	Ec_SetRGBAi(&LPrefs->ActiveEdgeRGBAiColor,            0xF000, 0xF000, 0x0000, 0xFFFF);
	Ec_SetRGBAi(&LPrefs->SelectedEdgeRGBAiColor,          0xFF00, 0x2800, 0x0000, 0xFFFF);

	Ec_SetRGBAi(&LPrefs->SelectedPolygonRGBAiColor,       0xF000, 0x9000, 0x2000, 0x6000);
	Ec_SetRGBAi(&LPrefs->ShadedSelectedPolygonRGBAiColor, 0xF000, 0x9000, 0x2000, 0x6000);

	Ec_SetRGBAi(&LPrefs->NormalRGBAiColor,                0xE000, 0xF000, 0x1000, 0xFFFF);
	Ec_SetRGBAi(&LPrefs->TangentRGBAiColor,               0xE000, 0x3000, 0x2000, 0xFFFF);
	Ec_SetRGBAi(&LPrefs->BoneRGBAiColor,                  0x0000, 0xA000, 0x8000, 0xFFFF);

	Ec_SetRGBAi(&LPrefs->LeftEyeRGBAiColor,               0xF000, 0x0000, 0x0000, 0xFFFF);
	Ec_SetRGBAi(&LPrefs->RightEyeRGBAiColor,              0x0000, 0xC000, 0xFF00, 0xFFFF);


// StatusPanel message color codes
//
	LPrefs->ButtonNameRGBAp8Color    = Ec_RGBA8p(0x003090FF);

	E_SetRGBA_s(LPrefs->VariableRGBAiColor,  0x0000, 0x3000, 0x9000, 0xFFFF);
	LPrefs->VariableRGBAp8Color = Ec_RGBAiToRGBAp8(&LPrefs->VariableRGBAiColor);

	LPrefs->OperationRGBAp8Color     = Ec_RGBA8p(0x0040C0FF);
	LPrefs->ErrorRGBAp8Color         = Ec_RGBA8p(0xC00000FF);
	LPrefs->WarningRGBAp8Color       = Ec_RGBA8p(0xE0E000FF);
	LPrefs->NodeBracketRGBAp8Color   = Ec_RGBA8p(0x0000C0FF);
	LPrefs->NodeNameRGBAp8Color      = Ec_RGBA8p(0x0000C0FF);
	LPrefs->ObjectBracketRGBAp8Color = Ec_RGBA8p(0x0080C0FF);
	LPrefs->ObjectNameRGBAp8Color    = Ec_RGBA8p(0x006060FF);
	LPrefs->XRGBAp8Color = Ec_RGBA8p(0x500000FF);
	LPrefs->YRGBAp8Color = Ec_RGBA8p(0x005000FF);
	LPrefs->ZRGBAp8Color = Ec_RGBA8p(0x000050FF);



// Display control flags
//
	LPrefs->ShowNodeCenters = TRUE;
	LPrefs->ShowPolygonCenters = FALSE;
	LPrefs->ShowPolygonNormals = FALSE;
	LPrefs->ShowVertexNormals = FALSE;
	LPrefs->ShowTangents = FALSE;
	LPrefs->ShowTags = TRUE;
	LPrefs->ShowSelectedPolygons = TRUE;
	LPrefs->ShowWireframeOfHiddenObjects = TRUE;
	LPrefs->ShowTextureProjectors = TRUE;
	LPrefs->ShowGridsInShadedMode = TRUE;
	LPrefs->ShowLampsInWireMode = TRUE;
	LPrefs->ShowLampsInSolidMode = TRUE;


	E_SetXYZ_s(LPrefs->DefCameraPos,    0.0, 12.0, 42.0);
	E_SetXYZ_s(LPrefs->DefInterestPos,  0.0,  0.0,  0.0);
	LPrefs->DefMag = 1.0 / 15.0;LPrefs->DefFieldOfView = 36.0;
	LPrefs->DefZNear = 0.01;
	LPrefs->DefZFar = 32767.0;

// Camera control
//
	LPrefs->DollyStep1 = 0.36;LPrefs->DollyStep2 = 0.8;LPrefs->DollyStep3 = 2.0;
	LPrefs->BankStep = 0.25;
	LPrefs->OrbitStepLongitude = 0.25;LPrefs->OrbitStepLatitude = 0.25;
	LPrefs->DollyStep = 1.0;LPrefs->TrackStep = 1.0;

	LPrefs->FrameToTime = 0.25;
	LPrefs->FrameToTimeStep = 0.02;

	LPrefs->CameraFlyToVMOrtho      = TRUE;
	LPrefs->CameraFlyToVMSchematics = TRUE;
	LPrefs->CameraFlyToVMAnimation  = TRUE;
	LPrefs->MouseWheelZoomAndDolly  = TRUE;
	LPrefs->MouseWheelForwardZoomIn = TRUE;
	LPrefs->CameraZoomToPointer     = TRUE;


// TimePanel settings
//
	LPrefs->CurrentFrame = 0.0;
	LPrefs->StartFrame = 0.0;
	LPrefs->EndFrame = 299.0;
	LPrefs->FramesPerSecond = 30.0;


	E_SetXYZ_s(LPrefs->ScalingStep,      1.1,  1.1,  1.1);
	E_SetXYZ_s(LPrefs->RotatingStep,     0.5,  0.5,  0.5);
	E_SetXYZ_s(LPrefs->TranslatingStep,  0.01, 0.01, 0.01);
	LPrefs->DefEyeDistance = 1.0;
	E_SetXYZ_s(LPrefs->DefGridSizeDisplay,    1.0, 1.0, 1.0);
	E_SetXYZ_s(LPrefs->DefSubGridSizeDisplay, 0.5, 0.5, 0.5);
	LPrefs->GridOn = TRUE;
	LPrefs->GridCount.X = 100;
	LPrefs->GridCount.Y = 100;
	LPrefs->GridCount.Z = 100;
	LPrefs->UseEdgeSteps = FALSE;
	LPrefs->EdgeSteps = 4;

	LPrefs->GridXYOn = FALSE;LPrefs->GridXZOn = TRUE;LPrefs->GridYZOn = FALSE;
	LPrefs->AxesOn = TRUE;LPrefs->RulerOn = FALSE;
	LPrefs->UseQuickExposeBuffer = FALSE;
	LPrefs->UseTransformBuffer = FALSE;
	LPrefs->UseCameraSpeedupBuffer = FALSE;
	LPrefs->SelectingEnabled = TRUE;
	LPrefs->DeletingEnabled = TRUE;


// Default paths to NULL, which means "relative to launch-path"
//
	LPrefs->LogFile = NULL;
	LPrefs->ResourcePath = NULL;
	LPrefs->PluginPath = NULL;
	LPrefs->PluginPath3DIO = NULL;
	LPrefs->PluginPathImageIO = NULL;
	LPrefs->TexturePath = NULL;

	LPrefs->OptimizeFor = EOptimizeSPEED;

	LPrefs->SelectNewModels = TRUE;

	LPrefs->Schematics_Orientation = E3dVERTICAL;
	LPrefs->Schematics_LinkStyle = E3dCURVED;

	LPrefs->Schematics_ShowObjects = TRUE;
	LPrefs->Schematics_ShowPolyGroups = TRUE;
	LPrefs->Schematics_AutoFrame = TRUE;

	LPrefs->DeleteSelection_DeleteSelectedPoints = FALSE;
	LPrefs->DeleteSelection_DeleteSelectedEdges  = FALSE;
	LPrefs->DeleteSelection_DeleteUnusedVertices = TRUE;

	LPrefs->Duplicate_Flags = E3dCopyMODEL_UNIQUE_NAME|E3dCopyOBJECT_UNIQUE_NAME|E3dCopyLODS|E3dCopyCOMPONENTS|E3dCopyANIMATIONS|E3dCopyEDGES;

// Animation
//
	LPrefs->NewAnimKeyEaseInOut = 0.2;

	LPrefs->RedrawOnRendererChange = TRUE;
}
