//================================================================================================
// EQX game engine
//
// Minimal libE3D use sample.
// It initializes libE3D, loads a 3D hierarchy from a file and saves it into another file.
//
// AUTHOR:  Gabor Nagy
// DATE:    2019-Jun-23 03:22:29
//
// This sample code is freely modifiable and distributable
//
// EQUINOX-3D and its components are Copyright (C) 1994-2021 by Gabor Nagy. All rights reserved.
//================================================================================================
#include <assert.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>

#include <EMalloc.h>
#include <EPlugins.h>

#include <E3D/Plugins.h>
#include <E3D/Scene.h>

#include <E3D/UI/Prefs.h>

#include "common.h"

// The main 3D Scene
//
E3dScene*  E3d_Scene = NULL;


//========================================================================================
// Function, used by libE3D to print messages in the GUI (EQUINOX-3D's StatusBar).
// We are a command-line tool, so we'll just print the message to stdout.
//========================================================================================
E3dID E3dp_PrintMessage(const int PType, const int PRemoveInMillisedonds, const char* PMessage, ...)
{
	va_list  LAL;
	va_start(LAL, PMessage);         // PMessage is the last argument before the varargs
		vprintf(PMessage, LAL);
	va_end(LAL);
	printf("\n");fflush(stdout);

	return(0);
}


//========================================================
// Main
//========================================================
int main(const int PArgs_Count, const char* const* PArgs)
{
	E_MallocInit();


	E3dp_PreferencesInit(PArgs_Count, PArgs);

// Initialize paths
//
	E3d_LaunchPath   = E_LaunchPath(PArgs[0]);
	E_ExecutableName = EStrDup(PArgs[0]);

// Initialize working versions of paths (the ones saved in the preferences may be NULL, indicating "use default"
// (relative to the launch location)
//
	if(E3dp_Prefs.ResourcePath) E3dp_Prefs.WorkingResourcePath = EStrDup(E3dp_Prefs.ResourcePath);
	else
	{
		E3dp_Prefs.WorkingResourcePath = EStrDup(E3d_LaunchPath);
	}


	{
// Initialize lib2D and libE3D
//
		Ep_Init();
		E3d_Init();

// Load 3D I/O plugins
//
		if(E3d_LaunchPath)
		{
			char  LTmpStr[MAXPATHLEN];sprintf(LTmpStr, "%s/3DIO", E3d_LaunchPath);
			E3d_3DIOReadStatus = E3d_PluginsLoad(LTmpStr, "Formats.list");
		}

// Create the main Scene
//
		E3dScene*  LScene = E3dScene_New("default");if(LScene == NULL) return(-1);
		E3d_Scene = LScene;


// Read a Model hierarchy from a file and save it into a different file
//
//		const char  LInFileName[]  = "data/modeltest.hrc";
		const char  LInFileName[]  = "data/hrctest.hrc";
		const char  LOutFileName[] = "data/out.hrc";
		E3dNode*  LNode;const int  LReadResult = E3d_HierarchyReadFromFile(LInFileName, E3dFileFormatAUTOMATIC, NULL, &LNode);if(LReadResult == E_SUCCESS)
		{
			if(LNode)
			{
				if(LNode->Child) printf("Loaded hierarchy from %s\n", LInFileName);
				else printf("Loaded Model from %s\n", LInFileName);
				fflush(stdout);
			}
			LNode->FileName = EStrDup(LInFileName);

			E3dNodeHrc_UpdateMaterialInheritance(LNode);
			E3dNodeHrc_UpdateConnections(LNode);
			E3dScene_NodeHrcAdd(LScene, LNode, 0);    // Add the Model or hierarchy to the Scene

			const int  LSaveResult = E3d_HierarchyWriteToFile(LOutFileName,  NULL, 0,  LNode, NULL, NULL);if(LSaveResult == E_SUCCESS)
			{
				printf("Saved hierarchy to %s\n", LOutFileName);fflush(stdout);
			}
			else { printf("Could not save %s: %s\n", LOutFileName, E_ResultToString(LSaveResult));fflush(stdout); }
		}
		else { printf("Error reading input file: %s\n", E_ResultToString(LReadResult));fflush(stdout); }
	}

	return(0);
}
