//================================================================
// EQUINOX-3D plugin
//
// Common code for picking controlling Splines
//
// AUTHOR:  Gabor Nagy
// DATE:    2019-Jun-23 18:54:24
//
// Copyright (C) 1994-2021 By Gabor Nagy. All rights reserved.
//================================================================


static void E3d_ScalingBySplineInit(E3dScalingBySpline* PP)
{
	PP->Active     = FALSE;
	PP->OrigSpline = NULL;
	PP->Model      = NULL;
}


static void E3d_ScalingBySplineFreeContents(E3dScalingBySpline* PP, EInterface* PInputInterface)
{
	if(PP->OrigSpline)
	{
		if(PInputInterface) E3dBaseObject_RemoveOutput((E3dBaseObject*)PP->OrigSpline, PInputInterface);
		E_DeleteChk(PP->OrigSpline);
		E_DeleteChk(PP->Model);
	}
}


static inline void E3d_ScalingBySplineCopyContents(const E3dScalingBySpline* PSrc, E3dScalingBySpline* PDst,  EInterface* PInputInterface)
{
	E3d_ScalingBySplineInit(PDst);PDst->Active = PSrc->Active;
	if(PSrc->OrigSpline)
	{
		if(PInputInterface) E3dBaseObject_AddOutput((E3dBaseObject*)PSrc->OrigSpline, PInputInterface);
		E_Assign(PSrc->OrigSpline, PDst->OrigSpline);
		E_Assign(PSrc->Model,  PDst->Model);
	}
}


//================================================================
// Pick scaling Spline callback
//================================================================
static void _PickCB_ScaleBySpline(E3dPickCallbackStruct* PPickStruct, EPointer PClientData,  const EIndex PNumOfItems, E3dItem* PPickedItems)
{
	XeApp_Sync(E3dp_XeApp);

	if(PPickStruct->Reason == E3dCR_ABORT)
	{
		_TmpInfo = NULL;
		_TmpPanel = NULL;
		_TmpScalingBySpline = NULL;
		_TmpScalingBySplineIndex = -1;
		return;
	}

// Cancel
//
	if(PPickStruct->Button == Button3)
	{
		E3dp_PickEnd();
		_TmpInfo = NULL;
		_TmpPanel = NULL;
		_TmpScalingBySpline = NULL;
		_TmpScalingBySplineIndex = -1;
		return;
	}

	E3dScalingBySpline*  LScalingBySpline = _TmpScalingBySpline;

	switch(PPickStruct->Button)
	{
		case Button1:
			if(PNumOfItems && _TmpInfo)
			{
				E3dItem*  LItem = PPickedItems;
				E3dInfo*  LInfo = _TmpInfo;

//char s[256];printf("%s() %s:%d [%s]\n", __FUNCTION__, __FILE__, __LINE__, E3d_ItemFlagsToString(LItem->Item.Type, s));fflush(stdout);

				if(LItem->Item.Type == E3dItemSPLINE || LItem->Item.Type == E3dItemSPLINE_SEGMENT || LItem->Item.Type == E3dItemMODEL)
				{
					E3dModel*   LModel  = NULL;
					E3dSpline*  LSpline = NULL;

					if(LItem->Item.Type == E3dItemMODEL)
					{
						E3dObjItem*  LObjItem = (E3dObjItem*)LItem;

						LModel = (E3dModel*)(LObjItem->Node);

						EIndex  LObjC = 0;const EIndex  LObjN = LModel->Objects_Count;E3dObject**  LObjects = LModel->Objects;for(;LObjC < LObjN;LObjC++)
						{
							E3dObject*  LObject = LObjects[LObjC];

							if(EObject_IsSubClass(LObject, E3dSpline)) LSpline = (E3dSpline*)LObject;
						}

					}
					else
					{
						E3dObjItem*  LObjItem = (E3dObjItem*)LItem;
						E3dObject*   LObject = LObjItem->Object;

						if(EObject_IsSubClass(LObject, E3dSpline))
						{
							LSpline = (E3dSpline*)LObject;
							LModel  = LObject->Models[0];
//							LObject->Selection = E3dObjectSelOBJECT;
						}
					}

					if(LInfo && LSpline)
					{
//printf("%s() %s:%d  %p %p Picked %s/%s\n", __FUNCTION__, __FILE__, __LINE__, _TmpInfo, LScalingBySpline, LModel->Name, LSpline->Name);fflush(stdout);
						if(LSpline != LScalingBySpline->OrigSpline)
						{
							if(LScalingBySpline->OrigSpline) E3dBaseObject_RemoveOutput((E3dBaseObject*)LScalingBySpline->OrigSpline, &LInfo->InputInterface);
							E_Assign(LSpline, LScalingBySpline->OrigSpline);

							if(LScalingBySpline->OrigSpline) E3dBaseObject_AddOutput((E3dBaseObject*)LScalingBySpline->OrigSpline, &LInfo->InputInterface);
						}
						if(LModel != LScalingBySpline->Model) E_Assign(LModel, LScalingBySpline->Model);

						E3dp_PrintMessage(ELogMESSAGE, 3000, "Picked %s/%s", LModel->Name, LSpline->Name);
						EChangeFlags  LChanged = 0;

						if(_TmpPanel && !LScalingBySpline->Active)
						{
							XeComponentClassDrawer*  LDrawer = XeInfoPanel_GetFolderByClass((XeInfoPanel*)_TmpPanel, TmpComponentClass);
							E3dResourcesGUI*         LResourcesGUI = (E3dResourcesGUI*)(LDrawer->ResourcesGUI);

							LScalingBySpline->Active = TRUE;
							XeToggleButton_SetState(LResourcesGUI->ScaleBySplineWs[_TmpScalingBySplineIndex],  TRUE, FALSE);
						}
						TmpUpdateObject;
					}

					E3dp_PickEnd();

					E3dWindows_RedrawByChangeFlags(TmpCHANGEFLAGS);
				}
				_TmpInfo = NULL;
				_TmpPanel = NULL;
				_TmpScalingBySpline = NULL;
				_TmpScalingBySplineIndex = -1;
			}
		break;
	}
}



#define _PickSpline(mText) do\
{\
	_TmpInfo = LInfo;\
	_TmpPanel = LPanel;\
	const E3dpPrefs*  LPrefs = &E3dp_Prefs;\
	const uint32_t    LLabelRGBAp8Color = Xe_ForegroundRGBA8pColor;\
	E3dp_PickRequest(XeString_NewVa(TRUE,\
	 XeFONT, XeFontDEFAULT, XeCOLOR_RGBA, LLabelRGBAp8Color,  XeTEXT,  "Extrude along Spline    ",\
	 E3dp_LMouseButtonS,     XeTEXT, mText,\
	 E3dp_RMouseButtonEscS,  XeTEXT, ": cancel",\
	 0), E3dItemSPLINE|E3dItemMODEL,  _PickCB_ScaleBySpline, NULL,  0, FALSE);\
} while(0)


// Create GUI
//static inline void _ScaleUBySplineGUI(E3dInfo* PInfo, E3dResourcesGUI* PResourcesGUI, Widget PParentW)
#define _ScaleBySplineGUI(mCoordinate, mInfo, mResourcesGUI, mParentW)\
{\
	EXeStart;\
	EXeSetArg(XeNshadowThickness, 0);\
	EXeSetArg(XeNhighlightThickness, 0);\
	EXeSetArg(XeNset, mInfo->Scale##mCoordinate##BySpline.Active);\
	EXeSetArg(XeNwidthAdjustable, FALSE);EXeSetArg(XeNalignmentX, XeAlignBEGINNING);\
	XeToolTip8(_Scale##mCoordinate##ToolTip);\
	mResourcesGUI->ScaleBySplineWs[E3dSCALESPL_##mCoordinate] = LW = XeWidget_ManagedNew("Scale "#mCoordinate" by Spline", XeToggleButton_class, mParentW, Args, ArgCnt);\
	XeAddCallback(LW, XeNvalueChangedCallback, _DCB_Scale##mCoordinate##BySpline, (EPointer)LPanel);\
\
	EXeStart;\
	EXeSetArg(XeNmarginSizeX, 0);EXeSetArg(XeNspacingLeft, 6);\
	EXeSetArg(XeNshadowThickness, 0);EXeSetArg(XeNlabelType, XeIMAGE);EXeSetArg(XeNimageSizeYPtr, &Xe_IconSize);EXeSetImages2(_Xe_Theme->Image_Find);\
	EXeSetArg(XeNwidthAdjustable, FALSE);EXeSetArg(XeNalignmentX, XeAlignBEGINNING);\
	EXeSetArg(XeNheightAdjustable, FALSE);\
	XeToolTip8("Select scaling Spline (so you can find it).");\
	mResourcesGUI->ScaleBySplineShowWs[E3dSCALESPL_##mCoordinate] = LW = XeWidget_ManagedNew("Show", XePushButton_class, mParentW, Args, ArgCnt);\
	XeAddCallback(LW, XeNactivateCallback, _DCB_Scale##mCoordinate##BySplineShow, (EPointer)LPanel);\
\
	EXeStart;\
	EXeSetArg(XeNmarginSizeX, 0);\
	EXeSetArg(XeNshadowThickness, 0);EXeSetArg(XeNlabelType, XeIMAGE);EXeSetArg(XeNimageSizeYPtr, &Xe_IconSize);EXeSetImages2(_Xe_Theme->Image_Replace);\
	EXeSetArg(XeNwidthAdjustable, FALSE);EXeSetArg(XeNalignmentX, XeAlignBEGINNING);\
	EXeSetArg(XeNwidthAdjustable, FALSE);EXeSetArg(XeNalignmentX, XeAlignBEGINNING);\
	EXeSetArg(XeNheightAdjustable, FALSE);\
	XeToolTip8("Pick a new Spline to scale by.");\
	mResourcesGUI->ScaleBySplinePickWs[E3dSCALESPL_##mCoordinate] = LW = XeWidget_ManagedNew("Pick", XePushButton_class, mParentW, Args, ArgCnt);\
	XeAddCallback(LW, XeNactivateCallback, _DCB_Scale##mCoordinate##BySplinePick, (EPointer)LPanel);\
}
