//================================================================
// EQUINOX-3D(TM) plugin
//
// - Duplicate selection
// - Duplicate with repeat
//
// AUTHOR:  Gabor Nagy
// DATE:    2017-Sep-15 22:51:59
//
// Copyright (C) 1994-2021 by Gabor Nagy. All rights reserved.
//================================================================
#include <assert.h>
#include <math.h>
#include <stdlib.h>
#include <stdio.h>

#include <ELists.h>
#include <EMalloc.h>
#include <EPlugins.h>

#include <Xe/Matrix.h>
#include <Xe/Scale.h>
#include <Xe/Separator.h>
#include <Xe/PushButton.h>
#include <Xe/ToggleButton.h>

#include <Xe/Shell.h>
#include <Xe/Dialog.h>

#include <E3D/Matrix.h>
#include <E3D/Select.h>

#include <E3D/UI/Panel.h>
#include <E3D/UI/Panels.h>
#include <E3D/UI/StatusPanel.h>  // For E3dp_PrintMessage


#include "Duplicate.h"

char*  E_PluginInfo = "Duplicating\n - Copy selection\n - Duplicate with repeat\n - Duplicate over shape (Spline, Surface etc.)";

enum
{
	CR_DUPLICATE = XeCR_CUSTOM0,
	CR_INSTANCE,
	CR_SET_DUPLICATES,
	CR_SET_INSTANCE,
};


static EPlugin*      _Plugin = NULL;
static EPluginLock*  _MainDialogBoxLock = NULL;

static EguiItem      _MenuButtons[8];static int       _MenuButtons_Count = 0;
static EguiItem      _ToolBarButtons[2];static int  _ToolBarButtons_Count = 0;


// Duplicate
//
static Widget        _DialogW = NULL;

static Widget        _InstanceW = NULL, _ModelUniqueNamesW, _ObjectUniqueNamesW, _CopyMaterialsW, _CopyLODsW, _CopyConstructionHistoryW, _CopyAnimationsW;


static const EBool   _CopyBranchUnderSameParent = TRUE;
static const EBool   _ObjectCopyCreateNewNode = FALSE;
static EBool         _SelectNew = TRUE;


//================================
// Duplicate selection
//================================
static void _DoDuplicate(E3dScene* PScene, const EChangeFlags PFlags, const EBool PInstance, const EBool PSelectNew)
{
	E3dNode**  LSelectedNodes;const EIndex  LNumOfSelectedNodes = E3dScene_GetSelectedNodes(PScene,  &LSelectedNodes, NULL,  E3dNodeSelNODE|E3dNodeSelBRANCH_ROOT|E3dNodeSelOBJECT, FALSE);

	if(LNumOfSelectedNodes > 0)
	{
		EChangeFlags  LChanged = 0;
		EListI(,E3dNode*, LNodesToSelect);

		EIndex  LMC = 0;for(;LMC < LNumOfSelectedNodes;LMC++)
		{
			E3dNode*  LNode = LSelectedNodes[LMC];
			E3dNode*  LOldRootNode = LNode;while(LOldRootNode->Parent) LOldRootNode = LOldRootNode->Parent;const EIndex  LOldRootNodeIndexInScene = EList_PointerGetIndex(PScene->RootNodes, LOldRootNode);

//printf("%s() %s:%d [[%s]][%s] %d\n", __FUNCTION__, __FILE__, __LINE__, LOldRootNode->Name, LNode->Name, LNode->Selection);fflush(stdout);

			switch(LNode->Selection)
			{
				case E3dNodeSelBRANCH_ROOT:
					{
						E3dNode*  LNewRootNode = E3dNodeHrc_BranchCopy(PScene, LNode, PFlags & (~E3dCopyNODE_COMPONENTS), PInstance, E3dItemALL, FALSE, TRUE);if(LNewRootNode)
						{
							if(_CopyBranchUnderSameParent && LNode->Parent)
							{
								E3dNode*  LChild = LNode->Parent->Child;
								EIndex    LIndex = 0;while(LChild != LNode && LChild->NextSibling) { LChild = LChild->NextSibling;LIndex++; }
								E3dNode_ConnectChild(LNode->Parent, LNewRootNode, LIndex + 1, FALSE, FALSE);

								E3dNodeHrc_UpdateConnections(LOldRootNode);
								E3dNodeHrc_UpdateMatrices(LOldRootNode, TRUE);
								E3dNodeHrc_UpdateMaterialInheritance(LOldRootNode);


								E3dScene_ObjectsUpdateModelList(PScene);
								E3dScene_LayoutOnSchematics(PScene, TRUE);
								E3dp_PrintMessage(ELogMESSAGE, 5000, "Branch %s cloned to %s", LNode->Name, LNewRootNode->Name);

								if(PSelectNew)
								{
									EList_PointerAdd(LNodesToSelect, LNewRootNode);
									if(PSelectNew) LChanged |= E3dChgSELECTION;
								}

								LChanged |= E3dNodeHrc_GetTransformChangeFlagsByObjects(LNewRootNode) | E3dChgHIERARCHY;

// Clone Node info after the Matrices had been updated
//
								if(PFlags & E3dCopyCOMPONENTS) LChanged |= E3dNodeHrc_CloneInfo(PScene, LNode, LNewRootNode, E3dCopyNODE_COMPONENTS);
							}
							else
							{
								E3dScene_NodeHrcInsert(PScene, LNewRootNode, LOldRootNodeIndexInScene + 1, FALSE);
								if(!PInstance) E3dNodeHrc_UpdateForDisplay(LNewRootNode, ~(E3dUF_REMOVE_STRIPS|E3dUF_VERTEX_NORMAL|E3dUF_POLYGON_NORMAL));

								E3dScene_ObjectsUpdateModelList(PScene);
								E3dScene_LayoutOnSchematics(PScene, TRUE);
								E3dp_PrintMessage(ELogMESSAGE, 5000, "Branch %s cloned to %s", LNode->Name, LNewRootNode->Name);

								LChanged |= E3dNodeHrc_GetTransformChangeFlagsByObjects(LNewRootNode) | E3dChgHIERARCHY;

								if(PSelectNew)
								{
									EList_PointerAdd(LNodesToSelect, LNewRootNode);
									LChanged |= E3dChgSELECTION;
								}

// Clone Node info after the Matrices had been updated
//
								if(PFlags & E3dCopyCOMPONENTS) LChanged |= E3dNodeHrc_CloneInfo(PScene, LNode, LNewRootNode, E3dCopyNODE_COMPONENTS);
							}
						}
					}
				break;

				case E3dNodeSelNODE:
					{
						E3dNode*  LNewNode = E3dNode_Copy(PScene, LNode, PFlags & (~E3dCopyNODE_COMPONENTS), PInstance, E3dItemALL, FALSE, FALSE);
						if(LNewNode)
						{
							if(LNode->Parent)
							{
								E3dNode*  LChild = LNode->Parent->Child;
								EIndex    LIndex = 0;while(LChild != LNode && LChild->NextSibling) { LChild = LChild->NextSibling;LIndex++; }
								E3dNode_ConnectChild(LNode->Parent, LNewNode, LIndex + 1, FALSE, FALSE);

								E3dNodeHrc_UpdateConnections(LOldRootNode);
								E3dNodeHrc_UpdateMatrices(LOldRootNode, TRUE);
								E3dNodeHrc_UpdateMaterialInheritance(LOldRootNode);
							}
							else { E3dNodeHrc_UpdateMatrices(LNewNode, TRUE);E3dScene_NodeHrcInsert(PScene, LNewNode, LOldRootNodeIndexInScene + 1, FALSE); }

							E3dScene_ObjectsUpdateModelList(PScene);
							E3dScene_LayoutOnSchematics(PScene, TRUE);
							E3dp_PrintMessage(ELogMESSAGE, 5000, "Model %s cloned to %s", LNode->Name, LNewNode->Name);

							if(PSelectNew)
							{
								EList_PointerAdd(LNodesToSelect, LNewNode);
								if(PSelectNew) LChanged |= E3dChgSELECTION;
							}

							LChanged |= E3dNodeHrc_GetTransformChangeFlagsByObjects(LNewNode) | E3dChgHIERARCHY;

// Clone Node info after the Matrices had been updated
//
							if(PFlags & E3dCopyCOMPONENTS) LChanged |= E3dNodeHrc_CloneInfo(PScene, LNode, LNewNode, E3dCopyNODE_COMPONENTS);
						}
					}
				break;

				case E3dNodeSelOBJECT:
					if(_ObjectCopyCreateNewNode || PInstance)
					{
						E3dNode*  LNewRootNode = E3dNode_Copy(PScene, LNode, E3dp_Prefs.Duplicate_Flags, PInstance, E3dItemALL, FALSE, TRUE);if(LNewRootNode)
						{
							E3dNodeHrc_UpdateMatrices(LNewRootNode, TRUE);

							E3dScene_NodeHrcInsert(PScene, LNewRootNode, LOldRootNodeIndexInScene + 1, FALSE);

							E3dScene_ObjectsUpdateModelList(PScene);
							E3dScene_LayoutOnSchematics(PScene, TRUE);
							if(PInstance) E3dp_PrintMessage(ELogMESSAGE, 5000, "Model %s's selected Object(s) instantiated under %s", LNode->Name, LNewRootNode->Name);
							else E3dp_PrintMessage(ELogMESSAGE, 5000, "Model %s's selected Object(s) cloned to %s", LNode->Name, LNewRootNode->Name);

							LChanged |= E3dNodeHrc_GetTransformChangeFlagsByObjects(LNewRootNode) | E3dChgHIERARCHY;

							if(PSelectNew)
							{
								EList_PointerAdd(LNodesToSelect, LNewRootNode);
								LChanged |= E3dChgSELECTION;
							}
						}
					}
					else
					{
						E3dModel*  LModel = E_DynamicCast(LNode, E3dModel);if(LModel)
						{
							EIndex  LObjC = 0;const EIndex  LObjN = LModel->Objects_Count;for(;LObjC < LObjN;LObjC++)
							{
								const E3dObject*  LObject = LModel->Objects[LObjC];if(LObject->Selection == E3dObjectSelOBJECT)
								{
									E3dObject*  LNewObject = E3dObject_Copy(LObject, E3dp_Prefs.Duplicate_Flags, E3d_Scene);if(LNewObject)
									{
//printf("%s() %s:%d <%p:%s>\n", __FUNCTION__, __FILE__, __LINE__, LObject, LObject->Name);fflush(stdout);
										if(E3dp_Prefs.Duplicate_Flags & E3dCopyOBJECT_UNIQUE_NAME) EObject_SetName(LNewObject, E3dScene_GetUniqueObjectName(PScene, LObject->Name), FALSE);
										else EObject_SetName(LNewObject, EStrDup(LObject->Name), FALSE);

										E3dModel_AppendObject(LModel, LNewObject);
										E3dObject_UpdateForDisplay(LNewObject, ~(E3dUF_REMOVE_STRIPS|E3dUF_TRIANGULATE|E3dUF_EDGES|E3dUF_VERTEX_NORMAL|E3dUF_POLYGON_NORMAL), PScene);
										{ E3dMesh*  LMesh = E_DynamicCast(LNewObject, E3dMesh);if(LMesh) E3dMesh_EdgeFlagsUpdate(LMesh); }

										E3dScene_ObjectsUpdateModelList(PScene);
										E3dScene_LayoutOnSchematics(PScene, TRUE);

										LChanged |= E3dNodeHrc_GetTransformChangeFlagsByObjects(LNode) | E3dChgHIERARCHY;
									}

								}
							}
						}
					}
				break;
			}
		}

		EFree(LSelectedNodes);

		if(LNodesToSelect)
		{
// Unselect everything
//
			{
				EIndex  LC = 0;const EIndex  LN = PScene->RootNodes_Count;E3dNode**  LRootNodes = PScene->RootNodes;for(;LC < LN;LC++)
				{
					E3dNode*  LNode = LRootNodes[LC];for(;LNode;LNode = LNode->Next)
					{
						LNode->Selection = E3dNodeSelNONE;
						if(EObject_IsSubClass(LNode, E3dModel)) E3dModel_UnselectObjects(PScene, (E3dModel*)LNode,  NULL, 0,  E3dSF_ALL, TRUE);
					}
				}
			}


			EIndex  LC = 0;for(;LC < LNodesToSelect_Count;LC++)
			{
				E3dSel_NodeSelectBranch(PScene, LNodesToSelect[LC], FALSE, LC == LNodesToSelect_Count - 1);  // Only call select callbacks at the end (when LC == LNodesToSelect_Count - 1)
			}
			EFree(LNodesToSelect);
		}


		E3dWindows_RedrawByChangeFlags(LChanged);
		E3dTransformPanel_Update(NULL,  E3dDoALL, E3dDoALL, E3dDoALL,  -1,  E3dTrPanelALL);
	}
	else { E3dp_PrintMessage(ELogWARNING, 5000, "Select Model(s) or Object(s) to duplicate");return; }
}


//========================================
// Duplicate-immediate dialog callback
//========================================
static void _DCB_OK(Widget PW, EPointer PClientData, EPointer PCallData)
{
	_DoDuplicate(E3d_Scene, E3dp_Prefs.Duplicate_Flags, XeToggleButton_GetState(_InstanceW), _SelectNew);
	Xe_WidgetUnmap(_DialogW);
	E_PluginLockDel(_Plugin, &_MainDialogBoxLock, FALSE);
}


//========================================
// Duplicate-immediate dialog callback
//========================================
static void _DCB_DuplicateClose(Widget PW, EPointer PClientData, EPointer PCallData)
{
	Xe_WidgetUnmap(_DialogW);
	E_PluginLockDel(_Plugin, &_MainDialogBoxLock, FALSE);
}


//========================================
// Duplicate-immediate dialog callback
//========================================
static void _DCB_Duplicate(Widget PW, EPointer PClientData, EPointer PCallData)
{
	unsigned int  LFlag = (EPointerInt)PClientData;

	if(((XeToggleButtonCallbackStruct*)PCallData)->Set) E3dp_Prefs.Duplicate_Flags |= LFlag;
	else E3dp_Prefs.Duplicate_Flags &= ~LFlag;
}


//========================================
// Duplicate-immediate dialog callback
//========================================
static void _DCB_DuplicateInstance(Widget PW, EPointer PClientData, EPointer PCallData)
{
	Widget  LOptionWs[] = { _ObjectUniqueNamesW, _CopyMaterialsW, _CopyLODsW, _CopyConstructionHistoryW, _CopyAnimationsW };

	if(((XeToggleButtonCallbackStruct*)PCallData)->Set) XeWidgets_Unmanage(LOptionWs, sizeof(LOptionWs) / sizeof(Widget));
	else XeWidgets_Manage(LOptionWs, sizeof(LOptionWs) / sizeof(Widget));
}


//========================================
// Duplicate-immediate dialog callback
//========================================
static void _DCB_DuplicateSelectNew(Widget PW, EPointer PClientData, EPointer PCallData)
{
	_SelectNew = ((XeToggleButtonCallbackStruct*)PCallData)->Set;
}


static int _MainDialogBoxLockDel(EPlugin* PPlugin, EPluginLock* PLock)
{
	Xe_WidgetUnmap(_DialogW);

	return(E_SUCCESS);
}


//================================================
// Pop up the options dialog box
//================================================
static void _PopupDialog(const EBool PInstance)
{
	if(_MainDialogBoxLock == NULL) _MainDialogBoxLock = E_PluginLockAdd(_Plugin, _MainDialogBoxLockDel);

	if(_DialogW == NULL)
	{
		EArg  Args[256];int  ArgCnt;

		EXeStart;
		EXeSetArg(XeNallowShellResize, TRUE);
		EXeSetArg(XeNbuttonsStretch, FALSE);
		EXeSetArg(XeNworkAreaType, XeMATRIX);
		EXeSetArg(XeNwmDecorations, XeWMDecorCLOSE|XeWMDecorBORDER|XeWMDecorTITLE);
		EXeSetArg(XeNtitle, "Duplicate options");EXeSetArg(XeNiconName, "Duplicate");
		Widget  LDialogW = XeDialog_New(XeDIALOG_TEMPLATE, "Duplicate", Eqx_TopShellW,  Args, ArgCnt);if(LDialogW == NULL) return;
		XeAddCallback(LDialogW, XeNwmCallback, _DCB_DuplicateClose, (EPointer)0);

		_DialogW = LDialogW;

		Widget  LMessageBoxW = XeDialog_GetChild(LDialogW, XeDIALOG_MAIN);
		EXeStart;
		EXeSetArg(XeNworkAreaMarginSizeX, 6);
		EXeSetArg(XeNworkAreaMarginSizeY, 6);
		Xe_SetValues(LMessageBoxW,  Args, ArgCnt);
		{ Widget  LW = XeDialog_AddButton("Do it", XePushButton_class, LMessageBoxW, TRUE);XeAddCallback(LW, XeNactivateCallback, _DCB_OK, (EPointer)XeCR_OK); }


		Widget  LTopMatrixW = XeDialog_GetChild(LDialogW, XeDIALOG_WORK_AREA);

		EXeStart;
		EXeSetArg(XeNorientation, XeVERTICAL);
		EXeSetArg(XeNmarginTop,   0);
		EXeSetArg(XeNmarginSizeX, 0);
		Xe_SetValues(LTopMatrixW,  Args, ArgCnt);

		Widget  LW;
		EXeStart;
		EXeSetArg(XeNshadowThickness, 0);
		EXeSetArg(XeNhighlightThickness, 0);
		EXeSetArg(XeNheightAdjustable, FALSE);
		_InstanceW = LW = XeWidget_ManagedNew("Instance", XeToggleButton_class, LTopMatrixW,  Args, ArgCnt);
		XeToggleButton_SetState(LW, PInstance, FALSE);
		XeAddCallback(LW, XeNvalueChangedCallback, _DCB_DuplicateInstance, (EPointer)NULL);

		LW = XeWidget_ManagedNew("Select new copy", XeToggleButton_class, LTopMatrixW,  Args, ArgCnt);
		XeAddCallback(LW, XeNvalueChangedCallback, _DCB_DuplicateSelectNew, (EPointer)NULL);
		XeToggleButton_SetState(LW, _SelectNew, FALSE);

		EXeStart;
		EXeSetArg(XeNshadowThickness, 0);
		EXeSetArg(XeNhighlightThickness, 0);
		EXeSetArg(XeNspacingTop, 8);
		EXeSetArg(XeNheightAdjustable, FALSE);
		_ModelUniqueNamesW = LW = XeWidget_ManagedNew("Generate unique names for Models", XeToggleButton_class, LTopMatrixW,  Args, ArgCnt);
		XeAddCallback(LW, XeNvalueChangedCallback, _DCB_Duplicate, (EPointer)E3dCopyMODEL_UNIQUE_NAME);
		XeToggleButton_SetState(LW, (EBool)((E3dp_Prefs.Duplicate_Flags & E3dCopyMODEL_UNIQUE_NAME) != 0), FALSE);

		EXeStart;
		EXeSetArg(XeNshadowThickness, 0);
		EXeSetArg(XeNhighlightThickness, 0);
		EXeSetArg(XeNheightAdjustable, FALSE);
		_ObjectUniqueNamesW = LW = XeWidget_ManagedNew("Generate unique names for Objects", XeToggleButton_class, LTopMatrixW,  Args, ArgCnt);
		XeAddCallback(LW, XeNvalueChangedCallback, _DCB_Duplicate, (EPointer)E3dCopyOBJECT_UNIQUE_NAME);
		XeToggleButton_SetState(LW, (EBool)((E3dp_Prefs.Duplicate_Flags & E3dCopyOBJECT_UNIQUE_NAME) != 0), FALSE);

		_CopyMaterialsW = LW = XeWidget_New("Copy Materials", XeToggleButton_class, LTopMatrixW,  Args, ArgCnt);
//		Xe_SetValuesVa(LW,  XeNmarginTop, 8,  NULL);
		XeAddCallback(LW, XeNvalueChangedCallback, _DCB_Duplicate, (EPointer)E3dCopyMATERIALS);
		XeToggleButton_SetState(LW, (EBool)((E3dp_Prefs.Duplicate_Flags & E3dCopyMATERIALS) != 0), FALSE);

		_CopyLODsW = LW = XeWidget_New("Copy Geometry levels of detail", XeToggleButton_class, LTopMatrixW,  Args, ArgCnt);
		XeAddCallback(LW, XeNvalueChangedCallback, _DCB_Duplicate, (EPointer)E3dCopyLODS);
		XeToggleButton_SetState(LW, (EBool)((E3dp_Prefs.Duplicate_Flags & E3dCopyLODS) != 0), FALSE);

		_CopyConstructionHistoryW = LW = XeWidget_New("Copy construction history", XeToggleButton_class, LTopMatrixW,  Args, ArgCnt);
		XeAddCallback(LW, XeNvalueChangedCallback, _DCB_Duplicate, (EPointer)E3dCopyCOMPONENTS);
		XeToggleButton_SetState(LW, (EBool)((E3dp_Prefs.Duplicate_Flags & E3dCopyCOMPONENTS) != 0), FALSE);

		_CopyAnimationsW = LW = XeWidget_New("Copy animations", XeToggleButton_class, LTopMatrixW,  Args, ArgCnt);
		XeAddCallback(LW, XeNvalueChangedCallback, _DCB_Duplicate, (EPointer)E3dCopyANIMATIONS);
		XeToggleButton_SetState(LW, (EBool)((E3dp_Prefs.Duplicate_Flags & E3dCopyANIMATIONS) != 0 ), FALSE);

		if(!PInstance)
		{
			Widget  LOptionWs[] = { _ObjectUniqueNamesW, _CopyMaterialsW, _CopyLODsW, _CopyConstructionHistoryW, _CopyAnimationsW };

			XeWidgets_Manage(LOptionWs, sizeof(LOptionWs) / sizeof(Widget));
		}

		XeShell_Raise(_DialogW);
		XeShell_CenterToWidget(_DialogW, Eqx_TopShellW);
	}
	else
	{
		Widget  LOptionWs[] = { _ObjectUniqueNamesW, _CopyMaterialsW, _CopyLODsW, _CopyConstructionHistoryW, _CopyAnimationsW };

		if(PInstance) XeWidgets_Unmanage(LOptionWs, sizeof(LOptionWs) / sizeof(Widget));
		else XeWidgets_Manage(LOptionWs, sizeof(LOptionWs) / sizeof(Widget));

		if(XeShell_DisplayStatus(_DialogW) != XeDISPLAYED)
		{
			XeToggleButton_SetState(_InstanceW, PInstance, FALSE);

			XeShell_Raise(_DialogW);
		}
	}
}


//========================================
// Duplicate menu callback
//========================================
static void _MCB(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dScene*  LScene = E3d_Scene;

	switch((EPointerInt)PClientData)
	{
		case CR_DUPLICATE:
			if(((XePushButtonCallbackStruct*)PCallData)->Reason == XeCR_ACTIVATE_OPTIONS) _PopupDialog(FALSE);
			else _DoDuplicate(LScene, E3dp_Prefs.Duplicate_Flags, FALSE, _SelectNew);
		break;

		case CR_INSTANCE:
			if(((XePushButtonCallbackStruct*)PCallData)->Reason == XeCR_ACTIVATE_OPTIONS) _PopupDialog(TRUE);
			else _DoDuplicate(LScene, E3dp_Prefs.Duplicate_Flags, TRUE, _SelectNew);
		break;
	}
}


//========================================
// Key callback
//========================================
static void _KCB(const EKeyCode PKey, const unsigned int PModifiers, const int PKeyEventType, EPointer PClientData)
{
	switch((EPointerInt)PClientData)
	{
		case CR_DUPLICATE:

		case CR_INSTANCE:
			{
				XePushButtonCallbackStruct  LCBS;

				LCBS.Reason = XeCR_ACTIVATE;
				_MCB(NULL, PClientData, &LCBS);
			}
		break;
	}
}


//========================================
// Entry point of the plugin
//========================================
int E_PluginInit(EPlugin* PPlugin)
{
	_Plugin = PPlugin;

	if(_DuplicateWithRepeatInit())
	{
		_MenuButtons[_MenuButtons_Count++] = XePushButton_Add("Menu->Clone", "Immediate", '\n', "D", "D", TRUE, NULL, _MCB, (EPointer)CR_DUPLICATE);
		E_KeySetCallback("Duplicate selection",            'D', 0,          EKeyPRESS, _KCB, (EPointer)CR_DUPLICATE);

		_MenuButtons[_MenuButtons_Count++] = XePushButton_Add("Menu->Clone", "New instance", '\0', "Shift+D", "Shift+D", TRUE, NULL, _MCB, (EPointer)CR_INSTANCE);
		E_KeySetCallback("Duplicate: create new instance", 'D', EShiftMask, EKeyPRESS, _KCB, (EPointer)CR_INSTANCE);
	}

	return(E_SUCCESS);
}


//========================================
// Exit method of the plugin
//========================================
int E_PluginExit(const EBool PExiting)
{
	if(!PExiting || E3dp_GUIFreeOnExit)
	{
		int  LC;

		for(LC = 0;LC < _MenuButtons_Count;LC++) { XeItemFree(_MenuButtons[LC]); } _MenuButtons_Count = 0;
		for(LC = 0;LC < _ToolBarButtons_Count;LC++) { XeItemFree(_ToolBarButtons[LC]); } _ToolBarButtons_Count = 0;
	}

	E_KeyRemoveCallback('D', 0,          EKeyPRESS, _KCB, (EPointer)CR_DUPLICATE);
	E_KeyRemoveCallback('D', EShiftMask, EKeyPRESS, _KCB, (EPointer)CR_INSTANCE);

	_DuplicateWithRepeatFree();

	return(E_SUCCESS);
}
