//================================================================
// EQUINOX-3D(TM) plugin
//
// - Edit selected Node, Object, PolyGroup etc
//
// AUTHOR:  Gabor Nagy
// DATE:    2019-Jun-23 23:27:46
//
// Copyright (C) 1994-2021 By Gabor Nagy. All rights reserved.
//================================================================
#include <stdio.h>

#include <EMalloc.h>
#include <EPlugins.h>

#include <E3D/Scene.h>
#include <E3D/Editing.h>
#include <E3D/Deleting.h>
#include <E3D/SplineOps.h>

#include <E3D/UI/Panel.h>
#include <E3D/UI/Panels.h>
#include <E3D/UI/StatusPanel.h>  // For E3dp_PrintMessage

#include <Xe/TextField.h>

#include <Xe/Shell.h>
#include <Xe/Dialog.h>
#include <Xe/ToggleButton.h>

#include "Edit.h"

#include "../EQXExt.h"


EPlugin*             _EditInfoPlugin = NULL;
static EPluginLock*  _NodeDialogLock = NULL;


static EguiItem          _MenuButtons[16];static int  _MenuButtons_Count = 0;
static EScriptFunction*  _ScriptFunctions[32];static int  _ScriptFunctions_Count = 0;


// Node dialog box
//
static Widget         _NodeDialogW = NULL;
static E3dNodePanel*  _NodePanel = NULL;


static void     _CB_Scene(ECoreObject* PScn, EPointer PClientData, EPointer PCallData);

static EBool    _NodeSelectionCallbackAdded = FALSE;





enum
{
	CR_EDIT_SELECTION = XeCR_CUSTOM0
};


static void _NodeDialogClose(void)
{
	E3dNodePanel*  LPanel = _NodePanel;
	E3dNode*       LNode = LPanel->Node;if(LNode)
	{
		{
			EIndex  LC = 0;for(;LC < LNode->Components_Count;LC++)
			{
				E3dNodeComponent*  LInfo = LNode->Components[LC];

//printf("%s() %s:%d  [%s] {%s}\n", __FUNCTION__, __FILE__, __LINE__, LNode->Name, LInfo->Class->Name);fflush(stdout);
				if(LInfo->Class && LInfo->Class->GUI) LInfo->Class->GUI(LNode, LInfo, LPanel, NULL, E_CLOSE);
			}
		}

		Xe_WidgetUnmap(_NodeDialogW);

		EChangeFlags	LChanged = E3dNode_LockDel(LNode, &E3dNode_EditLock);

		E_Delete(LPanel->Node);

		if(LChanged) E3dWindows_RedrawByChangeFlags(LChanged);
	}

	E3dScene_CallbackRemove(E3d_Scene, _CB_Scene, (EPointer)0);
	_NodeSelectionCallbackAdded = FALSE;
}


//========================================
// Node dialog callback
//========================================
static void _DCB_NodeClose(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E_PluginLockDel(_EditInfoPlugin, &_NodeDialogLock, FALSE);
}


static int _NodeDialogLockDel(EPlugin* PPlugin, EPluginLock* PLock)
{
	_NodeDialogClose();

	return(E_SUCCESS);
}


//========================================
// Edit Node attributes
//========================================
static void _NodePopupDialog(E3dNode* PNode, E3dNodeComponentClass* POpenFolderForThisComponentClass)
{
	if(_NodeDialogLock == NULL) _NodeDialogLock = E_PluginLockAdd(_EditInfoPlugin, _NodeDialogLockDel);

	E3dNodePanel*  LPanel;

	if(_NodeDialogW == NULL)
	{
		EArg  Args[256];int  ArgCnt;

		EXeStart;
		EXeSetArg(XeNallowShellResize, TRUE);
		EXeSetArg(XeNcreateButtons, FALSE);
		EXeSetArg(XeNworkAreaType, XeMATRIX);
		EXeSetArg(XeNminSizeX, 240);
		EXeSetArg(XeNcreateButtons, FALSE);
		EXeSetArg(XeNwmDecorations, XeWMDecorCLOSE|XeWMDecorBORDER|XeWMDecorRESIZEH|XeWMDecorTITLE);
		EXeSetArg(XeNtitle, "Edit Node attributes");EXeSetArg(XeNiconName, "Node");
		Widget  LDialogW = XeDialog_New(XeDIALOG_TEMPLATE, "Node", E3dp_TopShellW,  Args, ArgCnt);if(LDialogW == NULL) return;
		XeAddCallback(LDialogW, XeNwmCallback, _DCB_NodeClose, (EPointer)_NodePanel);

		_NodeDialogW = LDialogW;

		Widget  LTopMatrixW = XeDialog_GetChild(LDialogW, XeDIALOG_WORK_AREA);
		EXeStart;
		EXeSetArg(XeNorientation, XeVERTICAL);
		EXeSetArg(XeNspacingY, 2);
		Xe_SetValues(LTopMatrixW,  Args, ArgCnt);

		_NodePanel = LPanel = E3dNodePanel_Create(LTopMatrixW, PNode, -1, POpenFolderForThisComponentClass);
//		E3dNodePanel_Update(_NodePanel, PNode, -1, NULL);

		XeShell_Raise(_NodeDialogW);
		XeShell_CenterToWidget(_NodeDialogW, E3dp_TopShellW);

		if(E3dNode_LockAdd(PNode, &E3dNode_EditLock))  E3dWindows_RedrawByChangeFlags(E3dChgLOCK_UNLOCK);
	}
	else
	{
		EBool  LRedraw = FALSE;

		LPanel = _NodePanel;

// Is the dialog-box already up?
//
		if(LPanel->Node == PNode)
		{
			E3dNodePanel_Update(LPanel, PNode, -1, POpenFolderForThisComponentClass);

// Prevent the dialog box stealing keyboard focus, by only raising it, if it's not visible
//
			if(XeShell_DisplayStatus(_NodeDialogW) != XeDISPLAYED) XeShell_Raise(_NodeDialogW);

			return;
		}
		else if(LPanel->Node)
		{
			E3dNode_LockDel(LPanel->Node, &E3dNode_EditLock);
			LRedraw = TRUE;
		}

		E3dNodePanel_Update(_NodePanel, PNode, -1, POpenFolderForThisComponentClass);
		if(E3dNode_LockAdd(LPanel->Node, &E3dNode_EditLock)) LRedraw = TRUE;

		if(LRedraw) E3dWindows_RedrawByChangeFlags(E3dChgLOCK_UNLOCK);
	}

	if(Eqx_Prefs.SelectAndFocusOnName)
	{
		XeShell_Raise(_NodeDialogW);
		XeDisplay*  LDisplay = Xe_GetDisplay(LPanel->NameW);
		if(!E_IsKeyDown(LDisplay, 0))
		{
			Xe_SetFocus(LPanel->NameW, TRUE, TRUE);
			XeTextField_SelectAll(LPanel->NameW);
		}
	}

	if(!_NodeSelectionCallbackAdded)
	{
		E3dScene_CallbackAdd(E3d_Scene, E3dChgSELECTION, _CB_Scene, (EPointer)0);
		_NodeSelectionCallbackAdded = TRUE;
	}
}


//========================================
// Edit Node attributes
//========================================
static void _Node_Edit(E3dNode* PNode, E3dNodeComponentClass* POpenFolderForThisComponentClass)
{
	EBool  LDoEdit = TRUE;

	if(PNode->Components)
	{
		E3dNodeComponentClass*  LClass = PNode->Components[0]->Class;if(LClass)
		{
			if(LClass->Edit)
			{
				const int  LResult = LClass->Edit((E3dNode*)PNode, PNode->Components[0]);

				switch(LResult)
				{
					case E3dEDITP_NOTHING:
					case E3dEDITP_INFO_ADDED:
					break;

					default:  LDoEdit = FALSE;break;
				}
			}
		}
	}

	if(LDoEdit) _NodePopupDialog(PNode, POpenFolderForThisComponentClass);
}


//========================================================================================
// Return TRUE, if PNode's attributes are being edited (the dialog box is displayed)
//========================================================================================
static EBool _Node_IsEdited(E3dNode* PNode)
{
	if(_NodePanel && _NodePanel->Node == PNode) return(TRUE);
	return(FALSE);
}


//========================================
// Menu callback
//========================================
static void _MCB_Edit(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dItem  LSelectedItem;

	switch((EPointerInt)PClientData)
	{
		case CR_EDIT_SELECTION:
			switch(E3dScene_GetSingleSelection(E3d_Scene, &LSelectedItem, E3dItemALL, 0))
			{
				case E3dItemMODEL:
					_Node_Edit(LSelectedItem.ObjItem.Node, NULL);
				break;

				case E3dItemOBJECT:
					{
						E3dObjItem*               LObjItem = &LSelectedItem.ObjItem;
						E3dObject*                LObject = LObjItem->Object;
						E3dObjectComponentClass*  LOpenFolderForThisComponentClass = NULL;
						if(LObject->Components_Count == 1) LOpenFolderForThisComponentClass = LObject->Components[0]->Class;

						_Object_Edit((E3dModel*)LObjItem->Node, LObject, 0, NULL, -1,  LOpenFolderForThisComponentClass);
					}
				break;

				case E3dItemFACE_CONTOUR:
					{
						E3dSplineItem*  LSplineItem = &LSelectedItem.SplineItem;

						_Object_Edit((E3dModel*)(LSplineItem->Node), LSplineItem->Object, 0, (EObject*)(LSplineItem->Spline), 0,  NULL);
					}
				break;

				case E3dItemSPLINE_SEGMENT:
					{
						E3dSplineItem*  LSplineItem = &LSelectedItem.SplineItem;
						E3dObject*      LObject = LSplineItem->Object;

						if(EObject_IsSubClass(LObject, E3dSpline)) _Object_Edit((E3dModel*)(LSplineItem->Node), LObject, 0, NULL, LSplineItem->PointIndex,  NULL);
						else if(EObject_IsSubClass(LObject, E3dFace)) _Object_Edit((E3dModel*)(LSplineItem->Node), LObject, 0, (EObject*)(LSplineItem->Spline), LSplineItem->PointIndex,  NULL);
					}
				break;

				case E3dItemPOLYGROUP:
					{
						E3dPolyItem*  LPolyItem = &LSelectedItem.PolyItem;

						_Edit_PolyGroupPopupDialog(LPolyItem->PolyGroup, NULL, TRUE);
					}
				break;
			}
		break;
	}
}


//========================================
// Key callback	
//========================================
static void _KCB_Edit(const EKeyCode PKey, const unsigned int PModifiers, const int PKeyEventType, EPointer PClientData)
{
	_MCB_Edit(NULL, PClientData, NULL);
}



//========================================
// Scene changed callback
//========================================
static void _CB_Scene(ECoreObject* PScn, EPointer PClientData, EPointer PCallData)
{
	switch(((E3dCallbackStruct*)PCallData)->Reason)
	{
		case E3dCR_SELECTION_CHANGED:
			{
				E3dNodePanel*  LPanel = _NodePanel;if(LPanel)
				{
					if(LPanel->FollowSelection)
					{
						if(LPanel->Node)
						{
							E3dItem  LSelectedItem;

							switch(E3dScene_GetSingleSelection(E3d_Scene, &LSelectedItem, E3dItemALL, 0))
							{
								case E3dItemMODEL:
									_Node_Edit(LSelectedItem.ObjItem.Node, NULL);
								break;
							}
						}
					}
				}
			}
		break;
	}
}









//-------------------------------------------------------------------------------------------------------
// Deleting
//-------------------------------------------------------------------------------------------------------


static EOperatorClass*  _OperatorDelete_class           = NULL;
static EOperatorClass*  _OperatorDeleteAnimations_class = NULL;

// E3dFINISHME this avoids calling E3dMesh_PolyGroupsEdgesFromMeshEdges() after Mesh Edges were added / removed, so it may be faster at the cost of using more memory to store the removed Mesh Edge indices for UNDO.
// Implementation of this is not quite finished yet. It only remaps Mesh Edge indices in PolyGroups with added/removed Edges and/or Polygons...
// We need to do this in all PolyGroups of an affected Mesh!
// Test and see if the speed-up is worth finishing the implementation.
//
//#define E3dRESTORE_POLYGROUP_EDGES  1

//#define E3dDebugEDGES              1
//#define E3dDebugDELETE_VERTICES    1


//========================================
// Free an array of ObjectAndIndex
//========================================
static void _ObjectAndIndexListFree(EObjectAndIndex*** PListP, EIndex* PCountP)
{
	if(PListP)
	{
		EObjectAndIndex**  LList = *PListP;

		EIndex  LC = 0;const EIndex  LN = *PCountP;for(;LC < LN;LC++)
		{
			EObjectAndIndex*  LEntry = LList[LC];
			EObject*          LObject = LEntry->Object;

			_EM_Delete(LObject, &LEntry->Object);
			EFree(LEntry);
		}
		EFree(LList);*PListP = NULL;*PCountP = 0;
	}
}


static void _OpObjectFreeContents(E3dOpObject* POpObject)
{
	E_DeleteChk(POpObject->Object);

	_ObjectAndIndexListFree(&POpObject->Models, &POpObject->Models_Count);

	if(POpObject->SubObjects)
	{
		EIndex  LSOC = 0;const EIndex  LSON = POpObject->SubObjects_Count;E3dOpSubObject*  LOpSubObject = POpObject->SubObjects;for(;LSOC < LSON;LSOC++, LOpSubObject++)
		{
			E_Delete(LOpSubObject->SubObject);
		}
		EFree(POpObject->SubObjects);
	}
}


#define _PolyGroupPrintEdges(mPolyGroup) do\
{\
	const EIndex  LPGEN = mPolyGroup->Edges_Count;\
printf("%s() %s:%d  PolyGroup Edges: %3d:\n", __FUNCTION__, __FILE__, __LINE__, (int)LPGEN);fflush(stdout);\
	EIndex*  LPolyGroupEdgeIndices = mPolyGroup->Edges;\
\
	EIndex  LPGEC = 0;for(;LPGEC < LPGEN;LPGEC++)\
	{\
printf(" %s%d %d%s|", ENormalCode, (int)LPGEC, (int)LPolyGroupEdgeIndices[LPGEC], ENormalCode);fflush(stdout);\
	}\
printf("\n");fflush(stdout);\
} while(0)


//========================================
// Initialize ObjectsEdit Operator
//========================================
static void _EditObjects_Init(EOperator* POperator)
{
	E3dOperatorEditObjects*  LOperator = (E3dOperatorEditObjects*)POperator;

	EList_Init(LOperator->Objects);
}


//========================================
// Free ObjectsEdit Operator
//========================================
static int _EditObjects_Free(EOperator* POperator)
{
	E3dOperatorEditObjects*  LOperator = (E3dOperatorEditObjects*)POperator;
	const EIndex             LObjects_Count = LOperator->Objects_Count;if(LObjects_Count)
	{
		EIndex  LC = 0;for(;LC < LObjects_Count;LC++)
		{
			E3dOpObject*  LOpObject = LOperator->Objects[LC];_OpObjectFreeContents(LOpObject);EFree(LOpObject);
		}
		EList_Free(LOperator->Objects);
	}

	return(E_SUCCESS);
}


//================================================
// EditObjects Operator proc
//================================================
static EChangeFlags _EditObjects_Proc(EOps* POps, EOperator* POperator, const int POperation)
{
	E3dScene*                LScene = E3d_Scene;
	E3dOperatorEditObjects*  LOperator = (E3dOperatorEditObjects*)POperator;
	const EIndex             LObjects_Count = LOperator->Objects_Count;

	EListAI(, EOpNode, LOpNodes);
	EChangeFlags  LChanged = 0;

	switch(POperation)
	{
		case EOpDO:
			{
				if(LObjects_Count)
				{
					EIndex  LC = 0;for(;LC < LObjects_Count;LC++)
					{
						E3dOpObject*  LOpObject = LOperator->Objects[LC];
						E3dObject*    LObject = LOpObject->Object;

						if(LOpObject->SubObjects)
						{
							if(EObject_IsSubClass(LObject, E3dFace))
							{
								E3dFace*  LFace = (E3dFace*)LObject;
								EIndex    LFaceHolesRemoved = 0;

								EIndex  LSOC = 0;const EIndex  LSON = LOpObject->SubObjects_Count;E3dOpSubObject*  LOpSubObject = LOpObject->SubObjects;for(;LSOC < LSON;LSOC++, LOpSubObject++)
								{
									E3dSpline*  LHole = E_DynamicCast(LOpSubObject->SubObject, E3dSpline);if(LHole)
									{
										E3dFace_RemoveHole(LFace, LHole);

										LFaceHolesRemoved++;
									}
								}

								if(LFaceHolesRemoved)
								{
// Free tesselation and E3dGL_FaceDraw() or E3d_GetRenderTriangles() etc. will call E3dFace_Tesselate() when necessary
// This will speed things up when there are no Windows with Solid display modes
//
//printf("%s() %s:%d <%s>\n", __FUNCTION__, __FILE__, __LINE__, LFace->Name);fflush(stdout);
									E3dFace_FreeTesselation(LFace);
									LChanged |= E3dChgFACE;

									EOpNode*  LOpNode = EList_StructAddAClr(LOpNodes, 4);if(LOpNode)
									{
										LOpNode->Object  = (ECoreObject*)LObject;
										LOpNode->Changed = E3dChgFACE;
									}
								}
							}
						}
						else
						{
							LChanged |= E3dObject_GetChangeFlagsByClass(LObject) | E3dChgHIERARCHY;

							EIndex  LMC = 0;const EIndex  LMN = LOpObject->Models_Count;for(;LMC < LMN;LMC++)
							{
								E3dModel_RemoveObject((E3dModel*)(LOpObject->Models[LMC]->Object), LObject);
							}

							if(LObject->Models_Count == 0) E3dObject_FreeRenderData(LObject);
						}
					}
				}
			}
		break;

		case EOpUNDO:
			{
				if(LObjects_Count)
				{
					EIndex  LC = 0;for(;LC < LObjects_Count;LC++)
					{
						E3dOpObject*  LOpObject = LOperator->Objects[LC];
						E3dObject*    LObject = LOpObject->Object;

						if(LOpObject->SubObjects)
						{
							if(EObject_IsSubClass(LObject, E3dFace))
							{
								E3dFace*  LFace = (E3dFace*)LObject;
								EIndex    LFaceHolesAdded = 0;

								EIndex  LSOC = 0;const EIndex  LSON = LOpObject->SubObjects_Count;E3dOpSubObject*  LOpSubObject = LOpObject->SubObjects;for(;LSOC < LSON;LSOC++, LOpSubObject++)
								{
									E3dSpline*  LHole = E_DynamicCast(LOpSubObject->SubObject, E3dSpline);if(LHole)
									{
										E3dFace_InsertHole(LFace, LHole, LOpSubObject->OriginalIndex);
										LFaceHolesAdded++;

/*
										if(EList_ObjectInsert(LFace->Holes, LOpSubObject->OriginalIndex, LHole))
										{
											LFaceHolesAdded++;
										}
*/
									}
								}

								if(LFaceHolesAdded)
								{
// Free tesselation and E3dGL_FaceDraw() or E3d_GetRenderTriangles() etc. will call E3dFace_Tesselate() when necessary
// This will speed things up when there are no Windows with Solid display modes
//
									E3dFace_FreeTesselation(LFace);
									LChanged |= E3dChgFACE;

									EOpNode*  LOpNode = EList_StructAddAClr(LOpNodes, 4);if(LOpNode)
									{
										LOpNode->Object  = (ECoreObject*)LObject;
										LOpNode->Changed = E3dChgFACE;
									}
								}
							}
						}
						else
						{
							EIndex  LMC = 0;const EIndex  LMN = LOpObject->Models_Count;for(;LMC < LMN;LMC++)
							{
								EObjectAndIndex*  LEntry = LOpObject->Models[LMC];
								E3dModel*         LModel = (E3dModel*)(LEntry->Object);

								E3dModel_InsertObject(LModel, LObject, LEntry->Index);

								{
									E3dNode*  LRootNode = (E3dNode*)LModel;

									while(LRootNode->Parent) LRootNode = LRootNode->Parent;

									E3dNodeHrc_UpdateMaterialInheritance(LRootNode);
								}

								E3dScene_MaterialsAddFromObject(LScene, LObject);
							}

							E3dObject_UpdateForDisplay(LObject, ~(E3dUF_EDGES|E3dUF_REMOVE_STRIPS|E3dUF_TRIANGULATE|E3dUF_CALL_CHANGEPROCS), LScene);
							LChanged |= E3dObject_GetChangeFlagsByClass(LObject) | E3dChgHIERARCHY;
						}
					}
				}
			}
		break;
	}


	if(LOpNodes)
	{
		LChanged |= E3dObjects_CallOutputs(ECbINTERNAL|ECbUSER, LOpNodes, LOpNodes_Count);
		EFree(LOpNodes);
	}

	return(LChanged);
}


//========================================
// Free E3dOpPolyGroup
//========================================
static void _OpPolyGroupFreeContents(E3dOpPolyGroup* POpPolyGroup)
{
	if(POpPolyGroup->PolygonsToModify)
	{
		EIndex  LPC = 0;const EIndex  LPN = POpPolyGroup->PolygonsToModify_Count;E3dOpPolygon*  LDVPolygon = POpPolyGroup->PolygonsToModify;for(;LPC < LPN;LPC++, LDVPolygon++)
		{
			EFree(LDVPolygon->VertexNodesToModify_Indices);
			EFree(LDVPolygon->VertexNodesToModify);
		}
		EFree(POpPolyGroup->PolygonsToModify);
	}

	if(POpPolyGroup->PolygonsToDelete)
	{
		EIndex  LPC = 0;const EIndex  LPN = POpPolyGroup->PolygonsToDelete_Count;E3dPolygon*  LPolygon = POpPolyGroup->PolygonsToDelete;for(;LPC < LPN;LPC++, LPolygon++)
		{
			E3dPolygon_FreeContents(LPolygon);
		}
		EFree(POpPolyGroup->PolygonsToDelete);
	}
	EFreeChk(POpPolyGroup->PolygonsToDelete_Indices);

	if(POpPolyGroup->PolygonsToAdd)
	{
		EIndex  LPC = 0;const EIndex  LPN = POpPolyGroup->PolygonsToAdd_Count;E3dPolygon*  LPolygon = POpPolyGroup->PolygonsToAdd;for(;LPC < LPN;LPC++, LPolygon++)
		{
			E3dPolygon_FreeContents(LPolygon);
		}
		EFree(POpPolyGroup->PolygonsToAdd);
	}
	EFreeChk(POpPolyGroup->PolygonsToAdd_Indices);


	EList_FreeChk(POpPolyGroup->EdgeIndicesToDelete);
	EFreeChk(POpPolyGroup->EdgeIndicesToDelete_Indices);

	E_Delete(POpPolyGroup->PolyGroup);
}


static void _AddDeleteVertices_Init(EOperator* POperator)
{
	E3dOperatorAddDeleteVertices*  LOperator = (E3dOperatorAddDeleteVertices*)POperator;

	LOperator->MeshNewT = NULL;

	LOperator->Vertices_Indices = NULL;
	EList_Init(LOperator->Vertices);
	E3dVertexAttribute_ListInit(&LOperator->VertexAttributes);
	LOperator->VertexSize = 0;

	EList_Init(LOperator->OpPolyGroups);

	LOperator->EdgesToDelete_Indices = NULL;EList_Init(LOperator->EdgesToDelete);
	LOperator->EdgesToAdd_Indices    = NULL;EList_Init(LOperator->EdgesToAdd);

	LOperator->DeleteMeshVertices = TRUE;
}


static int _AddDeleteVertices_Free(EOperator* POperator)
{
	E3dOperatorAddDeleteVertices*  LOperator = (E3dOperatorAddDeleteVertices*)POperator;

//printf("%s() %s:%d  %p:%s\n", __FUNCTION__, __FILE__, __LINE__, LOperator, LOperator->Name);fflush(stdout);
	EFreeChkClr(LOperator->Vertices_Indices);
	E3dVertices_Free((E3dVertex*)LOperator->Vertices, LOperator->VertexSize, LOperator->Vertices_Count, &LOperator->VertexAttributes);LOperator->Vertices = NULL;

	E3dVertexAttribute_ListFree(&LOperator->VertexAttributes);
	LOperator->VertexSize = 0;

	if(LOperator->OpPolyGroups)
	{
		EIndex  LSOC = 0;const EIndex  LSON = LOperator->OpPolyGroups_Count;for(;LSOC < LSON;LSOC++)
		{
			_OpPolyGroupFreeContents(LOperator->OpPolyGroups + LSOC);
		}
		EFree(LOperator->OpPolyGroups);
	}

	EFreeChk(LOperator->EdgesToDelete_Indices);EList_FreeChk(LOperator->EdgesToDelete);
	EFreeChk(LOperator->EdgesToAdd_Indices);   EList_FreeChk(LOperator->EdgesToAdd);

	E_DeleteChk(LOperator->MeshNewT);

	return(E_SUCCESS);
}


//================================================
// DeleteVertices Operator proc
//================================================
static EChangeFlags _DeleteVertices_Proc(EOps* POps, EOperator* POperator, const int POperation)
{
	E3dOperatorAddDeleteVertices*  LOperator = (E3dOperatorAddDeleteVertices*)POperator;
	EChangeFlags                   LChanged = 0;
	E3dMesh*                       LMesh = LOperator->MeshNewT;
	E3dScene*                      LScene = E3d_Scene;

	switch(POperation)
	{
		case EOpDO:
			if(LOperator->Vertices_Count || LOperator->EdgesToDelete_Count || LOperator->EdgesToAdd_Count)
			{
				E3dMesh_FreeCaches(LMesh);

				if(LOperator->OpPolyGroups)
				{
					EIndex  LPGC = 0;const EIndex  LPGN = LOperator->OpPolyGroups_Count;E3dOpPolyGroup*  LOpPolyGroup = LOperator->OpPolyGroups;for(;LPGC < LPGN;LPGC++, LOpPolyGroup++)
					{
						E3dPolyGroup*            LPolyGroup = LOpPolyGroup->PolyGroup;
						const E3dVertexNodeSize  LVertexNodeSize = LPolyGroup->VertexNodeSize;

						EIndex  LPC = 0;const EIndex  LPN = LOpPolyGroup->PolygonsToModify_Count;E3dOpPolygon*  LDVPolygon = LOpPolyGroup->PolygonsToModify;for(;LPC < LPN;LPC++, LDVPolygon++)
						{
							E3dPolygon*           LPolygon = LPolyGroup->Polygons + LDVPolygon->PolygonIndex;
							const E3dVertexNode*  LVertexNodes = LPolygon->VertexNodes;
							const EIndex*         LVertexNodesToDelete_Indices = LDVPolygon->VertexNodesToModify_Indices;

#ifdef E3dDebugDELETE_VERTICES
printf("\n%s() %s:%d  %sDO%s   Poly %d  ", __FUNCTION__, __FILE__, __LINE__, EGreenCode, ENormalCode, (int)(LPolygon - LPolyGroup->Polygons));fflush(stdout);
E3dPolygon_Print(LPolygon, LVertexNodeSize);printf("\n");fflush(stdout);
#endif
							const EIndex  LOrigNumOfExteriorVertices = LPolygon->NumOfExteriorVertices;
							EIndex  LC = 0;const EIndex  LN = LDVPolygon->VertexNodesToModify_Count;for(;LC < LN;LC++)
							{
								const EIndex  LVertexNodeIndex = LVertexNodesToDelete_Indices[LC];
								if(LVertexNodeIndex < LOrigNumOfExteriorVertices) LPolygon->NumOfExteriorVertices -= 1;

								const E3dVertexNode*  LVertexNode = EStruct_ByIndex(LVertexNodes, LVertexNodeSize, LVertexNodeIndex);
#ifdef E3dDebugDELETE_VERTICES
printf("  VertexNode %d (%d)", (int)LVertexNodeIndex, (int)LVertexNode->VertexID);fflush(stdout);
#endif
								if(LVertexNode->VertexID >= 0) LPolygon->NumOfVertices -= 1;
							}
#ifdef E3dDebugDELETE_VERTICES
printf("\n");fflush(stdout);
#endif

							ELst_StructsBackupAndRemove(&LPolygon->VertexNodes, &LPolygon->NumOfVertexNodes,  LDVPolygon->VertexNodesToModify_Indices, LDVPolygon->VertexNodesToModify_Count, &LDVPolygon->VertexNodesToModify,  LVertexNodeSize);

							E3d_PolygonCorrectHoleVertexCounts(LPolygon, LVertexNodeSize);

#ifdef E3dDebugDELETE_VERTICES
E3dPolygon_Print(LPolygon, LVertexNodeSize);printf("\n");fflush(stdout);
printf("%s() %s:%d    LPolygon %p  VertexNodes %d  Vertices %d  ExteriorVertices %d\n", __FUNCTION__, __FILE__, __LINE__, LPolygon, (int)LPolygon->NumOfVertexNodes, (int)LPolygon->NumOfVertices, (int)LPolygon->NumOfExteriorVertices);fflush(stdout);
#endif
							E3dPolygon_UpdateNormal(LPolygon, LVertexNodeSize,  LMesh->Vertices, LMesh->VertexSize);
							E3dPolygon_Triangulate(LPolygon, LVertexNodeSize,  LMesh->Vertices, LMesh->VertexSize,  TRUE);
						}

						if(LOpPolyGroup->PolygonsToDelete_Indices)
						{
							EList_StructsBackupAndRemove(LPolyGroup->Polygons, LOpPolyGroup->PolygonsToDelete,  sizeof(E3dPolygon));

//							LRemovedPolysInMesh += LOpPolyGroup->PolygonsToDelete_Count;
						}

					}
				}

//printf("%s() %s:%d DO0\n", __FUNCTION__, __FILE__, __LINE__);fflush(stdout);E3dMesh_EdgesPrint(LMesh);

				EList_StructsBackupAndRemoveA(LMesh->Edges, LOperator->EdgesToDelete,  sizeof(E3dEdge));
				EList_StructsInsertFromBackupA(LMesh->Edges, LOperator->EdgesToAdd,  sizeof(E3dEdge));

/*
{
printf("%s() %s:%d  [%s] V %lld, deleting %lld\n", __FUNCTION__, __FILE__, __LINE__, LMesh->Name, (long long)LMesh->Vertices_Count, (long long)LOperator->Vertices_Count);fflush(stdout);
	const EIndex  LValuesPerLine = 10;
	EIndex  LVC = 0;for(;LVC < LOperator->Vertices_Count;LVC++)
	{
		const EIndex  LIndex = LOperator->Vertices_Indices[LVC];
		const EIndex  LMod = LVC % LValuesPerLine;
		const EBool   LLineStart = LMod == 0;
		const EBool   LLineEnd   = LMod == LValuesPerLine - 1;

		if(LLineStart) printf("%s%6lld%s: ", ECyanCode, (long long)LVC, ENormalCode);

		if(LIndex < 0 || LIndex >= LMesh->Vertices_Count) printf(" %s%6lld%s", ERedCode, (long long)LIndex, ENormalCode);
		else printf(" %6lld", (long long)LIndex);

		if(LLineEnd) printf("\n");

		fflush(stdout);
	}
printf("\n");fflush(stdout);
}
*/

				if(LOperator->DeleteMeshVertices)
				{
					EIndex*  LOldToNewVertexTable = ELst_DelCreateRemappingTable(LMesh->Vertices_Count, LOperator->Vertices_Indices, LOperator->Vertices_Count);if(LOldToNewVertexTable)
					{
						E3dMesh_EdgesAndPolyGroupsRemapVertexIndices(LMesh, LOldToNewVertexTable);

						ELst_StructsBackupAndRemove(&LMesh->Vertices, &LMesh->Vertices_Count,  LOperator->Vertices_Indices, LOperator->Vertices_Count, &LOperator->Vertices,  LOperator->VertexSize);

						EFree(LOldToNewVertexTable);
//printf("%s() %s:%d\n", __FUNCTION__, __FILE__, __LINE__);fflush(stdout);if(E3dMesh_Check(LMesh, 0) != E_SUCCESS) assert(0);
					}
				}

				E3dMesh_PolyGroupsEdgesFromMeshEdges(LMesh, E3dp_Prefs.OptimizeFor == EOptimizeSPEED, FALSE);
				E3dObject_UpdateForDisplay((E3dObject*)LMesh, ~(E3dUF_TRIANGULATE|E3dUF_EDGES|E3duf_CALLS), LScene);

				LChanged |= E3dChgMESHSHAPE;

				if(LChanged)
				{
					EOpNode*  LOpNode = EList_StructAddAClr(POps->OpNodes, 4);if(LOpNode)
					{
						LOpNode->Object  = (ECoreObject*)LMesh;
						LOpNode->Changed = E3dChgMESHSHAPE;
					}
				}
			}
		break;

		case EOpUNDO:
			if(LOperator->Vertices_Count || LOperator->EdgesToDelete_Count || LOperator->EdgesToAdd_Count)
			{
				E3dMesh_FreeCaches(LMesh);

				if(LOperator->DeleteMeshVertices)
				{
					EIndex*  LOldToNewVertexTable = ELst_AddCreateRemappingTable(LMesh->Vertices_Count, LOperator->Vertices_Indices, LOperator->Vertices_Count);if(LOldToNewVertexTable)
					{
						EList_StructsInsertFromBackup(LMesh->Vertices, LOperator->Vertices,  LOperator->VertexSize);
						E3dVertices_Free((E3dVertex*)LOperator->Vertices, LOperator->VertexSize, LOperator->Vertices_Count, &LOperator->VertexAttributes);LOperator->Vertices = NULL;

						E3dMesh_EdgesAndPolyGroupsRemapVertexIndices(LMesh, LOldToNewVertexTable);
						EFree(LOldToNewVertexTable);
					}
				}

				if(LOperator->OpPolyGroups)
				{
					EIndex  LPGC = 0;const EIndex  LPGN = LOperator->OpPolyGroups_Count;E3dOpPolyGroup*  LOpPolyGroup = LOperator->OpPolyGroups;for(;LPGC < LPGN;LPGC++, LOpPolyGroup++)
					{
						E3dPolyGroup*            LPolyGroup = LOpPolyGroup->PolyGroup;
						const E3dVertexNodeSize  LVertexNodeSize = LPolyGroup->VertexNodeSize;

						if(LOpPolyGroup->PolygonsToDelete_Indices)
						{
							EList_StructsInsertFromBackup(LPolyGroup->Polygons, LOpPolyGroup->PolygonsToDelete,  sizeof(E3dPolygon));
							EFreeClr(LOpPolyGroup->PolygonsToDelete);  // Free struct backup copies to make sure that _EditObjectsFree won't free Polygons' VertexNodes (we only need PolygonsToDelete_Indices to perform the DO action again)

//							LAddedPolysInMesh += LOpPolyGroup->PolygonsToDelete_Count;
						}

						EIndex  LPC = 0;const EIndex  LPN = LOpPolyGroup->PolygonsToModify_Count;E3dOpPolygon*  LDVPolygon = LOpPolyGroup->PolygonsToModify;for(;LPC < LPN;LPC++, LDVPolygon++)
						{
							E3dPolygon*  LPolygon = LPolyGroup->Polygons + LDVPolygon->PolygonIndex;

//printf("%s() %s:%d  %sUNDO%s Poly %d", __FUNCTION__, __FILE__, __LINE__, EGreenCode, ENormalCode, (int)(LPolygon - LPolyGroup->Polygons));fflush(stdout);

							ELst_StructsInsertFromBackup(&LPolygon->VertexNodes, &LPolygon->NumOfVertexNodes, LDVPolygon->VertexNodesToModify_Indices, LDVPolygon->VertexNodesToModify_Count, &LDVPolygon->VertexNodesToModify,  LVertexNodeSize);
							EFreeClr(LDVPolygon->VertexNodesToModify);  // Free struct backup copies to make sure that _DeleteVerticesFree won't free Polygons' VertexNodes

							const E3dVertexNode*  LVertexNodes = LPolygon->VertexNodes;

							const EIndex  LOrigNumOfExteriorVertices = LDVPolygon->OrigNumOfExteriorVertices;
							EIndex  LC = 0;const EIndex  LN = LDVPolygon->VertexNodesToModify_Count;for(;LC < LN;LC++)
							{
								const EIndex  LVertexNodeIndex = LDVPolygon->VertexNodesToModify_Indices[LC];
//printf("  VertexNode %d", (int)LVertexNodeIndex);fflush(stdout);
								if(LVertexNodeIndex < LOrigNumOfExteriorVertices) LPolygon->NumOfExteriorVertices += 1;

								const E3dVertexNode*  LVertexNode = EStruct_ByIndex(LVertexNodes, LVertexNodeSize, LVertexNodeIndex);
								if(LVertexNode->VertexID >= 0) LPolygon->NumOfVertices += 1;
							}
//printf("  VN %d  Vtx %d  ExtV %d\n", (int)LPolygon->NumOfVertexNodes, (int)LPolygon->NumOfVertices, (int)LPolygon->NumOfExteriorVertices);fflush(stdout);

							E3d_PolygonCorrectHoleVertexCounts(LPolygon, LVertexNodeSize);

//E3dPolygon_Print(LPolygon, LVertexNodeSize);printf("\n");fflush(stdout);

							E3dPolygon_UpdateNormal(LPolygon, LVertexNodeSize,  LMesh->Vertices, LMesh->VertexSize);
							E3dPolygon_Triangulate(LPolygon, LVertexNodeSize,  LMesh->Vertices, LMesh->VertexSize,  TRUE);

//E3dPolygon_Print(LPolygon, LVertexNodeSize);printf("\n");fflush(stdout);
						}
					}
				}


				EList_StructsRemoveA(          LMesh->Edges, LOperator->EdgesToAdd);
				EList_StructsInsertFromBackupA(LMesh->Edges, LOperator->EdgesToDelete,  sizeof(E3dEdge));

				E3dMesh_PolyGroupsEdgesFromMeshEdges(LMesh, E3dp_Prefs.OptimizeFor == EOptimizeSPEED, FALSE);
				E3dObject_UpdateForDisplay((E3dObject*)LMesh, ~(E3dUF_TRIANGULATE|E3dUF_EDGES|E3duf_CALLS), LScene);

				LChanged |= E3dChgMESHSHAPE;

				if(LChanged)
				{
					EOpNode*  LOpNode = EList_StructAddAClr(POps->OpNodes, 4);if(LOpNode)
					{
						LOpNode->Object  = (ECoreObject*)LMesh;
						LOpNode->Changed = E3dChgMESHSHAPE;
					}
				}
			}
		break;
	}

	return(LChanged);
}


//========================================
// Initialize EditMesh Operator
//========================================
static void _EditMesh_Init(EOperator* POperator)
{
	E3dOperatorEditMesh*  LOperator = (E3dOperatorEditMesh*)POperator;

	LOperator->MeshNewT = NULL;

	EList_Init(LOperator->OpPolyGroups);

	EList_Init(LOperator->EdgesToDelete);   LOperator->EdgesToDelete_Indices = NULL;
	EList_Init(LOperator->EdgesToAdd);      LOperator->EdgesToAdd_Indices = NULL;

	E3dVertexAttribute_ListInit(&LOperator->VertexAttributes);
	LOperator->VertexSize = 0;

	EList_Init(LOperator->VerticesToDelete);LOperator->VerticesToDelete_Indices = NULL;
	EList_Init(LOperator->VerticesToAdd);   LOperator->VerticesToAdd_Indices = NULL;
}


//========================================
// Free ObjectsEdit Operator
//========================================
static int _EditMesh_Free(EOperator* POperator)
{
	E3dOperatorEditMesh*  LOperator = (E3dOperatorEditMesh*)POperator;

//printf("%s() %s:%d  %p:%s\n", __FUNCTION__, __FILE__, __LINE__, LOperator, LOperator->Name);fflush(stdout);
	E_DeleteChk(LOperator->MeshNewT);

	if(LOperator->OpPolyGroups)
	{
		EIndex  LSOC = 0;const EIndex  LSON = LOperator->OpPolyGroups_Count;for(;LSOC < LSON;LSOC++)
		{
			_OpPolyGroupFreeContents(LOperator->OpPolyGroups + LSOC);
		}
		EFreeClr(LOperator->OpPolyGroups);
	}

	EFreeChk(LOperator->EdgesToDelete_Indices);EList_FreeChk(LOperator->EdgesToDelete);
	EFreeChk(LOperator->EdgesToAdd_Indices);   EList_FreeChk(LOperator->EdgesToAdd);

	EFreeChk(LOperator->VerticesToDelete_Indices);
	E3dVertices_Free((E3dVertex*)LOperator->VerticesToDelete, LOperator->VertexSize, LOperator->VerticesToDelete_Count, &LOperator->VertexAttributes);LOperator->VerticesToDelete = NULL;

	EFreeChk(LOperator->VerticesToAdd_Indices);
	E3dVertices_Free((E3dVertex*)LOperator->VerticesToAdd, LOperator->VertexSize, LOperator->VerticesToAdd_Count, &LOperator->VertexAttributes);LOperator->VerticesToAdd = NULL;

	E3dVertexAttribute_ListFree(&LOperator->VertexAttributes);
	LOperator->VertexSize = 0;

	return(E_SUCCESS);
}


//================================================
// EditObjects Operator proc
//================================================
static EChangeFlags _EditMesh_Proc(EOps* POps, EOperator* POperator, const int POperation)
{
	E3dScene*             LScene = E3d_Scene;
	E3dOperatorEditMesh*  LOperator = (E3dOperatorEditMesh*)POperator;

	switch(POperation)
	{
		case EOpDO:
			{
				EChangeFlags  LChanged = 0;
				E3dMesh*      LMesh = LOperator->MeshNewT;

//ETime  LPrev_ms = E_GetMilliseconds();

//printf("%s() %s:%d <%s>\n", __FUNCTION__, __FILE__, __LINE__, LMesh->Name);fflush(stdout);

#ifdef E3dDebugEDGES
printf("%s() %s:%d  <%s>  %sDO%s  LOperator %p LOperator->EdgesToDelete_Indices %p\n", __FUNCTION__, __FILE__, __LINE__,  LMesh->Name,  EGreenCode, ENormalCode,  LOperator, LOperator->EdgesToDelete_Indices);fflush(stdout);
#endif

				E3dMesh_FreeCaches(LMesh);

				EIndex*  LOldToNewVertexTable = ELst_AddCreateRemappingTable(LMesh->Vertices_Count, LOperator->VerticesToAdd_Indices, LOperator->VerticesToAdd_Count);if(LOldToNewVertexTable)
				{
					EList_StructsInsertFromBackup(LMesh->Vertices, LOperator->VerticesToAdd,  LOperator->VertexSize);
					E3dVertices_Free((E3dVertex*)LOperator->VerticesToAdd, LOperator->VertexSize, LOperator->VerticesToAdd_Count, &LOperator->VertexAttributes);LOperator->VerticesToAdd = NULL;

					E3dMesh_EdgesAndPolyGroupsRemapVertexIndices(LMesh, LOldToNewVertexTable);
					EFree(LOldToNewVertexTable);
				}

				EList_StructsBackupAndRemoveA(LMesh->Edges, LOperator->EdgesToDelete,  sizeof(E3dEdge));
				EList_StructsInsertFromBackupA(LMesh->Edges, LOperator->EdgesToAdd,  sizeof(E3dEdge));

				{
					EIndex  LPolyGroupsRemoved = 0, LPolyGroupsAdded = 0, LRemovedPolysInMesh = 0, LAddedPolysInMesh = 0;

					EIndex  LSOC = 0;const EIndex  LSON = LOperator->OpPolyGroups_Count;E3dOpPolyGroup*  LOpPolyGroup = LOperator->OpPolyGroups;for(;LSOC < LSON;LSOC++, LOpPolyGroup++)
					{
						E3dPolyGroup*  LPolyGroup = LOpPolyGroup->PolyGroup;if(LPolyGroup)
						{
							switch(LOpPolyGroup->Operation)
							{
								case EEditMODIFY:
									{
										const E3dVertexNodeSize  LVertexNodeSize = LPolyGroup->VertexNodeSize;

// Polygons to modify (add/delete VertexNodes etc.)
//
										{
											EIndex  LPC = 0;const EIndex  LPN = LOpPolyGroup->PolygonsToModify_Count;E3dOpPolygon*  LDVPolygon = LOpPolyGroup->PolygonsToModify;for(;LPC < LPN;LPC++, LDVPolygon++)
											{
												E3dPolygon*  LPolygon = LPolyGroup->Polygons + LDVPolygon->PolygonIndex;

//printf("%s() %s:%d  %sDO%s  %sPoly %3d%s ", __FUNCTION__, __FILE__, __LINE__, EGreenCode, ENormalCode, EGreenCode, (int)(LPolygon - LPolyGroup->Polygons), ENormalCode);fflush(stdout);E3dPolygon_Print(LPolygon, LVertexNodeSize);printf("\n");fflush(stdout);

												ELst_StructsInsertFromBackup(&LPolygon->VertexNodes, &LPolygon->NumOfVertexNodes, LDVPolygon->VertexNodesToModify_Indices, LDVPolygon->VertexNodesToModify_Count, &LDVPolygon->VertexNodesToModify,  LVertexNodeSize);

//printf("%s() %s:%d  %sFREEING %p%s\n", __FUNCTION__, __FILE__, __LINE__, EYellowCode, LDVPolygon->VertexNodesToModify, ENormalCode);fflush(stdout);
												EFreeClr(LDVPolygon->VertexNodesToModify);  // Free struct backup copies to make sure that _DeleteVerticesFree won't free Polygons' VertexNodes

												const E3dVertexNode*  LVertexNodes = LPolygon->VertexNodes;

												const EIndex  LOrigNumOfExteriorVertices = LDVPolygon->OrigNumOfExteriorVertices;
												EIndex  LC = 0;const EIndex  LN = LDVPolygon->VertexNodesToModify_Count;for(;LC < LN;LC++)
												{
													const EIndex  LVertexNodeIndex = LDVPolygon->VertexNodesToModify_Indices[LC];
//printf("  Adding VertexNode %d", (int)LVertexNodeIndex);fflush(stdout);

													const E3dVertexNode*  LVertexNode = EStruct_ByIndex(LVertexNodes, LVertexNodeSize, LVertexNodeIndex);const EIndex  LVertexID = LVertexNode->VertexID;
													if(LVertexNodeIndex <= LOrigNumOfExteriorVertices && LVertexID >= 0) LPolygon->NumOfExteriorVertices += 1;
													if(LVertexID >= 0) LPolygon->NumOfVertices += 1;
												}
//printf("\n");fflush(stdout);

												E3d_PolygonCorrectHoleVertexCounts(LPolygon, LVertexNodeSize);

//printf("%s() %s:%d               ", __FUNCTION__, __FILE__, __LINE__);fflush(stdout);E3dPolygon_Print(LPolygon, LVertexNodeSize);printf("\n");fflush(stdout);

												E3dPolygon_UpdateNormal(LPolygon, LVertexNodeSize,  LMesh->Vertices, LMesh->VertexSize);
												E3dPolygon_Triangulate(LPolygon, LVertexNodeSize,  LMesh->Vertices, LMesh->VertexSize,  TRUE);

//printf("%s() %s:%d", __FUNCTION__, __FILE__, __LINE__);fflush(stdout);E3dPolygon_Print(LPolygon, LVertexNodeSize);printf("\n");fflush(stdout);
											}
										}

// Polygons to delete
//
										if(LOpPolyGroup->PolygonsToDelete_Indices)
										{
											EList_StructsBackupAndRemove(LPolyGroup->Polygons, LOpPolyGroup->PolygonsToDelete,  sizeof(E3dPolygon));

											LRemovedPolysInMesh += LOpPolyGroup->PolygonsToDelete_Count;
										}

// Polygons to add
//
										if(LOpPolyGroup->PolygonsToAdd)
										{
											const E3dPolygon*  LSrcPolygon = LOpPolyGroup->PolygonsToAdd;
											const EIndex       LPolygonsToAdd_Count = LOpPolyGroup->PolygonsToAdd_Count;

											LAddedPolysInMesh += LPolygonsToAdd_Count;

											if(LOpPolyGroup->PolygonsToAdd_Indices)
											{
												const EIndex*  LPolygonsToAdd_Indices = LOpPolyGroup->PolygonsToAdd_Indices;
												EList_StructsInsertFromBackup(LPolyGroup->Polygons, LOpPolyGroup->PolygonsToAdd,  sizeof(E3dPolygon));EFreeClr(LOpPolyGroup->PolygonsToAdd);  // Free struct backup copies to make sure that _EditObjectsFree won't free Polygons' VertexNodes

// We must duplicate the Polygon's VertexNodes and ITriangles because other tools may reallocate them, which would leave dangling pointers in LOpPolyGroup->PolygonsToAdd
//
												{
													EIndex  LPC = 0;for(;LPC < LPolygonsToAdd_Count;LPC++, LSrcPolygon++)
													{
														E3dPolygon*  LDstPolygon = LPolyGroup->Polygons + LPolygonsToAdd_Indices[LPC];

														{
															const EIndex  LSize = LSrcPolygon->NumOfVertexNodes * LVertexNodeSize;
															EMallocT(E3dVertexNode*, LSize, LDstPolygon->VertexNodes);memcpy(LDstPolygon->VertexNodes, LSrcPolygon->VertexNodes, LSize);
														}

														if(LSrcPolygon->ITriangles_Count)
														{
															const EIndex  LSize = sizeof(E3dITriangle) * LSrcPolygon->ITriangles_Count;
															EMallocT(E3dITriangle*, LSize, LDstPolygon->ITriangles);memcpy(LDstPolygon->ITriangles, LSrcPolygon->ITriangles, LSize);
														}
													}
												}
											}
											else
											{
// If PolygonsToAddIndices = NULL, add Polygons to the end
//
												E3dPolygon*  LTmp;EResizeNN(E3dPolygon, LPolyGroup->Polygons_Count + LPolygonsToAdd_Count, LPolyGroup->Polygons, LTmp);if(LTmp)
												{
													memcpy(LPolyGroup->Polygons + LPolyGroup->Polygons_Count, LOpPolyGroup->PolygonsToAdd, sizeof(E3dPolygon) * LPolygonsToAdd_Count);

// We must duplicate the Polygon's VertexNodes and ITriangles because other tools may reallocate them, which would leave dangling pointers in LOpPolyGroup->PolygonsToAdd
//
													{
														EIndex  LPC = 0;E3dPolygon*  LDstPolygon = LPolyGroup->Polygons + LPolyGroup->Polygons_Count;for(;LPC < LPolygonsToAdd_Count;LPC++, LSrcPolygon++, LDstPolygon++)
														{
															{
																const EIndex  LSize = LSrcPolygon->NumOfVertexNodes * LVertexNodeSize;
																EMallocT(E3dVertexNode*, LSize, LDstPolygon->VertexNodes);memcpy(LDstPolygon->VertexNodes, LSrcPolygon->VertexNodes, LSize);
															}

															if(LSrcPolygon->ITriangles_Count)
															{
																const EIndex  LSize = (EIndex)sizeof(E3dITriangle) * LSrcPolygon->ITriangles_Count;
																EMallocT(E3dITriangle*, LSize, LDstPolygon->ITriangles);memcpy(LDstPolygon->ITriangles, LSrcPolygon->ITriangles, LSize);
															}
														}
													}

													LPolyGroup->Polygons_Count += LPolygonsToAdd_Count;
												}
											}
										}


#ifdef E3dRESTORE_POLYGROUP_EDGES

// If we're deleting Mesh Edges, back up and remove the corresponding PolyGroup Mesh Edge indices
//
										if(LOperator->EdgesToDelete && LMesh->PolyGroups_Count > 1)
										{
											const EIndex  LEDN = LOperator->EdgesToDelete_Count;const EIndex*  LED = LOperator->EdgesToDelete_Indices;

// Update PolyGroup Edge indices
//
											if(LOpPolyGroup->EdgeIndicesToDelete_Indices == NULL)
											{
#ifdef E3dDebugEDGES
printf("%s() %s:%d  LSOC %2d  LOpPolyGroup->EdgeIndicesToDelete_Indices empty\n", __FUNCTION__, __FILE__, __LINE__, (int)LSOC);fflush(stdout);
#endif
// Build list of indices in LPolyGroup->Edges to the indices of Mesh Edges to be deleted
//
												EIndex*  LEdgeIndicesToDelete;ENewN(LEDN, LEdgeIndicesToDelete);
												const EIndex*  LPolyGroupEdges = LPolyGroup->Edges;const EIndex  LPGEN = LPolyGroup->Edges_Count;
												EIndex  LEdgeIndicesToDelete_Count = 0;
												EIndex  LEDC = 0;for(;LEDC < LEDN;LEDC++)
												{
													const EIndex  LMeshEdgeIndex = LED[LEDC];
													EIndex  LPGEC = 0;for(;LPGEC < LPGEN;LPGEC++)
													{
														if(LMeshEdgeIndex == LPolyGroupEdges[LPGEC])
														{
															LEdgeIndicesToDelete[LEdgeIndicesToDelete_Count++] = LPGEC;
														}
													}
												}
												if(LEdgeIndicesToDelete_Count < LEDN) EResizeNT(EIndex, LEdgeIndicesToDelete_Count, LEdgeIndicesToDelete);

												LOpPolyGroup->EdgeIndicesToDelete_Indices = LEdgeIndicesToDelete;
												LOpPolyGroup->EdgeIndicesToDelete_Count   = LEdgeIndicesToDelete_Count;
											}

printf("%s() %s:%d  %sGREEN%s\n", __FUNCTION__, __FILE__, __LINE__, EGreenCode, ENormalCode);fflush(stdout);
#ifdef E3dDebugEDGES
{
	_PolyGroupPrintEdges(LPolyGroup);

	const EIndex  LTN = LOpPolyGroup->EdgeIndicesToDelete_Count;
	printf("%s() %s:%d   LEDN %d  LEdgeIndicesToDelete_Count %d\n", __FUNCTION__, __FILE__, __LINE__, (int)LEDN, (int)LTN);fflush(stdout);
	{
		EIndex  LTC = 0;for(;LTC < LTN;LTC++)
		{
			printf(" %d", (int)LOpPolyGroup->EdgeIndicesToDelete_Indices[LTC]);fflush(stdout);
		}
	}
printf("\n");fflush(stdout);
}
#endif

											EList_IndicesBackupAndRemove(LPolyGroup->Edges, LOpPolyGroup->EdgeIndicesToDelete,  TRUE);

// Remap (compact) LPolyGroup->Edges index list after the given Edge indices were removed
//
											{
												const EIndex  LPGEN = LPolyGroup->Edges_Count;

#ifdef E3dDebugEDGES
printf("%s() %s:%d  PolyGroup Edges: %3d:\n", __FUNCTION__, __FILE__, __LINE__, (int)LPGEN);fflush(stdout);
#endif
												EIndex*  LPolyGroupEdgeIndices = LPolyGroup->Edges;

// LOperator->EdgesToDelete_Indices is in ascending order, but we need to do this in descending order...
//
												EIndex  LEDC = LEDN - 1;for(;LEDC >= 0;LEDC--)
												{
													const EIndex  LEI = LED[LEDC];
//printf("%s() %s:%d  PolyGroup removing Mesh Edge %3d:\n", __FUNCTION__, __FILE__, __LINE__, (int)LEI);fflush(stdout);
													EIndex  LPGEC = 0;for(;LPGEC < LPGEN;LPGEC++)
													{
														const EIndex  LPI = LPolyGroupEdgeIndices[LPGEC];if(LPI > LEI)
														{
															LPolyGroupEdgeIndices[LPGEC] -= 1;
//printf(" %d %s%d%s|", (int)LPGEC, EYellowCode, (int)LPolyGroupEdgeIndices[LPGEC], ENormalCode);fflush(stdout);
														}
														else if(LPI == LEI)
														{
//printf(" %s%d %d%s|", ERedCode, (int)LPGEC, (int)LPolyGroupEdgeIndices[LPGEC], ENormalCode);fflush(stdout);
														}
														else
														{
//printf(" %d %d|", (int)LPGEC, (int)LPolyGroupEdgeIndices[LPGEC]);fflush(stdout);
														}
													}
//printf("\n");fflush(stdout);
												}
											}
										}
#endif  // E3dRESTORE_POLYGROUP_EDGES
									}
								break;

								case EEditADD:
									{
										E3dPolyGroup*  LNewPolyGroup = E3dPolyGroup_Copy(LPolyGroup, NULL,  NULL,  ECopyNAME|E3dCopyPOLYGONS);

printf("%s() %s:%d  %lld %lld\n", __FUNCTION__, __FILE__, __LINE__, (long long)LPolyGroup->Polygons_Count, (long long)LNewPolyGroup->Polygons_Count);fflush(stdout);
										if(E3dMesh_AppendPolyGroup(LMesh, LNewPolyGroup) == E_SUCCESS)
										{
printf("%s() %s:%d\n", __FUNCTION__, __FILE__, __LINE__);fflush(stdout);
											E3dM_ObjectUpdateDrawingMaterial((E3dObject*)LMesh, LNewPolyGroup->Material, LNewPolyGroup,DrawingMaterial);
											LPolyGroupsAdded++;
										}
									}
								break;

								case EEditDEL:
									if(E3dMesh_PolyGroupDel(LMesh, LPolyGroup) == E_SUCCESS) LPolyGroupsRemoved++;
								break;
							}
						}
					}


					if(LPolyGroupsRemoved || LPolyGroupsAdded || LAddedPolysInMesh || LRemovedPolysInMesh || LOperator->EdgesToAdd || LOperator->EdgesToDelete_Indices)
					{
//printf("%s() %s:%d  LPolyGroupsRemoved %lld  LAddedPolysInMesh %lld  LRemovedPolysInMesh %lld  EdgesToAdd %p EdgesToDelete_Indices %p\n", __FUNCTION__, __FILE__, __LINE__, (long long)LPolyGroupsRemoved, (long long)LAddedPolysInMesh, (long long)LRemovedPolysInMesh, LOperator->EdgesToAdd, LOperator->EdgesToDelete_Indices);fflush(stdout);
						EUpdateFlags  LUpdateFlags = 0;
						if(LPolyGroupsRemoved || LPolyGroupsAdded || LAddedPolysInMesh || LRemovedPolysInMesh) LUpdateFlags = ~(E3dUF_TRIANGULATE|E3duf_CALLS);
						else LUpdateFlags = ~(E3dUF_REMOVE_STRIPS|E3duf_NORMALS|E3duf_CALLS);

// If the caller specified Edges to add and/or delete, it means it took care of tracking Edges, so we don't need to build them from the Polygons
// (Check EdgesTo*_Indices (vs. LOperator->Edges*) because LOperator->EdgesTo* may get freed).
//
//										if(LOperator->EdgesToAdd || LOperator->EdgesToDelete_Indices)
						{
							LUpdateFlags &= ~E3dUF_EDGES;
#ifndef E3dRESTORE_POLYGROUP_EDGES
							E3dMesh_PolyGroupsEdgesFromMeshEdges(LMesh, E3dp_Prefs.OptimizeFor == EOptimizeSPEED, FALSE);
#endif
						}

						E3dObject_UpdateForDisplay((E3dObject*)LMesh, LUpdateFlags, LScene);
						LChanged |= E3dChgMESHSHAPE;
					}
				}

//{ const ETime  LNow_ms = E_GetMilliseconds();char  LTStr[256];EStr_MillisecondsToString(LNow_ms - LPrev_ms, LTStr, FALSE);LPrev_ms = LNow_ms;printf("%s() %s:%d  %s\n", __FUNCTION__, __FILE__, __LINE__, LTStr);fflush(stdout); }

				if(LChanged && POps)
				{
					EOpNode*  LOpNode = EList_StructAddAClr(POps->OpNodes, 4);if(LOpNode)
					{
						LOpNode->Object  = (ECoreObject*)LMesh;
						LOpNode->Changed = E3dChgMESHSHAPE;
					}
				}

				return(LChanged);
			}
		break;

		case EOpUNDO:
			{
				EChangeFlags  LChanged = 0;

				E3dMesh*  LMesh = LOperator->MeshNewT;

				E3dMesh_FreeCaches(LMesh);

				EList_StructsRemoveA(          LMesh->Edges, LOperator->EdgesToAdd);
				EList_StructsInsertFromBackupA(LMesh->Edges, LOperator->EdgesToDelete,  sizeof(E3dEdge));

				EIndex  LPolyGroupsAdded = 0, LPolyGroupsRemoved = 0, LAddedPolysInMesh = 0, LRemovedPolysInMesh = 0;

				EIndex  LSOC = 0;const EIndex  LSON = LOperator->OpPolyGroups_Count;E3dOpPolyGroup*  LOpPolyGroup = LOperator->OpPolyGroups;for(;LSOC < LSON;LSOC++, LOpPolyGroup++)
				{
					E3dPolyGroup*  LPolyGroup = LOpPolyGroup->PolyGroup;if(LPolyGroup)
					{
						switch(LOpPolyGroup->Operation)
						{
							case EEditMODIFY:
								{
									const E3dVertexNodeSize  LVertexNodeSize = LPolyGroup->VertexNodeSize;

// Polygons to add
//
									if(LOpPolyGroup->PolygonsToAdd)
									{
										const EIndex  LPolygonsToAdd_Count = LOpPolyGroup->PolygonsToAdd_Count;

										if(LPolyGroup->Polygons_Count > LPolygonsToAdd_Count)  // Sanity check
										{
											LRemovedPolysInMesh += LPolygonsToAdd_Count;

											if(LOpPolyGroup->PolygonsToAdd_Indices)
											{
												const EIndex*  LPolygonsToAdd_Indices = LOpPolyGroup->PolygonsToAdd_Indices;

// Free Polygons (we have a backup in the Operator)
//
												{
													EIndex  LPC = 0;for(;LPC < LPolygonsToAdd_Count;LPC++)
													{
														E3dPolygon*  LPolygon = LPolyGroup->Polygons + LPolygonsToAdd_Indices[LPC];
														E3dPolygon_FreeContents(LPolygon);
													}
												}

// Just remove, no backup (we have a backup in the Operator)
//
												EList_StructsRemove(LPolyGroup->Polygons, LPolygonsToAdd);
											}
											else
											{
// If PolygonsToAddIndices = NULL, Polygons were added to the end
//
												const EIndex  LNewCount = LPolyGroup->Polygons_Count - LPolygonsToAdd_Count;

// Free Polygons (we have a backup in the Operator)
//
												{
													EIndex  LPC = 0;E3dPolygon*  LPolygon = LPolyGroup->Polygons + LNewCount;for(;LPC < LPolygonsToAdd_Count;LPC++, LPolygon++)
													{
														E3dPolygon_FreeContents(LPolygon);
													}
												}

												E3dPolygon*  LTmp;EResizeNN(E3dPolygon, LNewCount, LPolyGroup->Polygons, LTmp);
												if(LTmp) LPolyGroup->Polygons_Count = LNewCount;
											}
										}
									}


// Polygons to delete
//
									if(LOpPolyGroup->PolygonsToDelete_Indices)
									{
										EList_StructsInsertFromBackup(LPolyGroup->Polygons, LOpPolyGroup->PolygonsToDelete,  sizeof(E3dPolygon));
										EFreeClr(LOpPolyGroup->PolygonsToDelete);  // Free struct backup copies to make sure that _EditObjectsFree won't free Polygons' VertexNodes (we only need PolygonsToDelete_Indices to perform the DO action again)

										LAddedPolysInMesh += LOpPolyGroup->PolygonsToDelete_Count;
									}


// Polygons to modify (add/delete VertexNodes etc.)
//
									{
										EIndex  LPC = 0;const EIndex  LPN = LOpPolyGroup->PolygonsToModify_Count;E3dOpPolygon*  LDVPolygon = LOpPolyGroup->PolygonsToModify;for(;LPC < LPN;LPC++, LDVPolygon++)
										{
											E3dPolygon*           LPolygon = LPolyGroup->Polygons + LDVPolygon->PolygonIndex;
											const E3dVertexNode*  LVertexNodes = LPolygon->VertexNodes;
											const EIndex*         LVertexNodesToDelete_Indices = LDVPolygon->VertexNodesToModify_Indices;

//printf("%s() %s:%d  %sUNDO%s Poly %d  ", __FUNCTION__, __FILE__, __LINE__, EGreenCode, ENormalCode, (int)(LPolygon - LPolyGroup->Polygons));fflush(stdout);

//E3dPolygon_Print(LPolygon, LVertexNodeSize);printf("\n");fflush(stdout);

											const EIndex  LOrigNumOfExteriorVertices = LPolygon->NumOfExteriorVertices;
											EIndex  LC = 0;const EIndex  LN = LDVPolygon->VertexNodesToModify_Count;for(;LC < LN;LC++)
											{
												const EIndex  LVertexNodeIndex = LVertexNodesToDelete_Indices[LC];
//printf("  VertexNode %d", (int)LVertexNodeIndex);fflush(stdout);

												const E3dVertexNode*  LVertexNode = EStruct_ByIndex(LVertexNodes, LVertexNodeSize, LVertexNodeIndex);const EIndex  LVertexID = LVertexNode->VertexID;
												if(LVertexNodeIndex <= LOrigNumOfExteriorVertices && LVertexID >= 0) LPolygon->NumOfExteriorVertices -= 1;
												if(LVertexID >= 0) LPolygon->NumOfVertices -= 1;
											}
//printf("\n");fflush(stdout);


											ELst_StructsBackupAndRemove(&LPolygon->VertexNodes, &LPolygon->NumOfVertexNodes,  LDVPolygon->VertexNodesToModify_Indices, LDVPolygon->VertexNodesToModify_Count, &LDVPolygon->VertexNodesToModify,  LVertexNodeSize);

											E3d_PolygonCorrectHoleVertexCounts(LPolygon, LVertexNodeSize);

//E3dPolygon_Print(LPolygon, LVertexNodeSize);printf("\n");fflush(stdout);


//printf("   VertexNodes %d  Vertices %d  ExteriorVertices %d\n", (int)LPolygon->NumOfVertexNodes, (int)LPolygon->NumOfVertices, (int)LPolygon->NumOfExteriorVertices);fflush(stdout);

											E3dPolygon_UpdateNormal(LPolygon, LVertexNodeSize,  LMesh->Vertices, LMesh->VertexSize);
											E3dPolygon_Triangulate(LPolygon, LVertexNodeSize,  LMesh->Vertices, LMesh->VertexSize,  TRUE);
										}
									}

#ifdef E3dDebugEDGES
			printf("%s() %s:%d  %sUNDO%s  PolyGroup Edges: %3d:\n", __FUNCTION__, __FILE__, __LINE__, ECyanCode, ENormalCode, (int)LPolyGroup->Edges_Count);fflush(stdout);
#endif

// If we're restoring Mesh Edges, restore the corresponding PolyGroup Mesh Edge indices
//
#ifdef E3dRESTORE_POLYGROUP_EDGES
									if(LOperator->EdgesToDelete_Indices && LMesh->PolyGroups_Count > 1)
									{
										const EIndex  LEDN = LOperator->EdgesToDelete_Count;const EIndex*  LED = LOperator->EdgesToDelete_Indices;

// Update PolyGroup Edge indices
//
										if(LOpPolyGroup->EdgeIndicesToDelete_Indices)
										{
// Remap (expand) LPolyGroup->Edges index list after the given Edge indices were restored
//
											{
												const EIndex  LPGEN = LPolyGroup->Edges_Count;
												EIndex*  LPolyGroupEdgeIndices = LPolyGroup->Edges;

// LOperator->EdgesToDelete_Indices is in ascending order, which is what we need here
//
												EIndex  LEDC = 0;for(;LEDC < LEDN;LEDC++)
												{
													const EIndex  LEI = LED[LEDC];
//printf("%s() %s:%d  PolyGroup re-adding Mesh Edge %3d:\n", __FUNCTION__, __FILE__, __LINE__, (int)LEI);fflush(stdout);
													EIndex  LPGEC = 0;for(;LPGEC < LPGEN;LPGEC++)
													{
														const EIndex  LPI = LPolyGroupEdgeIndices[LPGEC];if(LPI >= LEI)
														{
															LPolyGroupEdgeIndices[LPGEC] += 1;
//printf(" %d %s%d%s|", (int)LPGEC, EYellowCode, (int)LPolyGroupEdgeIndices[LPGEC], ENormalCode);fflush(stdout);
														}
														else if(LPI == LEI)
														{
//printf(" %s%d %d%s|", ERedCode, (int)LPGEC, (int)LPolyGroupEdgeIndices[LPGEC], ENormalCode);fflush(stdout);
														}
														else
														{
//printf(" %d %d|", (int)LPGEC, (int)LPolyGroupEdgeIndices[LPGEC]);fflush(stdout);
														}
													}
//printf("\n");fflush(stdout);
												}
											}

											EList_IndicesInsertFromBackup(LPolyGroup->Edges, LOpPolyGroup->EdgeIndicesToDelete);
										}
									}
#endif  // E3dRESTORE_POLYGROUP_EDGES
								}
							break;

							case EEditADD:  // Undo adding this PolyGroup
								if(LOpPolyGroup->OriginalIndex >= 0 && LOpPolyGroup->OriginalIndex < LMesh->PolyGroups_Count)
								{
									E3dPolyGroup*  LNewPolyGroup = LMesh->PolyGroups[LOpPolyGroup->OriginalIndex];
									if(E3dMesh_PolyGroupDel(LMesh, LNewPolyGroup) == E_SUCCESS) LPolyGroupsRemoved++;
								}
							break;

							case EEditDEL:  // Undo deleting this PolyGroup
								{
									if(EList_ObjectInsert(LMesh->PolyGroups, LOpPolyGroup->OriginalIndex, LPolyGroup))
									{
										E3dScene_MaterialAdd(LScene, LPolyGroup->Material);
										if(LPolyGroup->Material != LPolyGroup->DrawingMaterial) E3dScene_MaterialAdd(LScene, LPolyGroup->DrawingMaterial);
										LPolyGroupsAdded++;
									}
								}
							break;
						}
					}
				}



				{
					EIndex*  LOldToNewVertexTable = ELst_DelCreateRemappingTable(LMesh->Vertices_Count, LOperator->VerticesToAdd_Indices, LOperator->VerticesToAdd_Count);if(LOldToNewVertexTable)
					{
						ELst_StructsBackupAndRemove(&LMesh->Vertices, &LMesh->Vertices_Count,  LOperator->VerticesToAdd_Indices, LOperator->VerticesToAdd_Count, &LOperator->VerticesToAdd,  LOperator->VertexSize);

//printf("%s() %s:%d %sUNDO1%s\n", __FUNCTION__, __FILE__, __LINE__, EMagentaCode, ENormalCode);fflush(stdout);E3dMesh_EdgesPrint(LMesh);
						E3dMesh_EdgesAndPolyGroupsRemapVertexIndices(LMesh, LOldToNewVertexTable);
//printf("%s() %s:%d %sUNDO2%s\n", __FUNCTION__, __FILE__, __LINE__, EMagentaCode, ENormalCode);fflush(stdout);E3dMesh_EdgesPrint(LMesh);
						EFree(LOldToNewVertexTable);
					}
				}



				if(LPolyGroupsAdded || LPolyGroupsRemoved || LAddedPolysInMesh || LRemovedPolysInMesh || LOperator->EdgesToAdd || LOperator->EdgesToDelete_Indices)
				{
//printf("%s() %s:%d  LPolyGroupsAdded %lld  LAddedPolysInMesh %lld  LRemovedPolysInMesh %lld  EdgesToAdd %p EdgesToDelete_Indices %p\n", __FUNCTION__, __FILE__, __LINE__, (long long)LPolyGroupsAdded, (long long)LAddedPolysInMesh, (long long)LRemovedPolysInMesh, LOperator->EdgesToAdd, LOperator->EdgesToDelete_Indices);fflush(stdout);
					EUpdateFlags  LUpdateFlags = 0;
					if(LPolyGroupsAdded || LPolyGroupsRemoved || LAddedPolysInMesh || LRemovedPolysInMesh) LUpdateFlags = ~(E3dUF_TRIANGULATE|E3duf_CALLS);
					else LUpdateFlags = ~(E3dUF_REMOVE_STRIPS|E3duf_NORMALS|E3duf_CALLS);

//printf("%s() %s:%d  <%s>  LOperator %p LOperator->EdgesToDelete_Indices %p\n", __FUNCTION__, __FILE__, __LINE__, LMesh->Name, LOperator, LOperator->EdgesToDelete_Indices);fflush(stdout);

// If the caller specified Edges to add and/or delete, it means it took care of tracking Edges, so we don't need to build them from the Polygons
// (Check EdgesTo*_Indices (vs. LOperator->Edges*) because LOperator->EdgesTo* may get freed).
//
//					if(LOperator->EdgesToAdd || LOperator->EdgesToDelete_Indices)
					{
						LUpdateFlags &= ~E3dUF_EDGES;
#ifndef E3dRESTORE_POLYGROUP_EDGES
//_PolyGroupPrintEdges(LPolyGroup);
						E3dMesh_PolyGroupsEdgesFromMeshEdges(LMesh, E3dp_Prefs.OptimizeFor == EOptimizeSPEED, FALSE);
//_PolyGroupPrintEdges(LPolyGroup);
#endif
					}

					E3dObject_UpdateForDisplay((E3dObject*)LMesh, LUpdateFlags, LScene);
					LChanged |= E3dChgMESHSHAPE;
				}

				if(LChanged && POps)
				{
					EOpNode*  LOpNode = EList_StructAddAClr(POps->OpNodes, 4);if(LOpNode)
					{
						LOpNode->Object  = (ECoreObject*)LMesh;
						LOpNode->Changed = E3dChgMESHSHAPE;
					}
				}

				return(LChanged);
			}
		break;
	}
	return(0);
}


//================================================================================================
// Correct selections:
// Make sure that Models' Selection is E3dNodeSelOBJECT if any of their Objects are selected
//================================================================================================
static void _SceneUpdateSelections(E3dScene* PScene)
{
	EIndex  LRC = 0;const EIndex  LRN = PScene->RootNodes_Count;for(;LRC < LRN;LRC++)
	{
		E3dNode*  LNode = PScene->RootNodes[LRC];for(;LNode;LNode = LNode->Next)
		{
			E3dModel*  LModel = E_DynamicCast(LNode, E3dModel);if(LModel && LNode->Selection == 0)
			{
				EIndex  LOC = 0;const EIndex  LON = LModel->Objects_Count;for(;LOC < LON;LOC++)
				{
					E3dObject*  LObject = LModel->Objects[LOC];if(LObject->Selection)
					{
						LNode->Selection = E3dNodeSelOBJECT;
						break;
					}
				}
			}
		}
	}
}


//========================================================================================
// Check if a Node has selected Geometries, PolyGroups, Polygons, Edges or Vertices
//========================================================================================
static EBool _NodeCheckForSelectedComponents(const E3dNode* PNode,  const uint8_t PDeleteComponentsFlags, E3dObject*** PObjectsWithSelectedComponentsP, EIndex* PObjectsWithSelectedComponents_CountP, EIndex* PObjectsWithSelectedComponents_CountAllocatedP)
{
	EIndex  LHasObjectsWithSelectedComponent = 0;

	E3dModel*  LModel = E_DynamicCast(PNode, E3dModel);if(LModel)
	{
		const EBool        LNodeSelected = E3dNode_IsSelected(PNode);

		if(LNodeSelected || PNode->Selection == E3dNodeSelOBJECT)
		{
			E3dObject* const*  LObjects = LModel->Objects;

			EIndex  LObjC = 0;const EIndex  LObjN = LModel->Objects_Count;for(;LObjC < LObjN;LObjC++)
			{
				E3dObject*  LObject = LObjects[LObjC];

				if(LNodeSelected || LObject->Selection)
				{

//printf("%s() %s:%d <%s>\n", __FUNCTION__, __FILE__, __LINE__, LObject->Name);fflush(stdout);


// Check if LObject is already in PObjectsWithSelectedComponentsP.
// If so, don't bother checking its Face contours, Polygons etc.
//
					if(ELst_PointerGetIndex(*PObjectsWithSelectedComponentsP, *PObjectsWithSelectedComponents_CountP, LObject) < 0)
					{
						if(EObject_IsSubClass(LObject, E3dSpline))
						{
							const E3dSpline*  LSpline = (const E3dSpline*)LObject;

							if(PDeleteComponentsFlags & E3dDeleteSELECTED_POINTS)
							{
								const EIndex  LSplineSelectedCVs_Count = E3dSpline_NumOfSelectedCVs(LSpline);if(LSplineSelectedCVs_Count > 0)
								{
printf("%s() %s:%d <%s> SelectedCVs_Count %lld\n", __FUNCTION__, __FILE__, __LINE__, LSpline->Name, (long long)LSplineSelectedCVs_Count);fflush(stdout);
									EIndex  LI;(void)LI;ELst_ObjectAddAChk(PObjectsWithSelectedComponentsP, PObjectsWithSelectedComponents_CountP, PObjectsWithSelectedComponents_CountAllocatedP,  4, LObject,  LI);
									LHasObjectsWithSelectedComponent++;
								}
							}
						}
						else if(EObject_IsSubClass(LObject, E3dFace))
						{
							const E3dFace*  LFace = (const E3dFace*)LObject;

							if(LObject->Selection == E3dObjectSelFACE_CONTOUR)
							{
								if(LFace->Exterior->Selection == E3dObjectSelOBJECT)
								{
									if(PDeleteComponentsFlags & E3dDeleteSELECTED_POINTS && E3dSpline_NumOfSelectedCVs(LFace->Exterior) > 0)
									{
										EIndex  LI;(void)LI;ELst_ObjectAddAChk(PObjectsWithSelectedComponentsP, PObjectsWithSelectedComponents_CountP, PObjectsWithSelectedComponents_CountAllocatedP,  4, LObject,  LI);
										LHasObjectsWithSelectedComponent++;
									}
//								else E3dp_PrintMessage(ELogWARNING, 5000, Lng_CanNotDeleteFaceExterior);
								}

								EIndex  LHC = 0;for(;LHC < LFace->Holes_Count;LHC++)
								{
									const E3dSpline*  LHole = LFace->Holes[LHC];

									if(LHole->Selection == E3dObjectSelOBJECT)
									{
										EIndex  LI;(void)LI;ELst_ObjectAddAChk(PObjectsWithSelectedComponentsP, PObjectsWithSelectedComponents_CountP, PObjectsWithSelectedComponents_CountAllocatedP,  4, LObject,  LI);

										LHasObjectsWithSelectedComponent++;
									}
								}
							}
							else if(LObject->Selection && PDeleteComponentsFlags & E3dDeleteSELECTED_POINTS)
							{
								if(E3dSpline_NumOfSelectedCVs(LFace->Exterior) > 0)
								{
									EIndex  LI;(void)LI;ELst_ObjectAddAChk(PObjectsWithSelectedComponentsP, PObjectsWithSelectedComponents_CountP, PObjectsWithSelectedComponents_CountAllocatedP,  4, LObject,  LI);
								}

								EIndex  LHC = 0;for(;LHC < LFace->Holes_Count;LHC++)
								{
									const E3dSpline*  LHole = LFace->Holes[LHC];

									if(E3dSpline_NumOfSelectedCVs(LHole) > 0)
									{
										EIndex  LI;(void)LI;ELst_ObjectAddAChk(PObjectsWithSelectedComponentsP, PObjectsWithSelectedComponents_CountP, PObjectsWithSelectedComponents_CountAllocatedP,  4, LObject,  LI);
										LHasObjectsWithSelectedComponent++;
									}
								}
							}
						}
						else if(EObject_IsSubClass(LObject, E3dMesh))
						{
							EBool  LMeshAdded = FALSE;

//printf("%s() %s:%d  <%s%s%s>\n", __FUNCTION__, __FILE__, __LINE__, EGreenCode, LObject->Name, ENormalCode);fflush(stdout);
							const E3dMesh*  LMesh = (E3dMesh*)LObject;if(LMesh->PolyGroups_Count)
							{
								if(PDeleteComponentsFlags & (E3dDeleteSELECTED_POLYGONS|E3dDeleteSELECTED_POLYGROUPS))
								{
//printf("%s() %s:%d  <%s%s%s>\n", __FUNCTION__, __FILE__, __LINE__, EGreenCode, LObject->Name, ENormalCode);fflush(stdout);
									EIndex  LPGC = 0;const EIndex  LPGN = LMesh->PolyGroups_Count;E3dPolyGroup* const*  LPolyGroups = LMesh->PolyGroups;for(;LPGC < LPGN;LPGC++)
									{
										const E3dPolyGroup*  LPolyGroup = LPolyGroups[LPGC];

//#define E3dDEBUG  1

										const EIndex  LNumOfSelectedPolygons = PDeleteComponentsFlags & E3dDeleteSELECTED_POLYGONS ? E3dPolyGroup_NumOfSelectedPolygons(LPolyGroup) : 0;

#ifdef E3dDEBUG
//printf("%s() %s:%d  %s/%s PolyGroup %d LPolyGroup->Selected %d SelectedPolys %d  LPolyGroup->Flags %d\n", __FUNCTION__, __FILE__, __LINE__, LMesh->Name, LPolyGroup->Name, LPGC, LPolyGroup->Selected, E3dPolyGroup_NumOfSelectedPolygons(LPolyGroup), LPolyGroup->Flags);fflush(stdout);
#endif

										if(LPolyGroup->Selected || LNumOfSelectedPolygons > 0)
										{
											EIndex  LI;(void)LI;ELst_ObjectAddAChk(PObjectsWithSelectedComponentsP, PObjectsWithSelectedComponents_CountP, PObjectsWithSelectedComponents_CountAllocatedP, 4, LObject,  LI);
											LMeshAdded = TRUE;

#ifdef E3dDEBUG
	printf("%s() %s:%d  %s/%s  selected Polys %d\n", __FUNCTION__, __FILE__, __LINE__, LMesh->Name, LPolyGroup->Name, (int)LNumOfSelectedPolygons);fflush(stdout);
#endif

//printf("%s() %s:%d Polygons\n", __FUNCTION__, __FILE__, __LINE__);fflush(stdout);

											break;
										}
									}
								}

// No selected Polygons, check for selected Edges
//
								if(PDeleteComponentsFlags & E3dDeleteSELECTED_EDGES)
								{
//printf("%s() %s:%d EDGES\n", __FUNCTION__, __FILE__, __LINE__);fflush(stdout);
									if(!LMeshAdded)
									{
										const EIndex  LNumOfSelectedEdges = E3dMesh_NumOfSelectedEdges(LMesh);if(LNumOfSelectedEdges > 0)
										{
//printf("%s() %s:%d Edges\n", __FUNCTION__, __FILE__, __LINE__);fflush(stdout);
											EIndex  LI;(void)LI;ELst_ObjectAddAChk(PObjectsWithSelectedComponentsP, PObjectsWithSelectedComponents_CountP, PObjectsWithSelectedComponents_CountAllocatedP, 4, LObject,  LI);

											LMeshAdded = TRUE;
										}
									}
								}

// No selected Edges, check for selected Vertices
//
								if(!LMeshAdded && PDeleteComponentsFlags & (E3dDeleteSELECTED_POINTS|E3dDisconnectSELECTED_POINTS))
								{
printf("%s() %s:%d VERTS\n", __FUNCTION__, __FILE__, __LINE__);fflush(stdout);
									if(!LMeshAdded)
									{
										const EIndex  LNSelectedVertices = E3dMesh_NumOfSelectedVertices(LMesh);if(LNSelectedVertices > 0)
										{
printf("%s() %s:%d Vertices\n", __FUNCTION__, __FILE__, __LINE__);fflush(stdout);
											EIndex  LI;(void)LI;ELst_ObjectAddAChk(PObjectsWithSelectedComponentsP, PObjectsWithSelectedComponents_CountP, PObjectsWithSelectedComponents_CountAllocatedP, 4, LObject,  LI);

											LMeshAdded = TRUE;
										}
									}
								}
							}

							if(LMeshAdded) LHasObjectsWithSelectedComponent++;
						}

// Object is selected, but its components are not, return TRUE, but don't add it to PObjectsWithSelectedComponentsP,
// because the Object will be deleted and removed from its Models
//
//					if(LObject->Selection == E3dObjectSelOBJECT) LHasObjectsWithSelectedComponent++;
					}
				}
			}
		}
	}

	return(LHasObjectsWithSelectedComponent);
}


//========================================================================================
// Add selected Objects (Geometries, Lights etc.) from a Model to a Delete operator if
// they're not in the list PObjectsWithSelectedComponentsP.
//========================================================================================
static EChangeFlags _NodeDeleteSelectedObjects(E3dOperatorEditObjects** POperatorP, E3dNode* PNode,  E3dObject*** PObjectsWithSelectedComponentsP, EIndex* PObjectsWithSelectedComponents_CountP, EIndex* PObjectsWithSelectedComponents_CountAllocatedP)
{
	EChangeFlags  LChanged = 0;

printf("%s() %s:%d  [%s]\n", __FUNCTION__, __FILE__, __LINE__, PNode->Name);fflush(stdout);
	E3dModel*     LModel = E_DynamicCast(PNode, E3dModel);if(LModel)
	{
		E3dObject**  LObjects = LModel->Objects;
		EBool        LModelSelected;

		switch(LModel->Selection)
		{
			case E3dNodeSelNODE:
			case E3dNodeSelBRANCH:
			case E3dNodeSelBRANCH_ROOT:
				LModelSelected = TRUE;
			break;

			default:
				LModelSelected = FALSE;
			break;
		}

(void)LModelSelected;

		EIndex  LObjC = 0;for(;LObjC < LModel->Objects_Count;LObjC++)
		{
			E3dObject*  LObject = LObjects[LObjC];
			EBool       LObjSelected = FALSE;

			switch(LObject->Selection)
			{
				case E3dObjectSelOBJECT:  LObjSelected = TRUE;break;
			}

			if(EObject_IsSubClass(LObject, E3dCamera))
			{
				E3dCamera*  LCamera = (E3dCamera*)LObject;
				EIndex  LPC = 0;for(;LPC < E3dp_Panels_Count;LPC++)
				{
					E3dPanel*  L3DPanel = E3dp_Panels[LPC];
					EIndex  LWC = 0;const EIndex  LWN = L3DPanel->Windows_Count;for(;LWC < LWN;LWC++)
					{
						E3dpWindow*  L3DWindow = E_DynamicCast(L3DPanel->Windows[LWC], E3dpWindow);if(L3DWindow)
						{
							EIndex  LVC = 0;for(;LVC < L3DWindow->Views_Count;LVC++)
							{
								E3dView*  LView = L3DWindow->Views[LVC];if(LCamera == LView->VCamera)
								{
									LObjSelected = FALSE;
									E3dp_PrintMessage(ELogWARNING, 5000, E3dLng_Can_not_delete_active_Camera);
								}
							}
						}
					}
				}
			}

			if(LObjSelected)
			{
				if(ELst_PointerGetIndex(*PObjectsWithSelectedComponentsP, *PObjectsWithSelectedComponents_CountP, LObject) < 0)
				{
					if(*POperatorP == NULL) *POperatorP = (E3dOperatorEditObjects*)EOp_OperatorNew(E3dOperatorEditObjects_class);

					E3dOpObject*      LOpObject = E3dEditObjects_AddOpObject(*POperatorP, LObject);
					EObjectAndIndex*  LModelRec;ENew(LModelRec);

					EList_PointerAdd(LOpObject->Models, LModelRec);

					E_AssignNoChkNoFree((EObject*)LModel, LModelRec->Object);
					LModelRec->Index = LObjC;

					LChanged |= E3dObject_GetChangeFlagsByClass(LObject) | E3dChgHIERARCHY;
				}
			}
		}
	}

	return(LChanged);
}


static EIndex _Spline_GetSelectedCVIndices(const E3dSpline* PSpline, EIndex** PRet)
{
	EIndex*       LIndices = NULL;
	const EIndex  LCVN = PSpline->CVs_Count;
	EIndex        LN = 0;

	{
		EIndex  LCVC = 0;const E3dSplineCV*  LCV = PSpline->CVs;for(;LCVC < LCVN;LCVC++, LCV++)
		{
			if(LCV->CVFlags & E3dCV_TAGGED) LN++;
		}
	}

	if(PRet)
	{
		if(LN)
		{
			ENewN(LN, LIndices);

			EIndex  LCVC = 0, LN1 = 0;const E3dSplineCV*  LCV = PSpline->CVs;for(;LCVC < LCVN;LCVC++, LCV++)
			{
				if(LCV->CVFlags & E3dCV_TAGGED)
				{
					LIndices[LN1++] = LCVC;
				}
			}
		}

		*PRet = LIndices;
	}

	return(LN);
}


//================================================================================================
// Collect indices of selected Edges in a Mesh
//
// Argument
//  PMesh      Pointer to the Mesh
//  PIndicesP  Pointer to store the resulting array
//
// Description
//  This function returns the number of selected Polygons in the given PolyGroup.
//
// Return value
//  The number of selected Edges.
//================================================================================================
static EIndex _MeshGetSelectedEdgeIndices(E3dMesh* PMesh, EIndex** PIndicesP, const EBool PKeepWingedEdgesAndVertexLinks)
{
	EListAI(, EIndex, LIndices);

// Count selected Polygons
//
	const EIndex  LEN = PMesh->Edges_Count;
	EIndex  LSelectedEdges_Count = 0;const E3dEdge*  LSEdge = NULL;

	{
		EIndex  LEC = 0;const E3dEdge*  LEdge = PMesh->Edges;for(;LEC < LEN;LEC++, LEdge++)
		{
			if(LEdge->Flags & E3dEdgeSELECTED)
			{
				LSelectedEdges_Count++;
				LSEdge = LEdge;
			}
		}
	}

	EBool  LWeCreatedWingedEdges = FALSE, LWeCreatedVertexLinks = FALSE;

	EIndex  LBoundaryEdgesC = 0, LPolyGroupSeparatorEdgesC = 0;

	if(LSelectedEdges_Count > 1)
	{
// If there are more than one selected Edges, build WingedEdges
//
		if(PMesh->WingedEdges == NULL) { E3dMesh_CreateWingedEdgesFromEdges(PMesh, TRUE, FALSE);LWeCreatedWingedEdges = TRUE;LWeCreatedVertexLinks = TRUE; }
		else if(PMesh->VertexLinks == NULL) { E3dMesh_VertexLinksBuild(PMesh);LWeCreatedVertexLinks = TRUE; }

		EIndex  LEC = 0;const E3dEdge*  LEdge = PMesh->Edges;const E3dWingedEdge*  LWingedEdge = PMesh->WingedEdges;for(;LEC < LEN;LEC++, LEdge++, LWingedEdge++)
		{
			if(LEdge->Flags & E3dEdgeSELECTED)
			{
				if(LWingedEdge->Polygons_Count == 2)
				{
					const E3dEdgePolygonInfo*  LEdgeDescriptors = LWingedEdge->Polygons;

					if(LEdgeDescriptors[0].PolyGroupIndex == LEdgeDescriptors[1].PolyGroupIndex)
					{
						EList_IndexAddA(LIndices, 64, LEC);
					}
					else LPolyGroupSeparatorEdgesC++;  // PolyGroup separating Edge
				}
				else LBoundaryEdgesC++;  // Boundary Edge
			}
		}

		if(LBoundaryEdgesC) E3dp_PrintMessage(ELogWARNING, 5000, E3dLng_Can_not_delete_boundary_Edge_Try_deleting_Vertices);
		else if(LPolyGroupSeparatorEdgesC) E3dp_PrintMessage(ELogWARNING, 5000, E3dLng_Can_not_delete_Edge_that_separates_2_PolyGroups);
	}
	else if(LSelectedEdges_Count == 1)
	{
// If there's only one selected Edge, get info about it via E3dMesh_EdgeGetPolygons().
//
		unsigned int  LDescription;E3dEdgePolygonInfo*  LEdgeDescriptors;const EIndex  LNumOfPolygonsOnEdge = E3dMesh_EdgeGetPolygons(PMesh, LSEdge, &LEdgeDescriptors, &LDescription);

		if(LNumOfPolygonsOnEdge != 2)
		{
			EFreeChk(LEdgeDescriptors);
			E3dp_PrintMessage(ELogWARNING, 5000, E3dLng_Can_not_delete_boundary_Edge_Try_deleting_Vertices);
			return(E3dEdgeNOT_SHARED_BY_2_POLYS);
		}

		if(LEdgeDescriptors[0].PolyGroupIndex != LEdgeDescriptors[1].PolyGroupIndex)
		{
			EFreeChk(LEdgeDescriptors);
			E3dp_PrintMessage(ELogWARNING, 5000, E3dLng_Can_not_delete_Edge_that_separates_2_PolyGroups);
			return(E3dEdgePOLYS_IN_DIFFERENT_POLYGROUPS);
		}

		EList_IndexAdd(LIndices, LSEdge - PMesh->Edges);
	}


// We must free WingedEdges and VertexLinks of PSrcMesh to avoid data corruption with stale data
//
	if(!PKeepWingedEdgesAndVertexLinks)
	{
		if(LWeCreatedWingedEdges) E3dMesh_WingedEdgesFree(PMesh);
		if(LWeCreatedVertexLinks) E3dMesh_VertexLinksFree(PMesh);
	}

	*PIndicesP = LIndices;

	if(LBoundaryEdgesC) return(E3dEdgeNOT_SHARED_BY_2_POLYS);
	else if(LPolyGroupSeparatorEdgesC) return(E3dEdgePOLYS_IN_DIFFERENT_POLYGROUPS);
	else return(LIndices_Count);
}


//========================================================================================
// Add Objects and their selected components to a Delete operator.
//========================================================================================
static EChangeFlags _ObjectsDeleteSelectedComponents(EOps* POps, E3dOperatorEditObjects** POperatorEditObjectsP,  const uint8_t PDeleteComponentsFlags,  E3dObject** PObjects, const EIndex PObjects_Count)
{
	EChangeFlags  LChanged = 0;

	EIndex  LObjC = 0;for(;LObjC < PObjects_Count;LObjC++)
	{
		E3dObject*  LObject = PObjects[LObjC];

		if(EObject_IsSubClass(LObject, E3dSpline))
		{
			const E3dSpline*  LSpline = (const E3dSpline*)LObject;

			if(PDeleteComponentsFlags & E3dDeleteSELECTED_POINTS)
			{
				EIndex*  LSelectedCVIndices = NULL;

				const EIndex  LSelectedCVIndices_Count = _Spline_GetSelectedCVIndices(LSpline, &LSelectedCVIndices);if(LSelectedCVIndices_Count > 0)
				{
					E3dOperatorAddDelSplineCVs*  LOperator = (E3dOperatorAddDelSplineCVs*)EOp_OperatorNew(E3d_OperatorDeleteSplineCVs_class);

					E_AssignToMember(LSpline, LOperator,Object);
					E_AssignToMember(LSpline, LOperator,Spline);
					LOperator->CVs_Indices = LSelectedCVIndices;LOperator->CVs = NULL;LOperator->CVs_Count = LSelectedCVIndices_Count;

					EList_ObjectAdd(POps->Operators, LOperator);
				}
			}
		}
		else if(EObject_IsSubClass(LObject, E3dFace))
		{
			E3dFace*  LFace = (E3dFace*)LObject;

			if(LObject->Selection == E3dObjectSelFACE_CONTOUR)
			{
				if(LFace->Exterior->Selection == E3dObjectSelOBJECT)
				{
					E3dp_PrintMessage(ELogWARNING, 5000, E3dLng_Can_not_delete_Face_exterior);
				}

				EIndex  LHC = 0;for(;LHC < LFace->Holes_Count;LHC++)
				{
					E3dSpline*  LHole = LFace->Holes[LHC];

					if(LHole->Selection == E3dObjectSelOBJECT)
					{
						if(*POperatorEditObjectsP == NULL) *POperatorEditObjectsP = (E3dOperatorEditObjects*)EOp_OperatorNew(E3dOperatorEditObjects_class);
						E3dOpObject*     LOpObject    = E3dEditObjects_AddOpObject(*POperatorEditObjectsP, LObject);
						E3dOpSubObject*  LOpSubObject = EList_StructAdd(LOpObject->SubObjects);if(LOpSubObject)
						{
printf("%s() %s:%d <%s> %d\n", __FUNCTION__, __FILE__, __LINE__, LHole->Name, LHole->RefCnt);fflush(stdout);
							E3dEditObjects_OpSubObjectInit(LOpSubObject, LHC, (EObject*)LHole);
printf("%s() %s:%d <%s> %d\n", __FUNCTION__, __FILE__, __LINE__, LHole->Name, LHole->RefCnt);fflush(stdout);
						}
					}
				}
			}
		}
		else if(EObject_IsSubClass(LObject, E3dMesh))
		{
			E3dMesh*  LMesh = (E3dMesh*)LObject;
			EBool     LHasSelectedPolygons = FALSE, LHasSelectedPolyGroups = FALSE;

			if(PDeleteComponentsFlags & (E3dDeleteSELECTED_POLYGROUPS|E3dDeleteSELECTED_POLYGONS))
			{
				EIndex  LPGC = 0;const EIndex  LPGN = LMesh->PolyGroups_Count;E3dPolyGroup* const*  LPolyGroups = LMesh->PolyGroups;for(;LPGC < LPGN;LPGC++)
				{
					const E3dPolyGroup*  LPolyGroup = LPolyGroups[LPGC];
					if(PDeleteComponentsFlags & E3dDeleteSELECTED_POLYGROUPS && LPolyGroup->Selected) LHasSelectedPolyGroups = TRUE;

					if(PDeleteComponentsFlags & E3dDeleteSELECTED_POLYGONS)
					{
						const EIndex  LNumOfSelectedPolygons = E3dPolyGroup_NumOfSelectedPolygons(LPolyGroup);if(LNumOfSelectedPolygons > 0)
						{
							LHasSelectedPolygons = TRUE;
							break;
						}
					}
				}
			}

			E3dOperatorAddDeleteVertices*  LOperatorDeleteVertices = NULL;

			EIndex*  LSelectedEdgeIndices = NULL;EIndex  LSelectedEdgeIndices_Count;
			if(LHasSelectedPolygons)
			{
				LSelectedEdgeIndices_Count = 0;
			}
			else
			{
// No selected Polygons, check for selected PolyGroups
//
				if(LHasSelectedPolyGroups) LSelectedEdgeIndices_Count = 0;
				else
				{
					if(PDeleteComponentsFlags & E3dDeleteSELECTED_EDGES) LSelectedEdgeIndices_Count = _MeshGetSelectedEdgeIndices(LMesh, &LSelectedEdgeIndices, TRUE);
					else LSelectedEdgeIndices_Count = 0;

					if(LSelectedEdgeIndices_Count < 1)
					{
// No Edges to delete, check for Vertices
//
						EBool   LWeCreatedVertexLinks = FALSE;
						EIndex  LVerticesToDelete_Count = 0;

						if(PDeleteComponentsFlags & (E3dDeleteSELECTED_POINTS|E3dDisconnectSELECTED_POINTS))
						{
							const EIndex  LNSelectedVertices = E3dMesh_NumOfSelectedVertices(LMesh);if(LNSelectedVertices)
							{
								if(LMesh->VertexLinks == NULL) { E3dMesh_VertexLinksBuild(LMesh);LWeCreatedVertexLinks = TRUE; }

								if(LMesh->VertexLinks)
								{
									const EIndex  LVN = LMesh->Vertices_Count;const E3dVertexSize  LMeshVertexSize = LMesh->VertexSize;

									{
										EIndex  LVC = 0;const E3dVertex*  LVertex = LMesh->Vertices;const E3dVertexLink*  LVertexLink = LMesh->VertexLinks;for(;LVC < LVN;LVC++, EStruct_Inc(LVertex, LMeshVertexSize), LVertexLink++)
										{
											if(LVertex->Flags & E3dVtxSELECTED)
											{
												if(LVertexLink->EdgeEnds_Count == 0 || LVertexLink->EdgeEnds_Count == 2) LVerticesToDelete_Count++;
											}
										}
									}


									EIndex*  LVerticesToDelete_Indices;ENewN(LVerticesToDelete_Count, LVerticesToDelete_Indices);
									EIndex   LVCC = 0;

									{
										EIndex  LVC = 0;const E3dVertex*  LVertex = LMesh->Vertices;const E3dVertexLink*  LVertexLink = LMesh->VertexLinks;for(;LVC < LVN;LVC++, EStruct_Inc(LVertex, LMeshVertexSize), LVertexLink++)
										{
											if(LVertex->Flags & E3dVtxSELECTED)
											{
												if(LVertexLink->EdgeEnds_Count == 0 || LVertexLink->EdgeEnds_Count == 2) LVerticesToDelete_Indices[LVCC++] = LVC;
											}
										}
									}

									if(LWeCreatedVertexLinks) E3dMesh_VertexLinksFree(LMesh);

//printf("%s() %s:%d  LVerticesToDelete_Count %d\n", __FUNCTION__, __FILE__, __LINE__, (int)LVerticesToDelete_Count);fflush(stdout);

									if(LVerticesToDelete_Count)
									{
										E3dOperatorAddDeleteVertices*  LOperator = (E3dOperatorAddDeleteVertices*)EOp_OperatorNew(E3dOperatorDeleteVertices_class);
										LOperatorDeleteVertices = LOperator;

										LOperator->Vertices_Indices = LVerticesToDelete_Indices;
										LOperator->Vertices_Count   = LVerticesToDelete_Count;

										E_AssignToMember(LMesh, LOperator,MeshNewT);
										EResourceList_Copy(&LMesh->VertexAttributes, &LOperator->VertexAttributes,  0);LOperator->VertexSize = LMesh->VertexSize;

										E3d_DeleteVerticesAddAffectedPolygons(LOperator);
										E3d_DeleteVerticesAddAffectedEdges(LOperator);

										if((PDeleteComponentsFlags & E3dDeleteSELECTED_POINTS) == 0) LOperator->DeleteMeshVertices = FALSE;
										EList_ObjectAdd(POps->Operators, LOperator);
									}
								}
								else
								{
printf("%s() %s:%d <%s> %sNO VertexLinks%s\n", __FUNCTION__, __FILE__, __LINE__, LMesh->Name, ERedCode, ENormalCode);fflush(stdout);
								}
							}
						}

					}
				}
			}

			if(LSelectedEdgeIndices_Count < 0)
			{
printf("%s() %s:%d  Deleting Edges %sIMPLEMENT ME%s\n", __FUNCTION__, __FILE__, __LINE__, EYellowCode, ENormalCode);fflush(stdout);
			}
			else
			{
//printf("%s() %s:%d  %d\n", __FUNCTION__, __FILE__, __LINE__, (int)LSelectedEdgeIndices_Count);fflush(stdout);

				E3dOperatorEditMesh*  LOperatorEditMesh = (E3dOperatorEditMesh*)EOp_OperatorNew(E3dOperatorEditMesh_class);

				EBool  LDeletingPolygons = FALSE;
				const int  LResult = E3dMesh_DeleteComponents(LMesh,  LSelectedEdgeIndices, LSelectedEdgeIndices_Count,  TRUE,  LOperatorEditMesh,  NULL,  &LDeletingPolygons);switch(LResult)
				{
					case E_SUCCESS:
						EList_ObjectAdd(POps->Operators, LOperatorEditMesh);
						if(LDeletingPolygons)
						{
							if(E3dp_Prefs.DeleteSelection_DeleteUnusedVertices)
							{
								const E3dVertexUsage*  LVertexUsageList = LMesh->VertexUsageList;
								const EBool            LHadVertexUsageList = LVertexUsageList != NULL ? TRUE : FALSE;

								if(LVertexUsageList == NULL) LVertexUsageList = E3dMesh_GetVertexUsageList(LMesh, FALSE, FALSE);
								if(LVertexUsageList)
								{
									const EIndex  LVN = LMesh->Vertices_Count;
									EIndex*  LVertexRefCounts;ENewN(LVN, LVertexRefCounts);
									{
										EIndex  LVC = 0;const E3dVertexUsage*  LVertexUsage = LVertexUsageList;for(;LVC < LVN;LVC++, LVertexUsage++)
										{
											EIndex  LVertexRefCount = LVertexUsage->VertexNodes_Count;
											EIndex  LVVC = 0;const EIndex  LVVN = LVertexRefCount;for(;LVVC < LVVN;LVVC++)
											{
												const E3dPolygon*  LPolygon = LVertexUsage->Polygons[LVVC];
												if(LPolygon->Flags & E3dPolygonSELECTED) LVertexRefCount--;
											}
											if(LVertexRefCount < 0) LVertexRefCount = 0;
											LVertexRefCounts[LVC] = LVertexRefCount;
										}
									}

									if(!LHadVertexUsageList) E3dMesh_VertexUsageListFree(LMesh);

									EListI(, EIndex, LVerticesToDeleteIndices);

									{
										EIndex  LVC = 0;for(;LVC < LVN;LVC++)
										{
											if(LVertexRefCounts[LVC] == 0) LVerticesToDeleteIndices_Count++;
										}
									}

//									E3d_MeshRemoveUnusedVertices(POps, LMesh);

									if(LVerticesToDeleteIndices_Count)
									{
										ENewN(LVerticesToDeleteIndices_Count, LVerticesToDeleteIndices);
										{
											EIndex  LVC = 0, LVCT = 0;for(;LVC < LVN;LVC++)
											{
												if(LVertexRefCounts[LVC] == 0) LVerticesToDeleteIndices[LVCT++] = LVC;
											}
										}

										E3dOperatorAddDeleteVertices*  LOperatorDeleteUnusedVertices = (E3dOperatorAddDeleteVertices*)EOp_OperatorNew(E3dOperatorDeleteVertices_class);if(LOperatorDeleteUnusedVertices)
										{
											E_AssignToMember(LMesh, LOperatorDeleteUnusedVertices,MeshNewT);
											EResourceList_Copy(&LMesh->VertexAttributes, &LOperatorDeleteUnusedVertices->VertexAttributes,  0);LOperatorDeleteUnusedVertices->VertexSize = LMesh->VertexSize;
											LOperatorDeleteUnusedVertices->Vertices_Indices = LVerticesToDeleteIndices;
											LOperatorDeleteUnusedVertices->Vertices_Count   = LVerticesToDeleteIndices_Count;

											EList_ObjectAdd(POps->Operators, LOperatorDeleteUnusedVertices);
										}
									}

									EFree(LVertexRefCounts);
								}
							}
						}
					break;

					case E_NOTHING_TO_DO:
						E_Delete(LOperatorEditMesh);
//						if(LOperatorDeleteVertices->Vertices_Count == 0)
						if(LOperatorDeleteVertices == NULL)
						{
							E3dp_PrintMessage(ELogWARNING, 5000, "Internal error : Mesh %s marked for component deletion, but no selected components found", LMesh->Name);
printf("%s() %s:%d  %sMesh %s marked for component deletion, but no selected components found.%s\n", __FUNCTION__, __FILE__, __LINE__, ERedCode, LMesh->Name, ENormalCode);fflush(stdout);
						}
					break;
				}
			}

			if(E3dp_Prefs.OptimizeFor == EOptimizeMEMORY)
			{
				E3dMesh_WingedEdgesFree(LMesh);
				E3dMesh_VertexLinksFree(LMesh);
			}
		}
	}

	return(LChanged);
}


//========================================
// Initialize DeleteSelection Operator
//========================================
static void _DeleteSelection_Init(EOperator* POperator)
{
	E3dOperatorDelete*  LOperator = (E3dOperatorDelete*)POperator;

	EList_Init(LOperator->Hierarchies);
	EList_Init(LOperator->Branches);
}


//========================================
// Free DeleteSelection Operator
//========================================
static int _DeleteSelection_Free(EOperator* POperator)
{
	E3dOperatorDelete*  LOperator = (E3dOperatorDelete*)POperator;

// Free Hierarchies
//
	_ObjectAndIndexListFree(&LOperator->Hierarchies, &LOperator->Hierarchies_Count);


// Free Branches
//
	if(LOperator->Branches)
	{
		EIndex  LC = 0;const EIndex  LN = LOperator->Branches_Count;for(LC = 0;LC < LN;LC++)
		{
			E3dBranch*  LEntry = LOperator->Branches[LC];
			E_DeleteChk(LEntry->Node);
			E_DeleteChk(LEntry->Parent);
			EFree(LEntry);
		}
		EList_Free(LOperator->Branches);
	}

	return(E_SUCCESS);
}


//================================================
// DeleteSelection Operator proc
//================================================
static EChangeFlags _DeleteSelection_Proc(EOps* POps, EOperator* POperator, const int POperation)
{
	E3dScene*           LScene = E3d_Scene;
	E3dOperatorDelete*  LOperator = (E3dOperatorDelete*)POperator;


	switch(POperation)
	{
		case EOpDO:
			{
				EChangeFlags  LChanged = 0;

				if(LOperator->Branches_Count)
				{
					E3dNode**  LRootNodes = NULL;EIndex  LRootNodes_Count = 0;

					{
						EIndex  LC = 0;const EIndex  LN = LOperator->Branches_Count;for(;LC < LN;LC++)
						{
							E3dBranch*  LEntry = LOperator->Branches[LC];
							E3dNode*    LNode = (E3dNode*)(LEntry->Node);

							E3dNode*  LRootNode = LNode;while(LRootNode->Parent) LRootNode = LRootNode->Parent;
							EList_PointerAddChk(LRootNodes, LRootNode);

							E3dNode_DisconnectFromParent(LNode, FALSE);
							E3dNodeHrc_CallRemoveProcs(LNode);
							E3dNode_FreeRenderData(LNode);
							LChanged |= E3dNodeHrc_GetTransformChangeFlagsByObjects(LNode) | E3dChgHIERARCHY;
						}
					}

					if(LRootNodes)
					{
						EIndex  LC = 0;for(;LC < LRootNodes_Count;LC++)
						{
							E3dNode*  LRootNode = LRootNodes[LC];

// Although NodeHrcBranchDisconnectFromParent updates link members (Prev, Next etc.),
// we still need to call NodeHrcUpdateConnections to update PopN and PushN
//
							E3dNodeHrc_UpdateConnections(LRootNode);
						}
						EFree(LRootNodes);
					}
				}


				if(LOperator->Hierarchies_Count)
				{
					{
						EIndex  LC = 0;const EIndex  LN = LOperator->Hierarchies_Count;for(;LC < LN;LC++)
						{
							E3dNode*  LRootNode = (E3dNode*)(LOperator->Hierarchies[LC]->Object);

							if(E3dScene_NodeHrcRemove(LScene, LRootNode, &LChanged, FALSE))
							{
								LChanged |= E3dChgHIERARCHY;
								LChanged |= E3dNodeHrc_GetTransformChangeFlagsByObjects(LRootNode);
							}
						}
					}

// Call Scene callbacks
//
					{
						E3dSceneNodesCallbackStruct  LCBS;
						EIndex                       LN = LOperator->Hierarchies_Count;
						E3dNode**                    LList;ENewN(LN, LList);

						EIndex  LC = 0;for(;LC < LN;LC++)
						{
							E_AssignNoChkNoFree((E3dNode*)(LOperator->Hierarchies[LC]->Object), LList[LC]);
						}

						LCBS.Reason = E3dCR_HIERARCHY_REMOVED;
						LCBS.Nodes = LList;LCBS.NumOfNodes = LN;
						E3dScene_CallCallbacks((ECoreObject*)LScene,  LScene->Callbacks, LScene->Callbacks_Count,  &LCBS);

						ELst_ObjectsFree(&LList, &LN);
					}
				}

				_SceneUpdateSelections(LScene);
				return(LChanged);
			}
		break;

		case EOpUNDO:
			{
				EChangeFlags  LChanged = 0;

				if(LOperator->Branches_Count)
				{
					E3dNode**  LRootNodes = NULL;EIndex  LRootNodes_Count = 0;

					{
						EIndex  LC = 0;const EIndex  LN = LOperator->Branches_Count;for(;LC < LN;LC++)
						{
							E3dBranch*  LEntry = LOperator->Branches[LC];
							E3dNode*    LNode = LEntry->Node;
							E3dNode*    LParent = LEntry->Parent;


//printf("%s() %s:%d Parent [%s]  Restoring Branch root [%s] at index %d\n", __FUNCTION__, __FILE__, __LINE__, LParent->Name, LNode->Name, LEntry->Index);fflush(stdout);
							E3dNode_ConnectChild(LParent, LNode, LEntry->Index, FALSE, FALSE);

							while(LParent->Parent) LParent = LParent->Parent;
							EList_PointerAddChk(LRootNodes, LParent);
						}
					}

					if(LRootNodes)
					{
						EIndex  LC = 0;for(LC = 0;LC < LRootNodes_Count;LC++)
						{
							E3dNode*  LRootNode = LRootNodes[LC];

printf("%s() %s:%d [%s]\n", __FUNCTION__, __FILE__, __LINE__, LRootNode->Name);fflush(stdout);
							E3dNodeHrc_UpdateConnections(LRootNode);
							E3dNodeHrc_UpdateMatrices(LRootNode, TRUE);
							E3dNodeHrc_UpdateMaterialInheritance(LRootNode);
							E3dNodeHrc_CallNodeInfoUpdateProcs(LRootNode, E3dChgHIERARCHY, FALSE);
						}
						EFree(LRootNodes);
					}

// Add back Materials to the Scene
//
					{
						EIndex  LC = 0;const EIndex  LN = LOperator->Branches_Count;for(;LC < LN;LC++)
						{
							E3dBranch*  LEntry = LOperator->Branches[LC];
							E3dNode*    LNode = LEntry->Node;

							E3dScene_MaterialsAddFromNodeHrc(LScene, LNode);
						}
					}



// Now that the hierarchies are up-to-date, we can call E3dNodeHrc_GetTransformChangeFlagsByObjects
//
					{
						EIndex  LC = 0;const EIndex  LN = LOperator->Branches_Count;for(;LC < LN;LC++)
						{
							E3dBranch*  LEntry = LOperator->Branches[LC];
							E3dNode*    LNode = LEntry->Node;

							LChanged |= E3dNodeHrc_GetTransformChangeFlagsByObjects(LNode) | E3dChgHIERARCHY;
						}
					}
				}


				if(LOperator->Hierarchies_Count)
				{
					{
						LChanged |= E3dChgHIERARCHY;

						EIndex  LC = 0;const EIndex  LN = LOperator->Hierarchies_Count;for(;LC < LN;LC++)
						{
							EObjectAndIndex*  LEntry = LOperator->Hierarchies[LC];
							E3dNode*          LRootNode = (E3dNode*)(LEntry->Object);
							E3dNode*          LNode;

							E3dScene_NodeHrcInsert(LScene, LRootNode, LEntry->Index, FALSE);
							for(LNode = LRootNode;LNode;LNode = LNode->Next) E3dNode_UpdateForDisplay(LNode, E3duf_FULLGEO);

							E3dScene_MaterialsAddFromNodeHrc(LScene, LRootNode);

							LChanged |= E3dNodeHrc_GetTransformChangeFlagsByObjects(LRootNode);
						}
					}

// Call Scene callbacks
//
					{
						E3dSceneNodesCallbackStruct  LCBS;
						EIndex                       LN = LOperator->Hierarchies_Count;
						E3dNode**                    LList;ENewN(LN, LList);

						EIndex  LC = 0;for(;LC < LN;LC++)
						{
							E_AssignNoChkNoFree((E3dNode*)(LOperator->Hierarchies[LC]->Object), LList[LC]);
						}

						LCBS.Reason = E3dCR_HIERARCHY_ADDED;
						LCBS.Nodes = LList;LCBS.NumOfNodes = LN;
						E3dScene_CallCallbacks((ECoreObject*)LScene,  LScene->Callbacks, LScene->Callbacks_Count,  &LCBS);

						ELst_ObjectsFree(&LList, &LN);
					}
				}

				_SceneUpdateSelections(LScene);

				return(LChanged);
			}
		break;
	}
	return(0);
}


typedef struct
{
	E3dAnimation*         Animation;
	E3dAnimationChannel*  Channel;
	EIndex                ChannelIndex;
} E3dChannelRec;


typedef struct
{
	EOperatorCore;

	E3dScene*      Scene;
	E3dAnimStack*  AnimStack;

	EListA(, EObjectAndIndex*, Animations);
	EListA(, E3dChannelRec*, Channels);
} E3dOperatorDeleteAnimations;


//========================================
// Initialize DeleteAnimations Operator
//========================================
static void _DeleteAnimationsInit(EOperator* POperator)
{
	E3dOperatorDeleteAnimations*  LOperator = (E3dOperatorDeleteAnimations*)POperator;

	LOperator->AnimStack = NULL;

	EList_InitA(LOperator->Animations);
	EList_InitA(LOperator->Channels);
}


//========================================
// Free DeleteAnimations Operator
//========================================
static int _DeleteAnimationsFree(EOperator* POperator)
{
	E3dOperatorDeleteAnimations*  LOperator = (E3dOperatorDeleteAnimations*)POperator;

	if(LOperator->Animations)
	{
		EIndex  LC = 0;const EIndex  LN = LOperator->Animations_Count;for(;LC < LN;LC++)
		{
			E_Delete(LOperator->Animations[LC]->Object);
			EFree(LOperator->Animations[LC]);
		}

		EList_FreeA(LOperator->Animations);
	}

	if(LOperator->Channels)
	{
		EIndex  LC = 0;const EIndex  LN = LOperator->Channels_Count;for(;LC < LN;LC++)
		{
			E3dChannelRec*  LEntry = LOperator->Channels[LC];

			E_Delete(LEntry->Animation);
			E_Delete(LEntry->Channel);
			EFree(LEntry);
		}

		EList_FreeA(LOperator->Channels);
	}

	E_DeleteChk(LOperator->AnimStack);

	return(E_SUCCESS);
}


//================================================
// DeleteAnimations Operator proc
//================================================
static EChangeFlags _DeleteAnimationsProc(EOps* POps, EOperator* POperator, const int POperation)
{
	E3dOperatorDeleteAnimations*  LOperator = (E3dOperatorDeleteAnimations*)POperator;
	E3dAnimStack*                 LAnimStack = LOperator->AnimStack;
	const EIndex                  LNumOfAnimations = LOperator->Animations_Count;
	const EIndex                  LNumOfChannels = LOperator->Channels_Count;


	switch(POperation)
	{
		case EOpDO:
			{
				EChangeFlags  LChanged = 0;

				if(LNumOfAnimations)
				{
					EIndex  LC = 0;for(;LC < LNumOfAnimations;LC++)
					{
						EObjectAndIndex*  LEntry = LOperator->Animations[LC];
						E3dAnimation*     LAnimation = (E3dAnimation*)(LEntry->Object);

						E3dAnimStack_AnimationRemove(LAnimStack, LAnimation, FALSE);

						LChanged |= E3dChgANIMATION;
					}
				}


				if(LNumOfChannels)
				{
					EIndex  LC = 0;for(;LC < LNumOfChannels;LC++)
					{
						E3dChannelRec*  LEntry = LOperator->Channels[LC];

						E3dAnimation_RemoveChannel(LEntry->Animation, LEntry->Channel);

						LChanged |= E3dChgANIMATION;
					}
				}
				return(LChanged);
			}
		break;

		case EOpUNDO:
			{
				EChangeFlags  LChanged = 0;

				if(LNumOfAnimations)
				{
					EIndex  LC = 0;for(;LC < LNumOfAnimations;LC++)
					{
						EObjectAndIndex*  LEntry = LOperator->Animations[LC];
						E3dAnimation*     LAnimation = (E3dAnimation*)(LEntry->Object);

						EList_ObjectInsertA(LAnimStack->Animations, 8, LEntry->Index, LAnimation);

						LChanged |= E3dChgANIMATION;
					}
				}


				if(LNumOfChannels)
				{
					EIndex  LC = 0;for(;LC < LNumOfChannels;LC++)
					{
						E3dChannelRec*        LEntry = LOperator->Channels[LC];
						E3dAnimation*         LAnimation = LEntry->Animation;
						E3dAnimationChannel*  LChannel = LEntry->Channel;

						EList_ObjectInsertA(LAnimation->Channels, LEntry->ChannelIndex, 4, LChannel);

						LChanged |= E3dChgANIMATION;
					}
				}

				return(LChanged);
			}
		break;
	}

	return(0);
}


//================================================================================================
// If the target Object of an animation is deleted, that animation must also be deleted
//================================================================================================
static EBool _AnimationIndirectlySelectedForDeletion(E3dAnimation* PAnimation)
{
	if(EObject_IsSubClass(PAnimation, E3dKeyFrameAnimation))
	{
		E3dKeyFrameAnimation*  LKeyFrameAnimation = (E3dKeyFrameAnimation*)PAnimation;
		EIndex                 LCC = 0;const EIndex  LCN = LKeyFrameAnimation->Channels_Count;for(;LCC < LCN;LCC++)
		{
			if(E3dAnimationChannel_TargetSelected(LKeyFrameAnimation->Channels[LCC])) return(TRUE);
		}
	}

	return(FALSE);
}


//========================================
// Delete selection
//========================================
static EChangeFlags _DeleteSelection(E3dScene* PScene, const uint16_t PDeleteComponentsFlags)
{
	EChangeFlags  LChanged = 0;
	EOps          LOps;EOp_OpsInit(&LOps);

// Nodes, Objects etc.
//
	{
		E3dOperatorDelete*       LOperatorNodes = (E3dOperatorDelete*)EOp_OperatorNew(_OperatorDelete_class);
		E3dOperatorEditObjects*  LOperatorEditObjects = NULL;

		EListAI(, E3dObject*, LObjectsWithSelectedComponents);  // Objects with selected components (Meshes with selected PolyGroups/Polygons etc.)

		{
			E3dNode**  LRootNodes = PScene->RootNodes;
			EIndex     LC = 0;const EIndex  LN = PScene->RootNodes_Count;for(;LC < LN;LC++)
			{
				E3dNode*  LRootNode = LRootNodes[LC];
				EBool     LHierarchyDeletable = TRUE;

				{
					E3dNode*  LNode = LRootNode;for(;LNode;LNode = LNode->Next)
					{
						if(_NodeCheckForSelectedComponents(LNode,  PDeleteComponentsFlags, &LObjectsWithSelectedComponents, &LObjectsWithSelectedComponents_Count, &LObjectsWithSelectedComponents_CountAllocated) > 0)
						{
							LHierarchyDeletable = FALSE;
						}
					}
				}

				if(LHierarchyDeletable)
				{
					switch(LRootNode->Selection)
					{
						case E3dNodeSelBRANCH_ROOT:  break;
						case E3dNodeSelNODE:         if(LRootNode->Child) LHierarchyDeletable = FALSE;break;
						default:                     LHierarchyDeletable = FALSE;break;
					}

					if(LHierarchyDeletable)
					{
						if(E3d_IsActiveCameraNode(LRootNode)) { LHierarchyDeletable = FALSE;E3dp_PrintMessage(ELogWARNING, 5000, E3dLng_Can_not_delete_active_Camera); }
						else if(E3d_IsActiveCameraInterestNode(LRootNode)) { LHierarchyDeletable = FALSE;E3dp_PrintMessage(ELogWARNING, 5000, E3dLng_Cant_delete_interest_point_of_Camera); }
					}
				}

				if(LHierarchyDeletable)
				{
					EObjectAndIndex*  LEntry;ENew(LEntry);
					EList_PointerAdd(LOperatorNodes->Hierarchies, LEntry);

// Store this Hierarchy in the Delete Operator
//
					E_AssignNoChkNoFree((EObject*)LRootNode, LEntry->Object);
					LEntry->Index = LC;
					LChanged |= E3dNodeHrc_GetTransformChangeFlagsByObjects(LRootNode) | E3dChgHIERARCHY;
					E3dNodeHrc_CallRemoveProcs(LRootNode);
				}
			}
		}


// Now check the remaining hierarchies for Node, Branch or Object / Component to delete
//
		{
			E3dNode**  LRootNodes = PScene->RootNodes;
			EIndex     LC = 0;const EIndex  LN = PScene->RootNodes_Count;for(;LC < LN;LC++)
			{
				E3dNode*  LRootNode = LRootNodes[LC];
				EBool     LBranchDeleted = FALSE;

				E3dNode*  LNode = LRootNode;for(;LNode;)
				{
					if(LNode->Locks_Count == 0)
					{
						switch(LNode->Selection)
						{
							case E3dNodeSelBRANCH_ROOT:
								if(LNode->Parent)
								{
// Store this Branch in the Delete Operator
//
									E3dBranch*  LEntry;ENew(LEntry);
									EList_PointerAdd(LOperatorNodes->Branches, LEntry);

									E_AssignNoChkNoFree(LNode, LEntry->Node);
									E_AssignNoChkNoFree(LNode->Parent, LEntry->Parent);
									LEntry->Index = E3dNode_GetChildIndex(LNode);
								}
								LNode = LNode->Next;
							break;

							case E3dNodeSelNODE:
								if(LNode->Parent)
								{
									EBool  LDelThis = TRUE;

									if(E3d_IsActiveCameraNode(LNode)) { LDelThis = FALSE;E3dp_PrintMessage(ELogWARNING, 5000, E3dLng_Can_not_delete_active_Camera); }
									else if(E3d_IsActiveCameraInterestNode(LNode)) { LDelThis = FALSE;E3dp_PrintMessage(ELogWARNING, 5000, E3dLng_Cant_delete_interest_point_of_Camera); }

									if(LDelThis)
									{
printf("%s() %s:%d  [%s] %04x\n", __FUNCTION__, __FILE__, __LINE__, LNode->Name, PDeleteComponentsFlags);fflush(stdout);
										const EChangeFlags  LChangedT = PDeleteComponentsFlags & E3dDeleteSELECTED_OBJECTS  ? _NodeDeleteSelectedObjects(&LOperatorEditObjects, LNode,  &LObjectsWithSelectedComponents, &LObjectsWithSelectedComponents_Count, &LObjectsWithSelectedComponents_CountAllocated) : 0;

//printf("%s() %s:%d SEL_NODE\n", __FUNCTION__, __FILE__, __LINE__);fflush(stdout);
										LChanged |= LChangedT;
										if(LChangedT) LNode = LNode->Next;
										else if(PDeleteComponentsFlags & E3dDeleteSELECTED_NODES)
										{
											if(LNode->Child == NULL)
											{
// Store this Branch in the Delete Operator
//
												E3dBranch*	LEntry;ENew(LEntry);
												EList_PointerAdd(LOperatorNodes->Branches, LEntry);

												E_AssignNoChkNoFree(LNode, LEntry->Node);
												E_AssignNoChkNoFree(LNode->Parent, LEntry->Parent);
												LEntry->Index = E3dNode_GetChildIndex(LNode);

												LNode = LNode->Next;
											}
											else
											{
												E3dp_PrintMessage(ELogWARNING, 10000, "Node: '%s' has children!", LNode->Name);
												LNode = LNode->Next;
											}
										}
									}
									else LNode = LNode->Next;
								}
								else LNode = LNode->Next;
							break;

							case E3dNodeSelOBJECT:
printf("%s() %s:%d  [%s] %04x\n", __FUNCTION__, __FILE__, __LINE__, LNode->Name, PDeleteComponentsFlags);fflush(stdout);
								if(PDeleteComponentsFlags & E3dDeleteSELECTED_OBJECTS) LChanged |= _NodeDeleteSelectedObjects(&LOperatorEditObjects, LNode,  &LObjectsWithSelectedComponents, &LObjectsWithSelectedComponents_Count, &LObjectsWithSelectedComponents_CountAllocated);
								LNode = LNode->Next;
							break;

							default:
								LNode = LNode->Next;
							break;
						}
					}
					else
					{
						E3dp_PrintMessage(ELogWARNING, 10000, "Node '%s' is locked (it's probably being edited), thus can not be deleted.", LNode->Name);
						LNode = LNode->Next;
					}
				}

				if(LBranchDeleted)
				{
//printf("%s() root [%s]\n", __FUNCTION__, LRootNode->Name);fflush(stdout);
					E3dNodeHrc_UpdateConnections(LRootNode);
					E3dNodeHrc_UpdateMatrices(LRootNode, TRUE);
					E3dNodeHrc_CallNodeInfoUpdateProcs(LRootNode, E3dChgHIERARCHY, FALSE);
				}
			}
		}


		_ObjectsDeleteSelectedComponents(&LOps, &LOperatorEditObjects,  PDeleteComponentsFlags,  LObjectsWithSelectedComponents, LObjectsWithSelectedComponents_Count);


		if(LOperatorNodes->Hierarchies_Count || LOperatorNodes->Branches_Count)
		{
			EList_ObjectAdd(LOps.Operators, LOperatorNodes);
		}
		else E_Delete(LOperatorNodes);    // Free Operator struct (it's empty)



		if(LOperatorEditObjects && LOperatorEditObjects->Objects_Count)
		{
			EList_ObjectAdd(LOps.Operators, LOperatorEditObjects);
		}
		else E_DeleteChk(LOperatorEditObjects);    // Free Operator struct (it's empty)


		EList_ObjectsFree(LObjectsWithSelectedComponents);
	}



// Animations
//
	if(PScene->CurrentAnimStack)
	{
		E3dAnimStack*  LAnimStack = PScene->CurrentAnimStack;

		E3dOperatorDeleteAnimations*  LOperatorAnimationsAndChannels = NULL;
		EIndex  LC = 0;const EIndex  LN = LAnimStack->Animations_Count;for(;LC < LN;LC++)
		{
			E3dAnimation*  LAnimation = LAnimStack->Animations[LC];

			if(LAnimation->Selection == E3dNodeSelNODE || _AnimationIndirectlySelectedForDeletion(LAnimation))
			{
				if(LOperatorAnimationsAndChannels == NULL)
				{
					LOperatorAnimationsAndChannels = (E3dOperatorDeleteAnimations*)EOp_OperatorNew(_OperatorDeleteAnimations_class);
					E_AssignToMember(LAnimStack, LOperatorAnimationsAndChannels,AnimStack);
				}

				EObjectAndIndex*  LEntry;ENew(LEntry);
				EList_PointerAddA(LOperatorAnimationsAndChannels->Animations, 8, LEntry);

				E_AssignNoChkNoFree((EObject*)LAnimation, LEntry->Object);
				LEntry->Index = LC;
			}
			else
			{
				if(EObject_IsSubClass(LAnimation, E3dKeyFrameAnimation))
				{
					E3dKeyFrameAnimation*  LKeyFrameAnimation = (E3dKeyFrameAnimation*)LAnimation;
					EIndex                 LCC = 0;const EIndex  LCN = LKeyFrameAnimation->Channels_Count;for(;LCC < LCN;LCC++)
					{
						E3dKeyFrameAnimationChannel*  LChannel = (E3dKeyFrameAnimationChannel*)(LKeyFrameAnimation->Channels[LCC]);
						E3dSpline*                    LSpline = LChannel->Curve;if(LSpline)
						{
							switch(LSpline->Selection)
							{
								case 0:
								break;

								default:
									{
										if(LOperatorAnimationsAndChannels == NULL)
										{
											LOperatorAnimationsAndChannels = (E3dOperatorDeleteAnimations*)EOp_OperatorNew(_OperatorDeleteAnimations_class);
											E_AssignToMember(LAnimStack, LOperatorAnimationsAndChannels,AnimStack);
										}

										E3dChannelRec*  LEntry;ENewClr(LEntry);
										EList_PointerAddA(LOperatorAnimationsAndChannels->Channels, 8, LEntry);

										E_Assign(LAnimation, LEntry->Animation);
										E_Assign((E3dAnimationChannel*)LChannel, LEntry->Channel);
										LEntry->ChannelIndex = LCC;
									}
								break;
							}
						}
					}
				}
			}
		}

		if(LOperatorAnimationsAndChannels) EList_ObjectAdd(LOps.Operators, LOperatorAnimationsAndChannels);
	}

	if(LOps.Operators_Count)
	{
		E3dp_PickStop();

// Run Operators
//
		LChanged |= EOp_OpsRun(&LOps, EOpDO);
		if(LOps.OpNodes) E3dObjects_CallOutputs(ECbINTERNAL|ECbUSER, LOps.OpNodes, LOps.OpNodes_Count);

// Add OperatorList to undo stack
//
		E3dp_UndoStackAddOps(&LOps, E3dLng_Delete_selection);
		EOp_OpsFreeContents(&LOps);

		E3dp_PickResume();
	}

	return(LChanged);
}



static Widget    _DeleteSelectionDialogW = NULL;
static Widget    _DeleteSelection_DeleteSelectedPointsW = NULL, _DeleteSelection_DeleteSelectedEdgesW = NULL, _DeleteSelection_DeleteUnusedVerticesW = NULL;


//========================================
// Settings dialog callback
//========================================
static void _DCB_DeleteSelectionClose(Widget PW, EPointer PClientData, EPointer PCallData)
{
	Xe_WidgetUnmap(_DeleteSelectionDialogW);
}


//========================================
//========================================
static void _DCB_DeleteSelection_DeleteSelectedPoints(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dp_Prefs.DeleteSelection_DeleteSelectedPoints = ((XeToggleButtonCallbackStruct*)PCallData)->Set;
}

static void _DCB_DeleteSelectionDeleteSelectedEdges(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dp_Prefs.DeleteSelection_DeleteSelectedEdges = ((XeToggleButtonCallbackStruct*)PCallData)->Set;
}

static void _DCB_DeleteSelection_DeleteUnusedVertices(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dp_Prefs.DeleteSelection_DeleteUnusedVertices = ((XeToggleButtonCallbackStruct*)PCallData)->Set;
}


static void _MCB_Delete_Selection(Widget PW, EPointer PClientData, EPointer PCallData)
{
	if(PCallData && ((XePushButtonCallbackStruct*)PCallData)->Reason == XeCR_ACTIVATE_OPTIONS)
	{
		if(_DeleteSelectionDialogW == NULL)
		{
			Widget  LW;
			EArg    Args[256];int  ArgCnt;

			EXeStart;
			EXeSetArg(XeNallowShellResize, TRUE);
			EXeSetArg(XeNbuttonsStretch, FALSE);
			EXeSetArg(XeNworkAreaType, XeMATRIX);
			EXeSetArg(XeNwmDecorations, XeWMDecorCLOSE|XeWMDecorBORDER|XeWMDecorTITLE);
			EXeSetArg(XeNtitle, "Delete selection options");EXeSetArg(XeNiconName, "Delete selection");
			Widget  LDialogW = Xe_TopDialogNew(E3dp_XeApp, XeDIALOG_TEMPLATE, "Delete selection",  Args, ArgCnt);if(LDialogW == NULL) return;
			XeAddCallback(LDialogW, XeNwmCallback, _DCB_DeleteSelectionClose, (EPointer)0);

			_DeleteSelectionDialogW = LDialogW;

			Widget  LTopMatrixW = XeDialog_GetChild(LDialogW, XeDIALOG_WORK_AREA);

			EXeStart;
			EXeSetArg(XeNorientation, XeVERTICAL);
			EXeSetArg(XeNmarginSizeX, 4);
			Xe_SetValues(LTopMatrixW, Args, ArgCnt);

			EXeStart;
			EXeSetArg(XeNshadowThickness, 0);
			EXeSetArg(XeNhighlightThickness, 0);
			EXeSetArg(XeNheightAdjustable, FALSE);
			EXeSetArg(XeNset, E3dp_Prefs.DeleteSelection_DeleteSelectedPoints);
			XeToolTip8("Delete selected Mesh Vertices or Spline CVs (in addition to selected Polygons) when the 'Del' key is pressed.\nCtrl + Del will also do this.");
			_DeleteSelection_DeleteSelectedPointsW = LW = XeWidget_ManagedNew("Delete selected Points", XeToggleButton_class, LTopMatrixW,  Args, ArgCnt);
			XeAddCallback(LW, XeNvalueChangedCallback, _DCB_DeleteSelection_DeleteSelectedPoints, (EPointer)NULL);

			EXeStart;
			EXeSetArg(XeNshadowThickness, 0);
			EXeSetArg(XeNhighlightThickness, 0);
			EXeSetArg(XeNheightAdjustable, FALSE);
			EXeSetArg(XeNset, E3dp_Prefs.DeleteSelection_DeleteSelectedEdges);
			XeToolTip8("Delete selected Mesh Edges (in addition to selected Polygons) when the 'Del' key is pressed.\nShift + Del will also do this.");
			_DeleteSelection_DeleteSelectedEdgesW = LW = XeWidget_ManagedNew("Delete selected Edges", XeToggleButton_class, LTopMatrixW,  Args, ArgCnt);
			XeAddCallback(LW, XeNvalueChangedCallback, _DCB_DeleteSelectionDeleteSelectedEdges, (EPointer)NULL);

			EXeStart;
			EXeSetArg(XeNshadowThickness, 0);
			EXeSetArg(XeNhighlightThickness, 0);
			EXeSetArg(XeNheightAdjustable, FALSE);
			EXeSetArg(XeNset, E3dp_Prefs.DeleteSelection_DeleteUnusedVertices);
			XeToolTip8("When deleting selected Polygons in a Mesh, also delete the unused Vertices.");
			_DeleteSelection_DeleteUnusedVerticesW = LW = XeWidget_ManagedNew("Delete unused Vertices", XeToggleButton_class, LTopMatrixW,  Args, ArgCnt);
			XeAddCallback(LW, XeNvalueChangedCallback, _DCB_DeleteSelection_DeleteUnusedVertices, (EPointer)NULL);

			XeShell_Raise(LDialogW);
			XeShell_CenterToWidget(LDialogW, Eqx_TopShellW);
		}
		else
		{
			XeToggleButton_SetState(_DeleteSelection_DeleteSelectedPointsW, E3dp_Prefs.DeleteSelection_DeleteSelectedPoints, FALSE);
			XeToggleButton_SetState(_DeleteSelection_DeleteSelectedEdgesW,  E3dp_Prefs.DeleteSelection_DeleteSelectedEdges,  FALSE);
			XeToggleButton_SetState(_DeleteSelection_DeleteUnusedVerticesW, E3dp_Prefs.DeleteSelection_DeleteUnusedVertices, FALSE);
			XeShell_Raise(_DeleteSelectionDialogW);
		}
	}
	else
	{
		E3dScene*     LScene = E3d_Scene;
		EChangeFlags  LSceneChangeFlags = 0;

		const uint16_t  LInDeleteComponentsFlags = (EPointerInt)PClientData;
		uint16_t        LDeleteComponentsFlags = 0;

//printf("%s() %s:%d  %p  %d %d\n", __FUNCTION__, __FILE__, __LINE__, PW, E3dp_Prefs.DeleteSelection_DeleteSelectedPoints, E3dp_Prefs.DeleteSelection_DeleteSelectedEdges);fflush(stdout);
		if(PW)
		{
// Called from menu
//
			switch(LInDeleteComponentsFlags)
			{
				case E3dDisconnectSELECTED_POINTS:
					LDeleteComponentsFlags = E3dDisconnectSELECTED_POINTS;
				break;

				default:
					if(E3dp_Prefs.DeleteSelection_DeleteSelectedPoints)  LDeleteComponentsFlags |= E3dDeleteSELECTED_POINTS;
					if(E3dp_Prefs.DeleteSelection_DeleteSelectedEdges)   LDeleteComponentsFlags |= E3dDeleteSELECTED_EDGES;
					if(E3dp_Prefs.DeleteSelection_DeleteUnusedVertices)  LDeleteComponentsFlags |= E3dDeleteUNUSED_POINTS;
				break;
			}
		}
		else
		{
// Called from hot-key
//
			LDeleteComponentsFlags = (uint32_t)LInDeleteComponentsFlags & (E3dDeleteSELECTED_POINTS | E3dDeleteSELECTED_EDGES);
		}

		LSceneChangeFlags |= _DeleteSelection(LScene, LDeleteComponentsFlags);

		if(LSceneChangeFlags & E3dChgHIERARCHY)
		{
			E3dScene_ObjectsUpdateModelList(LScene);
			E3dScene_LayoutOnSchematics(E3d_Scene, TRUE);
		}

		E3dWindows_RedrawByChangeFlags(LSceneChangeFlags);

		if(LSceneChangeFlags & E3dChgHIERARCHY) E3dTransformPanel_Update(NULL, E3dDoALL, E3dDoALL, E3dDoALL,  -1,  E3dTrPanelALL);
	}
}


static void _KCB_Delete_Selection(const EKeyCode PKey, const unsigned int PModifiers, const int PKeyEventType, EPointer PClientData)
{
	E3dScene*  LScene = E3d_Scene;

	const uint16_t  LDeleteComponentsFlags = (EPointerInt)PClientData;

	const EChangeFlags  LSceneChangeFlags = _DeleteSelection(LScene, LDeleteComponentsFlags);

	if(LSceneChangeFlags & E3dChgHIERARCHY)
	{
		E3dScene_ObjectsUpdateModelList(LScene);
		E3dScene_LayoutOnSchematics(E3d_Scene, TRUE);
	}

	E3dWindows_RedrawByChangeFlags(LSceneChangeFlags);

	if(LSceneChangeFlags & E3dChgHIERARCHY) E3dTransformPanel_Update(NULL, E3dDoALL, E3dDoALL, E3dDoALL,  -1,  E3dTrPanelALL);
}


static EChangeFlags _Disconnect_SelectedPoints(E3dScene* PScene, EPointer POptions)
{
	E3dScene*  LScene = E3d_Scene;

	const EChangeFlags  LSceneChangeFlags = _DeleteSelection(LScene, E3dDisconnectSELECTED_POINTS);

	return(LSceneChangeFlags);
}


//================================================
// Delete Animations from the selected objects
//================================================
static void _MCB_Delete_Animations_from_selection(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dScene*      LScene = E3d_Scene;
	EChangeFlags   LChanged = 0;
	EIndex         LDeleted = 0;
	EOps           LOps;EOp_OpsInit(&LOps);

	EIndex  LASC = 0;const EIndex  LASN = LScene->AnimStacks_Count;for(;LASC < LASN;LASC++)
	{
		const E3dAnimStack*  LAnimStack = LScene->AnimStacks[LASC];const EIndex  LAN = LAnimStack->Animations_Count;

		EBool  LStackHasAnimToDelete = FALSE;

		{
			EIndex  LAC = 0;for(;LAC < LAN;LAC++)
			{
				E3dAnimation*  LAnimation = LAnimStack->Animations[LAC];

				if(LAnimation->Removable && LAnimation->Class == E3dKeyFrameAnimation_class)
				{
					EIndex  LCC = 0;const EIndex  LCN = LAnimation->Channels_Count;for(;LCC < LCN;LCC++)
					{
						E3dKeyFrameAnimationChannel*  LChannel = (E3dKeyFrameAnimationChannel*)(LAnimation->Channels[LCC]);

						if(LChannel->Curve && LChannel->TargetObject)
						{
							if(E3dAnimationChannel_TargetSelected((E3dAnimationChannel*)LChannel)) { LStackHasAnimToDelete = TRUE;goto Found; }
						}
					}
				}
			}
		}

Found:
		if(LStackHasAnimToDelete)
		{
			E3dOperatorDeleteAnimations*  LOperatorAnimationsAndChannels = (E3dOperatorDeleteAnimations*)EOp_OperatorNew(_OperatorDeleteAnimations_class);
			E_AssignToMember(LAnimStack, LOperatorAnimationsAndChannels,AnimStack);

			EIndex  LAC = 0;for(;LAC < LAN;LAC++)
			{
				E3dAnimation*  LAnimation = LAnimStack->Animations[LAC];
				EBool          LDelete = FALSE;

				if(LAnimation->Removable && LAnimation->Class == E3dKeyFrameAnimation_class)
				{
					EIndex  LCC = 0;const EIndex  LCN = LAnimation->Channels_Count;for(;LCC < LCN;LCC++)
					{
						E3dKeyFrameAnimationChannel*  LChannel = (E3dKeyFrameAnimationChannel*)(LAnimation->Channels[LCC]);

						if(LChannel->Curve && LChannel->TargetObject)
						{
							if(E3dAnimationChannel_TargetSelected((E3dAnimationChannel*)LChannel))
							{
								EObjectAndIndex*  LEntry;ENew(LEntry);
								EList_PointerAddA(LOperatorAnimationsAndChannels->Animations, 8, LEntry);

								E_AssignNoChkNoFree((EObject*)LAnimation, LEntry->Object);
								LEntry->Index = LAC;LDelete = TRUE;
							}
						}
					}
					if(LDelete) LDeleted++;
				}
			}


			EList_ObjectAdd(LOps.Operators, LOperatorAnimationsAndChannels);
		}
	}

	if(LOps.Operators_Count)
	{
// Run Operators
//
		LChanged |= EOp_OpsRun(&LOps, EOpDO);

// Add OperatorList to undo stack
//
		E3dp_UndoStackAddOps(&LOps, E3dLng_Delete_selection);
		EOp_OpsFreeContents(&LOps);
	}

	if(LChanged) E3dWindows_RedrawByChangeFlags(LChanged);

	if(LDeleted) E3dp_PrintMessage(ELogMESSAGE, 10000, "Animation(s) deleted: %d", LDeleted);
	else E3dp_PrintMessage(ELogMESSAGE, 10000, "No animations deleted");
}


//========================================
// Delete all Animations
//========================================
static void _MCB_Delete_All_animations(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dScene*     LScene = E3d_Scene;
	EChangeFlags  LChanged = 0;
	EIndex        LDeleted = 0;
	EOps          LOps;EOp_OpsInit(&LOps);

	EIndex  LASC = 0;const EIndex  LASN = LScene->AnimStacks_Count;for(;LASC < LASN;LASC++)
	{
		const E3dAnimStack*  LAnimStack = LScene->AnimStacks[LASC];const EIndex  LAN = LAnimStack->Animations_Count;

		EBool  LStackHasAnimToDelete = FALSE;

		{
			EIndex  LAC = 0;for(;LAC < LAN;LAC++)
			{
				E3dAnimation*  LAnimation = LAnimStack->Animations[LAC];

				if(LAnimation->Removable && LAnimation->Class == E3dKeyFrameAnimation_class) { LStackHasAnimToDelete = TRUE;goto Found; }
			}
		}

Found:
		if(LStackHasAnimToDelete)
		{
			E3dOperatorDeleteAnimations*  LOperatorAnimationsAndChannels = (E3dOperatorDeleteAnimations*)EOp_OperatorNew(_OperatorDeleteAnimations_class);
			E_AssignToMember(LAnimStack, LOperatorAnimationsAndChannels,AnimStack);

			EIndex  LAC = 0;for(;LAC < LAN;LAC++)
			{
				E3dAnimation*  LAnimation = LAnimStack->Animations[LAC];

				if(LAnimation->Removable && LAnimation->Class == E3dKeyFrameAnimation_class)
				{
					EObjectAndIndex*  LEntry;ENew(LEntry);
					EList_PointerAddA(LOperatorAnimationsAndChannels->Animations, 8, LEntry);

					E_AssignNoChkNoFree((EObject*)LAnimation, LEntry->Object);
					LEntry->Index = LAC;

					LDeleted++;
				}
			}

			EList_ObjectAdd(LOps.Operators, LOperatorAnimationsAndChannels);
		}
	}

	if(LOps.Operators_Count)
	{
// Run Operators
//
		LChanged |= EOp_OpsRun(&LOps, EOpDO);

// Add OperatorList to undo stack
//
		E3dp_UndoStackAddOps(&LOps, E3dLng_Delete_selection);
		EOp_OpsFreeContents(&LOps);
	}

	if(LChanged) E3dWindows_RedrawByChangeFlags(LChanged);

	if(LDeleted) E3dp_PrintMessage(ELogMESSAGE, 10000, "Animation(s) deleted: %d", LDeleted);
	else E3dp_PrintMessage(ELogMESSAGE, 10000, "No animations deleted");
}


//========================================
// Delete scene or selection
//========================================
static void _MCB_Delete_All(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dScene*  LScene = E3d_Scene;
	E3dScene_Delete_All(LScene);
}


//========================================
// Delete scene or selection
//========================================
static void _MCB_Delete_All_hierarchies(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dScene*     LScene = E3d_Scene;
	EChangeFlags  LChanged = 0;

	if(LScene->RootNodes_Count)
	{
		E3dNode**  LRootNodes;ENewN(LScene->RootNodes_Count, LRootNodes);
		EIndex     LRootNodes_Count = LScene->RootNodes_Count;

		memcpy(LRootNodes, LScene->RootNodes, sizeof(E3dNode*) * LScene->RootNodes_Count);

		EIndex  LC = 0;for(;LC < LRootNodes_Count;LC++)
		{
			E3dNode*  LRootNode = LRootNodes[LC];

			if(!E3d_IsActiveCameraNode(LRootNode))
			{
				if(LRootNode->Locks_Count == 0)
				{
					LChanged |= E3dNodeHrc_GetTransformChangeFlagsByObjects(LRootNode) | E3dChgHIERARCHY;
					E3dScene_NodeHrcRemove(LScene, LRootNode, &LChanged, TRUE);
				}
				else
				{
					E3dp_PrintMessage(ELogWARNING, 10000, "Node '%s' is locked (it's probably being edited), thus can not be deleted", LRootNode->Name);
				}
			}
		}

		EFree(LRootNodes);
		E3dScene_ObjectsUpdateModelList(LScene);
		E3dScene_LayoutOnSchematics(LScene, TRUE);
		E3dTransformPanel_Update(NULL, E3dDoALL, E3dDoALL, E3dDoALL,  -1,  E3dTrPanelALL);
		E3dWindows_RedrawByChangeFlags(LChanged);
	}
}


//========================================
// Menu callback
//========================================
static void _MCB_Delete_Empty_animations(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dScene*      LScene = E3d_Scene;
	E3dAnimStack*  LAnimStack = LScene->CurrentAnimStack;if(LAnimStack == NULL) return;

	EIndex	LAC = 0;for(;LAC < LAnimStack->Animations_Count;)
	{
		E3dAnimation*  LAnim = LAnimStack->Animations[LAC];
		EBool          LAnimRemoved = FALSE;

		if(LAnim->Removable && LAnim->Class == E3dKeyFrameAnimation_class)
		{
			E3dKeyFrameAnimation*	LAnimation = (E3dKeyFrameAnimation*)LAnim;

// Delete empty Channels
//
			if(LAnimation->Channels_Count)
			{
				EIndex  LChC = 0;for(;LChC < LAnimation->Channels_Count;)
				{
					E3dKeyFrameAnimationChannel*  LChannel = (E3dKeyFrameAnimationChannel*)(LAnimation->Channels[LChC]);
					EBool                         LChannelRemoved = FALSE;


// Remove Channel if it has no Curve / KeyFrames
//
					if(LChannel->Curve)
					{
printf("%s() %s:%d  CVs_Count %lld\n", __FUNCTION__, __FILE__, __LINE__, (long long)(LChannel->Curve->CVs_Count));fflush(stdout);
						if(LChannel->Curve->CVs_Count == 0)
						{
							E3dAnimation_RemoveChannel(LAnim, (E3dAnimationChannel*)LChannel);LChannelRemoved = TRUE;
						}
					}
					else
					{
						E3dAnimation_RemoveChannel(LAnim, (E3dAnimationChannel*)LChannel);LChannelRemoved = TRUE;
					}
					printf("%s() LChannelRemoved %d\n", __FUNCTION__, LChannelRemoved);fflush(stdout);
					if(!LChannelRemoved) LChC++;
				}
			}

// Now, delete LAnim, if it has no (remaining) Channels
//
			if(LAnimation->Channels_Count == 0)
			{
				E3dScene_AnimationRemove(LScene, LAnim, FALSE);LAnimRemoved = TRUE;
			}

		}

		if(!LAnimRemoved) LAC++;
	}
}


//========================================
// Entry point of the plugin
//========================================
int E_PluginInit(EPlugin* PPlugin)
{
	_EditInfoPlugin = PPlugin;

	E3dNode_EditFunc      = _Node_Edit;
	E3dNode_IsEditedFunc  = _Node_IsEdited;

	E3dPointSet_class->GUI = _PointSet_GUI;
	E3dMesh_class->GUI     = _Mesh_GUI;

	E3dPolyGroup_IsEditedFunc = _PolyGroup_IsEdited;
	E3dPolyGroup_UpdateDialogFunc = _PolyGroup_UpdateDialog;

	E3dObjectEdit_Init();

	_MenuButtons[_MenuButtons_Count++] = XePushButton_Add("Menu->Info", "Edit selection", '\0', "~Shift ~Ctrl <Key>E", "E", FALSE, NULL, _MCB_Edit, (EPointer)CR_EDIT_SELECTION);
	E_KeySetCallback("Edit selection", 'e', 0, EKeyPRESS, _KCB_Edit, (EPointer)CR_EDIT_SELECTION);

	{
		static EOperatorClass  LClassDesc[] =
		{
			{
				.Name=        "Delete selection",
				.StructSize=  sizeof(E3dOperatorDelete),
				.Init=        _DeleteSelection_Init,
				.Free=        _DeleteSelection_Free,
				.Proc=        _DeleteSelection_Proc
			}
		};
		_OperatorDelete_class = EOp_OperatorClassRegister(LClassDesc);
	}

	{
		static EOperatorClass  LClassDesc[] =
		{
			{
				.Name=        "Delete animations and channels",
				.StructSize=  sizeof(E3dOperatorDeleteAnimations),
				.Init=        _DeleteAnimationsInit,
				.Free=        _DeleteAnimationsFree,
				.Proc=        _DeleteAnimationsProc
			}
		};
		_OperatorDeleteAnimations_class = EOp_OperatorClassRegister(LClassDesc);
	}

// Operator for deleting Objects, or modifying a Face:
// Face
// - Delete hole
//
	{
		static EOperatorClass  LClassDesc[] =
		{
			{
				.Name=        "Edit Objects",
				.StructSize=  sizeof(E3dOperatorEditObjects),
				.Init=        _EditObjects_Init,
				.Free=        _EditObjects_Free,
				.Proc=        _EditObjects_Proc
			}
		};
		E3dOperatorEditObjects_class = EOp_OperatorClassRegister(LClassDesc);
	}


// Operator for modifying a Mesh
// -Add/delete Edges
// -Add/delete Polygons
//
	{
		EOperatorClass  LClassDesc[] =
		{
			{
				.Name=        "Edit Mesh",
				.StructSize=  sizeof(E3dOperatorEditMesh),
				.Init=        _EditMesh_Init,
				.Free=        _EditMesh_Free,
				.Proc=        _EditMesh_Proc
			}
		};
		E3dOperatorEditMesh_class = EOp_OperatorClassRegister(LClassDesc);
	}

	{
		EOperatorClass  LClassDesc[] =
		{
			{
				.Name=        "Delete Vertices",
				.StructSize=  sizeof(E3dOperatorAddDeleteVertices),
				.Init=        _AddDeleteVertices_Init,
				.Free=        _AddDeleteVertices_Free,
				.Proc=        _DeleteVertices_Proc
			}
		};
		E3dOperatorDeleteVertices_class = EOp_OperatorClassRegister(LClassDesc);
	}

	_MenuButtons[_MenuButtons_Count++] = XePushButton_Add("Menu->Delete", "Delete selection", '\0', "~Shift ~Ctrl <Key>Delete", "Del", TRUE, NULL, _MCB_Delete_Selection, (EPointer)0);
	const EKeyModifiers  LCtrlMask = Xe_CommandKeyForCtrl ? E_CommandModifierMask : EControlMask;
	E_KeySetCallback("Delete selection", EKeyDELETE,    0,                    EKeyPRESS, _KCB_Delete_Selection, (EPointer)(E3dDeleteSELECTED_NODES|E3dDeleteSELECTED_OBJECTS|E3dDeleteSELECTED_POLYGROUPS|E3dDeleteSELECTED_POLYGONS|E3dDeleteSELECTED_EDGES|E3dDeleteSELECTED_POINTS));
	E_KeySetCallback("Delete selection", EKeyDELETE,    LCtrlMask,            EKeyPRESS, _KCB_Delete_Selection, (EPointer)E3dDeleteSELECTED_POINTS);
	E_KeySetCallback("Delete selection", EKeyDELETE,    EShiftMask,           EKeyPRESS, _KCB_Delete_Selection, (EPointer)E3dDeleteSELECTED_EDGES);
	E_KeySetCallback("Delete selection", EKeyDELETE,    LCtrlMask|EShiftMask, EKeyPRESS, _KCB_Delete_Selection, (EPointer)(E3dDeleteSELECTED_POINTS|E3dDeleteSELECTED_EDGES));
	E_KeySetCallback("Delete selection", EKeyBACKSPACE, 0,                    EKeyPRESS, _KCB_Delete_Selection, (EPointer)(E3dDeleteSELECTED_NODES|E3dDeleteSELECTED_OBJECTS|E3dDeleteSELECTED_POLYGROUPS|E3dDeleteSELECTED_POLYGONS|E3dDeleteSELECTED_EDGES|E3dDeleteSELECTED_POINTS));
	E_KeySetCallback("Delete selection", EKeyBACKSPACE, LCtrlMask,            EKeyPRESS, _KCB_Delete_Selection, (EPointer)E3dDeleteSELECTED_POINTS);
	E_KeySetCallback("Delete selection", EKeyBACKSPACE, EShiftMask,           EKeyPRESS, _KCB_Delete_Selection, (EPointer)E3dDeleteSELECTED_EDGES);
	E_KeySetCallback("Delete selection", EKeyBACKSPACE, LCtrlMask|EShiftMask, EKeyPRESS, _KCB_Delete_Selection, (EPointer)(E3dDeleteSELECTED_POINTS|E3dDeleteSELECTED_EDGES));

	_MenuButtons[_MenuButtons_Count++] = XePushButton_Add("Menu->Delete", "All hierarchies",               '\0', NULL, NULL, FALSE, NULL, _MCB_Delete_All_hierarchies,           (EPointer)0);
	_MenuButtons[_MenuButtons_Count++] = XePushButton_Add("Menu->Delete", "All",                           '\0', NULL, NULL, FALSE, NULL, _MCB_Delete_All,                       (EPointer)0);
	_MenuButtons[_MenuButtons_Count++] = XePushButton_Add("Menu->Delete", "Animations from the selection", '\0', NULL, NULL, FALSE, NULL, _MCB_Delete_Animations_from_selection, (EPointer)0);
	_MenuButtons[_MenuButtons_Count++] = XePushButton_Add("Menu->Delete", "Empty animations",              '\0', NULL, NULL, FALSE, NULL, _MCB_Delete_Empty_animations,          (EPointer)0);
	_MenuButtons[_MenuButtons_Count++] = XePushButton_Add("Menu->Delete", "All animations",                '\0', NULL, NULL, FALSE, NULL, _MCB_Delete_All_animations,            (EPointer)0);
	_MenuButtons[_MenuButtons_Count++] = XePushButton_Add("Menu->Delete", "Disconnect Vertices",           '\0', NULL, NULL, FALSE, NULL, _MCB_Delete_Selection,                 (EPointer)E3dDisconnectSELECTED_POINTS);
	E3dScene_ScriptFunctionRegister("Disconnect_SelectedPoints", "Delete selected Edges", "Delete selected Edges on selected Meshes.", _Disconnect_SelectedPoints,  _ScriptFunctions + _ScriptFunctions_Count++);

	return(E_SUCCESS);
}


//========================================
// Exit method of the plugin
//========================================
int E_PluginExit(const EBool PExiting)
{
	if(!PExiting || E3dp_GUIFreeOnExit)
	{
		{ int  LC = 0;for(;LC < _MenuButtons_Count;LC++) { XeItemFree(_MenuButtons[LC]); } _MenuButtons_Count = 0; }
		{ int  LC = 0;for(;LC < _ScriptFunctions_Count;LC++) { E3dScene_ScriptFunctionRemove(_ScriptFunctions[LC]); } _ScriptFunctions_Count = 0; }

// XeFIXME
//		XeWidget_Free(_NodeDialogW);  // Don't free top-level shells
	}

	if(_NodePanel) { E3dNodePanel_Free(_NodePanel);_NodePanel = NULL; }

	E3dNode_EditFunc     = NULL;
	E3dNode_IsEditedFunc = NULL;

	E3dPointSet_class->GUI = NULL;
	E3dMesh_class->GUI     = NULL;

	E3dPolyGroup_IsEditedFunc = NULL;
	E3dPolyGroup_UpdateDialogFunc = NULL;

	E3dObjectEdit_Cleanup();

	E_KeyRemoveCallback('e', 0, EKeyPRESS, _KCB_Edit, (EPointer)CR_EDIT_SELECTION);


	const EKeyModifiers  LCtrlMask = Xe_CommandKeyForCtrl ? E_CommandModifierMask : EControlMask;
	E_KeyRemoveCallback(EKeyDELETE,    0,                    EKeyPRESS, _KCB_Delete_Selection, (EPointer)(E3dDeleteSELECTED_NODES|E3dDeleteSELECTED_OBJECTS|E3dDeleteSELECTED_POLYGROUPS|E3dDeleteSELECTED_POLYGONS|E3dDeleteSELECTED_EDGES|E3dDeleteSELECTED_POINTS));
	E_KeyRemoveCallback(EKeyDELETE,    LCtrlMask,            EKeyPRESS, _KCB_Delete_Selection, (EPointer)E3dDeleteSELECTED_POINTS);
	E_KeyRemoveCallback(EKeyDELETE,    EShiftMask,           EKeyPRESS, _KCB_Delete_Selection, (EPointer)E3dDeleteSELECTED_EDGES);
	E_KeyRemoveCallback(EKeyDELETE,    LCtrlMask|EShiftMask, EKeyPRESS, _KCB_Delete_Selection, (EPointer)(E3dDeleteSELECTED_POINTS|E3dDeleteSELECTED_EDGES));
	E_KeyRemoveCallback(EKeyBACKSPACE, 0,                    EKeyPRESS, _KCB_Delete_Selection, (EPointer)(E3dDeleteSELECTED_NODES|E3dDeleteSELECTED_OBJECTS|E3dDeleteSELECTED_POLYGROUPS|E3dDeleteSELECTED_POLYGONS|E3dDeleteSELECTED_EDGES|E3dDeleteSELECTED_POINTS));
	E_KeyRemoveCallback(EKeyBACKSPACE, LCtrlMask,            EKeyPRESS, _KCB_Delete_Selection, (EPointer)E3dDeleteSELECTED_POINTS);
	E_KeyRemoveCallback(EKeyBACKSPACE, EShiftMask,           EKeyPRESS, _KCB_Delete_Selection, (EPointer)E3dDeleteSELECTED_EDGES);
	E_KeyRemoveCallback(EKeyBACKSPACE, LCtrlMask|EShiftMask, EKeyPRESS, _KCB_Delete_Selection, (EPointer)(E3dDeleteSELECTED_POINTS|E3dDeleteSELECTED_EDGES));

	if(_OperatorDelete_class)           { EOp_OperatorClassDeactivate(_OperatorDelete_class);          _OperatorDelete_class           = NULL; }
	if(_OperatorDeleteAnimations_class) { EOp_OperatorClassDeactivate(_OperatorDeleteAnimations_class);_OperatorDeleteAnimations_class = NULL; }

	if(E3dOperatorEditObjects_class)    { EOp_OperatorClassDeactivate(E3dOperatorEditObjects_class);       E3dOperatorEditObjects_class    = NULL; }
	if(E3dOperatorEditMesh_class)       { EOp_OperatorClassDeactivate(E3dOperatorEditMesh_class);          E3dOperatorEditMesh_class       = NULL; }
	if(E3dOperatorDeleteVertices_class) { EOp_OperatorClassDeactivate(E3dOperatorDeleteVertices_class);E3dOperatorDeleteVertices_class = NULL; }
	return(E_SUCCESS);
}
