//================================================================================================
// Plugin for EQUINOX-3D
//
// Distribute points on a Mesh.
//
// AUTHOR:  Gabor Nagy
// DATE:    2019-Jun-23 03:36:02
//
// Copyright (C) 2014-2021 by Gabor Nagy. All rights reserved.
//================================================================================================
#include <assert.h>
#include <stdio.h>

#include <ELists.h>
#include <EMalloc.h>
#include <EPlugins.h>

#include <E3D/Scene.h>

#include <E3D/UI/Panel.h>
#include <E3D/UI/Panels.h>

#include <Xe/PushButton.h>
#include <Xe/Scale.h>


char*             E_PluginInfo = "Distribute points on a Mesh";

static EPlugin*   _Plugin = NULL;

static EguiItem	  _ToolBarButtons[16];static int  _ToolBarButtons_Count = 0;


//================================================
// ToolBar icon callback handler
//================================================
static void _CB(Widget PW, EPointer PClientData, EPointer PCallData)
{
	const EIndex  LNumOfPoints = 100000;

	E3dObject**  LObjects;EIndex  LNumOfObjects = E3dScene_GetObjects(E3d_Scene, &LObjects, E3dItemMESH, E3dSELECTED, FALSE);if(LNumOfObjects)
	{
		EIndex  LObjC = 0;for(;LObjC < LNumOfObjects; LObjC++)
		{
			E3dObject*  LObject = LObjects[LObjC];

			if(EObject_IsSubClass(LObject, E3dMesh))
			{
				E3dMesh*  LMesh = (E3dMesh*)LObject;

				E3dCoordinate*  LPoints = E3dMesh_GenerateSurfacePoints(LMesh, NULL,  LNumOfPoints, 42, FALSE,  0, 0,  NULL, NULL);

				E3d_VisualizePointsWithPointSet((E3dCoord3*)LPoints, LNumOfPoints, "Points");
				EFree(LPoints);
			}
		}
		EFree(LObjects);
	}

	E3dWindows_RedrawByChangeFlags(E3dChgPOINTSET);
}


//========================================
// Entry point of the plugin
//========================================
int E_PluginInit(EPlugin* PPlugin)
{
	_Plugin = PPlugin;

// Use default plugin icon
//
	Xe_PushButtonAddImg3(_ToolBarButtons[_ToolBarButtons_Count++], Plugin, "Tool->Create", "Points", '\0', NULL, NULL, FALSE, "Distribute points on a Mesh", _CB, (EPointer)0);

	return(E_SUCCESS);
}


//========================================
// Exit method of the plugin
//========================================
int E_PluginExit(const EBool PExiting)
{
	if(!PExiting || E3dp_GUIFreeOnExit)
	{
		int  LC = 0;for(;LC < _ToolBarButtons_Count;LC++) { XeItemFree(_ToolBarButtons[LC]); }
		_ToolBarButtons_Count = 0;
	}

	return(E_SUCCESS);
}
