//================================================================
// File I/O
//
// Plugin for EQUINOX-3D
//
// AUTHOR:  Gabor Nagy
// DATE:    2018-Apr-18 11:16:27
//
// Copyright (C) 1994-2021 by Gabor Nagy. All rights reserved.
//================================================================
#include <stdio.h>

#include <EMalloc.h>
#include <EPlugins.h>

#include <Xe/Label.h>
#include <Xe/Matrix.h>
#include <Xe/Menu.h>
#include <Xe/PushButton.h>
#include <Xe/ToggleButton.h>

#include <Xe/Dialog.h>
#include <Xe/Shell.h>
#include <Xe/FSBox.h>

#include <E3D/IO.h>
#include <E3D/Matrix.h>

#include <E3D/UI/ModeStack.h>    // E3dFIXME : for E3dp_ResetWorkModes (we shouldn't need it, it's just a work-around)
#include <E3D/UI/Panel.h>
#include <E3D/UI/StatusPanel.h>  // For E3dp_PrintMessage
#include <E3D/UI/TimePanel.h>


#include "../EQXExt.h"


static EPlugin*      _Plugin = NULL;
static EPluginLock*  _SaveSceneDialogLock = NULL;


static EIOModuleGUI*  _SaveSceneFormatWidgets = NULL;


static EguiItem  E3d_MenuButton = NULL,
                 E3d_SaveSceneMenuButton = NULL;



// "Save Scene" dialog box
//
static Widget    _SaveSceneDialogW = NULL;
static Widget    _SaveFormatsW = NULL, _FromExtensionBW = NULL;



/*
static Widget  _SaveSelectedHierarchiesDialogW = NULL;

//========================================================
// Save Selected Hierarchies dialog callback handler
//========================================================
static void _FCB_SaveSelectedHierarchies(Widget PW, EPointer PClientData, EPointer PCallData)
{
	switch((EPointerInt)PClientData)
	{
		case XeCR_OK:
			{
				char*  LStringPtr = XeFSBox_GetFile(_SaveSelectedHierarchiesDialogW);if(LStringPtr)
				{
					const char*  LExtension = EStr_GetFileNameExtension(LStringPtr);

					EIndex  LC = 0;const EIndex  LN = E3d_Scene->RootNodes_Count;for(;LC < LN;LC++)
					{
						E3dNode*  LNode = E3d_Scene->RootNodes[LC];if(LNode->Selection == E3dNodeSelBRANCH_ROOT)
						{
							char  LPathName[MAXPATHLEN + 1];EStr_GetPathName(LStringPtr, MAXPATHLEN, LPathName);
							char  LFileName[MAXPATHLEN + 1];

//							sprintf(LFileName, "%s%s%s", LPathName, LNode->Name, LExtension);
							sprintf(LFileName, "%s%s", LPathName, LNode->Name);
//printf("Saving: [%s]\n", LFileName);fflush(stdout);

							char  LFileNameRet[MAXPATHLEN + 1];
							if((E3d_HierarchyWriteToFile(LFileName, LFileNameRet, MAXPATHLEN, LNode, E3dFileFormatFROM_EXTENSION, NULL)) == E_SUCCESS)
							{
								E3dp_PrintMessage(ELogMESSAGE, 10000, "Saved: %s\n", LFileName);
							}
						}
					}

					EFree(LStringPtr);
				}

				XeDialog_GetValuesVa(_SaveSelectedHierarchiesDialogW,  XeNfilterList, Eqx_Prefs.ModelFilterList,  XeNpath, Eqx_ModelPath,  NULL);
			}
		break;

		case XeCR_CANCEL:
			Xe_WidgetUnmap(_SaveSelectedHierarchiesDialogW);
			XeDialog_GetValuesVa(_SaveSelectedHierarchiesDialogW,  XeNfilterList, Eqx_Prefs.ModelFilterList,  XeNpath, Eqx_ModelPath,  NULL);
		break;
	}
}


//========================================
// Save Selected hierarchies
//========================================
static void EqxMCB_SaveSelectedHierarchies(Widget PW, EPointer PClientData, EPointer PCallData)
{
	if(Eqx_ModelPath[0] == 0) EStr_CurrentDir(Eqx_ModelPath);

	if(_SaveSelectedHierarchiesDialogW)
	{
		XeShell_Raise(_SaveSelectedHierarchiesDialogW);return;
	}
	else
	{
		EArg  Args[32];int  ArgCnt;

		if(Eqx_ModelPath[0] == 0) EStr_CurrentDir(Eqx_ModelPath);

		EXeStart;
		EXeSetArg(XeNminSizeX, 380);EXeSetArg(XeNminSizeY, 400);
		EXeSetArg(XeNfileListVisibleItemNum, 22);
		EXeSetArg(XeNtitle, "EQUINOX-3D: Save hierarchies");
		EXeSetArg(XeNpath, Eqx_ModelPath);
		EXeSetArg(XeNfilterListVisibleItemNum, 4);
		EXeSetArg(XeNcustomItemType, XeMATRIX);
		EXeSetArg(XeNcustomItemPlacement, XePLACE_ABOVE_FILENAME);
		Widget  LDialogW = XeFileSelectionDialog_New(Eqx_TopShellW,  Args, ArgCnt);if(LDialogW)
		{
			E3d_SaveSelectedHierarchiesDialogW = LDialogW;

			XeAddCallbackN(LDialogW, XeNokCallback, _FCB_SaveSelectedHierarchies, (EPointer)XeCR_OK);
			XeAddCallback(LDialogW, XeNwmCallback, _FCB_SaveSelectedHierarchies, (EPointer)XeCR_CANCEL);

			XeShell_CenterToWidget(LDialogW, Eqx_TopShellW);
			XeShell_Raise(LDialogW);
		}
	}
}
*/


//========================================
// Set Scene output file format
//========================================
static void _CB_3DSceneSaveFormat(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dIOModule*  LIOModule = (E3dIOModule*)(((XeOptionMenuCallbackStruct*)PCallData)->Value);
	char          LFName[MAXPATHLEN + 1];
	EBool         LCorrected = FALSE;

	Eqx_Prefs.SceneSaveFormat = LIOModule;

	E3dp_IOModuleShowGUI(LIOModule, _SaveSceneFormatWidgets, E3d_3DSceneFileFormats_Count, NULL);

	Eqx_Prefs.SceneFormatFromExtension = FALSE;
	XeToggleButton_SetState(_FromExtensionBW, FALSE, FALSE);

	char*  LFileName = XeFSBox_GetFile(_SaveSceneDialogW);if(LFileName)
	{
// See if the file name already has any of the extensions specified for any 3D-Scene format
//
		{
			const int  LLen = (int)strlen(LFileName);
			EBool      LFound = FALSE;

			int  LC = 0;const int  LN = E3d_3DSceneFileFormats_Count;for(;LC < LN;LC++)
			{
				E3dIOModule*  LTFileFormat = E3d_3DSceneFileFormats[LC];if(LTFileFormat->FileNameExtensions)
				{
					int  LC1 = 0;for(;LTFileFormat->FileNameExtensions[LC1] != NULL;LC1++)
					{
						const char*  LExt = LTFileFormat->FileNameExtensions[LC1];
						const int    LEL = (int)strlen(LExt);

						if(LEL > 0)
						{
// Extension matched
//
							if(LLen > LEL && EStr_NEqual(LExt, LFileName + LLen - LEL, LEL))
							{
								LFound = TRUE;
								LFileName[LLen - LEL] = 0;
								sprintf(LFName, "%s%s", LFileName, LIOModule->FileNameExtensions[0]);
								LCorrected = TRUE;
								LC = LN;
								break;
							}
						}
					}
				}
			}
			if(!LFound)
			{
				sprintf(LFName, "%s%s", LFileName, LIOModule->FileNameExtensions[0]);LCorrected = TRUE;
			}
		}

		if(LCorrected) XeFSBox_SetFile(_SaveSceneDialogW, LFName);

		EFree(LFileName);
	}
}


//========================================
// Save Scene FSBox callbacks
//========================================
static void _FCB_SaveScene(Widget PW, EPointer PClientData, EPointer PCallData)
{
	switch((EPointerInt)PClientData)
	{
		case XeCR_OK:
			{
				char*  LStringPtr = XeFSBox_GetFile(_SaveSceneDialogW);if(LStringPtr)
				{
					E3dp_Prefs.CurrentFrame    = E3dp_TimePanel->Frame;
					E3dp_Prefs.StartFrame      = E3dp_TimePanel->StartFrame;
					E3dp_Prefs.EndFrame        = E3dp_TimePanel->EndFrame;
					E3dp_Prefs.FramesPerSecond = E3dp_TimePanel->FramesPerSecond;

//{ EIndex  LC = 0;for(;LStringPtr[LC];LC++) printf("%02x ", LStringPtr[LC]);printf("\n");fflush(stdout); }

					E3dScene*  LScene = E3d_Scene;
					char       LActualName[MAXPATHLEN + 1];
					int        LResult;
					if(Eqx_Prefs.SceneFormatFromExtension) LResult = E3d_SceneWriteToFile(LStringPtr, LActualName, MAXPATHLEN, LScene, E3dFileFormatFROM_EXTENSION, NULL);
					else LResult = E3d_SceneWriteToFile(LStringPtr, LActualName, MAXPATHLEN, LScene, Eqx_Prefs.SceneSaveFormat, NULL);

					switch(LResult)
					{
						case E_SUCCESS:
							E3dp_PrintMessage(ELogMESSAGE, 10000, "Scene saved as '%s'", LActualName);
							Xe_WidgetUnmap(_SaveSceneDialogW);
							E_PluginLockDel(_Plugin, &_SaveSceneDialogLock, FALSE);
							Eqx_SetTitle(LActualName);
						break;

						case E_NO_IOFILTER_FOUND:
							if(Eqx_Prefs.SceneFormatFromExtension)
							{
								const char*  LExtension = EStr_GetFileNameExtension(LStringPtr);
								if(LExtension) E3dp_PrintMessage(ELogWARNING, 10000, "No exporter found for extension '%s'. Please correct it or turn off 'From extension'", LExtension);
								else E3dp_PrintMessage(ELogWARNING, 10000, "Please specify a file name extension or turn off 'From extension'");
							}
						break;

						default:  E3dp_PrintMessage(ELogMESSAGE, 10000, "%s '%s'", LStringPtr, E_ResultToString(LResult));break;
					}

					EFree(LStringPtr);
				}
			}
		break;

		case XeCR_CANCEL:
			Xe_WidgetUnmap(_SaveSceneDialogW);
			E_PluginLockDel(_Plugin, &_SaveSceneDialogLock, FALSE);
		break;

		case XeCR_APPLY:  // File name changed
			if(Eqx_Prefs.SceneFormatFromExtension)
			{
				char*  LStringPtr = XeFSBox_GetFile(_SaveSceneDialogW);if(LStringPtr)
				{
					const char*  LExtension = EStr_GetFileNameExtension(LStringPtr);if(LExtension)
					{
						E3dIOModule*  LIOModule = E3d_IOModuleByFileNameExtension(LExtension,  E3d_3DSceneFileFormats, E3d_3DSceneFileFormats_Count,  EWRITE);

						if(LIOModule) E3dp_IOModuleShowGUI(LIOModule, _SaveSceneFormatWidgets, E3d_3DSceneFileFormats_Count, _SaveFormatsW);
					}

					EFree(LStringPtr);
				}
			}
		break;
	}
}


//========================================
// Dialog-box callback handler
//========================================
static void _DCB_Close(Widget PW, EPointer PClientData, EPointer PCallData)
{
	Xe_WidgetUnmap(_SaveSceneDialogW);
	E_PluginLockDel(_Plugin, &_SaveSceneDialogLock, FALSE);
}


static int _SaveSceneDialogLockDel(EPlugin* PPlugin, EPluginLock* PLock)
{
	Xe_WidgetUnmap(_SaveSceneDialogW);

	return(E_SUCCESS);
}


//========================================
// Dialog-box callback handler
//========================================
static void _DCB_FromExtension(Widget PW, EPointer PClientData, EPointer PCallData)
{
	Eqx_Prefs.SceneFormatFromExtension = ((XeToggleButtonCallbackStruct*)PCallData)->Set;
}


//========================================
// Save Scene
//========================================
static void E3dMCB_SaveScene(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dp_ResetWorkModes();

	if(_SaveSceneDialogLock == NULL) _SaveSceneDialogLock = E_PluginLockAdd(_Plugin, _SaveSceneDialogLockDel);

	if(_SaveSceneDialogW == NULL)
	{
		EArg  Args[256];int  ArgCnt;

		if(Eqx_ModelPath[0] == 0) EStr_CurrentDir(Eqx_ModelPath);

		EXeStart;
		EXeSetArg(XeNminSizeX, 380);EXeSetArg(XeNminSizeY, 400);
		EXeSetArg(XeNcenterToWidget, Eqx_TopShellW);
		EXeSetArg(XeNfileListVisibleItemNum, 22);
		EXeSetArg(XeNtitle, "Save Scene");
		EXeSetArg(XeNpath, Eqx_ModelPath);
		EXeSetArg(XeNfilterListVisibleItemNum, 4);
		EXeSetArg(XeNcustomItemType, XeMATRIX);
		EXeSetArg(XeNcustomItemPlacement, XePLACE_ABOVE_FILENAME);
		Widget  LDialogW = XeFileSelectionDialog_New(Eqx_TopShellW,  Args, ArgCnt);if(LDialogW)
		{
			Widget            LW;
			XeOptionMenuRec*  LOptionMenuItems;
			int               LNumOfFormatsFound = 0;

			_SaveSceneDialogW = LDialogW;

			XeAddCallbackN(LDialogW, XeNokCallback,           _FCB_SaveScene, (EPointer)XeCR_OK);
			XeAddCallbackN(LDialogW, XeNcancelCallback,       _FCB_SaveScene, (EPointer)XeCR_CANCEL);
			XeAddCallbackN(LDialogW, XeNvalueChangedCallback, _FCB_SaveScene, (EPointer)XeCR_APPLY);
			XeAddCallback(LDialogW, XeNwmCallback,           _DCB_Close,     (EPointer)0);

			Widget  LTopMatrixW = XeFSBox_GetChild(LDialogW, XeDIALOG_CUSTOM_ITEM);
			EXeStart;
			EXeSetArg(XeNshadowType, XeShadowETCHED_IN);
			EXeSetArg(XeNorientation, XeVERTICAL);
			EXeSetArg(XeNshadowThickness, 2);
			EXeSetArg(XeNmarginSizeX, 7);EXeSetArg(XeNmarginSizeY, 5);
			EXeSetArg(XeNspacingX, 12);EXeSetArg(XeNspacingY, 7);
			Xe_SetValues(LTopMatrixW,  Args, ArgCnt);

			const EIndex  LN = E3d_3DSceneFileFormats_Count;

			{
				EIndex  LC = 0;for(;LC < LN;LC++)
				{
					E3dIOModule*  LIOModule = E3d_3DSceneFileFormats[LC];

					if(LIOModule->Write) LNumOfFormatsFound++;
				}
			}

			ENewN(LNumOfFormatsFound, LOptionMenuItems);

			{
				EIndex  LC = 0;XeOptionMenuRec*  LOptionMenuItem = LOptionMenuItems;for(;LC < LN;LC++)
				{
					E3dIOModule*	LIOModule = E3d_3DSceneFileFormats[LC];if(LIOModule->Write)
					{
						LOptionMenuItem->Name = EStrDup(LIOModule->Name);
						LOptionMenuItem->Callback = NULL;
						LOptionMenuItem->Value = (EPointer)LIOModule;
						LOptionMenuItem->ClientData = (EPointer)LIOModule;
						LOptionMenuItem->Icon = NULL;
						LOptionMenuItem++;
					}
				}
			}

			if(LNumOfFormatsFound > 0) Eqx_Prefs.SceneSaveFormat = E3d_3DSceneFileFormats[0];
			else Eqx_Prefs.SceneSaveFormat = NULL;


			{
				EXeStart;
				EXeSetArg(XeNorientation, XeHORIZONTAL);
				EXeSetArg(XeNspacingX, 12);
				EXeSetArg(XeNwidthAdjustable, FALSE);EXeSetArg(XeNalignmentX, XeAlignBEGINNING);
				Widget  LMatrixW = XeWidget_ManagedNew("LMatrixW", XeMatrixWidget_class, LTopMatrixW,  Args, ArgCnt);

				EXeStart;
				EXeSetArg(XeNbackgroundColorPtr, &Xe_BackgroundRGBAiColor);
				EXeSetArg(XeNforegroundColorPtr, &Xe_ForegroundRGBAiColor);
				EXeSetArg(XeNbuttonColorPtr, &Xe_ButtonRGBAiColor);
				EXeSetArg(XeNfontList, XeMenu_FontList);
				EXeSetArg(XeNlabelFontList, Xe_LabelFontList);
				EXeSetArg(XeNspacingX, 4);
				EXeSetArg(XeNmenuPaneBackgroundColor, &XeOptionMenu_RGBAiColor);
				EXeSetArg(XeNwidthAdjustable, FALSE);
				EXeSetArg(XeNalignmentX, XeAlignBEGINNING);
				_SaveFormatsW = XeOptionMenu_CreateAndInitManaged(LMatrixW, NULL, LOptionMenuItems, LNumOfFormatsFound, (EPointer)(Eqx_Prefs.SceneSaveFormat), "Format", _CB_3DSceneSaveFormat, NULL,  Args, ArgCnt);

				EIndex  LC = 0;XeOptionMenuRec*  LOptionMenuItem = LOptionMenuItems;for(;LC < LNumOfFormatsFound;LC++, LOptionMenuItem++)
				{
					if(LOptionMenuItem->Name) EFree(LOptionMenuItem->Name);
				}
				EFree(LOptionMenuItems);

				if(E3d_3DSceneFileFormats_Count)
				{
// E3dFIXME
// This should be freed when the app exits
//
					ENewN(E3d_3DSceneFileFormats_Count, _SaveSceneFormatWidgets);

					for(LC = 0;LC < LN;LC++)
					{
						E3dIOModule*	LIOModule = E3d_3DSceneFileFormats[LC];

						if(LIOModule->GUIForSave)
						{
							EXeStart;
							EXeSetArg(XeNorientation, XeVERTICAL);
							_SaveSceneFormatWidgets[LC].MatrixW = XeWidget_ManagedNew("3DIOMtx", XeMatrixWidget_class, LTopMatrixW,  Args, ArgCnt);
						}
						else
						{
							_SaveSceneFormatWidgets[LC].MatrixW = NULL;
						}

						_SaveSceneFormatWidgets[LC].IOModule = LIOModule;
					}

					E3dp_IOModuleShowGUI(E3d_3DSceneFileFormats[0], _SaveSceneFormatWidgets, E3d_3DSceneFileFormats_Count, NULL);
				}


				EXeStart;
				EXeSetArg(XeNshadowThickness, 0);
				EXeSetArg(XeNhighlightThickness, 0);
				EXeSetArg(XeNset, Eqx_Prefs.SceneFormatFromExtension);
				_FromExtensionBW = LW = XeWidget_ManagedNew("From extension", XeToggleButton_class, LMatrixW,  Args, ArgCnt);
				XeAddCallback(LW, XeNvalueChangedCallback, _DCB_FromExtension, (EPointer)NULL);
			}

			XeWidget_Manage(LTopMatrixW);

			XeShell_CenterToWidget(LDialogW, Eqx_TopShellW);
			XeShell_Raise(LDialogW);
		}

	}
	else XeShell_Raise(_SaveSceneDialogW);
}


//========================================
// Key callback
//========================================
static void E3dKCB_SaveScene(const EKeyCode PKey, const unsigned int PModifiers, const int PKeyEventType, EPointer PClientData)
{
	E3dMCB_SaveScene(NULL, PClientData, (EPointer)NULL);
}


typedef struct
{
	EpImage*  Image;
	EListA(, E3d2DTexture*, Textures);
	EBool  NeedsReload;
} E3dImageAndTxs;

static E3dImageAndTxs* _FindTx(E3dImageAndTxs* PS, const EIndex PNS, E3d2DTexture* PTexture)
{
	EIndex  LIC = 0;E3dImageAndTxs*  LS = PS;for(;LIC < PNS;LIC++, LS++)
	{
		EIndex  LTC = 0;for(;LTC < LS->Textures_Count;LTC++)
		{
			if(LS->Textures[LTC] == PTexture) return(LS);
		}
	}
	return(NULL);
}


//========================================
// Reload changed image files
//========================================
static void E3dMCB_ReloadChanged(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dScene*  LScene = E3d_Scene;

	EListAI(, E3dImageAndTxs, LImages);

	{
		EIndex  LC = 0;for(;LC < LScene->Textures_Count;LC++)
		{
			E3d2DTexture*  LTexture = LScene->Textures[LC];

			if(LTexture->Image && LTexture->Image->FileName)
			{
				E3dImageAndTxs*  LInTs = _FindTx(LImages, LImages_Count, LTexture);if(LInTs)
				{
					EList_PointerAddAChk(LInTs->Textures, 16, LTexture);
				}
				else
				{
					LInTs = EList_StructAddAClr(LImages, 16);
					LInTs->Image = (EpImage*)LTexture->Image;
					EList_PointerAddAChk(LInTs->Textures, 16, LTexture);
				}
			}
		}
	}

	if(LImages)
	{
		EIndex  LReloadC = 0;

		{
			EIndex  LIC = 0;E3dImageAndTxs*  LInTs = LImages;for(;LIC < LImages_Count;LIC++, LInTs++)
			{
				EpImage*  LImage = LInTs->Image;

				if(LImage->Loaded_ms >= 0)
				{
					struct stat  LFStat;if(lstat(LImage->FileName, &LFStat) == 0)
					{
						const ETime  LMTime_ms = ((long long)LFStat.st_mtime) * 1000;
//printf("%s() %s:%d  %s loaded at %lld  mtime %lld\n", __FUNCTION__, __FILE__, __LINE__, LImage->FileName, (long long)LImage->Loaded_ms, (long long)LMTime_ms);fflush(stdout);
						if(LMTime_ms > LImage->Loaded_ms) { LInTs->NeedsReload = TRUE;LReloadC++; }
					}
				}
			}
		}

		if(LReloadC > 0)
		{
			EBool  LRenderersStopped = FALSE;
			EIndex  LProggressC = 0;

			EIndex  LIC = 0;E3dImageAndTxs*  LInTs = LImages;for(;LIC < LImages_Count;LIC++, LInTs++)
			{
				if(LInTs->NeedsReload)
				{
					if(!LRenderersStopped) { E3d_RenderersStop();LRenderersStopped = TRUE; }

					E3dProgressIndicator_SetMessage("Reloading %s", LInTs->Textures[0]->FileName);E3dProgressIndicator_Set((float)LProggressC / (float)LReloadC, NULL, (EPointer)NULL);XeApp_Sync(E3dp_XeApp);

					E3dScene_TextureImageLoad(LScene, LInTs->Textures[0], E3dp_Prefs.TexturePath, E3dp_Prefs.CurrentFrame, TRUE,  NULL);LReloadC++;  // Do Image force reload for the first Texture, then just update the Texture for the rest
					EIndex  LTC = 1;for(;LTC < LInTs->Textures_Count;LTC++)
					{
						E3dScene_TextureImageLoad(LScene, LInTs->Textures[LTC], E3dp_Prefs.TexturePath, E3dp_Prefs.CurrentFrame, FALSE,  NULL);LReloadC++;
						E3dProgressIndicator_SetMessage("Reloading %s", LInTs->Textures[LTC]->FileName);E3dProgressIndicator_Set((float)LProggressC / (float)LReloadC, NULL, (EPointer)NULL);XeApp_Sync(E3dp_XeApp);
					}
				}
			}
			E3dProgressIndicator_Set((float)LProggressC / (float)LReloadC, NULL, (EPointer)NULL);XeApp_Sync(E3dp_XeApp);
			E3dProgressIndicator_SetMessage(NULL);E3dProgressIndicator_Set(-1.0, NULL, (EPointer)NULL);XeApp_Sync(E3dp_XeApp);
		}

		{
			EIndex  LIC = 0;E3dImageAndTxs*  LInTs = LImages;for(;LIC < LImages_Count;LIC++, LInTs++) EList_FreeA(LInTs->Textures);
		}

		EList_FreeA(LImages);

		E3dWindows_RedrawByChangeFlags(E3dChgMATERIALPARAM|E3dChgTEXTURE_MAPPING);
	}
}


//========================================
// Reload changed image files
//========================================
static void E3dKCB_ReloadChanged(const EKeyCode PKey, const unsigned int PModifiers, const int PKeyEventType, EPointer PClientData)
{
	E3dMCB_ReloadChanged(NULL, PClientData, (EPointer)NULL);
}


//========================================
// Entry point of the plugin
//========================================
int E_PluginInit(EPlugin* PPlugin)
{
	_Plugin = PPlugin;

//	E3d_MenuButton = XePushButton_Add("Menu->File", "Save selected hierarchies", '\0', NULL, NULL, FALSE, NULL, EqxMCB_SaveSelectedHierarchies, (EPointer)0);

	const EKeyModifiers  LCtrlMask = Xe_CommandKeyForCtrl ? E_CommandModifierMask : EControlMask;const char*  LCtrlName = Xe_CommandKeyForCtrl ? "Cmd" : "Ctrl";char                 LAccText[256];

	sprintf(LAccText, "Shift+%s+S", LCtrlName);E3d_SaveSceneMenuButton = XePushButton_Add("Menu/File", "Save scene", '\0', NULL, LAccText, FALSE, NULL, E3dMCB_SaveScene, (EPointer)0);
	E_KeySetCallback("Save scene", 'S', LCtrlMask|EShiftMask, EKeyPRESS, E3dKCB_SaveScene, (EPointer)0);

	sprintf(LAccText, "%s+R", LCtrlName);E3d_SaveSceneMenuButton = XePushButton_Add("Menu/File", "Reload changed", '\0', NULL, LAccText, FALSE, NULL, E3dMCB_ReloadChanged, (EPointer)0);
	E_KeySetCallback("Reload changed", 'R', LCtrlMask, EKeyPRESS, E3dKCB_ReloadChanged, (EPointer)0);

	return(E_SUCCESS);
}


//========================================
// Exit method of the plugin
//========================================
int E_PluginExit(const EBool PExiting)
{
	if(!PExiting || E3dp_GUIFreeOnExit)
	{
		XeItemFree(E3d_MenuButton);
		XeItemFree(E3d_SaveSceneMenuButton);
	}

	E_KeyRemoveCallback('S', EControlMask|EShiftMask, EKeyPRESS, E3dKCB_SaveScene, (EPointer)0);
	E_KeyRemoveCallback('R', EControlMask, EKeyPRESS, E3dKCB_ReloadChanged, (EPointer)0);

	return(E_SUCCESS);
}
