//================================================================
// E3D
//
// - Cg definitions
//
// AUTHOR:  Gabor Nagy
// DATE:    2019-Jun-23 19:15:55
//
// Copyright (C) 2014-2018 by Gabor Nagy. All rights reserved.
//================================================================
#ifndef _E3D_Cg_h
#define _E3D_Cg_h

#ifdef __APPLE__
#include <cgGL.h>
#else
#include <Cg/cgGL.h>
#endif

#include <E3D/GL/Draw.h>
#include "../PGPURenderer.h"

#define E3dThisShaderVariant  E3dCgShaderVariant
#define E3dThisVertexProgram  E3dCgVertexProgram

//=====================================================
// Resolve a uniform parameter from a Cg file
//=====================================================
#define E3dSLM_GetNamedParameterFrag(mMaterial, mParameterName, mParameterOut, mErrorOnFail) do\
{\
	E3dSL_DebugRec.ParameterName = mParameterName;\
	mParameterOut = cgGetNamedParameter(LFragmentProgram->CgProgram, mParameterName);if(mErrorOnFail && !E3dSLParameterValid(mParameterOut))\
	{\
		LShaderVariant->CreationResult = E_UNDEFINED_SYMBOL;\
		E3d_Log(ELogWARNING, "Cg", "Material [%s:%s] Unable to retrieve fragment program parameter '%s' from shader '%s'", (mMaterial)->Shader->Class->Name, (mMaterial)->Name, mParameterName, LShaderVariant->Signature);\
		if(_SaveShaderSourceOn & E3dSL_SYMBOL_RESOLVE_ERROR) E3dSL_FrgSaveSource(mMaterial, LShaderVariant, ".cg");\
/*		return(NULL);*/\
	}\
} while(0)


#if defined (__cplusplus) && (!(defined (__CPLUSLIBS)))
extern "C" {
#endif


#if E3dFLOAT_BITS == 64
	#define E3dSLSetMatrix3x3Parameterc  cgGLSetMatrixParameterdc
	#define E3dSLSetMatrix3x3Parameterr  cgGLSetMatrixParameterdr
	#define E3dSLSetMatrix4x4Parameterc  cgGLSetMatrixParameterdc
	#define E3dSLSetMatrix4x4Parameterr  cgGLSetMatrixParameterdr
#else
	#if E3dFLOAT_BITS == 32
		#define E3dSLSetMatrix3x3Parameterc  cgGLSetMatrixParameterfc
		#define E3dSLSetMatrix3x3Parameterr  cgGLSetMatrixParameterfr
		#define E3dSLSetMatrix4x4Parameterc  cgGLSetMatrixParameterfc
		#define E3dSLSetMatrix4x4Parameterr  cgGLSetMatrixParameterfr
	#else
ERROR
	#endif
#endif




//========================================================
// Resolve a uniform parameter from a vertex program
//========================================================
#define E3dSLM_GetNamedParameterVP(mVertexProgram, mParameterName, mParameterOut)\
do {\
	mParameterOut = cgGetNamedParameter(mVertexProgram->CgProgram, mParameterName);\
} while(0)


//========================================================
// Resolve a uniform parameter from a vertex program
//========================================================
#define E3dSLM_GetNamedParameterVPErrorCheck(mVertexProgram, mParameterName, mParameterOut)\
do {\
	mParameterOut = cgGetNamedParameter(mVertexProgram->CgProgram, mParameterName);if(!E3dSLParameterValid(mParameterOut))\
	{\
		if(E3dSL_DebugRec.Material)\
		{\
			E3dMaterial*  lMaterial = E3dSL_DebugRec.Material;\
			E3d_Log(ELogWARNING, "Cg", "Material [%s] {shader class:%s} Unable to retrieve vertex-program parameter '%s' from file '%s'", lMaterial->Name, lMaterial->Shader->Class->Name, mParameterName, LFileName);\
		}\
		else E3d_Log(ELogWARNING, "Cg", "Unable to retrieve vertex-program parameter '%s' from file '%s'", mParameterName, LFileName);\
		return(-1);\
	}\
} while(0)


//================================================================
// Resolve a vertex attribute parameter from a vertex program
//================================================================
#define E3dSLM_GetNamedAttributeVPErrorCheck(mVertexProgram, mParameterName, mParameterOut)  E3dSLM_GetNamedParameterVPErrorCheck(mVertexProgram, mParameterName, mParameterOut)


// Vertex program
//
typedef struct
{
	E3dStageProgramCore;

	CGprogram     CgProgram;

	CGparameter   Param_EyeLocalPosition,
	              TextureMapperMatrixParam[E3dSLMAX_NUM_OF_2DTEXTURES],  // Planar texture mappers use matrices
	              TexTangentParam[E3dSLMAX_NUM_OF_2DTEXTURES],           // Texture-space tangent for normalmap
	              Param_VtA_Tangent;

	CGparameter	  Param_VtU_ModelViewProjectionMatrix;

	CGparameter	  LocalToWorldMatrixParam,
	              LocalToWorldUnitVectorMatrixParam,
	              ModelViewMatrixParam,
	              ShadowMapProjMatrixParam[E3dSLMAX_NUM_OF_LIGHTS];
	CGparameter*  ResourceParameters;EIndex  NumOfResourceParameters;

	EBool         ResourceParametersStatic;
} E3dCgVertexProgram;


// Fragment program
//
typedef struct
{
	E3dStageProgramCore;

	CGprogram  CgProgram;
} E3dCgFragmentProgram;

typedef struct
{
	CGparameter  Param_Position, Param_SpotDirection, Param_SpotCutoffAngleCos, Param_SpotAngle, Param_SpotPenumbraAngle, Param_SpotExponent, Param_Color, Param_DiffuseSpecularFactors, Param_ShadowMap;
} E3dCgLight;

#define E3dSLLight  E3dCgLight

typedef struct
{
	E3dShaderVariantCore;

// Fragment program
//
	CGparameter  Param_EyeLocalPosition, Param_EyeWorldPosition,
	             AlphaParam,
	             Fresnel0_RGBParam;                // Fresnel reflectance at normal direction

	CGparameter  NumOfLightsParam;
	E3dCgLight   Lights[E3dSLMAX_NUM_OF_LIGHTS];

	CGparameter  TextureParam[E3dSLMAX_NUM_OF_2DTEXTURES],
	             TextureBRDFSamplesSqrtParam[E3dSLMAX_NUM_OF_2DTEXTURES], TextureEnvLightFactorParam[E3dSLMAX_NUM_OF_2DTEXTURES],
	             Eta[E3dSLMAX_NUM_OF_2DTEXTURES],     // Relative indices of refraction for refraction cube-maps
	             NormalMapParam[E3dSLMAX_NUM_OF_2DTEXTURES],
	             LocalToWorldUnitVectorMatrixParam,
	             FragmentUniformParams[E3dSLMAX_NUM_OF_PARAMETERS];

	CGparameter**  TextureBlendingParameters;     // For each [Texture][FragmentUniformParam]
} E3dCgShaderVariant;


typedef struct
{
	E3dShaderRendererBindingCore;

	E3dCgShaderVariant*  CgShader;
} E3dShaderCgBinding;



typedef struct
{
	E3dPGPURendererCore;
} E3dCgRenderer;



extern CGcontext	E3dCg_ShaderContext;

extern CGprofile	E3dCg_VertexProfile;
extern CGprofile	E3dCg_FragmentProfile;

extern void		E3dCg_FragmentProgramFree(E3dCgFragmentProgram* PProgram);

#if defined (__cplusplus) && (!(defined (__CPLUSLIBS)))
}	// Close scope of 'extern "C"' declaration which encloses file.
#endif

#endif	// _E3D_Cg_h
