# This script generates images to be used with EGUIM_AddPushButtonImg3_shadow
# It also expects separate Direct ("*-D.tif") and Indirect ("*-I.tif") illumination images,
# in order to apply more accurate indirect light on the ground plane
#!/bin/sh +xv
#set -o verbose
export IMAGE=$1
export VERBOSE=-v

# Copy alpha from image into a separate file
#$iSPL $VERBOSE  -i Icons/images/$IMAGE.sgi -oA tmp/.sgi

# Multily image with its alpha
#$iCM  $VERBOSE  -M   -i Icons/images/$IMAGE.sgi         -i tmp/$IMAGE.sgi  -o tmp/RGB.sgi

# Subtract the result from the image
#$iCM  $VERBOSE  -S   -i Icons/images/$IMAGE-shadow.sgi  -i tmp/$IMAGE.sgi  -o tmp/A.sgi

# Brighten up alpha, so shadows won't crop around the image edges
#$iCN $EITHREADS  $VERBOSE  -c 1.03  tmp/A.sgi  .sgi

# Merge image channels
#$iCM  $VERBOSE  -iRGB tmp/RGB.sgi  -iA tmp/A.sgi  -o tmp/A.sgi


# Ground rendered with Constant with shadow shader : direct lighting : black ground, indirect : 0.75 Diffuse.
# This preserves indirect light on the ground and eliminates the need for multiplying image by alpha
$iSPL $VERBOSE  -i Icons/images/$IMAGE.sgi -oA tmp/A-.sgi

# Multiply indirect light by 0.1 where alpha = 0 (on the ground)
$iCN $EITHREADS  $VERBOSE  -b 0.1 tmp/A-$IMAGE.sgi tmp/AO.sgi
$iCM  $VERBOSE  -M -i Icons/images/$IMAGE-I.tif  -i tmp/AO.sgi  -o tmp/New-I.tif

$iCM  $VERBOSE  -i Icons/images/$IMAGE-D.tif  -i tmp/New-I.tif  -o tmp/$IMAGE.tif
$iCN $EITHREADS  $VERBOSE  -g 1.3 tmp/$IMAGE.tif tmp/$IMAGE.sgi

# Brighten up alpha, so shadows won't crop around the image edges
$iCM  $VERBOSE  -S   -i Icons/images/$IMAGE-shadow.sgi  -i tmp/A-$IMAGE.sgi  -o tmp/A.sgi
$iCN $EITHREADS  $VERBOSE  -c 1.15 -g 1.8  tmp/A.sgi  .sgi
$iCM  $VERBOSE  -iRGB tmp/$IMAGE.sgi  -iA tmp/A.sgi  -o tmp/.sgi

$iS $EITHREADS   $VERBOSE  -s $2 $3  -gaussian -b 0.8  tmp/$IMAGE.sgi         tmp/$IMAGE.c
$iS $EITHREADS   $VERBOSE  -s $4 $5  -gaussian -b 0.8  tmp/$IMAGE.sgi         tmp/$IMAGE-Big.c


$iSPL $VERBOSE  -i Icons/images/$IMAGE-Active.sgi -oA tmp/.sgi
$iCM  $VERBOSE  -M   -i Icons/images/$IMAGE-Active.sgi         -i tmp/$IMAGE-Active.sgi  -o tmp/RGB.sgi
$iCM  $VERBOSE  -S   -i Icons/images/$IMAGE-Active-shadow.sgi  -i tmp/$IMAGE-Active.sgi  -o tmp/A.sgi
$iCN  $VERBOSE  -c 1.03  tmp/A.sgi  .sgi
$iCM  $VERBOSE  -iRGB tmp/RGB.sgi  -iA tmp/A.sgi  -o tmp/A.sgi

$iS $EITHREADS   $VERBOSE  -s $2 $3  tmp/RGBA.sgi         tmp/$IMAGE-Active.c
$iS $EITHREADS   $VERBOSE  -s $2 $3  tmp/RGBA.sgi         tmp/$IMAGE-Armed.c

cat tmp/$IMAGE.c tmp/$IMAGE-Active.c tmp/$IMAGE-Armed.c tmp/$IMAGE-Big.c > tmp.c
