//========================================================================
// Color library header file
//
// AUTHOR:  Gabor Nagy
// DATE:    2019-Jun-23 19:50:18
//
// Copyright (C) 1994-2021 by Gabor Nagy. All rights reserved.
//========================================================================
#ifndef	_EColor_h
#define	_EColor_h

#include <EClass.h>
#include <EResource.h>

#define EcAUTO_DEPTH     0

#define EcCOLOR_CHANGED  1

#define EcLumaR    0.212671
#define EcLumaG    0.715160
#define EcLumaB    0.072169

#define EcLumaI8R  54
#define EcLumaI8G  183
#define EcLumaI8B  19

#define EcLumaI13R  1742
#define EcLumaI13G  5859
#define EcLumaI13B  591

#define EcLumaI16R  13937
#define EcLumaI16G  46868
#define EcLumaI16B  4730


/*
#define EcLumaR   0.3086
#define EcLumaG   0.6094
#define EcLumaB   0.0820

#define EcLumaIR  79
#define EcLumaIG  156
#define EcLumaIB  21
*/

//--------------------------------
// Macros
//--------------------------------
#define Ec_ClampTo0_RGBf_s(m) do\
{\
	if(m.R < 0.0) m.R = 0.0;\
	if(m.G < 0.0) m.G = 0.0;\
	if(m.B < 0.0) m.B = 0.0;\
} while(0)

#define Ec_ClampTo0_3f(m) do\
{\
	int  lC = 0;for(;lC < 3;lC++) if(m[lC] < 0.0) m[lC] = 0.0;\
} while(0)

#define Ec_ClampTo0_1_RGBf_s(m)\
{\
	if(m.R < 0.0) m.R = 0.0;else if(m.R > 1.0) m.R = 1.0;\
	if(m.G < 0.0) m.G = 0.0;else if(m.G > 1.0) m.G = 1.0;\
	if(m.B < 0.0) m.B = 0.0;else if(m.B > 1.0) m.B = 1.0;\
} while(0)

#define Ec_ClampTo0_1_3f(m) do\
{\
	int  lC = 0;for(;lC < 3;lC++) { if(m[lC] < 0.0) m[lC] = 0.0;else if(m[lC] > 1.0) m[lC] = 1.0; }\
}

#define Ec_Float32ToUInt8(mIn, mOut) do\
{\
	float  lOut = (mIn) * 255.999;if(!EFinite(lOut)) lOut = 0.0;if(lOut < 0.0) lOut = 0.0;else if(lOut > 255.0) lOut = 255.999;\
	mOut = (uint8_t)lOut;\
} while(0)


#define EcM_LumaI8(r, g, b)      ((EcLumaI8R  * (r) + EcLumaI8G  * (g) + EcLumaI8B  * (b)) >> 8)

// We are using 13-bit luma weights to avoid overflow on signed int32_t, as we lose a bit to the sign and 2 bits because we are adding 3 values here
//
#define EcM_LumaI16(r, g, b)     ((EcLumaI13R * (r) + EcLumaI13G * (g) + EcLumaI13B * (b)) >> 13)
#define EcM_LumaI16to8(r, g, b)  ((EcLumaI8R  * (r) + EcLumaI8G  * (g) + EcLumaI8B  * (b)) >> 24)

// 16-bit luminance from RGB8 source. This gives better precision than simply up-shifting EcM_LumaI8 because it doest the weighted average of RGB on 14 bits.
// (not 16 bit, to avoid overflow on signed int32_t, as we lose a bit to the sign and 2 bits because we are adding 3 values here)
#define EcM_LumaI8to16(r, g, b)  ((EcLumaI13R * (r) + EcLumaI13G * (g) + EcLumaI13B * (b)) >> 5)

#define EcM_LumaF(r, g, b)       (EcLumaR * (r) + EcLumaG * (g) + EcLumaB * (b))
#define EcM_LumaRGBA(RGBA)       ((EcLumaI8R * (((RGBA) >> 24) & 0xFF) + EcLumaI8G * (((RGBA)>>16) & 0xFF) + EcLumaI8B * (((RGBA) >> 8) & 0xFF)) >> 8)
#define EcM_ABGRToRGBA(ABGR)     (((ABGR & 0xFF000000) >> 24) | ((ABGR & 0x00FF0000) >> 8) | ((ABGR & 0x0000FF00) << 8) | ((ABGR & 0x000000FF) << 24))



#define Ec_YUV8toRGB8_sp(mY, mU, mV,  mDst) do\
{\
	const float  lY = 1.1643 * ((float)mY - 16), lU = (float)mU - 128.0, lV = (float)mV - 128.0;\
	int  lR = lY                + 1.5980 * lV;if(lR < 0) lR = 0;if(lR > 255) lR = 255;\
	int  lG = lY - 0.39173 * lU - 0.8129 * lV;if(lG < 0) lG = 0;if(lG > 255) lG = 255;\
	int  lB = lY + 2.01700 * lU;              if(lB < 0) lB = 0;if(lB > 255) lB = 255;\
\
	mDst->R = lR;\
	mDst->G = lG;\
	mDst->B = lB;\
} while(0)

#define Ec_AlphaMult_s(mRGBA) do\
{\
	const float  lA = (mRGBA).A;\
	(mRGBA).R *= lA;\
	(mRGBA).G *= lA;\
	(mRGBA).B *= lA;\
} while(0)

#define Ec_AlphaMultSetA_s(mRGBA, mA) do\
{\
	(mRGBA).R *= mA;\
	(mRGBA).G *= mA;\
	(mRGBA).B *= mA;\
	(mRGBA).A = mA;\
} while(0)


//--------------------------------
// OpenGL color macros
//--------------------------------
#define EcM_glColorRGBAp8(c)      E3glColor4ub(EcM_R_RGBAp8(c), EcM_G_RGBAp8(c), EcM_B_RGBAp8(c), EcM_A_RGBAp8(c))
#define EcM_glColorRGBA8_s(c)     E3glColor4ub((c).R, (c).G, (c).B, (c).A)
#define EcM_glColorRGBAi_s(c)     E3glColor4f((float)((c).R) * 0.0000153,  (float)((c).G) * 0.0000153,  (float)((c).B) * 0.0000153,  (float)((c).A) * 0.0000153)
#define EcM_glColorRGBAi_p(c)     E3glColor4f((float)((c)->R) * 0.0000153, (float)((c)->G) * 0.0000153, (float)((c)->B) * 0.0000153, (float)((c)->A) * 0.0000153)

#define EcM_glColorRGBAf_s(c)     E3glColor4f((c).R,  (c).G,  (c).B,  (c).A)
#define EcM_glColorRGBAf_p(c)     E3glColor4f((c)->R, (c)->G, (c)->B, (c)->A)
#define EcM_glColorRGBAf_a(c)     E3glColor4fv(c)

#define EcM_glClearColorRGBAi(c)  E3glClearColor((float)((c).R) * 0.0000153, (float)((c).G) * 0.0000153, (float)((c).B) * 0.0000153, (float)((c).A) * 0.0000153)
#define EcM_glClearColorRGBf(c)   E3glClearColor((c).R, (c).G, (c).B, 1.0)
#define EcM_glClearColorRGBAf(c)  E3glClearColor((c).R, (c).G, (c).B, (c).A)



#if defined (__cplusplus) && (!(defined (__CPLUSLIBS)))
extern "C" {
#endif


// Packed integer color
// This is adjusted, depending on endianness, so it's always in RGBA order in memory.
//
typedef uint32_t  EcRGBAp8Color;


#if BYTE_ORDER == LITTLE_ENDIAN
#define EcM_PackRGBA8(r,g,b,a)  ((((EcRGBAp8Color)(r))) | (((EcRGBAp8Color)(g)) << 8) | (((EcRGBAp8Color)(b)) << 16) | ((EcRGBAp8Color)(a) << 24))

static inline uint32_t Ec_RGBA8p(const uint32_t P)
{
	const uint32_t  LR = (P >> 24) & 0xFF , LG = (P >> 16) & 0xFF, LB = (P >> 8) & 0xFF, LA = P & 0xFF;
	return(EcM_PackRGBA8(LR, LG, LB, LA));
}

#define EcM_R_RGBAp8(RGBA)      ((RGBA)         & 0xFF)
#define EcM_G_RGBAp8(RGBA)      (((RGBA) >>  8) & 0xFF)
#define EcM_B_RGBAp8(RGBA)      (((RGBA) >> 16) & 0xFF)
#define EcM_A_RGBAp8(RGBA)      (((RGBA) >> 24) & 0xFF)
#else
#define EcM_R_RGBAp8(RGBA)      (((RGBA) >> 24) & 0xFF)
#define EcM_G_RGBAp8(RGBA)      (((RGBA) >> 16) & 0xFF)
#define EcM_B_RGBAp8(RGBA)      (((RGBA) >> 8)  & 0xFF)
#define EcM_A_RGBAp8(RGBA)      ((RGBA)         & 0xFF)
#define EcM_PackRGBA8(r,g,b,a)  ((((EcRGBAp8Color)(r)) << 24) | (((EcRGBAp8Color)(g)) << 16) | (((EcRGBAp8Color)(b)) << 8) | ((EcRGBAp8Color)(a)))
#define Ec_RGBA8p(m)            m
#endif // LITTLE_ENDIAN

#define EcM_Gamma(mV, mGamma)  pow(mV, 1.0 / mGamma)



typedef struct
{
	uint8_t  R, G, B;
} EcRGB8Color;


// Integer-component RGBA color 
// Normal range: 0...0xFF
//
typedef struct
{
	uint8_t  R, G, B, A;
} EcRGBA8Color;



// Integer-component RGB color 
// Normal range: 0...0xFFFF
//
typedef struct
{
	int  R, G, B;
} EcRGBiColor;


// Integer-component RGBA color 
// Normal range: 0...0xFFFF
//
typedef struct
{
	int  R, G, B, A;
} EcRGBAiColor;


// Float-component color
// Normal range:	0.0...1.0
typedef struct
{
	float  R, G, B;
} EcRGBf32Color;

typedef struct
{
	unsigned short  R, G, B, A;
} EcRGBA16Color;

typedef struct
{
	unsigned short  R, G, B;
} EcRGB16Color;

// Float-component color
// Normal range:	0.0...1.0
typedef struct
{
	float  R, G, B, A;
} EcRGBAf32Color;


// RGB -> Spectrum and Spectrum -> RGB conversion type
//
enum
{
	EcINTERPOLATE = 1, EcCIE
};


#define EcMAX_NUM_OF_SPECTRAL_SAMPLES  32


typedef struct
{
	ECoreClassCore(ECore);

	EcRGBf32Color  RGB;
	uint8_t        RGBConversion;    // EcINTERPOLATE or EcCIE

// Spectrum
//
	float*         Wavelengths;float*  Values;EIndex  Values_Count;  // User-defined (may be non-uniformly sampled)

// Regularly sampled spectrum
//
	int            NumOfValidSamples;
	float          Samples[EcMAX_NUM_OF_SPECTRAL_SAMPLES];           // The size is hard-wired for now because of the temporary copies of shaders, used for texturing
//float*         Samples;int  NumOfSamples;
} EcColor;


// Color Matching function by wavelength from
// http://cvrl.ioo.ucl.ac.uk/database/text/tables/table3ss.htm
//
// wavelength (nm)  R, G, B contributions
//
typedef struct
{
	float  Wavelength, R, G, B;
} EcCIEEntry;


typedef struct
{
	float  Wavelength, Sensitivity;
} EcColorSensor;


enum
{
	EcsRGB = 1,
	EcAdobeRGB
};



static inline void ERes_ResourceSetColorRGBf(void* PStructAddress, const EResource* PResource,  const float PR, const float PG, const float PB)
{
	if(PResource)
	{
// If PStructAddress is NULL, the addresses are absolute. Otherwise they're relative to PStructAddress.
//
		uint8_t*  LVariableAddress;
		if(PStructAddress) LVariableAddress = (EPointerUInt)PStructAddress + (uint8_t*)(PResource->Address);
		else LVariableAddress = (uint8_t*)(PResource->Address);

		switch(PResource->Type & (~(EResDYNAMIC|EResCOUNT|EResCOUNT_FUNC)))
		{
			case EResCOLOR:
				{
					EcColor*  LAddress = (EcColor*)LVariableAddress;
					LAddress->RGB.R = PR;
					LAddress->RGB.G = PG;
					LAddress->RGB.B = PB;
				}
			break;
			default:  printf("%s() %s:%d  %sIMPLEMENT ME  %04x%s\n", __FUNCTION__, __FILE__, __LINE__, EYellowCode, PResource->Type & (~(EResDYNAMIC|EResCOUNT|EResCOUNT_FUNC)), ENormalCode);fflush(stdout);break;
		}
	}
}


extern const EcCIEEntry     Ec_CIE[];extern const int  Ec_CIE_count;

extern const EcRGBf32Color  Ec_CIE_Inv_380to720_16[];
extern const EcRGBf32Color  Ec_CIE_Inv_380to720_32[];
extern const EcRGBf32Color  Ec_CIE_Inv_380to720_64[];

extern EResource            EcColor_resources[];
extern ECoreClass           EcColor_class[];

extern void  Ec_GetForegroundColorRGBAi(const EcRGBAiColor* PBackgroundColor, EcRGBAiColor* PForegroundColor);

extern void  EcColor_Init(EcColor* PColor);
extern void  EcColor_InitRGB(EcColor* PColor,  const float PR, const float PG, const float PB);
extern void  EcColor_SetRGB(EcColor* PColor,  const float PR, const float PG, const float PB);

extern EcColor*  Ec_ResourceGetColor(void* PStructAddress, const EResource* PResource);
extern void      Ec_ResourceSetColorRGB(void* PStructAddress, const EResource* PResource,  const float PR, const float PG, const float PB);

extern void  EcColor_FreeContents(EcColor* PSpectrum);
extern void  EcColor_Copy(const EcColor* PS, EcColor* PD);


// Spectrum.c
//
extern void  Ec_SpectrumSample(EcColor* PColor,  const float PWavelengthMin, const float PWavelengthMax, const int PNSpectralSamples);
extern void  Ec_RGBToSpectrum(EcColor* PColor, const int PNSpectralSamples, const float PScalerIn);
extern void  Ec_RGBToSpectrumCIE(EcColor* PColor, const int PNSpectralSamples, const float PScalerIn);
extern void  Ec_RGBToSpectrumInterp(EcColor* PColor, const int PNSpectralSamples, const float PScalerIn);

extern void  Ec_RGBtoXYZInit(const uint16_t PColorSpace, float* PRGBtoXYZ);

extern void  Ec_SetRGBAi(EcRGBAiColor* PColor, const int PR, const int PG, const int PB, const int PA);
extern void  Ec_SetRGBf32(EcRGBf32Color* PColor, const float PR, const float PG, const float PB);
extern void  Ec_SetRGBAf32(EcRGBAf32Color* PColor, const float PR, const float PG, const float PB, const float PA);

extern int   Ec_AlphaMultRGBAi(const EcRGBAiColor* PSrc, EcRGBAiColor* PDst);


extern int   Ec_LumaRGBAi(const EcRGBAiColor* PColor);

extern EcRGBAp8Color    Ec_GainRGBAp8(const EcRGBAp8Color PColor, const float PGain);

extern EcRGB8Color*     Ec_GainRGB8(const EcRGB8Color* PColor, const float PGain, EcRGB8Color* PColorRet);
extern EcRGBAiColor*    Ec_GainRGBAi(const EcRGBAiColor* PColor, const float PGain, EcRGBAiColor* PColorRet);
extern EcRGBf32Color*   Ec_GainRGBf(const EcRGBf32Color* PColor, const float PGain, EcRGBf32Color* PColorRet);
extern EcRGBAf32Color*  Ec_GainRGBAf(const EcRGBAf32Color* PColor, const float PGain, EcRGBAf32Color* PColorRet);

extern EcRGBf32Color*   Ec_ContrastRGBf(const EcRGBf32Color* PColor, const float PContrast, EcRGBf32Color* PColorRet);
extern EcRGBAf32Color*  Ec_ContrastRGBAf(const EcRGBAf32Color* PColor, const float PContrast, EcRGBAf32Color* PColorRet);


extern EcRGBAiColor*    Ec_SaturateRGBAi(const EcRGBAiColor* PColor, const float PSaturation, EcRGBAiColor* PColorRet);
extern EcRGBf32Color*   Ec_SaturateRGBf(const EcRGBf32Color* PColor, const float PSaturation, EcRGBf32Color* PColorRet);
extern EcRGBAf32Color*  Ec_SaturateRGBAf(const EcRGBAf32Color* PColor, const float PSaturation, EcRGBAf32Color* PColorRet);

extern void             Ec_FColorShadeBy(const EcRGBf32Color* PColor, EcRGBf32Color* PColorLum, EcRGBf32Color* PDstColor);

extern EcRGB8Color*     Ec_LerpRGB8(const EcRGB8Color* PColor1, const EcRGB8Color* PColor2, const float PFactor, EcRGB8Color* PDst);
extern EcRGBiColor*     Ec_LerpRGBi(const EcRGBAiColor* PColor1, const EcRGBiColor* PColor2, const float PFactor, EcRGBiColor* PDst);
extern EcRGBf32Color*   Ec_LerpRGBf(const EcRGBf32Color* PColor1, const EcRGBf32Color* PColor2, const float PFactor, EcRGBf32Color* PDst);

extern EcRGBAiColor*    Ec_LerpRGBAi(const EcRGBAiColor* PColor1, const EcRGBAiColor* PColor2, const float PFactor, EcRGBAiColor* PDst);
extern EcRGBAf32Color*  Ec_LerpRGBAf(const EcRGBAf32Color* PColor1, const EcRGBAf32Color* PColor2, const float PFactor, EcRGBAf32Color* PDst);

extern double           Ec_GetAngle(const double PXPos, const double PYPos);

extern EcRGB8Color*     Ec_RGBfToRGB8(const float PR, const float PG, const float PB, EcRGB8Color* PResult);
extern EcRGBAp8Color    Ec_RGBfToRGBAp8(const float PR, const float PG, const float PB);
extern EcRGBAp8Color    Ec_RGBAfVToRGBAp8(const float PR, const float PG, const float PB, const float PA);
extern void             Ec_RGBAfToRGBA8(const EcRGBAf32Color* PSrcColor,  EcRGBA8Color* PDstColor);
extern void             Ec_RGBAfVToRGBA8(const float PR, const float PG, const float PB, const float PA,  EcRGBA8Color* PDstColor);

extern EcRGBAiColor     Ec_RGBAp8ToRGBAi(const EcRGBAp8Color PColor);
extern EcRGBAf32Color   Ec_RGBAp8ToRGBAf32(const EcRGBAp8Color PColor);

extern void             Ec_RGBf32ToRGBAi(const EcRGBf32Color* PRGBfColor, EcRGBAiColor* PRGBAiColor);

extern EcRGBAp8Color    Ec_RGBAiToRGBAp8(const EcRGBAiColor* PColor);
extern EcRGBAp8Color    Ec_RGBAiToRGBAp8AlphaMult(const EcRGBAiColor* PColor);

extern EcRGBAp8Color    Ec_RGBAfToRGBAp8(const EcRGBAf32Color* PColor);
extern void             Ec_RGBAf32ToRGBA8(const EcRGBAf32Color* PRGBAf32Color, EcRGBA8Color* PRGBA8Color);
extern void             Ec_RGBAf32ToRGBAi(const EcRGBAf32Color* PRGBAf32Color, EcRGBAiColor* PRGBAiColor);

extern EcRGBA8Color     Ec_RGBAiToRGBA8(const EcRGBAiColor* PSrc);
extern void             Ec_RGBAiToRGBAf32(const EcRGBAiColor* PRGBAiColor, EcRGBAf32Color* PDst);
extern void             Ec_RGBAiToRGBf32(const EcRGBAiColor* PRGBAiColor, EcRGBf32Color* PDst);

extern void             Ec_RGBAiToRGBA16(const EcRGBAiColor* PSrc, EcRGBA16Color* PDst);
extern void             Ec_RGBAiToRGB16(const EcRGBAiColor* PSrc, EcRGB16Color* PDst);


extern void             Ec_HSVtoRGB(double PHue, const double PSaturation, const double PValue, int* PR, int* PG, int* PB);
extern void             Ec_HSVtoRGBf(double PHue, const double PSaturation, const double PValue, float* PR, float* PG, float* PB);
extern void             Ec_HLStoRGB(double PHue, const double PLevel, const double PSaturation, int* PR, int* PG, int* PB);

extern void             Ec_RGBtoHSV(const EcRGBf32Color* PColor, float* PH, float* PS, float* PV);
extern void             Ec_RGBAtoHSV(const EcRGBAf32Color* PColor, float* PH, float* PS, float* PV);

extern void             Ec_Matrix3x3Invert_f32(const float* const PM, float* PDst);

#if defined (__cplusplus) && (!(defined (__CPLUSLIBS)))
}	// Close scope of 'extern "C"' declaration
#endif

#endif // _EColor_h
