//========================================================================
// Color library GUI
//
// AUTHOR:  Gabor Nagy
// DATE:    2019-Jun-23 08:58:25
//
// Copyright (C) 1994-2021 by Gabor Nagy. All rights reserved.
//========================================================================
#ifndef	_EColorPanel_h
#define	_EColorPanel_h

#include <Color/ColorWheel.h>
#include <Image/Graphs.h>

#if defined (__cplusplus) && (!(defined (__CPLUSLIBS)))
extern "C" {
#endif

extern EcRGBAiColor  Ec_DefaultColors[];

#define Ec_MAX_NUM_OF_COLORS	65536	// Max 16-bit palette

// Color definition modes
//
enum
{
	EcSLIDERS_RGBA = 0,
	EcSLIDERS_HSVA
};

// Color-component representation modes
//
enum
{
	EcINTEGER_0_255 = 0,
	EcINTEGER_0_65535,
	EcFLOAT_0_1
};


typedef struct EcColorPanel_S  EcColorPanel;

typedef void  (*EcColorPanelCallbackFunc)(EcColorPanel*, EPointer, EPointer);



struct EcColorPanel_S
{
	Widget  TopMatrixW, HTopMatrixW,
	        ColorCmdMatrixW, CopyColW, ExgColW, SpreadColW, LightM,
	        PaletteFrameW, PaletteDrawW,
	        DrawColorsFrameW, ColorSwatchDrawW,
	        RScaleW, GScaleW, BScaleW, IScaleW, AScaleW,
	        SpectrumPanelW,
	        SpectrumFrameW, SpectrumDrawW, SpectrumSpectrumW, SpectrumScaleW,
	        SpectrumXW, SpectrumYW;

	EpGraphDisplay*  SpectrumGraphDisplay;
	EpLineGraph*     SpectrumGraph;
	EpImage*         SpectrumBandsImage;
	float            SpectrumMin, SpectrumYMult;

	EcColorPanelCallbackFunc  ColorChangedCallback;
	EPointer                  ColorChangedCallbackClientData;
	int                       ComponentType;
	int                       NumOfAllocatedColors;    // Number of colors allocated
	int                       NumOfColors;             // Current number of colors and bitplane depth in the current palette
	EcRGBAiColor*             RGBAiColors;             // The array of colors in the palette
	EcRGBAf32Color*           RGBAfColors;             // The array of colors in the palette

	float                     WavelengthMin, WavelengthMax;  // In nanometers
	int                       NumOfSpectralSamples;          // For displaying spectrum of an RGB color
	EcColor                   Color;                         // RGB and spectrum for the current color

	float                     RGBtoXYZ[16], XYZtoRGB[16];

	EInt32_2                  ColorButtonSize;
	EInt32_2                  PaletteSize;
	int                       PaletteColumns, PaletteRows;

	int                       DrawColor0, DrawColor1, DrawColor2;
	EcColorWheel*             ColorWheel;
	EBool                     AlphaOn,
	                          Truecolor,               // All the drawing widgets have truecolor visual
	                          PaletteValid;

#ifdef EX11
	Colormap                  X_Colormap;              // Associated X Colormap of this ColorPanel
	Pixel*                    X_PixelValues;           // Pointer to the current array of X11 color pixel values
#endif
};


typedef struct
{
	int   reason;
	int   NumOfColors;  // Number of colors that were changed
	int*  ColorIDs;     // IDs of colors that were changed
} EcColorPanelCallbackStruct;


extern EcRGBAiColor  Ec_ThermoColorR, Ec_ThermoColorG, Ec_ThermoColorB, Ec_ThermoColorW;


extern EpImage*  Ec_RedRampImage;
extern EpImage*  Ec_GreenRampImage;
extern EpImage*  Ec_BlueRampImage;
extern EpImage*  Ec_IntensityRampImage;
extern EpImage*  Ec_AlphaRampImage;


extern EcColorPanel*  EcColorPanel_NewVa(Widget PParentW, char* PName, ...);
extern void           EcColorPanel_Free(EcColorPanel* PColorPanel);
extern void           EcColorPanel_Exit(void);

extern void    EcColorPanel_SetDrawColor(EcColorPanel* PColorPanel, const int PColorIndex, const int PColorID);

extern void    EcColorPanel_SetColor(EcColorPanel* PColorPanel, const int PColorIndex, const EcColor* PColor);

extern void    EcColorPanel_SetRGBfColor(EcColorPanel* PColorPanel, const int PColorIndex, const EcRGBf32Color* PRGBfColorIn);
extern void    EcColorPanel_SetRGBAfColor(EcColorPanel* PColorPanel, const int PColorIndex, const EcRGBAf32Color* PRGBAfColorIn);
extern void    EcColorPanel_SetRGBAiColor(EcColorPanel* PColorPanel, const int PColorIndex, const EcRGBAiColor* PRGBAiColorIn);

extern Widget  EcColorSwatch_NewRGBf(char* PName, Widget PParentW, Widget* PLabeWRet, EcRGBf32Color* PRGBfColor, XeCallbackFunc PCallback, EPointer PClientData);
extern Widget  EcColorSwatch_NewRGBAi(char* PName, Widget PParentW, Widget* PLabeWRet, EcRGBAiColor* PRGBAiColor, XeCallbackFunc PCallback, EPointer PClientData);
extern Widget  EcColorSwatch_NewRGBAf32(char* PName, Widget PParentW, Widget* PLabeWRet, EcRGBAf32Color* PRGBAf32Color, XeCallbackFunc PCallback, EPointer PClientData);

extern EBool   EcColorPanel_Configure(EcColorPanel* PColorPanel, const int PNumOfColors, const int PDepth, EcRGBAiColor* PColors);
extern EBool   EcColorPanel_Redraw(EcColorPanel* PColorPanel);

#if defined (__cplusplus) && (!(defined (__CPLUSLIBS)))
}	// Close scope of 'extern "C"' declaration
#endif

#endif // _EColorPanel_h
