//================================================================
// E3D library common things
//
// - Callbacks
// - Undoable operators core
// - System feature registry
//
// AUTHOR:  Gabor Nagy
// DATE:    2016-Dec-14 04:00:22
//
// Copyright (C) 1994-2021 by Gabor Nagy. All rights reserved.
//================================================================
#ifndef _E3D_h
#define _E3D_h

#ifndef _E3D_Types_h
#include <E3D/Types.h>
#endif


// Undoable operators
//
#ifndef _EUndo_h
#include <EUndo.h>
#endif


#if defined (__cplusplus) && (!(defined (__CPLUSLIBS)))
extern "C" {
#endif


// Feature classes
//
enum
{
	E3dFEATURE_HARDWARE_RENDERER = 1,
};

typedef struct
{
	int    Type;
	char*  Name;
} E3dFeature;


//----------------------------------------------------------------
// Function prototypes
//----------------------------------------------------------------
extern void           E3dScene_CallCallbacks(ECoreObject* PObject,  ECallbackRec* PCallbacks, const EIndex PNCallbacks,  EPointer PCallData);

extern E3dCoordinate  E3d_SnapCoordinateTo(const E3dCoordinate PF, const E3dCoordinate PS);
extern void           E3d_ColorShadeByLuma(EcRGBAiColor* PColorLum, const EcRGBAiColor* PColor, EcRGBAiColor* PColorRet);

extern EBool          E3d_RectangleVisible(const E3dMatrix PWorldToClipMatrix, const E3dCoordinate PCenterX, const E3dCoordinate PCenterY, const E3dCoordinate PSizeX, const E3dCoordinate PSizeY);

EList(extern, E3dFeature*, E3d_Features);

extern E3dFeature*  E3d_FeatureRegister(const int PType, const char* PName);
extern E3dFeature*  E3d_FeatureFindByName(const char* PName);
extern void         E3d_FeatureDeactivate(E3dFeature* PFeature);
extern void         E3d_FeatureRemove(E3dFeature* PFeature);
extern void         E3d_FeaturesFree(void);



#if defined (__cplusplus) && (!(defined (__CPLUSLIBS)))
}	// Close scope of 'extern "C"' declaration
#endif

#endif	// _E3D_h
