//================================================================
// E3D library
//
// - Interactive rendering
//
// AUTHOR:  Gabor Nagy
// DATE:    2019-Jun-23 15:14:12
//
// Copyright (C) 1994-2021 by Gabor Nagy. All rights reserved.
//================================================================
#ifndef _E3D_Draw_h
#define _E3D_Draw_h

#include <E3D/GL/E3GL.h>
#include <E3D/GL/Material.h>

extern EBool  E3dGL_DoubleBufferOK, E3dGL_ZBufferOK;


#if defined (__cplusplus) && (!(defined (__CPLUSLIBS)))
extern "C" {
#endif


//--------------------------------
// GL/SplineDraw.c
//--------------------------------
extern const unsigned int  E3d_FilledPoint_bits[];


//--------------------------------
// GL/FrameBuffer.c
//--------------------------------
extern GLenum              E3dGL_CheckFramebufferStatus(const char* PPrefix, const EBool PReportSuccess, const EBool PReportIncompleteDimensions, const EBool PPrint);


//--------------------------------
// GL/Scene.c
//--------------------------------
extern EBool            E3dGL_ExtensionsInitialized;
extern GLint            E3dGL_MAX_LIGHTS, E3dGL_MAX_TEXTURE_UNITS, E3dGL_MAX_TEXTURE_IMAGE_UNITS;
extern E3dGLCoordinate  E3dp_XYZAxis[][3];

extern void    E3dGL_Init(void);
extern EBool   E3dGL_ExtensionSupported(const char* PName);
extern EBool   E3dGL_SupportsGL12(void);


//--------------------------------
// GL/Renderers.c
//--------------------------------
extern void    E3dGL_GLRenderersRegister(void);
extern void    E3dGL_GLRenderersUnregister(void);


static inline void E3dGL_TexturesDisable(void)
{
#ifndef EGLES
	E3glDisable(GL_TEXTURE_1D);
	E3glDisable(GL_TEXTURE_2D);
	E3glDisable(GL_TEXTURE_3D);

	E3dGPUInfo*  LGPUInfo = &E3d_GPUInfo;
	if(LGPUInfo->Texture_Cubemap) E3glDisable(GL_TEXTURE_CUBE_MAP);

	#ifdef GL_TEXTURE_RECTANGLE
	E3glDisable(GL_TEXTURE_RECTANGLE);
	#endif
#endif // EGLES
}


//--------------------------------
// GL/Window.c
//--------------------------------
extern void  E3dGL_CheckGlobals(void);

extern void  _E3dOpenGLRenderer_GUI(ECoreObject* PObject, void* PPanel, void* PDrawer, const int PAction);
extern void  _E3dGLHiddenLine_GUI(ECoreObject* PObject, void* PPanel, void* PDrawer, const int PAction);
extern void  _E3dGLSolid_GUI(ECoreObject* PObject, void* PPanel, void* PDrawer, const int PAction);
extern void  _E3dGLShadedSolid_GUI(ECoreObject* PObject, void* PPanel, void* PDrawer, const int PAction);

extern void  E3dGL_TexturesDisableAll(void);

#ifdef EGLES
#define E3dGL_VertexPointer(mParam,  mSize, mType, mStride, mPtr) do\
{\
	if(mParam >= 0) { E3glVertexAttribPointer(mParam,  mSize,  mType,  FALSE,  mStride, mPtr);E3glEnableVertexAttribArray(mParam); }\
} while(0)

#define E3dGL_ColorPointer(mParam,  mSize, mType, mStride, mPtr) do\
{\
	if(mParam >= 0) { E3glVertexAttribPointer(mParam,  mSize,  mType,  FALSE,  mStride, mPtr);E3glEnableVertexAttribArray(mParam); }\
} while(0)

#define E3dGL_NormalPointer(mParam,  mType, mStride, mPtr) do\
{\
	if(mParam >= 0) { E3glVertexAttribPointer(mParam,  3,  mType,  FALSE,  mStride, mPtr);E3glEnableVertexAttribArray(mParam); }\
} while(0)

#define E3dGL_TexCoordPointer(mParam,  mSize, mType, mStride, mPtr) do\
{\
	if(mParam >= 0) { E3glVertexAttribPointer(mParam,  mSize,  mType,  FALSE,  mStride, mPtr);E3glEnableVertexAttribArray(mParam); }\
} while(0)

#else

#define E3dGL_VertexPointer(mParam,  mSize, mType, mStride, mPtr)    do { E3glVertexPointer(mSize, mType, mStride, mPtr);E3glEnableClientState(GL_VERTEX_ARRAY); } while(0)
#define E3dGL_ColorPointer(mParam,  mSize, mType, mStride, mPtr)     do { E3glColorPointer(mSize, mType, mStride, mPtr);E3glEnableClientState(GL_COLOR_ARRAY); } while(0)
#define E3dGL_NormalPointer(mParam,  mType, mStride, mPtr)           do { E3glNormalPointer(mType, mStride, mPtr);E3glEnableClientState(GL_NORMAL_ARRAY); } while(0)
#define E3dGL_TexCoordPointer(mParam,  mSize, mType, mStride, mPtr)  do { E3glTexCoordPointer(mSize, mType, mStride, mPtr);E3glEnableClientState(GL_TEXTURE_COORD_ARRAY); } while(0)
#endif  // EGLSL


#ifdef EGLES
	#define E3dGL_SetColorRGBAf_s(m)             glUniform4f(LGLShader->Param_FgU_Color,  (m).R, (m).G, (m).B, (m).A)
	#define E3dGL_SetColorRGBAf_p(m)             glUniform4f(LGLShader->Param_FgU_Color,  (m)->R, (m)->G, (m)->B, (m)->A)
	#define E3dGL_SetColorRGBAf_a(m)             glUniform4fv(LGLShader->Param_FgU_Color, 1,  m)
	#define E3dGL_SetColorRGBAf(mR, mG, mB, mA)  glUniform4f(LGLShader->Param_FgU_Color,  mR, mG, mB, mA)
	#define E3dGL_SetColorRGBAp8(c)              glUniform4ui(LGLShader->Param_FgU_Color,  EcM_R_RGBAp8(c), EcM_G_RGBAp8(c), EcM_B_RGBAp8(c), EcM_A_RGBAp8(c))
#else
	#define E3dGL_SetColorRGBAf_s(m)             EcM_glColorRGBAf_s(m)
	#define E3dGL_SetColorRGBAf_p(m)             EcM_glColorRGBAf_p(m)
	#define E3dGL_SetColorRGBAf_a(m)             EcM_glColorRGBAf_a(m)
	#define E3dGL_SetColorRGBAf(mR, mG, mB, mA)  E3glColor4f(mR, mG, mB, mA)
	#define E3dGL_SetColorRGBAp8(c)              E3glColor4ub(EcM_R_RGBAp8(c), EcM_G_RGBAp8(c), EcM_B_RGBAp8(c), EcM_A_RGBAp8(c))
#endif


extern void  E3dGL_CameraPickMatrix(const E3dCoordinate PX, const E3dCoordinate PY,  const E3dCoordinate PSizeX, const E3dCoordinate PSizeY,  const GLint PViewPort[4]);

// Post effects
//
extern const int  E3dGL_BloomFilterType;

extern E3dGLCoreShader*  E3dGL_ConstColorDrawPrep(E3dGPUContext* PGPUContext,  const EcRGBAf32Color* PColor);

#if defined (__cplusplus) && (!(defined (__CPLUSLIBS)))
}	// Close scope of 'extern "C"' declaration
#endif

#endif	// _E3D_Draw_h
