//================================================================
// E3D library
//
// OpenGL basics
//
// AUTHOR:  Gabor Nagy
// DATE:    2019-Jun-23 15:14:12
//
// Copyright (C) 1994-2021 by Gabor Nagy. All rights reserved.
//================================================================
#ifndef _E3D_E3GL_h
#define _E3D_E3GL_h

#ifndef _E3D_GPU_h
#include <E3D/GPU.h>
#endif

#include <Xe/XeGL.h>

#if defined (__cplusplus) && (!(defined (__CPLUSLIBS)))
extern "C" {
#endif

#ifdef EGLES
	#define  E3dGL_MatrixMode(m) do\
	{\
		switch(m)\
		{\
			case GL_MODELVIEW:   LGPUContext->MatrixStackIndex = E3dMatrixStackMODELVIEW;break;\
			case GL_PROJECTION:  LGPUContext->MatrixStackIndex = E3dMatrixStackPROJECTION;break;\
			default:\
				printf("%s() %s:%d  %sINVALID MATRIX MODE%s\n", __FUNCTION__, __FILE__, __LINE__, EYellowCode, ENormalCode);fflush(stdout);\
			break;\
		}\
	} while(0)
#else
	#define  E3dGL_MatrixMode  E3glMatrixMode
#endif

#if E3dGL_FLOAT_BITS == 64
	#define E3dGL_printf      XeGL_printfd
	#define E3dGL_printfML    XeGL_printfMLd
	#define E3dGL_wprintf     XeGL_wprintfd

	#define E3dGL_PushMatrix    E3glPushMatrix
	#define E3dGL_PopMatrix     E3glPopMatrix
	#define E3dGL_LoadIdentity  E3glLoadIdentity
	#define E3dGL_LoadMatrix    E3glLoadMatrixd
	#define E3dGL_MultMatrix    E3glMultMatrixd
	#define E3glMultMatrix      E3glMultMatrixd
	#define E3dGL_GetFloatv     E3glGetDoublev
	#define E3dGL_Translate     E3glTranslated
	#define E3dGL_Rotate        E3glRotated
	#define E3dGL_Scale         E3glScaled
	#define E3dGL_Ortho         E3glOrtho
	#define E3dGL_Frustum       E3glFrustum

	#define E3dGL_RasterPos2v  E3glRasterPos2dv

	#define E3dGL_Vertex2v     E3glVertex2dv
	#define E3dGL_Vertex3      E3glVertex3d
	#define E3dGL_Vertex3v     E3glVertex3dv
	#define E3dGL_Normal3      E3glNormal3d
	#define E3dGL_Normal3v     E3glNormal3dv
	#define E3dGL_TexCoord2    E3glTexCoord2d
	#define E3dGL_TexCoord2v   E3glTexCoord2dv
	#define E3dGL_RasterPos3   E3glRasterPos3d
	#define E3dGL_RasterPos3v  E3glRasterPos3dv
#else
	#if E3dGL_FLOAT_BITS == 32
		#define E3dGL_printf      XeGL_printff
		#define E3dGL_printfML    XeGL_printfMLf
		#define E3dGL_wprintf     XeGL_wprintff

#ifdef EGLES
		#define E3dGL_PushMatrix() do\
		{\
			const int  lIncrement = 16;\
			E3dMatrixStack*  lStack = LGPUContext->MatrixStacks + LGPUContext->MatrixStackIndex;\
			if(lStack->Pointer >= lStack->AllocatedDepth - 1) { EResizeNT(E3dMatrix, lStack->AllocatedDepth + lIncrement, lStack->Matrices);lStack->AllocatedDepth = lStack->AllocatedDepth + lIncrement; }\
			memcpy(lStack->Matrices[lStack->Pointer + 1], lStack->Matrices[lStack->Pointer], sizeof(E3dMatrix));\
			lStack->Pointer++;\
		} while(0)

		#define E3dGL_PopMatrix() do\
		{\
			E3dMatrixStack*  lStack = LGPUContext->MatrixStacks + LGPUContext->MatrixStackIndex;\
			if(lStack->Pointer > 0) lStack->Pointer--;\
		} while(0)

		#define E3dGL_LoadIdentity() do\
		{\
			E3dMatrixStack*  lStack = LGPUContext->MatrixStacks + LGPUContext->MatrixStackIndex;\
			E3dMatrix_LoadIdentity(lStack->Matrices[lStack->Pointer]);\
		} while(0)

		#define E3dGL_LoadMatrix(m) do\
		{\
			E3dMatrixStack*  lStack = LGPUContext->MatrixStacks + LGPUContext->MatrixStackIndex;\
			E3dMatrix_Copy(m, lStack->Matrices[lStack->Pointer]);\
		} while(0)

		#define E3dGL_MultMatrix(m) do\
		{\
			E3dMatrixStack*  lStack = LGPUContext->MatrixStacks + LGPUContext->MatrixStackIndex;\
			E3dMatrix_Mult4x4(m, lStack->Matrices[lStack->Pointer]);\
		} while(0)

		#define E3dGL_GetFloatv         E3glGetFloatv

		#define E3dGL_Translate(mX, mY, mZ) do\
		{\
			E3dMatrixStack*  lStack = LGPUContext->MatrixStacks + LGPUContext->MatrixStackIndex;\
			E3dMatrix_Translate(lStack->Matrices[lStack->Pointer],  mX, mY, mZ);\
		} while(0)

		#define E3dGL_Rotate(mA, mX, mY, mZ) do\
		{\
			E3dMatrixStack*  lStack = LGPUContext->MatrixStacks + LGPUContext->MatrixStackIndex;\
			E3dMatrix_Rotate(lStack->Matrices[lStack->Pointer],  mA,  mX, mY, mZ);\
		} while(0)

		#define E3dGL_Scale(mX, mY, mZ) do\
		{\
			E3dMatrixStack*  lStack = LGPUContext->MatrixStacks + LGPUContext->MatrixStackIndex;\
			E3dMatrix_Scale(lStack->Matrices[lStack->Pointer],  mX, mY, mZ);\
		} while(0)

		#define E3dGL_Ortho(mLeft, mRight, mBottom, mTop,  mZNear, mZFar) do\
		{\
			E3dMatrixStack*  lStack = LGPUContext->MatrixStacks + LGPUContext->MatrixStackIndex;\
			E3dMatrix_SetOrtho(lStack->Matrices[lStack->Pointer],  mLeft, mRight, mBottom, mTop,  mZNear, mZFar);\
		} while(0)

		#define E3dGL_Frustum(mLeft, mRight, mBottom, mTop,  mZNear, mZFar) do\
		{\
			E3dMatrixStack*  lStack = LGPUContext->MatrixStacks + LGPUContext->MatrixStackIndex;\
			E3dMatrix_SetFrustum(lStack->Matrices[lStack->Pointer],  mLeft, mRight, mBottom, mTop,  mZNear, mZFar);\
		} while(0)

#else
		#define E3dGL_PushMatrix    E3glPushMatrix
		#define E3dGL_PopMatrix     E3glPopMatrix
		#define E3dGL_LoadIdentity  E3glLoadIdentity
		#define E3dGL_LoadMatrix    E3glLoadMatrixf
		#define E3dGL_MultMatrix    E3glMultMatrixf
		#define E3dGL_GetFloatv     E3glGetFloatv
		#define E3dGL_Translate     E3glTranslatef
		#define E3dGL_Rotate        E3glRotatef
		#define E3dGL_Scale         E3glScalef
		#define E3dGL_Ortho         E3glOrtho
		#define E3dGL_Frustum       E3glFrustum
#endif

		#define E3glMultMatrix       E3glMultMatrixf

		#define E3dGL_RasterPos2v  E3glRasterPos2fv

		#define E3dGL_Vertex2v     E3glVertex2fv
		#define E3dGL_Vertex3      E3glVertex3f
		#define E3dGL_Vertex3v     E3glVertex3fv
		#define E3dGL_Normal3      E3glNormal3f
		#define E3dGL_Normal3v     E3glNormal3fv
		#define E3dGL_TexCoord2    E3glTexCoord2f
		#define E3dGL_TexCoord2v   E3glTexCoord2fv
		#define E3dGL_RasterPos3   E3glRasterPos3f
		#define E3dGL_RasterPos3v  E3glRasterPos3fv
	#else
		#if defined (__cplusplus) && (!(defined (__CPLUSLIBS)))
		#else
			ERROR
		#endif
	#endif
#endif // DOUBLE_PRECISION





// To force stenciled solid+wire, comment this out:
//
#if defined(GL_EXT_polygon_offset) || defined(GL_VERSION_1_1)
	#define E3dGL_POLYGON_OFFSET_SUPPORTED
#endif

#if defined(GL_EXT_polygon_offset) && !defined(GL_VERSION_1_1)
	#define glPolygonOffset(A, B)     glPolygonOffsetEXT(A, B)

// OpenGL 1.1's polygon offset can be different for each
// polygon mode primitive type.
// The EXT extension has only one offset.
//
	#define GL_POLYGON_OFFSET_FILL   GL_POLYGON_OFFSET_EXT
#endif

#ifndef sgi
	#define E3dGL_MULTITEXTURING  1
#endif

#if defined(GL_EXT_polygon_offset) || defined(GL_VERSION_1_1)
// To force stenciled solid+wire, comment this out:
//
	#define E3dGL_POLYGON_OFFSET_SUPPORTED
#endif


#ifndef GL_TEXTURE_RECTANGLE
	#ifdef GL_TEXTURE_RECTANGLE_ARB
		#define GL_TEXTURE_RECTANGLE  GL_TEXTURE_RECTANGLE_ARB
	#endif
#endif

#ifndef GL_TEXTURE_BINDING_RECTANGLE
	#ifdef GL_TEXTURE_BINDING_RECTANGLE_ARB
		#define GL_TEXTURE_BINDING_RECTANGLE  GL_TEXTURE_BINDING_RECTANGLE_ARB
	#endif
#endif




#define ME3dVertex3dot(v)  E3dGL_Vertex3((v).X, (v).Y, (v).Z)

#define ME3dVertex2i(v)  glVertex2i((v)[E3dX], (v)[E3dY])
#define ME3dVertex3i(v)  glVertex3i((v)[E3dX], (v)[E3dY], (v)[E3dZ])


#ifdef ERELEASE_BUILD
#define E3dGLDEBUG  0
#else
#define E3dGLDEBUG  1
#endif

#if E3dGLDEBUG
#define E3glActiveTexture(m)                          do { glActiveTexture(m);XeGLM_CheckStatusNL(FALSE); } while(0)
#define E3glAttachShader(m0, m1)                      do { glAttachShader(m0, m1);XeGLM_CheckStatusNL(FALSE); } while(0)
#define E3glBegin(m)                                  glBegin(m)      // Can't have any GL calls between glBegin and glEnd. That includes XeGLM_CheckStatusNL.
#define E3glBindBuffer(m0, m1)                        do { glBindBuffer(m0, m1);XeGLM_CheckStatusNL(FALSE); } while(0)
#define E3glBindTexture(m0, m1)                       do { glBindTexture(m0, m1);XeGLM_CheckStatusNL(FALSE); } while(0)
#define E3glBindFramebuffer(m0, m1)                   do { glBindFramebuffer(m0, m1);XeGLM_CheckStatusNL(FALSE); } while(0)
#define E3glBindRenderbuffer(m0, m1)                  do { glBindRenderbuffer(m0, m1);XeGLM_CheckStatusNL(FALSE); } while(0)
#define E3glBitmap(m0, m1, m2, m3, m4, m5, m6)        do { glBitmap(m0, m1, m2, m3, m4, m5, m6);XeGLM_CheckStatusNL(FALSE); } while(0)
#define E3glBlendFunc(m0, m1)                         do { glBlendFunc(m0, m1);XeGLM_CheckStatusNL(FALSE); } while(0)
#define E3glBlendColor(m0, m1, m2, m3)                do { glBlendColor(m0, m1, m2, m3);XeGLM_CheckStatusNL(FALSE); } while(0)
#define E3glBufferData(m0, m1, m2, m3)                do { glBufferData(m0, m1, m2, m3);XeGLM_CheckStatusNL(FALSE); } while(0)
#define E3glCallList(m)                               do { glCallList(m);XeGLM_CheckStatusNL(FALSE); } while(0)
#define E3glCallLists(m0, m1, m2)                     do { glCallLists(m0, m1, m2);XeGLM_CheckStatusNL(FALSE); } while(0)
#define E3glClear(m)                                  do { glClear(m);XeGLM_CheckStatusNL(FALSE); } while(0)
#define E3glClearColor(m0, m1, m2, m3)                do { glClearColor(m0, m1, m2, m3);XeGLM_CheckStatusNL(FALSE); } while(0)
#define E3glClearDepth(m)                             do { glClearDepth(m);XeGLM_CheckStatusNL(FALSE); } while(0)
#define E3glClearStencil(m)                           do { glClearStencil(m);XeGLM_CheckStatusNL(FALSE); } while(0)
#define E3glClientActiveTexture(m)                    do { glClientActiveTexture(m);XeGLM_CheckStatusNL(FALSE); } while(0)
#define E3glColor3d(m0, m1, m2)                       do { glColor3d(m0, m1, m2);XeGLM_CheckStatusNL(FALSE); } while(0)
#define E3glColor3f(m0, m1, m2)                       do { glColor3f(m0, m1, m2);XeGLM_CheckStatusNL(FALSE); } while(0)
#define E3glColor4ub(m0, m1, m2, m3)                  do { glColor4ub(m0, m1, m2, m3);XeGLM_CheckStatusNL(FALSE); } while(0)
#define E3glColor4ubv(m)                              do { glColor4ubv(m);XeGLM_CheckStatusNL(FALSE); } while(0)
#define E3glColor4f(m0, m1, m2, m3)                   do { glColor4f(m0, m1, m2, m3);XeGLM_CheckStatusNL(FALSE); } while(0)
#define E3glColor4fv(m)                               do { glColor4fv(m);XeGLM_CheckStatusNL(FALSE); } while(0)
#define E3glColorMask(m0, m1, m2, m3)                 do { glColorMask(m0, m1, m2, m3);XeGLM_CheckStatusNL(FALSE); } while(0)
#define E3glColorTable(m0, m1, m2, m3, m4, m5)        do { glColorTable(m0, m1, m2, m3, m4, m5);XeGLM_CheckStatusNL(FALSE); } while(0)
#define E3glCompileShader(m)                          do { glCompileShader(m);XeGLM_CheckStatusNL(FALSE); } while(0)
#define E3glCullFace(m)                               do { glCullFace(m);XeGLM_CheckStatusNL(FALSE); } while(0)
#define E3glDeleteLists(m0, m1)                       do { glDeleteLists(m0, m1);XeGLM_CheckStatusNL(FALSE); } while(0)
#define E3glDeleteProgram(m)                          do { glDeleteProgram(m);XeGLM_CheckStatusNL(FALSE); } while(0)
#define E3glDeleteTextures(m0, m1)                    do { glDeleteTextures(m0, m1);XeGLM_CheckStatusNL(FALSE); } while(0)
#define E3glDepthMask(m)                              do { glDepthMask(m);XeGLM_CheckStatusNL(FALSE); } while(0)
#define E3glDepthFunc(m)                              do { glDepthFunc(m);XeGLM_CheckStatusNL(FALSE); } while(0)
#define E3glDisable(m)                                do { glDisable(m);XeGLM_CheckStatusNL(FALSE); } while(0)
#define E3glDisableClientState(m)                     do { glDisableClientState(m);XeGLM_CheckStatusNL(FALSE); } while(0)
#define E3glDisableVertexAttribArray(m)               do { glDisableVertexAttribArray(m);XeGLM_CheckStatusNL(FALSE); } while(0)
#define E3glDrawBuffer(m)                             do { glDrawBuffer(m);XeGLM_CheckStatusNL(FALSE); } while(0)
#define E3glDrawArrays(m0, m1, m2)                    do { glDrawArrays(m0, m1, m2);XeGLM_CheckStatusNL(FALSE); } while(0)
#define E3glDrawElements(m0, m1, m2, m3)              do { glDrawElements(m0, m1, m2, m3);XeGLM_CheckStatusNL(FALSE); } while(0)
#define E3glDrawRangeElements(m0, m1, m2, m3, m4, m6) do { glDrawRangeElements(m0, m1, m2, m3, m4, m5);XeGLM_CheckStatusNL(FALSE); } while(0)
#define E3glDrawPixels(m0, m1, m2, m3, m4)            do { glDrawPixels(m0, m1, m2, m3, m4);XeGLM_CheckStatusNL(FALSE); } while(0)
#define E3glEnable(m)                                 do { glEnable(m);XeGLM_CheckStatusNL(FALSE); } while(0)
#define E3glEnableClientState(m)                      do { glEnableClientState(m);XeGLM_CheckStatusNL(FALSE); } while(0)
#define E3glEnableVertexAttribArray(m)                do { glEnableVertexAttribArray(m);XeGLM_CheckStatusNL(FALSE); } while(0)
#define E3glEnd()                                     do { glEnd();XeGLM_CheckStatusNL(FALSE); } while(0)
#define E3glEndList()                                 do { glEndList();XeGLM_CheckStatusNL(FALSE); } while(0)
#define E3glFinish()                                  do { glFinish();XeGLM_CheckStatusNL(FALSE); } while(0)
#define E3glFlush()                                   do { glFlush();XeGLM_CheckStatusNL(FALSE); } while(0)
#define E3glFramebufferTexture2D(m0, m1, m2, m3, m4)  do { glFramebufferTexture2D(m0, m1, m2, m3, m4);XeGLM_CheckStatusNL(FALSE); } while(0)
#define E3glFramebufferRenderbuffer(m0, m1, m2, m3)   do { glFramebufferRenderbuffer(m0, m1, m2, m3);XeGLM_CheckStatusNL(FALSE); } while(0)
#define E3glFrontFace(m)                              do { glFrontFace(m);XeGLM_CheckStatusNL(FALSE); } while(0)
#define E3glFrustum(m0, m1, m2, m3, m4, m5)           do { glFrustum(m0, m1, m2, m3, m4, m5);XeGLM_CheckStatusNL(FALSE); } while(0)
#define E3glGenBuffers(m0, m1)                        do { glGenBuffers(m0, m1);XeGLM_CheckStatusNL(FALSE); } while(0)
#define E3glGenRenderbuffers(m0, m1)                  do { glGenRenderbuffers(m0, m1);XeGLM_CheckStatusNL(FALSE); } while(0)
#define E3glGenTextures(m0, m1)                       do { glGenTextures(m0, m1);XeGLM_CheckStatusNL(FALSE); } while(0)
#define E3glGetBooleanv(m0, m1)                       do { glGetBooleanv(m0, m1);XeGLM_CheckStatusNL(FALSE); } while(0)
#define E3glGetDoublev(m0, m1)                        do { glGetDoublev(m0, m1);XeGLM_CheckStatusNL(FALSE); } while(0)
#define E3glGetFloatv(m0, m1)                         do { glGetFloatv(m0, m1);XeGLM_CheckStatusNL(FALSE); } while(0)
#define E3glGetIntegerv(m0, m1)                       do { glGetIntegerv(m0, m1);XeGLM_CheckStatusNL(FALSE); } while(0)
#define E3glGetPointerv(m0, m1)                       do { glGetPointerv(m0, m1);XeGLM_CheckStatusNL(FALSE); } while(0)
#define E3glGetProgramiv(m0, m1, m2)                  do { glGetProgramiv(m0, m1, m2);XeGLM_CheckStatusNL(FALSE); } while(0)
#define E3glHint(m0, m1)                              do { glHint(m0, m1);XeGLM_CheckStatusNL(FALSE); } while(0)
#define E3glIndexi(m)                                 do { glIndexi(m);XeGLM_CheckStatusNL(FALSE); } while(0)
#define E3glInitNames()                               do { glInitNames();XeGLM_CheckStatusNL(FALSE); } while(0)
#define E3glLightf(m0, m1, m2)                        do { glLightf(m0, m1, m2);XeGLM_CheckStatusNL(FALSE); } while(0)
#define E3glLightfv(m0, m1, m2)                       do { glLightfv(m0, m1, m2);XeGLM_CheckStatusNL(FALSE); } while(0)
#define E3glLightModeli(m0, m1)                       do { glLightModeli(m0, m1);XeGLM_CheckStatusNL(FALSE); } while(0)
#define E3glLightModelfv(m0, m1)                      do { glLightModelfv(m0, m1);XeGLM_CheckStatusNL(FALSE); } while(0)
#define E3glLineStipple(m0, m1)                       do { glLineStipple(m0, m1);XeGLM_CheckStatusNL(FALSE); } while(0)
#define E3glLineWidth(m)                              do { glLineWidth(m);XeGLM_CheckStatusNL(FALSE); } while(0)
#define E3glLinkProgram(m)                            do { glLinkProgram(m);XeGLM_CheckStatusNL(FALSE); } while(0)
#define E3glListBase(m)                               do { glListBase(m);XeGLM_CheckStatusNL(FALSE); } while(0)
#define E3glLoadIdentity()                            do { glLoadIdentity();XeGLM_CheckStatusNL(FALSE); } while(0)
#define E3glLoadMatrixf(m)                            do { glLoadMatrixf(m);XeGLM_CheckStatusNL(FALSE); } while(0)
#define E3glLoadName(m)                               do { glLoadName(m);XeGLM_CheckStatusNL(FALSE); } while(0)
#define E3glMaterialf(m0, m1, m2)                     do { glMaterialf(m0, m1, m2);XeGLM_CheckStatusNL(FALSE); } while(0)
#define E3glMaterialfv(m0, m1, m2)                    do { glMaterialfv(m0, m1, m2);XeGLM_CheckStatusNL(FALSE); } while(0)
#define E3glMatrixMode(m)                             do { glMatrixMode(m);XeGLM_CheckStatusNL(FALSE); } while(0)
#define E3glMultiTexCoord2fv(m0, m1)                  glMultiTexCoord2fv(m0, m1)  // Can't have any GL calls between glBegin and glEnd. That includes XeGLM_CheckStatusNL.
#define E3glMultMatrixf(m)                            do { glMultMatrixf(m);XeGLM_CheckStatusNL(FALSE); } while(0)
#define E3glNewList(m0, m1)                           do { glNewList(m0, m1);XeGLM_CheckStatusNL(FALSE); } while(0)
#define E3glNormal3fv(m)                              glNormal3fv(m)              // Can't have any GL calls between glBegin and glEnd. That includes XeGLM_CheckStatusNL.
#define E3glNormalPointer(m0, m1, m2)                 do { glNormalPointer(m0, m1, m2);XeGLM_CheckStatusNL(FALSE); } while(0)
#define E3glOrtho(m0, m1, m2, m3, m4, m5)             do { glOrtho(m0, m1, m2, m3, m4, m5);XeGLM_CheckStatusNL(FALSE); } while(0)
#define E3glPixelMapusv(m0, m1, m2)                   do { glPixelMapusv(m0, m1, m2);XeGLM_CheckStatusNL(FALSE); } while(0)
#define E3glPixelStorei(m0, m1)                       do { glPixelStorei(m0, m1);XeGLM_CheckStatusNL(FALSE); } while(0)
#define E3glPixelZoom(m0, m1)                         do { glPixelZoom(m0, m1);XeGLM_CheckStatusNL(FALSE); } while(0)
#define E3glPointSize(m)                              do { glPointSize(m);XeGLM_CheckStatusNL(FALSE); } while(0)
#define E3glPopMatrix()                               do { glPopMatrix();XeGLM_CheckStatusNL(FALSE); } while(0)
#define E3glPopName()                                 do { glPopName();XeGLM_CheckStatusNL(FALSE); } while(0)
#define E3glPolygonMode(m0, m1)                       do { glPolygonMode(m0, m1);XeGLM_CheckStatusNL(FALSE); } while(0)
#define E3glPolygonOffset(m0, m1)                     do { glPolygonOffset(m0, m1);XeGLM_CheckStatusNL(FALSE); } while(0)
#define E3glPushMatrix()                              do { glPushMatrix();XeGLM_CheckStatusNL(FALSE); } while(0)
#define E3glPushName(m)                               do { glPushName(m);XeGLM_CheckStatusNL(FALSE); } while(0)
#define E3glRasterPos2f(m0, m1)                       do { glRasterPos2f(m0, m1);XeGLM_CheckStatusNL(FALSE); } while(0)
#define E3glRasterPos2i(m0, m1)                       do { glRasterPos2i(m0, m1);XeGLM_CheckStatusNL(FALSE); } while(0)
#define E3glRasterPos2fv(m)                           do { glRasterPos2fv(m);XeGLM_CheckStatusNL(FALSE); } while(0)
#define E3glRasterPos3f(m0, m1, m2)                   do { glRasterPos3f(m0, m1, m2);XeGLM_CheckStatusNL(FALSE); } while(0)
#define E3glRasterPos3fv(m)                           do { glRasterPos3fv(m);XeGLM_CheckStatusNL(FALSE); } while(0)
#define E3glReadBuffer(m)                             do { glReadBuffer(m);XeGLM_CheckStatusNL(FALSE); } while(0)
#define E3glReadPixels(m0, m1, m2, m3, m4, m5, m6)    do { glReadPixels(m0, m1, m2, m3, m4, m5, m6);XeGLM_CheckStatusNL(FALSE); } while(0)
#define E3glRenderbufferStorage(m0, m1, m2, m3)       do { glRenderbufferStorage(m0, m1, m2, m3);XeGLM_CheckStatusNL(FALSE); } while(0)
static inline GLint E3glRenderMode(GLenum PMode) { const GLint  LResult = glRenderMode(PMode);XeGLM_CheckStatusNL(FALSE);return(LResult); }
#define E3glScalef(m0, m1, m2)                        do { glScalef(m0, m1, m2);XeGLM_CheckStatusNL(FALSE); } while(0)
#define E3glSelectBuffer(m0, m1)                      do { glSelectBuffer(m0, m1);XeGLM_CheckStatusNL(FALSE); } while(0)
#define E3glShadeModel(m)                             do { glShadeModel(m);XeGLM_CheckStatusNL(FALSE); } while(0)
#define E3glShaderSource(m0, m1, m2, m3)              do { glShaderSource(m0, m1, m2, m3);XeGLM_CheckStatusNL(FALSE); } while(0)
#define E3glTexCoord2f(m0, m1)                        glTexCoord2f(m0, m1)        // Can't have any GL calls between glBegin and glEnd. That includes XeGLM_CheckStatusNL.
#define E3glTexCoord2fv(m)                            glTexCoord2fv(m)            // Can't have any GL calls between glBegin and glEnd. That includes XeGLM_CheckStatusNL.
#define E3glTexCoordPointer(m0, m1, m2, m3)           do { glTexCoordPointer(m0, m1, m2, m3);XeGLM_CheckStatusNL(FALSE); } while(0)
#define E3glTexEnvf(m0, m1, m2)                       do { glTexEnvf(m0, m1, m2);XeGLM_CheckStatusNL(FALSE); } while(0)
#define E3glTexGenfv(m0, m1, m2)                      do { glTexGenfv(m0, m1, m2);XeGLM_CheckStatusNL(FALSE); } while(0)
#define E3glTexGeni(m0, m1, m2)                       do { glTexGeni(m0, m1, m2);XeGLM_CheckStatusNL(FALSE); } while(0)
#define E3glTexImage1D(m0, m1, m2, m3, m4, m5, m6, m7)  do { glTexImage1D(m0, m1, m2, m3, m4, m5, m6, m7);XeGLM_CheckStatusNL(FALSE); } while(0)
#define E3glTexImage2D(m0, m1, m2, m3, m4, m5, m6, m7, m8)  do { glTexImage2D(m0, m1, m2, m3, m4, m5, m6, m7, m8);XeGLM_CheckStatusNL(FALSE); } while(0)
#define E3glTexImage3D(m0, m1, m2, m3, m4, m5, m6, m7, m8, m9)  do { glTexImage3D(m0, m1, m2, m3, m4, m5, m6, m7, m8, m9);XeGLM_CheckStatusNL(FALSE); } while(0)
#define E3glTexParameterf(m0, m1, m2)                 do { glTexParameterf(m0, m1, m2);XeGLM_CheckStatusNL(FALSE); } while(0)
#define E3glTexParameterfv(m0, m1, m2)                do { glTexParameterfv(m0, m1, m2);XeGLM_CheckStatusNL(FALSE); } while(0)
#define E3glTexParameteri(m0, m1, m2)                 do { glTexParameteri(m0, m1, m2);XeGLM_CheckStatusNL(FALSE); } while(0)
#define E3glTexSubImage1D(m0, m1, m2, m3, m4, m5, m6) do { glTexSubImage1D(m0, m1, m2, m3, m4, m5, m6);XeGLM_CheckStatusNL(FALSE); } while(0)
#define E3glTexSubImage2D(m0, m1, m2, m3, m4, m5, m6, m7, m8)  do { glTexSubImage2D(m0, m1, m2, m3, m4, m5, m6, m7, m8);XeGLM_CheckStatusNL(FALSE); } while(0)
#define E3glTranslatef(m0, m1, m2)                    do { glTranslatef(m0, m1, m2);XeGLM_CheckStatusNL(FALSE); } while(0)
#define E3glUniform1i(m0, m1)                         do { glUniform1i(m0, m1);XeGLM_CheckStatusNL(FALSE); } while(0)
#define E3glUseProgram(m)                             do { glUseProgram(m);XeGLM_CheckStatusNL(FALSE); } while(0)
#define E3glVertex2i(m0, m1)                          glVertex2i(m0, m1)        // Can't have any GL calls between glBegin and glEnd. That includes XeGLM_CheckStatusNL.
#define E3glVertex2fv(m)                              glVertex2fv(m)            // Can't have any GL calls between glBegin and glEnd. That includes XeGLM_CheckStatusNL.
#define E3glVertex3f(m0, m1, m2)                      glVertex3f(m0, m1, m2)    // Can't have any GL calls between glBegin and glEnd. That includes XeGLM_CheckStatusNL.
#define E3glVertex3fv(m)                              glVertex3fv(m)            // Can't have any GL calls between glBegin and glEnd. That includes XeGLM_CheckStatusNL.
#define E3glVertexAttribPointer(m0, m1, m2, m3, m4, m5)   do { glVertexAttribPointer(m0, m1, m2, m3, m4, m5);XeGLM_CheckStatusNL(FALSE); } while(0)
#define E3glVertexPointer(m0, m1, m2, m3)             do { glVertexPointer(m0, m1, m2, m3);XeGLM_CheckStatusNL(FALSE); } while(0)
#define E3glColorPointer(m0, m1, m2, m3)              do { glColorPointer(m0, m1, m2, m3);XeGLM_CheckStatusNL(FALSE); } while(0)
#define E3glViewport(m0, m1, m2, m3)                  do { glViewport(m0, m1, m2, m3);XeGLM_CheckStatusNL(FALSE); } while(0)
//#define E3gl(m)                              do { gl(m);XeGLM_CheckStatusNL(FALSE); } while(0)
//#define E3gl(m0, m1)                         do { gl(m0, m1);XeGLM_CheckStatusNL(FALSE); } while(0)
//#define E3gl(m0, m1, m2)                     do { gl(m0, m1, m2);XeGLM_CheckStatusNL(FALSE); } while(0)
//#define E3gl(m0, m1, m2, m3)                 do { gl(m0, m1, m2, m3);XeGLM_CheckStatusNL(FALSE); } while(0)
//#define E3gl(m0, m1, m2, m3, m4)             do { gl(m0, m1, m2, m3, m4);XeGLM_CheckStatusNL(FALSE); } while(0)
//#define E3gl(m0, m1, m2, m3, m4, m5)         do { gl(m0, m1, m2, m3, m4, m5);XeGLM_CheckStatusNL(FALSE); } while(0)
#else
#define E3glActiveTexture(m)                          glActiveTexture(m)
#define E3glAttachShader(m0, m1)                      glAttachShader(m0, m1)
#define E3glBegin(m)                                  glBegin(m)
#define E3glBindBuffer(m0, m1)                        glBindBuffer(m0, m1)
#define E3glBindTexture(m0, m1)                       glBindTexture(m0, m1)
#define E3glBindFramebuffer(m0, m1)                   glBindFramebuffer(m0, m1)
#define E3glBindRenderbuffer(m0, m1)                  glBindRenderbuffer(m0, m1)
#define E3glBitmap(m0, m1, m2, m3, m4, m5, m6)        glBitmap(m0, m1, m2, m3, m4, m5, m6)
#define E3glBlendFunc(m0, m1)                         glBlendFunc(m0, m1)
#define E3glBlendColor(m0, m1, m2, m3)                glBlendColor(m0, m1, m2, m3)
#define E3glBufferData(m0, m1, m2, m3)                glBufferData(m0, m1, m2, m3)
#define E3glCallList(m)                               glCallList(m)
#define E3glCallLists(m0, m1, m2)                     glCallLists(m0, m1, m2)
#define E3glClear(m)                                  glClear(m)
#define E3glClearColor(m0, m1, m2, m3)                glClearColor(m0, m1, m2, m3)
#define E3glClearDepth(m)                             glClearDepth(m)
#define E3glClearStencil(m)                           glClearStencil(m)
#define E3glClientActiveTexture(m)                    glClientActiveTexture(m)
#define E3glColor3d(m0, m1, m2)                       glColor3d(m0, m1, m2)
#define E3glColor3f(m0, m1, m2)                       glColor3f(m0, m1, m2)
#define E3glColor4ub(m0, m1, m2, m3)                  glColor4ub(m0, m1, m2, m3)
#define E3glColor4ubv(m)                              glColor4ubv(m)
#define E3glColor4f(m0, m1, m2, m3)                   glColor4f(m0, m1, m2, m3)
#define E3glColor4fv(m)                               glColor4fv(m)
#define E3glColorMask(m0, m1, m2, m3)                 glColorMask(m0, m1, m2, m3)
#define E3glColorTable(m0, m1, m2, m3, m4, m5)        glColorTable(m0, m1, m2, m3, m4, m5)
#define E3glCompileShader(m)                          glCompileShader(m)
#define E3glCullFace(m)                               glCullFace(m)
#define E3glDeleteLists(m0, m1)                       glDeleteLists(m0, m1)
#define E3glDeleteProgram(m)                          glDeleteProgram(m)
#define E3glDeleteTextures(m0, m1)                    glDeleteTextures(m0, m1)
#define E3glDepthMask(m0)                             glDepthMask(m0)
#define E3glDepthFunc(m)                              glDepthFunc(m)
#define E3glDisable(m)                                glDisable(m)
#define E3glDisableClientState(m)                     glDisableClientState(m)
#define E3glDisableVertexAttribArray(m)               glDisableVertexAttribArray(m)
#define E3glDrawBuffer(m)                             glDrawBuffer(m)
#define E3glDrawArrays(m0, m1, m2)                    glDrawArrays(m0, m1, m2)
#define E3glDrawElements(m0, m1, m2, m3)              glDrawElements(m0, m1, m2, m3)
#define E3glDrawRangeElements(m0, m1, m2, m3, m4, m5) glDrawRangeElements(m0, m1, m2, m3, m4, m5)
#define E3glDrawPixels(m0, m1, m2, m3, m4)            glDrawPixels(m0, m1, m2, m3, m4)
#define E3glEnable(m)                                 glEnable(m)
#define E3glEnableClientState(m)                      glEnableClientState(m)
#define E3glEnableVertexAttribArray(m)                glEnableVertexAttribArray(m)
#define E3glEnd()                                     glEnd()
#define E3glEndList()                                 glEndList()
#define E3glFinish()                                  glFinish()
#define E3glFlush()                                   glFlush()
#define E3glFramebufferTexture2D(m0, m1, m2, m3, m4)  glFramebufferTexture2D(m0, m1, m2, m3, m4)
#define E3glFramebufferRenderbuffer(m0, m1, m2, m3)   glFramebufferRenderbuffer(m0, m1, m2, m3)
#define E3glFrontFace(m)                              glFrontFace(m)
#define E3glFrustum(m0, m1, m2, m3, m4, m5)           glFrustum(m0, m1, m2, m3, m4, m5)
#define E3glGenBuffers(m0, m1)                        glGenBuffers(m0, m1)
#define E3glGenRenderbuffers(m0, m1)                  glGenRenderbuffers(m0, m1)
#define E3glGenTextures(m0, m1)                       glGenTextures(m0, m1)
#define E3glGetBooleanv(m0, m1)                       glGetBooleanv(m0, m1)
#define E3glGetDoublev(m0, m1)                        glGetDoublev(m0, m1);
#define E3glGetFloatv(m0, m1)                         glGetFloatv(m0, m1)
#define E3glGetIntegerv(m0, m1)                       glGetIntegerv(m0, m1)
#define E3glGetPointerv(m0, m1)                       glGetPointerv(m0, m1)
#define E3glGetProgramiv(m0, m1, m2)                  glGetProgramiv(m0, m1, m2)
#define E3glHint(m0, m1)                              glHint(m0, m1)
#define E3glIndexi(m)                                 glIndexi(m)
#define E3glInitNames()                               glInitNames()
#define E3glLightf(m0, m1, m2)                        glLightf(m0, m1, m2)
#define E3glLightfv(m0, m1, m2)                       glLightfv(m0, m1, m2)
#define E3glLightModeli(m0, m1)                       glLightModeli(m0, m1)
#define E3glLightModelfv(m0, m1)                      glLightModelfv(m0, m1)
#define E3glLineStipple(m0, m1)                       glLineStipple(m0, m1)
#define E3glLineWidth(m)                              glLineWidth(m)
#define E3glLinkProgram(m)                            glLinkProgram(m)
#define E3glListBase(m)                               glListBase(m)
#define E3glLoadIdentity()                            glLoadIdentity()
#define E3glLoadMatrixf(m)                            glLoadMatrixf(m)
#define E3glLoadName(m)                               glLoadName(m)
#define E3glMaterialf(m0, m1, m2)                     glMaterialf(m0, m1, m2)
#define E3glMaterialfv(m0, m1, m2)                    glMaterialfv(m0, m1, m2)
#define E3glMatrixMode(m)                             glMatrixMode(m)
#define E3glMultiTexCoord2fv(m0, m1)                  glMultiTexCoord2fv(m0, m1)
#define E3glMultMatrixf(m)                            glMultMatrixf(m)
#define E3glNewList(m0, m1)                           glNewList(m0, m1)
#define E3glNormal3fv(m)                              glNormal3fv(m)
#define E3glNormalPointer(m0, m1, m2)                 glNormalPointer(m0, m1, m2)
#define E3glOrtho(m0, m1, m2, m3, m4, m5)             glOrtho(m0, m1, m2, m3, m4, m5)
#define E3glPixelMapusv(m0, m1, m2)                   glPixelMapusv(m0, m1, m2)
#define E3glPixelStorei(m0, m1)                       glPixelStorei(m0, m1)
#define E3glPixelZoom(m0, m1)                         glPixelZoom(m0, m1)
#define E3glPointSize(m)                              glPointSize(m)
#define E3glPolygonMode(m0, m1)                       glPolygonMode(m0, m1)
#define E3glPolygonOffset(m0, m1)                     glPolygonOffset(m0, m1)
#define E3glPopMatrix()                               glPopMatrix()
#define E3glPopName()                                 glPopName()
#define E3glPushMatrix()                              glPushMatrix()
#define E3glPushName(m)                               glPushName(m)
#define E3glRasterPos2f(m0, m1)                       glRasterPos2f(m0, m1)
#define E3glRasterPos2i(m0, m1)                       glRasterPos2i(m0, m1)
#define E3glRasterPos2fv(m)                           glRasterPos2fv(m)
#define E3glRasterPos3f(m0, m1, m2)                   glRasterPos3f(m0, m1, m2)
#define E3glRasterPos3fv(m)                           glRasterPos3fv(m)
#define E3glReadBuffer(m)                             glReadBuffer(m)
#define E3glReadPixels(m0, m1, m2, m3, m4, m5, m6)    glReadPixels(m0, m1, m2, m3, m4, m5, m6)
#define E3glRenderbufferStorage(m0, m1, m2, m3)       glRenderbufferStorage(m0, m1, m2, m3)
#define E3glRenderMode(m)                             glRenderMode(m)
#define E3glScalef(m0, m1, m2)                        glScalef(m0, m1, m2)
#define E3glSelectBuffer(m0, m1)                      glSelectBuffer(m0, m1)
#define E3glShadeModel(m)                             glShadeModel(m)
#define E3glShaderSource(m0, m1, m2, m3)              glShaderSource(m0, m1, m2, m3)
#define E3glTexCoord2f(m0, m1)                        glTexCoord2f(m0, m1)
#define E3glTexCoord2fv(m)                            glTexCoord2fv(m)
#define E3glTexCoordPointer(m0, m1, m2, m3)           glTexCoordPointer(m0, m1, m2, m3)
#define E3glTexEnvf(m0, m1, m2)                       glTexEnvf(m0, m1, m2)
#define E3glTexGenfv(m0, m1, m2)                      glTexGenfv(m0, m1, m2)
#define E3glTexGeni(m0, m1, m2)                       glTexGeni(m0, m1, m2)
#define E3glTexImage1D(m0, m1, m2, m3, m4, m5, m6, m7)    glTexImage1D(m0, m1, m2, m3, m4, m5, m6, m7)
#define E3glTexImage2D(m0, m1, m2, m3, m4, m5, m6, m7, m8)  glTexImage2D(m0, m1, m2, m3, m4, m5, m6, m7, m8)
#define E3glTexImage3D(m0, m1, m2, m3, m4, m5, m6, m7, m8, m9)  glTexImage3D(m0, m1, m2, m3, m4, m5, m6, m7, m8, m9)
#define E3glTexParameterf(m0, m1, m2)                 glTexParameterf(m0, m1, m2)
#define E3glTexParameterfv(m0, m1, m2)                glTexParameterfv(m0, m1, m2)
#define E3glTexParameteri(m0, m1, m2)                 glTexParameteri(m0, m1, m2)
#define E3glTexSubImage1D(m0, m1, m2, m3, m4, m5, m6) glTexSubImage1D(m0, m1, m2, m3, m4, m5, m6)
#define E3glTexSubImage2D(m0, m1, m2, m3, m4, m5, m6, m7, m8)  glTexSubImage2D(m0, m1, m2, m3, m4, m5, m6, m7, m8)
#define E3glTranslatef(m0, m1, m2)                    glTranslatef(m0, m1, m2)
#define E3glUniform1i(m0, m1)                         glUniform1i(m0, m1)
#define E3glUseProgram(m)                             glUseProgram(m)
#define E3glVertex2i(m0, m1)                          glVertex2i(m0, m1)
#define E3glVertex2fv(m)                              glVertex2fv(m)
#define E3glVertex3f(m0, m1, m2)                      glVertex3f(m0, m1, m2)
#define E3glVertex3fv(m)                              glVertex3fv(m)
#define E3glVertexAttribPointer(m0, m1, m2, m3, m4, m5)   glVertexAttribPointer(m0, m1, m2, m3, m4, m5)
#define E3glVertexPointer(m0, m1, m2, m3)             glVertexPointer(m0, m1, m2, m3)
#define E3glColorPointer(m0, m1, m2, m3)              glColorPointer(m0, m1, m2, m3)
#define E3glViewport(m0, m1, m2, m3)                  glViewport(m0, m1, m2, m3)
#endif

#define E3dFBO_GROW_ONLY   0x0001
#define E3dFBO_CREATE_PBO  0x0002

extern int     E3dGL_InitExtensions(void);


#if defined (__cplusplus) && (!(defined (__CPLUSLIBS)))
}	// Close scope of 'extern "C"' declaration
#endif

#endif	// _E3D_E3GL_h
