//================================================================
// E3D library
//
// - Interactive rendering
//
// AUTHOR:  Gabor Nagy
// DATE:    2018-Apr-01 15:14:12
//
// Copyright (C) 1994-2021 by Gabor Nagy. All rights reserved.
//================================================================
#ifndef _E3DGL_Material_h
#define _E3DGL_Material_h

#include <E3D/GL/E3GL.h>


#if defined (__cplusplus) && (!(defined (__CPLUSLIBS)))
extern "C" {
#endif

#ifndef sgi
static inline void _E3dGL_GetShaderLog(const GLuint PGLObject, const char* PPrefix, const char* PName)
{
	GLsizei  lLength;GLcharARB  lLog[8192];glGetShaderInfoLog(PGLObject, 8192, &lLength, lLog);if(lLength)
	{
		if(PName) printf("%s() %s:%d  %s%s%s %s\n%s%s%s", __FUNCTION__, __FILE__, __LINE__, ECyanCode, PPrefix, ENormalCode, PName, EYellowCode, lLog, ENormalCode);
		else printf("%s%s%s", EYellowCode, lLog, ENormalCode);
		fflush(stdout);
	}
}
#endif

#define _E3dGL_GetProgramLog(mObject, mPrefix, mName) do\
{\
	GLsizei  lLength;GLcharARB  lLog[8192];glGetProgramInfoLog(mObject, 8192, &lLength, lLog);if(lLength)\
	{\
		printf("%s() %s:%d %s%s%s %s\n%s%s%s", __FUNCTION__, __FILE__, __LINE__, ECyanCode, mPrefix, ENormalCode, mName, EYellowCode, lLog, ENormalCode);fflush(stdout);\
	}\
} while(0)




#define E3dGLShaderVariantCore\
	E3dShaderVariantCore;\
	GLuint  Program;\
\
	GLint   Param_VtA_Position, Param_VtA_Normal, Param_VtA_Color, Param_VtA_TexCoord0;\
	GLint   Param_VtU_ModelViewProjectionMatrix, Param_VtU_LocalToWorldMatrix

typedef struct
{
	E3dGLShaderVariantCore;

	GLint  Param_FgU_Color,
	       Param_FgU_Light0WorldPosition, Param_FgU_Light0Color;
} E3dGLCoreShader;


typedef struct
{
	E3dGLCoreShader  Sh_Unlit;
	E3dGLCoreShader  Sh_UnlitTextured;
	E3dGLCoreShader  Sh_Lambert;
	E3dGLCoreShader  Sh_LambertTextured;
} E3dGLGlobals;


extern E3dGLGlobals*  _GLGlobals;


#if defined (__cplusplus) && (!(defined (__CPLUSLIBS)))
}	// Close scope of 'extern "C"' declaration
#endif

#endif	// _E3DGL_Material_h
