//========================================================================
// Drawing macros etc.
//
// AUTHOR:  Gabor Nagy
// DATE:    2019-Jun-23 17:05:47
//
// Copyright (C) 1994-2021 by Gabor Nagy. All rights reserved.
//========================================================================
#ifndef _E3D_GL_MeshDraw_h
#define _E3D_GL_MeshDraw_h


#ifndef _E3D_Mesh_h
#include <E3D/Mesh.h>
#endif

#ifndef _E3D_Rendering_h
#include <E3D/Rendering.h>
#endif

#ifndef _E3D_PolygroupDraw_h
#include <E3D/GL/PolyGroupDraw.h>
#endif


static inline void E3dPolygon_GLDraw(const E3dPolygon* PPolygon, const E3dVertexNodeSize PVertexNodeSize,  const E3dVertex* PMeshVertices, const E3dVertexSize PMeshVertexSize)
{
#ifdef EGLES
if(E3d_Verbosity & E3dVbOPENGL) { printf("%s() %s:%d  %sGLES IMPLEMENT ME%s\n", __FUNCTION__, __FILE__, __LINE__, EYellowCode, ENormalCode);fflush(stdout); }
#else
	const E3dITriangle*  LITriangle = PPolygon->ITriangles;if(LITriangle == NULL)
	{
//if(PPolygon->NumOfVertices > 3) { printf("NO ITRIANGLES\n");fflush(stdout); }
		const E3dVertexNode*  LVertexNode = PPolygon->VertexNodes;
		E3glBegin(GL_POLYGON);
		EIndex  LVC = 0;const EIndex  LVN = PPolygon->NumOfVertexNodes;for(;LVC < LVN;LVC++)
		{
			const E3dVertex*  LVertex = EStruct_ByIndex(PMeshVertices, PMeshVertexSize, LVertexNode->VertexID);
			const E3dCoord3*  LVertexPosition = &LVertex->Position;

			E3dGL_Vertex3v((E3dCoordinate*)LVertexPosition);
			EStruct_Inc(LVertexNode, PVertexNodeSize);
		}
		E3glEnd();
	}
	else
	{
		const E3dVertexNode*  LVertexNodes = PPolygon->VertexNodes;

		E3glBegin(GL_TRIANGLES);
		EIndex  LPC = 0;const EIndex  LPN = PPolygon->ITriangles_Count;for(;LPC < LPN;LPC++)
		{
			const E3dCoord3*           LVertexPosition;
			const E3dVertexNodeIndex*  LVertexNodeIndices = LITriangle->VertexNodes;
			const E3dVertexNode*       LVertexNode;

			LVertexNode = EStruct_ByIndex(LVertexNodes, PVertexNodeSize, LVertexNodeIndices[0]);LVertexPosition = &(EStruct_ByIndex(PMeshVertices, PMeshVertexSize,  LVertexNode->VertexID)->Position);E3dGL_Vertex3v((E3dCoordinate*)LVertexPosition);
			LVertexNode = EStruct_ByIndex(LVertexNodes, PVertexNodeSize, LVertexNodeIndices[1]);LVertexPosition = &(EStruct_ByIndex(PMeshVertices, PMeshVertexSize,  LVertexNode->VertexID)->Position);E3dGL_Vertex3v((E3dCoordinate*)LVertexPosition);
			LVertexNode = EStruct_ByIndex(LVertexNodes, PVertexNodeSize, LVertexNodeIndices[2]);LVertexPosition = &(EStruct_ByIndex(PMeshVertices, PMeshVertexSize,  LVertexNode->VertexID)->Position);E3dGL_Vertex3v((E3dCoordinate*)LVertexPosition);

			LITriangle++;
		}
		E3glEnd();
	}
#endif
}


//----------------------------------------------------------------
// Draw selected Polygons on top, with blending
//----------------------------------------------------------------
static inline void E3dMesh_GLDrawSelectedPolygonsOnTop(const E3dMesh* PMesh, const E3dDisplayFlags PDisplayFlags)
{
#ifdef EGLES
if(E3d_Verbosity & E3dVbOPENGL) { printf("%s() %s:%d  %sGLES IMPLEMENT ME%s\n", __FUNCTION__, __FILE__, __LINE__, EYellowCode, ENormalCode);fflush(stdout); }
#else
	E3glDisable(GL_LIGHTING);
	E3dGL_TexturesDisable();
#endif

//	E3glEnable(GL_BLEND);E3glBlendFunc(GL_ONE, GL_ONE);
	E3glEnable(GL_BLEND);E3glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	const E3dVertex*  LMeshVertices = PMesh->Vertices;const E3dVertexSize  LMeshVertexSize = PMesh->VertexSize;
	E3dPolyGroup**    LPolyGroups = PMesh->PolyGroups;
	EIndex  LPGC = 0;const EIndex  LPGN = PMesh->PolyGroups_Count;for(;LPGC < LPGN;LPGC++)
	{
		E3dPolyGroup*  PPolyGroup = LPolyGroups[LPGC];if(PPolyGroup->Visible)
		{
			if(PPolyGroup->Selected || (PDisplayFlags & E3dDF_SELECTED) != 0)
			{
				EIndex  PPolyCnt = PPolyGroup->Polygons_Count;if(PPolyCnt > 0)
				{
					const E3dVertexNodeSize  LVertexNodeSize = PPolyGroup->VertexNodeSize;

					if(PPolyGroup->DoubleSided) E3glDisable(GL_CULL_FACE);
					else E3glEnable(GL_CULL_FACE);
					const E3dPolygon*  PPolygon = PPolyGroup->Polygons;
					do
					{
						if(PPolygon->Flags & E3dPolygonSELECTED) E3dPolygon_GLDraw(PPolygon, LVertexNodeSize,  LMeshVertices, LMeshVertexSize);

						PPolygon++;
					} while(--PPolyCnt);
				}
			}
		}
	}
	E3glDisable(GL_BLEND);
}



#if defined (__cplusplus) && (!(defined (__CPLUSLIBS)))
extern "C" {
#endif


//----------------------------------------
// DrawMesh.c
//----------------------------------------
extern void          E3dMeshPolyGroups_GLInvalidateVertexArrays(E3dMesh* PMesh);

extern void          E3dMesh_DrawEdges(E3dGPUContext* PGPUContext, E3dMesh* PMesh, const EcRGBAf32Color* PColor);

extern E3dDisplayFlags  E3dMesh_DrawCommon1(E3dGPUContext* PGLContest, E3dModel* PModel, E3dMesh* PMesh,  E3dRenderer* PRenderer, const E3dDrawContext* PDrawContext, const E3dDisplayFlags PDisplayFlags,  const unsigned char POpaqueTransparent, const EBool PDrawSelectedPolygonsInSolidMode);
extern void             E3dMesh_DrawComponents(E3dGPUContext* PGPUContext,  E3dModel* PModel, E3dMesh* PMesh,  const E3dDrawContext* PDrawContext, const E3dDisplayFlags PDisplayFlags,  const EBool PNormalGeometry, const unsigned char POpaqueTransparent, const EBool PDrawSelectedPolygonsInSolidMode);

extern void          E3dMesh_Draw(E3dGPUContext* PGPUContext, E3dModel* PModel, E3dMesh* PMesh,  E3dRenderer* PRenderer, const E3dDrawContext* PDrawContext, const E3dDisplayFlags PDisplayFlags,  const EBool PNormalGeometry, const char POpaqueTransparent, EBool LDrawSelectedPolygonsInShadedMode);
extern void          E3dMesh_PickDraw(E3dGPUContext* PGPUContext, E3dMesh* PMesh,  const E3dDrawContext* PDrawContext, const E3dDisplayFlags PDisplayFlags,  const unsigned int PPickTargets);

extern void          E3dMesh_DrawConstSolid(E3dGPUContext* PGPUContext, E3dMesh* PMesh);
extern void          E3dMesh_DrawColorSolid(E3dGPUContext* PGPUContext, E3dMesh* PMesh, const uint8_t POpaqueTransparent, const EBool PTransparencyOn);
extern void          E3dMesh_DrawColorSolidTextured(E3dGPUContext* PGPUContext, E3dMesh* PMesh, const uint8_t POpaqueTransparent, const EBool PTransparencyOn);

extern void          E3dPolyGroup_GLDrawVertexColorSolid(E3dGPUContext* PGPUContext, E3dPolyGroup* PPolyGroup,  const E3dVertex* PVertices, const E3dVertexSize PVertexSize, const EIndex PVertices_Count);

extern void          E3dMesh_DrawShadedSolid(E3dGPUContext* PGPUContext,  E3dModel* PModel, E3dMesh* PMesh,  E3dRenderer* PRenderer, const E3dDisplayFlags PDisplayFlags,  const unsigned char POpaqueTransparent, const EBool PDrawSelectedPolygonsInSolidMode);

extern void          E3dModel_DrawTextureMappers(E3dGPUContext* PGPUContext,  E3dModel* PModel,  const E3dDrawContext* PDrawContext, const E3dDisplayFlags PDisplayFlags);


#if defined (__cplusplus) && (!(defined (__CPLUSLIBS)))
}	// Close scope of 'extern "C"'
#endif

#endif	// _E3D_GL_MeshDraw_h
