//================================================================
// E3D
//
// - Basic type definitions
//
// AUTHOR:  Gabor Nagy
// DATE:    2019-Jun-23 14:03:08
//
// Copyright (C) 1994-2021 by Gabor Nagy. All rights reserved.
//================================================================
#ifndef	_E3D_GPU_h
#define	_E3D_GPU_h

#include <ETypes.h>

#if defined (__cplusplus) && (!(defined (__CPLUSLIBS)))
extern "C" {
#endif


extern const char  E3dAPI_OpenGL[], E3dAPI_Vulkan[], E3dAPI_Metal[];


typedef struct E3dShaderVariant_S  E3dShaderVariant;


#ifdef EGLES
#define EGLESContextPart\
	uint16_t           MatrixStackIndex;  /* E3dMatrixStackMODELVIEW or E3dMatrixStackPROJECTION */\
	EBool              LightingOn, TexturesOn;\
\
	E3dMatrixStack     MatrixStacks[2]
#else
#define EGLESContextPart
#endif

#define E3dGPUContextCore\
	const char*        API;\
\
	E3dShaderVariant*  CurrentShaderVariant;\
\
	EGLESContextPart

typedef struct E3dGPUContext_S  E3dGPUContext;

struct E3dGPUContext_S
{
	E3dGPUContextCore;
};


enum
{
	E3dFBOSharingPARTIAL = 1,
	E3dFBOSharingFULL
};

#define E3dGPU2DBufferCore\
	const char*  API;   /* E3dAPI_OpenGL, etc. */\
	int          RefCnt;\
	EInt32_2     Size

typedef struct
{
	E3dGPU2DBufferCore;
} E3dGPU2DBuffer;


#define E3dGPUFrameBufferCore\
	const char*      API;                 /* E3dAPI_OpenGL, etc. */\
	char*            Name;\
	EBool            Static;\
	EIndex_2         Size,                /* Virtual size (always <= AllocatedSize) this is separate, to minimize physical reallocation calls */\
	                 AllocatedSize;       /* Actual allocated size (via OpenGL etc.)  */\
\
	uint16_t         PixelFormat,         /* EpRGBA, EpRGB, EpGRAYA, EpGRAY */\
	                 ColorComponentType;  /* EResUINT8, EResFLOAT16 */\
\
	E3dGPU2DBuffer*  DepthBuffer

typedef struct
{
	E3dGPUFrameBufferCore;
} E3dGPUFrameBuffer;


typedef struct
{
	const char*  API;       // E3dAPI_OpenGL, E3dAPI_Vulkan etc.
	const char*  Vendor;
	const char*  Model;

// Rendering features supported by the hardware
//
	uint32_t       PixelType, PixelComponentType;
	uint8_t        VertexArrays,
	               Shadows,
	               Stereo,
	               VertexPrograms,
	               PixelPrograms;

	uint32_t       OpenGL12:1,
	               FrameBufferObject:1, Multisampling:1;

	float          Texture_MaximumAnisotropy;
	uint32_t       Texture_AnisotropicFiltering:1,
	               Texture_Float:1,
	               Texture_Rectangle:1,  // ATI / AMD GPUs don't have this
	               Texture_EdgeClamp:1, Texture_BorderClamp:1, Texture_Cubemap:1;
} E3dGPUInfo;

extern E3dGPUInfo  E3d_GPUInfo;


#define E3dGL_MATRIX_STACK_DEPTH  256

enum
{
	E3dMatrixStackMODELVIEW = 0,
	E3dMatrixStackPROJECTION
};


extern void  E3dGPUContext_Init(E3dGPUContext* PContext);
extern void  E3dGPUContext_FreeContents(E3dGPUContext* PContext);

extern void  E3dGPUContext_Free(E3dGPUContext* PContext);

extern EBool  E3dRenderers_UseGPUFrameBuffersForHDRPass;


// These are in graphics API-dependent source files
//
extern void             E3dGPU2DBuffer_Init(const char* PAPI, E3dGPU2DBuffer* PBuffer);
extern E3dGPU2DBuffer*  E3dGPU2DBuffer_New(void);
extern void             E3dGPU2DBuffer_Free(E3dGPU2DBuffer* PBuffer);

extern E3dGPUFrameBuffer*  E3dGPUFrameBuffer_New(const char* PName);
extern void                E3dGPUFrameBuffer_Init(E3dGPUFrameBuffer* PFBO, const char* PAPI, const char* PName, const EBool PStatic);
extern void                E3dGPUFrameBuffer_Free(E3dGPUFrameBuffer* PFBO);
extern int                 E3dGPUFrameBuffer_Resize(E3dGPUFrameBuffer* PFBO,  const EIndex PNewSizeX, const EIndex PNewSizeY,  const short PPixelFormat, const short PComponenType, E3dGPU2DBuffer* PDepthBuffer, const uint16_t PFlags);
extern int                 E3dGPUFrameBuffer_Save(E3dGPUFrameBuffer* PFBO, const char* PFileName, void* PFileFormat);
extern int                 E3dGPUFrameBuffer_IsDrawable(E3dGPUFrameBuffer* PFBO);

#if defined (__cplusplus) && (!(defined (__CPLUSLIBS)))
}	// Close scope of 'extern "C"' declaration
#endif


#endif	// _E3D_GPU_h
