//================================================================
// E3D
//
// General Geometry-related functions header file
//
// AUTHOR:  Gabor Nagy
// DATE:    2019-Jun-23 14:15:10
//
// Copyright (C) 1994-2021 by Gabor Nagy. All rights reserved.
//================================================================
#ifndef _E3D_Geometry_h
#define _E3D_Geometry_h


#ifndef _E3D_Object_h
#include <E3D/Object.h>
#endif

#ifndef _E3dMaterial
#define _E3dMaterial
typedef struct E3dMaterial_S   E3dMaterial;    // Forward-define E3dMaterial
typedef struct E3dMaterial_S*  E3dMaterialPtr;
EListA_t(E3dMaterialPtr);  // E3dMaterialPtrListA
#endif


#define E3dM_LODRefCmp(mIdx)  ((LGeometry->LODs[mIdx])->LODReference < lLODRef)

#define E3dM_GeometryGetLODForDrawing(mWorldToViewerMatrix, mSizeFactor, mModel, mGeometry) do\
{\
	E3dGeometry*         lGeometry = mGeometry;\
	const E3dCoordinate  lSizeFactor = mSizeFactor;\
	if(lGeometry->GetLOD) mGeometry = lGeometry->GetLOD(lGeometry, lSizeFactor);\
	else if(lGeometry->AutoLOD)\
	{\
		if(lGeometry->LODs)\
		{\
			E3dCoordinate  lXP, lYP, lZP;\
\
			E3dMatrix_Transform3x4_ss(mWorldToViewerMatrix,  mModel->LocalToWorldMatrix[M30], mModel->LocalToWorldMatrix[M31], mModel->LocalToWorldMatrix[M32],  lXP, lYP, lZP);\
			(void)lXP;(void)lYP;\
\
			const E3dCoordinate  lLODRef = -lZP * lSizeFactor;\
\
			if(lLODRef > lGeometry->LODs[0]->LODReference)\
			{\
				EIndex  l1Index;E_BinarySearch(lGeometry->LODs_Count, E3dM_LODRefCmp, l1Index);\
				if(l1Index >= lGeometry->LODs_Count) l1Index = mGeometry->LODs_Count - 1;\
				mGeometry = lGeometry->LODs[l1Index];\
			}\
		}\
	}\
	else\
	{\
		if(lGeometry->LODs)\
		{\
			if(lGeometry->LODStep) mGeometry = lGeometry->LODs[mGeometry->LODStep - 1];\
		}\
	}\
} while(0)


//------------------------------------------------------------------------
// Template functions
//------------------------------------------------------------------------

#define E3dTpl_ObjectComponentActivateWithCustomDefaultsCB_Geometry(mType, mPrefix, mComponentTemplate, mName,  mPlugin) do\
{\
	E3dObject*  lObject = NULL;\
\
	E3dModel*  lModel = E3d_ModelCustomObjectNewInit(E3d_Scene, mComponentTemplate, &lObject, E3dUF_ALL, mName, (E3dp_Prefs.Duplicate_Flags & E3dCopyMODEL_UNIQUE_NAME) != 0);if(lModel)\
	{\
		E3dScene*     lScene = E3d_Scene;\
		E3dGeometry*  lGeometry = (E3dGeometry*)lObject;\
\
		lGeometry->AutoLOD = FALSE;\
\
		E3dScene_NodeHrcAdd(lScene, (E3dNode*)lModel, E3dUF_CALL_SCENE_CALLBACKS);\
\
		E3dScene_ObjectsUpdateModelList(lScene);\
		E3dScene_LayoutOnSchematics(lScene, TRUE);\
	}\
\
	if(lObject)\
	{\
		if(((XePushButtonCallbackStruct*)PCallData)->Reason == XeCR_ACTIVATE_OPTIONS) E3dObject_Edit(lObject, (mPrefix##ComponentClass));\
		E3dWindows_RedrawByChangeFlags(E3dObject_GetChangeFlagsByClass(lObject)|E3dChgHIERARCHY);\
	}\
} while(0)


#if defined (__cplusplus) && (!(defined (__CPLUSLIBS)))
extern "C" {
#endif


#ifndef _E3dGeometry
#define _E3dGeometry
typedef struct E3dGeometry_S  E3dGeometry;    // Forward-define E3dGeometry
#endif

#ifndef _E3dModel
#define _E3dModel
typedef struct E3dModel_S     E3dModel;       // Forward-define E3dModel
#endif


typedef uint16_t  E3dEdgeFlags;

// Edge flags
//
// These two are saved into files, when Mesh->SaveEdges is TRUE
//
#define E3dEdgeSELECTED  0x0001    // Edge is selected
#define E3dEdgeVISIBLE   0x0002    // Edge is visible (e.g. Subdivision Surfaces may only show the ControlMesh Edges

#define E3dEdgeACTIVE    0x0004    // Mouse pointer is on Edge
#define E3dEdgeDRAWN     0x0008    // Edge has already been drawn
#define E3dEdgePASS      0x0010    // Temporary marker flag
#define E3dEdgeMARK0     0x0020    // Temporary marker flag
#define E3dEdgeFlagALL   0x003F

#define E3dEdgeDEFAULT   E3dEdgeVISIBLE

typedef struct
{
	E3dVertexIndex  Start, End;  // Index of starting and ending Vertex
	E3dEdgeFlags    Flags;
	float           Hardness;    // For subdivision surfaces
} E3dEdge;


typedef E3dGeometry*  (*E3dGetLOD)(E3dGeometry* PGeometry, const E3dCoordinate PSizeFactor);



// Fields of the generic Geometry structure (these are the first fields in all Geometry structures)
//
#define E3dGeometryCore\
	E3dObjectCore;\
	E3dGetLOD       GetLOD;                      /* For procedural (e.g. continous LOD on a subdiv. Mesh) LOD   */\
	E3dCoordinate   LODReference;                /* Level Of Detail "activation distance" for this Geometry     */\
	int             LODStep;                     /* Force displaying this LOD. 0: Original                      */\
	EList(, E3dGeometry*, LODs);                 /* Levels of detail beyond this one                            */\
\
	E3dBBox         BBox;                        /* Bounding box                     */\
	ETime           Shape_TimeStamp;             /* The last time the shape changed  */\
\
	EBool           AutoLOD;\
\
	EBool           CastsShadows;\
	EBool           ReceivesShadows;\
	EBool           BBoxUptodate




// Generic Geometry structure
//
struct E3dGeometry_S
{
	E3dGeometryCore;
};


extern E3dObjectClass	E3dGeometry_class[];


//--------------------------------------------------------
// Function prototypes
//--------------------------------------------------------
extern EIndex  E3d_GeometryCallDestroyProcs(E3dGeometry* PGeometry);

extern const char  E3dUI_Show[];

//--------------------------------------------------------
// E3D/GeometryAdv.c
//--------------------------------------------------------
extern EIndex  E3dGeometry_Map2DTextures(E3dGeometry* PGeometry);

extern int     E3dGeometry_Invert(E3dGeometry* PGeometry, EChangeFlags* PChangedP);

extern void    E3dGeometry_Transform(E3dGeometry* PGeometry, const E3dMatrix PMatrix);

extern EBool   E3dGeometry_GetBoundingBox(E3dGeometry* PGeometry, const unsigned int PWhatToInclude,  E3dCoord3* PBBMin, E3dCoord3* PBBMax);
extern EBool   E3dGeometry_GetTransformedBoundingBox(const E3dGeometry* PGeometry, const E3dMatrix PMatrix, const unsigned int PWhatToInclude,  E3dCoord3* PBBMin, E3dCoord3* PBBMax);
extern EBool   E3dGeometry_GetPerspectiveProjectedBoundingBox(E3dGeometry* PGeometry, const E3dMatrix PMatrix, const unsigned int PWhatToInclude,  E3dCoord2* PBBMin, E3dCoord2* PBBMax,  E3dCoord4* PMaxProjectedXYAbsValPoint);


extern int     E3dGeometry_Tesselate(E3dGeometry* PGeometry, const EBool PSelectionOnly, const EBool PDoConvex);
extern int     E3dGeometry_ReplaceTexturePaths(E3dGeometry* PGeometry, const char* PTexturePath);


#if defined (__cplusplus) && (!(defined (__CPLUSLIBS)))
}	// Close scope of 'extern "C"' declaration
#endif

#endif	// _E3D_Geometry_h
