//================================================================
// E3D library
//
// - Renderer settings panel
//
// AUTHOR:  Gabor Nagy
// DATE:    2019-Jun-23 00:22:50
//
// Copyright (C) 1994-2021 by Gabor Nagy. All rights reserved.
//================================================================
#ifndef _E3D_RendererSettings_h
#define _E3D_RendererSettings_h

#include <Xe/DropDownPanel.h>
#include <Xe/FSBox.h>

#ifdef __cplusplus
extern "C" {
#endif


#define	E3dRendererSetupRENDERER_SELECTOR  0x00000001
#define	E3dRendererSetupFILE_INFO          0x00000002
#define	E3dRendererSetupIMAGE_RESOLUTION   0x00000004
#define	E3dRendererSetupAREA_LIGHTS        0x00000100

typedef struct
{
	Widget  MatrixW;
	void*   ResourcesGUI;
} E3dAntiAliasingPanel;

typedef struct
{
	XeInfoPanelCore;

	E3dpWindow*           E3DWindow;

	XeFileSelectionPanel  FileSelectionPanel;

	XeDrawer              StereoSettingsDrawer, ImageSettingsDrawer, BackgroundColorDrawer, ImageStatsDrawer,
	                      AdvancedOptionsDrawer;

	Widget                TopMatrixW,
	                      FileNameTextW,
	                      StartFrameW, EndFrameW, FrameStepW, SaveDepthImageW, SaveShadowImageW,
	                      SizeXW, SizeYW, KeepRatioW,
	                      StereoModeW, StereoEyeDistanceW, StereoEye2DOffsetW,
	                      LensCorrectionW, LensCorrectionRW, LensCorrectionBW, MagnificationW,
	                      OrientNormalW, OrientRotateRightW, OrientVisoR1W,
	                      RMatrixW, HiddenObjectWireframeW, UnselectedObjectWireframeW, ShowSplinesW, ShowFacesW,
	                      ShowMeshVerticesW,
	                      ShowSplineCVsW, ShowSplineTangentsW, ShowSplineNormalsW,
	                      ProgressiveLabelW, ProgressiveMatrixW, Progressive1x1W, Progressive2x2W, Progressive4x4W, Progressive8x8W, Progressive16x16W, Progressive32x32W,
	                      HDRImageBW, ToneMapMatrixW, AdaptiveExposureW, AutoExposureSpotMeteringBW, GainW, GammaW,
	                      BackgroundColorSwatchW;

	Widget                AntiAliasingLabelW;
	EList(, E3dAntiAliasingPanel, AntiAliasingMatrixWs);

	EcColorPanel*         ColorPanel;
	Widget                CaptureNoneW, CaptureAABufferW, CaptureFinalW;

	E3dRendererSet*       RendererSet;
	float                 ImageAspectRatio;
	EBool                 KeepRatio;


	EBool                 HasFileSelectionPanel;
	EBool                 ApplyChangesToAllWindows;
} E3dRendererSettingsPanel;


//--------------------------------
// Panel/3DWindowSettings.c
//--------------------------------
extern E3dRendererSettingsPanel*  E3dp_RendererSettingsPanel;
extern E3dRendererSettingsPanel*  E3dpRendererSet_GetSettingsPanel(E3dRendererSet* PRendererSet);

extern E3dRendererSettingsPanel*  (*E3dpRendererSettingsPanel_CreateFunc)(Widget PTopMatrixW, E3dWindow* P3DWindow, E3dRendererSet* PRendererSet, const unsigned int PComponents);
extern void    (*E3dpRendererSettingsPanel_UpdateFunc)(E3dRendererSettingsPanel* PSettingsPanel, E3dWindow* P3DWindow, E3dRendererSet*);
extern void    (*E3dpRendererSettingsPanel_CloseFunc)(E3dRendererSettingsPanel* PSettingsPanel);
extern void    (*E3dpRendererSettingsPanel_FreeFunc)(E3dRendererSettingsPanel* PSettingsPanel);


//--------------------------------
// Plugins/Renderer/Settings.c
//--------------------------------
extern E3dRendererSettingsPanel*  E3dpRendererSettingsPanel_New(Widget PTopMatrixW, E3dWindow* P3DWindow, E3dRendererSet* PRendererSet, const unsigned int PComponents);
extern void    E3dpRendererSettingsPanel_Update(E3dRendererSettingsPanel* PSettingsPanel, E3dWindow* P3DWindow, E3dRendererSet* PRendererSet);
extern void    E3dpRendererSettingsPanel_Close(E3dRendererSettingsPanel* PSettingsPanel);
extern void    E3dpRendererSettingsPanel_Free(E3dRendererSettingsPanel* PSettingsPanel);


extern E3dAntiAliasingPanel*  E3dRendererSettingsPanel_GetAAPanel(E3dRendererSettingsPanel* PPanel, const EPointer PUserData);
extern void                   E3dRendererSettingsPanel_SwitchAAMatrix(E3dRendererSettingsPanel* PPanel, const EPointer PUserData);

extern void  _E3dCB_AntiAliasing_NAnisotropicTextureSamples(Widget PW, EPointer PClientData, EPointer PCallData);
extern void  _E3dDCB_NumOfSubFrames(Widget PW, EPointer PClientData, EPointer PCallData);

extern void  E3dGPURenderer_AntiAliasing_GUI(E3dRendererSettingsPanel* PPanel, XeComponentClassDrawer* PDrawer, E3dGPURenderer* PRenderer, const int PAction);

typedef struct
{
	Widget  RadiusW, StrengthW, PowerW;
} E3dBloomGUI;


extern void  E3dRenderer_BloomGUI(E3dRendererSettingsPanel* PPanel, E3dBloom* PBloom, E3dBloomGUI* PBloomGUI, Widget PParentW,  Widget* PTraverseWs, int* PTraverseN,  const int PAction);


#ifdef __cplusplus
}	// Close scope of 'extern "C"' declaration
#endif

#endif	// _E3D_RendererSettings_h
