//================================================================
// E3D library
//
// - Generic "Shape" header
//
// AUTHOR:  Gabor Nagy
// DATE:    2018-Apr-01 23:19:53
//
// Copyright (C) 1994-2021 by Gabor Nagy. All rights reserved.
//================================================================
#ifndef _E3D_Shape_h
#define _E3D_Shape_h

#ifndef _E3D_Face_h
#include <E3D/Face.h>
#endif

#ifndef _E3D_Model_h
#include <E3D/Model.h>
#endif

#ifndef _E3D_Mesh_h
#include <E3D/Mesh.h>
#endif

#ifndef _E3D_Polygon_h
#include <E3D/Polygon.h>
#endif


#if defined (__cplusplus) && (!(defined (__CPLUSLIBS)))
extern "C" {
#endif


#define E3dShapeSnapshotCore\
	E3dObject*  Object;\
\
	E3dMatrix   LocalToWorldMatrix;  /* Matrix used for translate operations on the Points  */\
	EBool       LocalToWorldMatrixIsNotIdentity

typedef struct
{
	E3dShapeSnapshotCore;
} E3dShapeSnapshot;


// These structures are used for taking "snapshots" of Vertices/Points/Spline keys etc. for interactive
// transforming and for Undo/Redo of transforming and deleting Points
//
typedef struct
{
	E3dShapeSnapshotCore;

	E3dSnapshotVertex*  SnapshotVertices;EIndex  NumOfPoints;
	int                 PointSize;
} E3dMeshVerticesSnapshot;


typedef struct
{
	E3dShapeSnapshotCore;

	E3dSnapshotSplineCV*  SnapshotCVs;EIndex  CVs_Count;
	unsigned int          CVSize;
	char                  SubCV;     // For Bezier curves
} E3dSplineCVsSnapshot;


extern void  E3d_ShapeSnapshotFree(E3dShapeSnapshot* PShapeSnapshot);
extern void  E3d_ShapeSnapshotsFree(E3dShapeSnapshot*** PShapeSnapshotsP, EIndex* PNumOfShapeSnapshotsP);


#if defined (__cplusplus) && (!(defined (__CPLUSLIBS)))
}	// Close scope of 'extern "C"' declaration
#endif

#endif  // _E3D_Shape_h
