//================================================================
// E3D library
//
// Textures
//
// AUTHOR:  Gabor Nagy
// DATE:    2019-Jun-23 03:17:42
//
// Copyright (C) 1994-2021 by Gabor Nagy. All rights reserved.
//================================================================
#ifndef _E3D_Texture_h
#define _E3D_Texture_h

#include <Image/Image.h>

#ifndef _E3D_Geometry_h
#include <E3D/Geometry.h>
#endif

#ifdef USEOpenGL
#include <Image/GL/Texture.h>
#endif


#define E3dST_OFF_TEXTURE  -1000000.0


// Texture update flags
//
#define E3dTXTUF_FILTERING  0x0001
#define E3dTXTUF_IMAGE      0x0002
#define E3dTXTUF_NORMALMAP  0x0004
#define E3dTXTUF_CDF        0x0008
#define E3dTXTUF_BLENDING   0x0010
#define E3dTXTUF_ALL        0x001F

//----------------------------------------
// 2D texture
//----------------------------------------
#define	E3dTXTF_PIXELINTERPOLATION  0x01
#define	E3dTXTF_SWAP_ST             0x02
#define	E3dTXTF_ALTERNATE_S         0x04
#define	E3dTXTF_ALTERNATE_T         0x08
#define	E3dTXTF_TRANSPARENCYBW      0x10
#define	E3dTXTF_CROPPING            0x20
#define	E3dTXTF_ALL                 0x3F



#if defined (__cplusplus) && (!(defined (__CPLUSLIBS)))
extern "C" {
#endif

typedef struct E3d2DTexture_S             E3d2DTexture;
typedef struct E3dTextureMapper_S         E3dTextureMapper;

#define E3dTextureMapperFreehand          E3dTextureMapper

typedef struct E3d2DTextureShaderClass_S  E3d2DTextureShaderClass;
typedef struct E3d2DTextureShader_S       E3d2DTextureShader;


typedef int  (*E3d2DTextureShaderGUIProc)(E3d2DTexture* PTexture, void* PTexturePanel, void* PTextureShaderClassGUI, const int PAction);
typedef int  (*E3d2DTextureShaderGetImageProc)(E3d2DTexture* PTexture, E3dTextureMapper* PMapper, const double PTime,  const EIndex PX, const EIndex PY,  const EIndex PSizeX, const EIndex PSizeY);


struct E3d2DTextureShaderClass_S
{
	EBaseClassCore(E3d2DTextureShader);

	E3d2DTextureShaderGUIProc       GUI;
	E3d2DTextureShaderGetImageProc  GetImage;
};


#define E3d2DTextureShaderCore\
 E3d2DTextureShaderClass*  Class;\
 EBool                     Animated, Animated_AbsoluteTime, Animated_Streaming

struct E3d2DTextureShader_S
{
	E3d2DTextureShaderCore;
};



enum
{
	E3dTXTFilterNEAREST = 0,
	E3dTXTFilterLINEAR
};


enum
{
	E3dTANGENT_SPACE = 1, E3dOBJECT_SPACE, E3dWORLD_SPACE
};


typedef struct E3dTextureBlender_S  E3dTextureBlender;

struct E3dTextureBlender_S
{
	EResource*  ShaderResource;
	float       Factor;     // Scaler
	uint32_t    Channels;   // EpRED, EpGREEN, EpBLUE, EpALPHA, EpINTENSITY
	uint16_t    Operation;  // E3dADD, E3dMUL

	void        (*Modulate)(const EcRGBAf32Color PTextureColor, const E3dTextureBlender* PBlender, void* PDst);
};

enum
{
	E3dTilingCLAMP = 1,
	E3dTilingREPEAT,
	E3dTilingALTERNATE
};


typedef void  (*E3dTextureSamplerProc)(const E3d2DTexture* PTexture, const E3dSTWithDiff* PST,  const char PTilingModeS, const char PTilingModeT,  E3dSurfacePointBase* PPoint, float* PResult);

#define E3dImageCDFSampleCore\
	float     CDFPosition;    /* CDF range start                   */\
	EIndex_2  PixelPosition   /* Pixel coordinates in the texture  */

typedef struct
{
	E3dImageCDFSampleCore;
} E3dImageCDFSample;

typedef struct
{
	E3dImageCDFSampleCore;

	E3dCoord3  Position, Normal;  // Position of sample in 3D
	float      Color[3];
} E3d3DCDFSample;

typedef struct
{
	EIndex_2         Size;
	E3d3DCDFSample*  Samples;
} E3d3DCDF;


typedef struct
{
	EpImageCore;
	EpPixelBuffer*  PixelBufferRGBf32;
	EpPixelBuffer*  PixelBufferRGBA8;
	EpPixelBuffer*  PixelBufferRGBX8;
	EpPixelBuffer*  PixelBufferGRAY8;
} E3dImage;


struct E3d2DTexture_S
{
	EObjectCore(EObject);

	double                 ImageLoadedForThisFrame;  // If the texture is animated (has a % character in the file name to indicate a printf-style number for the frame), the image was loaded for this frame

	E3d2DTextureShader*    Shader;             // Shader that creates the image. The default is NULL, meaning an image file, specified by the FileName field below

	char*                  FileName;           // Texture file name with path
	char*                  NormalMapFileName;  // NormalMap file name with path
	uint16_t               Flags;              // Pixel interpolation, Swap ST, ST alternation, crop

	E3dTextureSamplerProc  Sample;             // Sampling function, set by E3d2DTexture_UpdateForDisplay

	EChangeFlags           ChangedFlags;

#ifdef USEOpenGL
	EpGLTexture*           GLTexture_2D;       // For 2D textures
	EpGLTexture*           GLTexture_CUBE;     // For cube-maps

	int                    GLAlphaFunction;    // GL_ALWAYS means alpha-testing disabled
	float                  GLAlphaReference;

	EpPixelFormat          SLNormalMapPixelFormat;

	GLuint                 GLNormalMapIndex;
	int                    GLNormalMapValid;
	EpImage**              GLMIPMaps;EIndex  NumOfGLMIPMaps;
#endif // USEOpenGL

	E3dImage*              Image;              // EpImage of the texture

	E3dImage**             MIPMaps;EIndex	NumOfMIPMaps;
	EBool                  UseMIPMaps, AnisotropicFiltering;

// E3dFIXME Temporary for reading some old files
//
	E3dTextureMapper*      Mapper;

	EInt32_2               CropMin, CropMax;   // Cropping info


// Application of the Texture
//
	EList(, E3dTextureBlender, Blenders);      // One for each shader Resource

// Special blenders that don't map to Shader Resources
//
	E3dTextureBlender      Bump;               // Blender for bump-mapping
	E3dTextureBlender      Displacement;       // Blender for displacement-mapping

	EcRGBf32Color*         NormalMapPixels;    // For bump-mapping/normal-mapping
	EpImage*               NormalMapImage;     // Container for the above, just so we can treat it as an EpImage

	EBool                  ImageAsNormalMap;   // Use RGB image directly as a normal-map

	E3dCoord3              SchemPosition;      // Position on schematic view
	uint8_t                NormalMapType,      // Use RGB image directly as a normal-map 0, E3dTANGENT_SPACE, E3dOBJECT_SPACE or E3dWORLD_SPACE
	                       MagFilter;          // Magnifying filter


// Internals for rendering
//
	int                    ReflectiveChannelMask, TransparentChannelMask;
	EBool                  DoNormalMap, DoTangentSpace, DoBilinear;
	int                    MIPIndexDiff;

	int16_t                CDFImageSizeDivider;   // Scale down the Texture->Image by this factor, for building the CDF
	uint16_t               CDFImageBlurRadius;    // Blur radius passed to EpImage_Scale. Fixed point x100
	EpImage*               CDFRefImage;           // The image from which the CDF is built. May have a lower resolution than Texture->Image (and may be blurred), to minimize artifacts and to save memory.
	E3dImageCDFSample*     CDF;                   // Cumulative Distribution Function for emission Textures, for importance-sampling textured area lights.
	float                  EmissionFactor;        // Store it here as well, so we don't have to look it up when sampling the area light...
};


enum
{
	E3dTXMAP_NONE = 0,
	E3dTXMAP_POLYGROUP,
	E3dTXMAP_POLYGONS,	// Selected Polygons
	E3dTXMAP_GEOMETRY,
	E3dTXMAP_MODEL,
	E3dTXMAP_BRANCH,
	E3dTXMAP_TREE
};



typedef int  (*E3dTextureMapperUpdateProc)(E3dTextureMapper* PMapper, const E3dGeometry* PGeometry, const EObject* PShadeable);
typedef int  (*E3dTextureMapper3DToSTProc)(const E3dTextureMapper*, const E3dCoord3* P3DPosition, const E3dCoord3* PNormal, E3dST* PST);
typedef int  (*E3dTextureMapperSTTo3DProc)(const E3dTextureMapper*, const E3dST* PST, E3dCoord3* POrigin, E3dCoord3* PDirection, E3dCoordinate* PDistance, float* PProbabilityDensity);

typedef struct  E3dTextureMapperClass_S  E3dTextureMapperClass;


struct E3dTextureMapperClass_S
{
	EBaseClassCore(E3dTextureMapper);

	E3dTextureMapperUpdateProc  Update;       // Called when TextureMapper parameters or the Geometry change
	E3dTextureMapper3DToSTProc  MapPoint;     // 3D point -> texture coordinates
	E3dTextureMapperSTTo3DProc  InvMapPoint;  // Texture coordinates -> 3D point
	ECoreObjectGUIFunc          GUI;
};


#define E3dTextureMapperCore\
	EObjectCore(E3dTextureMapper);\
\
	E3d2DTexture*     MpTexture;\
\
	EList(, E3dEdge, Edges);                               /* For ST editing : Edges in texture-space  */\
	unsigned short*   PolygonFlags;EIndex	Polygons_Count;  /* For ST editing :  "Vertex" flags for each Polygon VertexNode ST, for the ST set corresponding to this Mapper. Layout : [Polygons_Count]                    */\
	unsigned short**  VertexFlags;                         /* For ST editing :  "Vertex" flags for each Polygon VertexNode ST, for the ST set corresponding to this Mapper. Layout : [Polygons_Count][NumOfVertexNodes]  */\
\
\
/* User settable  */\
\
	E3dCoord3         Translation;          /* Translation of texture (SOffset, TOffset, 0.0)  */\
	E3dCoord3         Scaling;              /* Scaling of texture (SScaling, TScaling, 1.0)  */\
	E3dEulerRotation  EulerRotation;        /* Rotation of texture  */\
	E3dCoord2         STScaling;            /* Size of application zone in S and T  */\
	E3dCoord2         STOffset;             /* Offset of application zone in S and T  */\
	EInt32_2          STCount;              /* Repetition of image in S and T  */\
	EBool             AlternateS, AlternateT;\
\
	int32_t           BRDFSamplesSqrt;      /* Sample the BRDF to get accurate specular lighting when used with reflection cube-maps. */\
	float             EnvLightFactor;\
\
\
/* Internal	*/\
\
	char              TilingS, TilingT;     /* Tiling mode : E3dTilingCLAMP, E3dTilingREPEAT or E3dTilingALTERNATE  */\
\
	int               MapOn;                /* Mapping target (PolyGroup, Mesh, Model, Branch etc)  */\
	E3dGeometry*      ProjectorGeometry;\
	E3dCoord3         ProjectorCenter;      /* Center of texture projector Geometry  */\
	E3dMatrix         ProjectorMatrix,      /* Matrix for drawing the texture-projector Geometry  */\
										LocalToMapperMatrix,\
										WorldToMapperMatrix,\
										MapperToWorldMatrix,\
										TangentRotatorMatrix;\
	E3d3DCDF          Emission3DCDF         /* Cached Cumulative Distribution Function, used if this is an emission Texture  */



// This gets attached to Shadeables (PolyGroup, Face etc.)
//
struct E3dTextureMapper_S
{
	E3dTextureMapperCore;
};


typedef struct
{
	E3dTextureMapperCore;

	E3dCoord3  Min, Max;    // Projector Geometry's boundaries
} E3dTextureMapperPlanar;


typedef struct
{
	E3dTextureMapperCore;

	EBool  SwapUV;
} E3dTextureMapperUV;


typedef struct
{
	E3dCoord3      Center;    // Center of minimal enclosing sphere
	E3dCoordinate  Radius;    // Radius of minimal enclosing sphere
} E3dWireSphere;


#define E3dMinSphereBasedTextureMapperCore\
	E3dTextureMapperCore;\
\
	E3dWireSphere	Sphere

typedef struct
{
	E3dMinSphereBasedTextureMapperCore;
} E3dMinSphereBasedTextureMapper;

extern int  E3d_MinSphereBasedTextureMapperUpdate(const E3dGeometry* PGeometry, E3dMinSphereBasedTextureMapper* PMapper);


//------------------------------------------------------------------------


//----------------------------------------
// 3D texture
//----------------------------------------
typedef struct
{
	int        RefCnt;           // Reference count
	char*      Name;
	char*      FileName;
	E3dCoord3  SchemPosition;    // Position on schematic view
} E3d3DTexture;
//------------------------------------------------------------------------


static inline EBool E3dTexture_IsBumpmap(E3d2DTexture* PTexture)
{
	if(PTexture->Bump.Factor > 0.001f || PTexture->Bump.Factor < -0.001f) return(TRUE);
	else return(FALSE);
}


#define _E3dM_NoTextureCoverage(m)  do { m[0] = m[1] = m[2] = m[3] = 0.0; } while(0)\


//
// Apply tiling to a coordinate, using a modulo
//
#define _M_Window(mVal, mModulo)     { if(mVal < 0) mVal = mModulo + (mVal % mModulo);else mVal = mVal % mModulo; }
#define _M_Alternate(mVal, mModulo)  { const EIndex  lVal = mVal;if(mVal < 0) mVal = mModulo + (mVal % mModulo);else mVal = mVal % mModulo;if((lVal / mModulo) & 1) mVal = mModulo - 1 - mVal; }
#define _M_Clamp(mVal, mModulo)      { if(mVal < 0) mVal = 0;else if(mVal >= mModulo) mVal = mModulo - 1; }


#define _E3d_Nearest1(mType, mComponentsPerPixel, mStride) do\
{\
	const mType*  lSrcPixels = (mType*)LSrcPixelBuffer->u8;\
	const mType*  lSrcPixel = lSrcPixels + (LTX + LTY * LSizeX) * mStride;\
\
	int  lC = 0;for(;lC < mComponentsPerPixel;lC++)\
	{\
		PResult[lC] = (float)(lSrcPixel[lC]);\
	}\
} while(0)


#define _E3d_Nearest1_n(mType, mComponentsPerPixel, mStride, mNormalizer) do\
{\
	const mType*  lSrcPixels = (mType*)LSrcPixelBuffer->u8;\
	const mType*  lSrcPixel = lSrcPixels + (LTX + LTY * LSizeX) * mStride;\
\
	int  lC = 0;for(;lC < mComponentsPerPixel;lC++)\
	{\
		PResult[lC] = (float)(lSrcPixel[lC]) * mNormalizer;\
	}\
} while(0)


static inline void _E3d_ImageColorNearest_AUTO_RGBAf32(const E3dImage* PImage,  const E3dCoordinate PFX, const E3dCoordinate PFY,  const char PTilingModeS, const char PTilingModeT,  float* PResult)
{
	const EIndex  LSizeX = PImage->Size.X, LSizeY = PImage->Size.Y;
	EIndex        LTX = (EIndex)PFX, LTY = (EIndex)PFY;

	switch(PTilingModeS)
	{
		case E3dTilingREPEAT:
			_M_Window(LTX, LSizeX);
		break;

		case E3dTilingALTERNATE:
			_M_Alternate(LTX, LSizeX);
		break;

		default:
			_M_Clamp(LTX, LSizeX);
		break;
	}

	switch(PTilingModeT)
	{
		case E3dTilingREPEAT:
			_M_Window(LTY, LSizeY);
		break;

		case E3dTilingALTERNATE:
			_M_Alternate(LTY, LSizeY);
		break;

		default:
			_M_Clamp(LTY, LSizeY);
		break;
	}


	const EpPixelBuffer*  LSrcPixelBuffer;
	if((LSrcPixelBuffer = EpImage_PixelBufferGetByFormat((EpImage*)PImage,  EpRGBA, 0, EResFLOAT32,  FALSE,  NULL)) != NULL)
	{
		_E3d_Nearest1(float, 4, 4);
	}
	else if((LSrcPixelBuffer = EpImage_PixelBufferGetByFormat((EpImage*)PImage,  EpRGB, 0, EResFLOAT32,  FALSE,  NULL)) != NULL)
	{
		_E3d_Nearest1(float, 3, 3);
		PResult[3] = 1.0;
	}
	else if((LSrcPixelBuffer = EpImage_PixelBufferGetByFormat((EpImage*)PImage,  EpRGBA, 0, EResUINT16,  FALSE,  NULL)) != NULL)
	{
		_E3d_Nearest1_n(unsigned short, 4, 4, EInv65536N);
	}
	else if((LSrcPixelBuffer = EpImage_PixelBufferGetByFormat((EpImage*)PImage,  EpRGB, 0, EResUINT16,  FALSE,  NULL)) != NULL)
	{
		_E3d_Nearest1_n(unsigned short, 3, 3, EInv65536N);
		PResult[3] = 1.0;
	}
	else if((LSrcPixelBuffer = EpImage_PixelBufferGetByFormat((EpImage*)PImage,  EpRGBA, 0, EResUINT8,  FALSE,  NULL)) != NULL)
	{
		_E3d_Nearest1_n(uint8_t, 4, 4, EInv256N);
	}
	else if((LSrcPixelBuffer = EpImage_PixelBufferGetByFormat((EpImage*)PImage,  EpRGB, 4, EResUINT8,  FALSE,  NULL)) != NULL)
	{
		_E3d_Nearest1_n(uint8_t, 3, 4, EInv256N);
		PResult[3] = 1.0;
	}
	else if((LSrcPixelBuffer = EpImage_PixelBufferGetByFormat((EpImage*)PImage,  EpGRAYA, 0, EResUINT16,  FALSE,  NULL)) != NULL)
	{
		_E3d_Nearest1_n(unsigned short, 2, 2, EInv65536N);
		PResult[3] = PResult[1];
		PResult[1] = PResult[2] = PResult[0];
	}
	else if((LSrcPixelBuffer = EpImage_PixelBufferGetByFormat((EpImage*)PImage,  EpGRAY, 0, EResUINT16,  FALSE,  NULL)) != NULL)
	{
		_E3d_Nearest1_n(unsigned short, 1, 2, EInv65536N);
		PResult[1] = PResult[2] = PResult[0];
		PResult[3] = 1.0;
	}
	else if((LSrcPixelBuffer = EpImage_PixelBufferGetByFormat((EpImage*)PImage,  EpGRAYA, 0, EResUINT8,  FALSE,  NULL)) != NULL)
	{
		_E3d_Nearest1_n(uint8_t, 2, 2, EInv256N);
		PResult[3] = PResult[1];
		PResult[1] = PResult[2] = PResult[0];
	}
	else if((LSrcPixelBuffer = EpImage_PixelBufferGetByFormat((EpImage*)PImage,  EpGRAY, 0, EResUINT8,  FALSE,  NULL)) != NULL)
	{
		_E3d_Nearest1_n(uint8_t, 1, 1, EInv256N);
		PResult[1] = PResult[2] = PResult[0];
		PResult[3] = 1.0;
	}
	else
	{
		{ char  s[1024];printf("%s() %s:%d  %sIMPLEMENT ME%s pixel formats %s\n", __FUNCTION__, __FILE__, __LINE__, EYellowCode, ENormalCode, EpImage_PixelFormatsToString((EpImage*)PImage, s));fflush(stdout); }
		PResult[0] = 0.0;
		PResult[1] = 0.0;
		PResult[2] = 0.0;
		PResult[3] = 1.0;
	}
}


//========================================
// Setup step for bi-linear filtering
//========================================
#define _E3d_Bilinear0(mX, mY,  mTilingS, mTilingT)\
	EIndex_2   LT;\
	EIndex_2   LNeighbor;\
	E3dCoord2  LFraction, LFractionInv;\
\
	LT.X = (EIndex)(mX);\
	LFraction.X = mX >= 0.0 ? mX - (E3dCoordinate)LT.X : 1.0 - ((E3dCoordinate)LT.X - mX);\
	if(LSizeX > 1)\
	{\
		if(LFraction.X < 0.5) LNeighbor.X = LT.X - 1;else LNeighbor.X = LT.X + 1;\
	}\
	else LNeighbor.X = LT.X;\
	LFractionInv.X = E3d_Abs((E3dCoordinate)0.5 - LFraction.X);LFraction.X = 1.0 - LFractionInv.X;\
\
\
	LT.Y = (EIndex)(mY);\
	LFraction.Y = mY >= 0.0 ? mY - (E3dCoordinate)LT.Y : 1.0 - ((E3dCoordinate)LT.Y - mY);\
	if(LSizeY > 1)\
	{\
		if(LFraction.Y < 0.5) LNeighbor.Y = LT.Y - 1;else LNeighbor.Y = LT.Y + 1;\
	}\
	else LNeighbor.Y = LT.Y;\
\
	LFractionInv.Y = 0.5 - LFraction.Y;LFractionInv.Y = E3d_Abs(LFractionInv.Y);LFraction.Y = 1.0 - LFractionInv.Y;\
\
	switch(mTilingS)\
	{\
		case E3dTilingREPEAT:\
			_M_Window(LNeighbor.X, LSizeX);\
			_M_Window(LT.X, LSizeX);\
		break;\
\
		case E3dTilingALTERNATE:\
			_M_Alternate(LNeighbor.X, LSizeX);\
			_M_Alternate(LT.X, LSizeX);\
		break;\
\
		default:\
			_M_Clamp(LNeighbor.X, LSizeX);\
			_M_Clamp(LT.X, LSizeX);\
		break;\
	}\
\
	switch(mTilingT)\
	{\
		case E3dTilingREPEAT:\
			_M_Window(LNeighbor.Y, LSizeY);\
			_M_Window(LT.Y, LSizeY);\
		break;\
\
		case E3dTilingALTERNATE:\
			_M_Alternate(LNeighbor.Y, LSizeY);\
			_M_Alternate(LT.Y, LSizeY);\
		break;\
\
		default:\
			_M_Clamp(LNeighbor.Y, LSizeY);\
			_M_Clamp(LT.Y, LSizeY);\
		break;\
	}

#define _E3d_Bilinear1(mType, mComponentsPerPixel, mStride, mT, mNeighbor, mFraction, mFractionInv)\
{\
	const mType*  lSrcPixels = (mType*)LSrcPixelBuffer->u8;\
	const mType*  lSrcPixel = lSrcPixels + (mT.X + mT.Y * LSizeX) * mStride;\
	const mType*  lSrcPixelXNeighbor = lSrcPixels + (mNeighbor.X + mT.Y * LSizeX) * mStride;\
	const mType*  lSrcPixelYNeighbor = lSrcPixels + (mT.X + mNeighbor.Y * LSizeX) * mStride;\
	const mType*  lSrcPixelXYNeighbor = lSrcPixels + (mNeighbor.X + mNeighbor.Y * LSizeX) * mStride;\
\
	int  lC = 0;for(;lC < mComponentsPerPixel;lC++)\
	{\
		const float  LV  = (float)(lSrcPixel[lC]) * mFraction.X + (float)(lSrcPixelXNeighbor[lC]) * mFractionInv.X;\
		const float  LV1 = (float)(lSrcPixelYNeighbor[lC]) * mFraction.X + (float)(lSrcPixelXYNeighbor[lC]) * mFractionInv.X;\
		PResult[lC] = (LV * mFraction.Y + LV1 * mFractionInv.Y);\
	}\
}


#define _E3d_Bilinear1_Normalize(mType, mComponentsPerPixel, mStride, mNormalizer,  mT, mNeighbor, mFraction, mFractionInv)\
{\
	const mType*  lSrcPixels = (mType*)LSrcPixelBuffer->u8;\
	const mType*  lSrcPixel = lSrcPixels + (mT.X + mT.Y * LSizeX) * mStride;\
	const mType*  lSrcPixelXNeighbor = lSrcPixels + (mNeighbor.X + mT.Y * LSizeX) * mStride;\
	const mType*  lSrcPixelYNeighbor = lSrcPixels + (mT.X + mNeighbor.Y * LSizeX) * mStride;\
	const mType*  lSrcPixelXYNeighbor = lSrcPixels + (mNeighbor.X + mNeighbor.Y * LSizeX) * mStride;\
\
	int  lC = 0;for(;lC < mComponentsPerPixel;lC++)\
	{\
		const float  LV  = (float)(lSrcPixel[lC]) * mFraction.X + (float)(lSrcPixelXNeighbor[lC]) * mFractionInv.X;\
		const float  LV1 = (float)(lSrcPixelYNeighbor[lC]) * mFraction.X + (float)(lSrcPixelXYNeighbor[lC]) * mFractionInv.X;\
		PResult[lC] = (LV * mFraction.Y + LV1 * mFractionInv.Y) * mNormalizer;\
	}\
}


static inline void _E3d_ImageColorBilinear_AUTO_RGBAf32(const E3dImage* PImage,
 const EIndex_2 PT,
 const E3dCoord2 PFraction, const E3dCoord2 PFractionInv,
 EIndex_2 PNeighbor,
 float* PResult)
{
	const EIndex    LSizeX = PImage->Size.X;

	const EpPixelBuffer*  LSrcPixelBuffer;
	if((LSrcPixelBuffer = EpImage_PixelBufferGetByFormat((EpImage*)PImage,  EpRGBA, 0, EResFLOAT32,  FALSE,  NULL)) != NULL)
	{
		_E3d_Bilinear1(float, 4, 4, PT, PNeighbor, PFraction, PFractionInv);
	}
	else if((LSrcPixelBuffer = EpImage_PixelBufferGetByFormat((EpImage*)PImage,  EpRGB, 0, EResFLOAT32,  FALSE,  NULL)) != NULL)
	{
		_E3d_Bilinear1(float, 3, 3, PT, PNeighbor, PFraction, PFractionInv);
		PResult[3] = 1.0;
	}
	else if((LSrcPixelBuffer = EpImage_PixelBufferGetByFormat((EpImage*)PImage,  EpGRAYA, 0, EResFLOAT32,  FALSE,  NULL)) != NULL)
	{
		_E3d_Bilinear1(float, 2, 2, PT, PNeighbor, PFraction, PFractionInv);
		PResult[3] = PResult[1];
		PResult[1] = PResult[2] = PResult[0];
	}
	else if((LSrcPixelBuffer = EpImage_PixelBufferGetByFormat((EpImage*)PImage,  EpGRAY, 0, EResFLOAT32,  FALSE,  NULL)) != NULL)
	{
		_E3d_Bilinear1(float, 1, 1, PT, PNeighbor, PFraction, PFractionInv);
		PResult[1] = PResult[2] = PResult[0];
		PResult[3] = 1.0;
	}
	else if((LSrcPixelBuffer = EpImage_PixelBufferGetByFormat((EpImage*)PImage,  EpRGBA, 0, EResUINT16,  FALSE,  NULL)) != NULL)
	{
		_E3d_Bilinear1_Normalize(uint16_t, 4, 4, EInv65536N,  PT, PNeighbor, PFraction, PFractionInv);
	}
	else if((LSrcPixelBuffer = EpImage_PixelBufferGetByFormat((EpImage*)PImage,  EpRGB, 0, EResUINT16,  FALSE,  NULL)) != NULL)
	{
		_E3d_Bilinear1_Normalize(uint16_t, 3, 3, EInv65536N,  PT, PNeighbor, PFraction, PFractionInv);
		PResult[3] = 1.0;
	}
	else if((LSrcPixelBuffer = EpImage_PixelBufferGetByFormat((EpImage*)PImage,  EpGRAYA, 0, EResUINT16,  FALSE,  NULL)) != NULL)
	{
		_E3d_Bilinear1_Normalize(uint16_t, 2, 2, EInv65536N,  PT, PNeighbor, PFraction, PFractionInv);
		PResult[3] = PResult[1];
		PResult[1] = PResult[2] = PResult[0];
	}
	else if((LSrcPixelBuffer = EpImage_PixelBufferGetByFormat((EpImage*)PImage,  EpGRAY, 0, EResUINT16,  FALSE,  NULL)) != NULL)
	{
		_E3d_Bilinear1_Normalize(unsigned short, 1, 1, EInv65536N,  PT, PNeighbor, PFraction, PFractionInv);
		PResult[1] = PResult[2] = PResult[0];
		PResult[3] = 1.0;
	}
	else if((LSrcPixelBuffer = EpImage_PixelBufferGetByFormat((EpImage*)PImage,  EpRGBA, 0, EResUINT8,  FALSE,  NULL)) != NULL)
	{
		_E3d_Bilinear1_Normalize(uint8_t, 4, 4, EInv256N,  PT, PNeighbor, PFraction, PFractionInv);
	}
	else if((LSrcPixelBuffer = EpImage_PixelBufferGetByFormat((EpImage*)PImage,  EpRGB, 4, EResUINT8,  FALSE,  NULL)) != NULL)
	{
		_E3d_Bilinear1_Normalize(uint8_t, 3, 4, EInv256N,  PT, PNeighbor, PFraction, PFractionInv);
		PResult[3] = 1.0;
	}
	else if((LSrcPixelBuffer = EpImage_PixelBufferGetByFormat((EpImage*)PImage,  EpGRAYA, 0, EResUINT8,  FALSE,  NULL)) != NULL)
	{
		_E3d_Bilinear1_Normalize(uint8_t, 2, 2, EInv256N,  PT, PNeighbor, PFraction, PFractionInv);
		PResult[3] = PResult[1];
		PResult[1] = PResult[2] = PResult[0];
	}
	else if((LSrcPixelBuffer = EpImage_PixelBufferGetByFormat((EpImage*)PImage,  EpGRAY, 0, EResUINT8,  FALSE,  NULL)) != NULL)
	{
		_E3d_Bilinear1_Normalize(uint8_t, 1, 1, EInv256N,  PT, PNeighbor, PFraction, PFractionInv);
		PResult[1] = PResult[2] = PResult[0];
		PResult[3] = 1.0;
	}
	else
	{
		{ char  s[1024];printf("%s() %s:%d  %sIMPLEMENT ME%s pixel formats %s\n", __FUNCTION__, __FILE__, __LINE__, EYellowCode, ENormalCode, EpImage_PixelFormatsToString((EpImage*)PImage, s));fflush(stdout); }
		PResult[0] = 0.0;
		PResult[1] = 0.0;
		PResult[2] = 0.0;
		PResult[3] = 1.0;
	}
}


EList_i32(extern, E3dTextureMapperClass*, E3d_TextureMapperClasses);


//----------------------------------------
// Global variables
//----------------------------------------
extern EObjectClass  E3d2DTexture_class[];

extern E3dImage*     E3d_NotFound2DTextureImage;
extern EResource     E3dTexture_resources[];
extern EResource     E3d_TextureOutputs[];


//----------------------------------------
// E3D/Texture.c
//----------------------------------------
EList(extern, E3d2DTextureShaderClass*, E3d_2DTextureShaderClasses);
extern EpImageClass  E3dImage_class[];


extern void             E3d2DTexture_Print(E3d2DTexture* P2DTexture, const int PIndent);

extern void             E3d2DTexture_Init(E3d2DTexture* P2DTexture);
extern E3d2DTexture*    E3d2DTexture_New(void);
extern E3d2DTexture*    E3d2DTexture_Copy(const E3d2DTexture* PTexture, const EBool PCopyCaches,  E3d2DTexture** P2DTextures, const EIndex PNumOfTextures);

extern int       E3d_CreateMIPMaps(const EpImage* PImage0,  EpImageClass* PClass, EpImage*** PMIPMapsP, EIndex* PNumOfMIPMapsP,  const EIndex PMinSizeX, const EIndex PMinSizeY);
extern void      E3d2DTexture_CreateCDF(E3d2DTexture* PTexture);


extern E3d2DTextureShaderClass*  E3d2DTextureShader_ClassRegister(E3d2DTextureShaderClass* PDescriptor);
extern void                      E3d2DTextureShader_ClassDeactivate(E3d2DTextureShaderClass* PClass);
extern E3d2DTextureShaderClass*  E3d2DTextureShader_ClassFindByName(const char* PName);
extern void                      E3d2DTextureShader_ClassesFree(void);

extern int     E3d_BumpmapToNormalmap8(EpImage* PSImage, EpRGB8Pixel* PNormalMap, E3dCoord3* PScale);
extern int     E3d_BumpmapToNormalmapu8(EpImage* PSImage, EpRGB8Pixel* PNormalMap, E3dCoord3* PScale);
extern int     E3d_BumpmapToNormalmap16(EpImage* PSImage, short* PNormalMap, E3dCoord3* PScale);
extern int     E3d_BumpmapToNormalmapf32(EpImage* PSImage, EcRGBf32Color* PNormalMap, E3dCoord3* PScale);

extern void    E3d2DTexture_UpdateNormalMap(E3d2DTexture* PTexture);

extern E3d2DTextureShader*  E3d2DTextureShader_New(E3d2DTextureShaderClass* PClass);
extern void                 E3d2DTextureShader_Free(E3d2DTextureShader* PTextureShader);



//----------------------------------------
// TextureMapping.c
//----------------------------------------
extern E3dTextureMapperClass*  E3d_BuiltInTextureMapperClasses[];

extern void  E3d_TextureMapperSetTiling(E3dTextureMapper* PMapper);

extern E3dTextureMapperClass  E3dTextureMapper_class[];
extern E3dTextureMapperClass  E3dTextureMapperFreehand_class[];



//----------------------------------------
// Adv/TextureMappingAdv.c
//----------------------------------------
extern E3dTextureMapperClass  E3dTextureMapperXY_class[];
extern E3dTextureMapperClass  E3dTextureMapperXZ_class[];
extern E3dTextureMapperClass  E3dTextureMapperYZ_class[];
extern E3dTextureMapperClass  E3dTextureMapperUV_class[];
extern E3dTextureMapperClass  E3dTextureMapperLayout_class[];
extern E3dTextureMapperClass  E3dTextureMapperCylindrical_class[];
extern E3dTextureMapperClass  E3dTextureMapperSpherical_class[];
extern E3dTextureMapperClass  E3dTextureMapperAngleMap_class[];
extern E3dTextureMapperClass  E3dTextureMapperMirrorSphere_class[];
extern E3dTextureMapperClass  E3dTextureMapperFishEye_class[];
extern E3dTextureMapperClass  E3dTextureMapperEnvMap_class[];

extern E3dTextureMapperClass  E3dTextureMapperCubeMap_class[];
extern E3dTextureMapperClass  E3dTextureMapperCubeMapReflection_class[];
extern E3dTextureMapperClass  E3dTextureMapperCubeMapRefraction_class[];

#ifndef ERELEASE_BUILD
typedef struct
{
	E3dTextureMapperCore;
} E3dTextureMapperPlanarReflection;

extern E3dTextureMapperClass  E3dTextureMapperPlanarReflection_class[];
#endif

extern const char         E3dTextureMapperPlanarReflection_classname[];


extern void               E3dTextureMapper_Init(E3dTextureMapper* PTextureMapper);
extern E3dTextureMapper*  E3dTextureMapper_New(E3dTextureMapperClass* PClass);

extern void               E3dTextureMapperPlanar_GetMatrix(const E3dTextureMapperPlanar* PMapper, const E3d2DTexture* PTexture, E3dMatrix PMatrix);

extern E3dTextureMapper*  E3dTextureMapper_Convert(E3dTextureMapper* PMapper, E3dTextureMapperClass* PNewClass);

extern void               E3d_TextureMappersUpdateWorldToMapperMatrix(const E3dCoordinate* PLocalToWorldMatrix, E3dTextureMapper** PMappers, const EIndex PNumOfMappers);

extern E3dTextureMapperClass*  E3dTextureMapper_ClassRegister(E3dTextureMapperClass* PClassTemplate);
extern void                    E3dTextureMapper_ClassDeactivate(E3dTextureMapperClass* PClass);
extern void                    E3d_TextureMapperClassesFree(void);

extern E3dTextureMapperClass*  E3dTextureMapper_ClassFindByName(const char* PClassName);
extern EBool                   E3d2DTexture_IsEmissive(E3d2DTexture* PTexture);


//----------------------------------------
// E3D/Sampler.c
//----------------------------------------
extern void  _E3d_TextureSample2D_bilinearSimple(const E3d2DTexture* PTexture, const E3dSTWithDiff* PST,  const char PTilingModeS, const char PTilingModeT, float* PResult);

extern void  E3d2DTexture_SetSampler(E3d2DTexture* PTexture, const EBool PMipMapsEnabled, const EBool PRendererHasAnisotropicFiltering);


//----------------------------------------
// GL/Texture.c
//----------------------------------------

// Possible responses from _SceneTextureGetGLImage()
//
enum
{
	E3dGLIMG_SUCCESS = 0,
	E3dGLIMG_ERROR
};

extern void           E3dGL_TextureSetFiltering(E3d2DTexture* PTexture, EpGLTexture* LGLTexture, GLenum LGLTarget, E3dTextureMapper* PMappe);
extern EpGLTexture*   E3dTexture_GLSetup(E3d2DTexture* P2DTexture, E3dTextureMapper* PMapper);
extern void           E3dGL_TextureSetImage(E3d2DTexture* PTexture, E3dTextureMapper* PMapper);
extern void           E3dGL_TextureFreeGLImages(E3d2DTexture* PTexture);
extern void           E3dGL_TextureFreeGLTextures(E3d2DTexture* PTexture);

extern void           E3dGL_TextureFilteringInvalidate(E3d2DTexture* PTexture);
extern int            E3dTextureMapper_GetGLTarget(const E3dTextureMapper* PMapper);

extern void           E3dGL_BindTextureToUnit(const GLuint PGLTextureEngine, const GLuint PGLTextureID);

extern EBool          E3d2DTexture_IsAnimated(E3d2DTexture* PTexture);




#if defined (__cplusplus) && (!(defined (__CPLUSLIBS)))
}	// Close scope of 'extern "C"' declaration
#endif

#endif	// _E3D_Texture_h
