//================================================================
// E3D
//
// - Basic type definitions
//
// AUTHOR:  Gabor Nagy
// DATE:    2019-Jun-23 14:03:08
//
// Copyright (C) 1994-2021 by Gabor Nagy. All rights reserved.
//================================================================
#ifndef	_E3D_Types_h
#define	_E3D_Types_h

#define E3D  1

#include <float.h>

#include <ETypes.h>
#include <Color/Color.h>

#define E3dWARN_PRECISION   1
#define E3dOBJECT_THREADS   1


#ifdef __SSE2__
//#define	E3dSIMD	1
#endif


#define E3dMAX_PRINTF_BUFFER_SIZE  4096

#define E3dSTARTF  -99999999.9
#define E3dENDF     99999999.9

//#define E3d64BIT_INTS

// Timer resolution
#define E3d_GetTime  E_GetMicroseconds


//----------------------------------------------------------------
// If the symbol E3dFLOAT_BITS is defined as 64, the 3D library
// will use double precision for all computations.
//----------------------------------------------------------------

#define E3dFLOAT_BITS     32
#define E3dGL_FLOAT_BITS  32

typedef int  E3dVertexSize;

// Basic operations
//
#define E3dADD  0x01
#define E3dSUB  0x02
#define E3dMUL  0x04
#define E3dDIV  0x08


// Transformation modes
//
enum
{
	E3dTmNONE = 0,
	E3dTmVIEW,
	E3dTmLOCAL,
	E3dTmGLOBAL,
	E3dTmPLANE,
	E3dTmREFERENCE,
	E3dTmMANIPULATOR
};


// Action codes for E3d_LayerFromScene etc.
enum
{
	E3dUPDATE = 0,  // Just update state
	E3dINSERT,
	E3dDELETE,
	E3dREORDER
};


//================================================================
// Set the value of a Resource in an EObject, from an E3dCoord3
//================================================================
#define E3d_SetResourceCoord3_s(mObject, mName,  mValue) do\
{\
	EResource*  lResource = EResource_FindByName(mObject->Class->Resources, mName);if(lResource)\
	{\
		switch(lResource->Type)\
		{\
			case E3dCOORD3_TYPE:\
				{\
					uint8_t*    lVariableAddress = (EPointerUInt)((uint8_t*)mObject) + (uint8_t*)(lResource->Address);\
					E3dCoord3*  lAddress = (E3dCoord3*)lVariableAddress;\
					*lAddress = mValue;\
				}\
			break;\
		}\
	}\
} while(0)


#define _E3d_PointSize(mIn, mOut) do\
{\
	mOut = mIn | 1;if(mOut > 25) mOut = 25;\
} while(0)



#if defined (__cplusplus) && (!(defined (__CPLUSLIBS)))
extern "C" {
#endif

typedef EIndex   E3dID;
typedef int16_t  E3dTxMapperIndex;
typedef int32_t  E3dDisplayFlags;
typedef int32_t  E3dViewModeFlags;

typedef struct
{
	EBool  Splines, Faces;
	EBool  MeshVertices, SplineCVs, SplineTangents, SplineNormals;

	EBool            HiddenObjectWireframe;      // E3dDM_Wireframe flags pulled out from HiddenObjectDisplayFlags, so it can be saved/loaded.
	EBool            UnselectedObjectWireframe, SelectedObjectWireframe;
} E3dShow;

#define E3dMAX_GRID_DEPTH  3   // 1-unit grid, 10-unit, 100-unig

typedef struct
{
	EcRGBAf32Color  BackgroundRGBAfColor,
	                GridRGBAfColors[E3dMAX_GRID_DEPTH],
	                AxisRGBAfColor,
	                WireRGBAfColor, HWireRGBAfColor, SelGeoWireRGBAfColor, SelSubGeoWireRGBAfColor, LockedItemRGBAfColor, TextureProjectorRGBAfColor,
	                VertexRGBAfColor, SelectedPointRGBAfColor, ActivePointRGBAfColor,
	                ActiveEdgeRGBAfColor, SelectedEdgeRGBAfColor,
	                Line1RGBAfColor, Line2RGBAfColor, HVertexRGBAfColor,
	                SelectedPolygonRGBAfColor, ShadedSelectedPolygonRGBAfColor,
	                NormalRGBAfColor,
	                TangentRGBAfColor,
	                BoneRGBAfColor;

	E3dShow         Show;

	EBool           UseZBuffer;
	int             AAMult;
} E3dDrawContext;



// Verbosity flags
enum
{
	E3dVbTEXTURES       = 0x0001,
	E3dVbTEXTUREMAPPING = 0x0002,
	E3dVbPOLYGONS       = 0x0004,
	E3dVbOPENGL         = 0x0008,
};

#define E3dVbALL (E3dVbTEXTURES|E3dVbTEXTUREMAPPING|E3dVbPOLYGONS|E3dVbOPENGL)

extern uint32_t  E3d_Verbosity;


// Some broken GLX implementations:
// - need a call to glXMakeCurrent before glXSwapBuffers
// - don't have glXMakeContextCurrent implemented (it's there, but it doesn't do anything)
//
#define E3dSUPPORT_BROKEN_GLX		1


// A generic NOTHING, NONE etc. value
//
#define E3dNONE			0

typedef int32_t   E3dVertexIndex;

#if E3dFLOAT_BITS == 64
#define E3dEPSILON        0.00000001  // TEST THIS
#define E3dMACHINE_ERROR  0.00000001
typedef double               E3dGLCoordinate;
#define E3dGLCoordinateType  GL_FLOAT

typedef double  E3dCoordinate;
typedef double  E3dWeight;
typedef double  E3dFColorComponent;

#define E3dCOORDINATE_TYPE  EResFLOAT64
#define E3dCOORD2_TYPE      EResFLOAT64_2
#define E3dCOORD3_TYPE      EResFLOAT64_3
#define E3dCOORD4_TYPE      EResFLOAT64_4

#define	E3dRADIANS_TO_NORMALIZED   0.15915494309189533577		// 1/(2 * PI)
#define	E3dRADIANS_TO_DEGREES     57.295779513082320877
#define E3dDEGREES_TO_RADIANS      0.017453292519943295769
#define E3dPI       3.14159265358979323846264
#define E3dE        2.718281828459045091  // The basis for the "natural" logarithm
#define E3dINV_PI  (1.0 / E3dPI)

#define E3dHUGE		1e36

#else

#define E3dEPSILON          0.00001f
#define E3dMACHINE_ERROR    0.00001f
typedef float                E3dGLCoordinate;
#define E3dGLCoordinateType  GL_FLOAT

typedef float  E3dCoordinate;
typedef float  E3dWeight;
typedef float  E3dFColorComponent;

#define E3dCOORDINATE_TYPE  EResFLOAT32
#define E3dCOORD2_TYPE      EResFLOAT32_2
#define E3dCOORD3_TYPE      EResFLOAT32_3
#define E3dCOORD4_TYPE      EResFLOAT32_4

#define	E3dRADIANS_TO_NORMALIZED   0.15915494f			// 1 / (2 * PI)
#define	E3dRADIANS_TO_DEGREES     57.2957795f
#define E3dDEGREES_TO_RADIANS	     0.017453292f
#define E3dPI       3.141592653
#define E3dE        2.718281828  // The basis for the "natural" logarithm
#define E3dINV_PI  (1.0 / E3dPI)

#define E3dHUGE  1e30

#endif // DOUBLE_PRECISION


#define E3dELEMENTS_IN_MATRIX	16

typedef E3dCoordinate  E3dMatrix[E3dELEMENTS_IN_MATRIX];



#define E3dOPAQUE       0x01
#define E3dTRANSPARENT  0x02


#define E3dSCALING      0x01
#define E3dROTATION     0x02
#define E3dTRANSLATION  0x04


// Callback reasons
//
enum
{
	E3dCR_HIERARCHY_ADDED = 1,
	E3dCR_HIERARCHY_REMOVED,
	E3dCR_HIERARCHY_CHANGED,
	E3dCR_NODE_SCALE,
	E3dCR_NODE_ROTATE,
	E3dCR_NODE_TRANSLATE,
	E3dCR_BRANCH_SCALE,
	E3dCR_BRANCH_ROTATE,
	E3dCR_BRANCH_TRANSLATE,
	E3dCR_TREE_SCALE,
	E3dCR_TREE_ROTATE,
	E3dCR_TREE_TRANSLATE,
	E3dCR_MATERIAL_ADDED,
	E3dCR_MATERIAL_REMOVED,
	E3dCR_MATERIAL_CHANGED,
	E3dCR_MATERIAL_NAME_CHANGED,
	E3dCR_2DTEXTURE_ADDED,
	E3dCR_2DTEXTURE_REMOVED,
	E3dCR_2DTEXTURE_CHANGED,
	E3dCR_2DTEXTURE_NAME_CHANGED,
	E3dCR_GEOMETRY_CHANGED,

	E3dCR_SELECTION_CHANGED,
	E3dCR_POLYGON_SELECTION,
	E3dCR_EDGE_SELECTION,
	E3dCR_POINTS_SELECTION,
	E3dCR_SPLINE_SEGMENT_SELECTED,

	E3dCR_HIDE_UNHIDE,
	E3dCR_LIGHT_CHANGED,
	E3dCR_CAMERA_CHANGED,
	E3dCR_WINDOW_REDRAW_START,
	E3dCR_WINDOW_REDRAW_END
};


#define E3dCR_ORIGIN_TRANSFORM  0x1000  // Flag, used in conjunction with E3dCR_NODE_SCALE etc.


#define E3dCallbackStructCore\
	int  Reason


// Generic Scene callback-struct
//
typedef struct
{
	E3dCallbackStructCore;
} E3dCallbackStruct;


// Callback-struct for 3DWindow redraws
//
typedef struct
{
	E3dCallbackStructCore;

	EChangeFlags	SceneChanged;
	EChangeFlags	RendererChanged;
} E3dDrawStartCallbackStruct;



// EUpdateFlags values
// These flags tell E3dObject_UpdateForDisplay() what should
// be updated in the rendering-specific internal structures.
//
#define E3dUF_LOCKUNLOCK                 0x000001    // Object was locked/unlocked, update wireframe color
#define E3dUF_NODE_TRANSFORM             0x000002    // Node transform changed
#define E3dUF_CV_POSITION                0x000004    // Spline CV position(s) changed

#define E3dUF_VERTEX_POSITION            0x000010    // Vertex position(s) changed
#define E3dUF_VERTEX_NORMAL              0x000020    // Vertex normal(s) changed
#define E3dUF_VERTEX_ST                  0x000040    // Update texture coordinates on Vertices
#define E3dUF_VERTEX_ATTRIBUTE           0x000080    // Vertex attribute(s) changed

#define E3dUF_POLYGON_NORMAL             0x000100    // Polygon normal(s) changed
#define E3dUF_REMOVE_STRIPS              0x000200    // Remove triangle strips
#define E3dUF_EDGES                      0x000400    // (Re)create Edge list
#define E3dUF_TRIANGULATE                0x000800    // (Re)create triangles
#define E3dUF_MATERIAL                   0x001000    // Update Materials
#define E3dUF_CONNECTIVITY               0x002000    // Mesh Polygon count/vertex list etc.
#define E3dUF_TESSELATION                0x004000    // Face->Vertices and Face->PolyGroup Polygons

#define E3dUF_CALL_CHANGEPROCS           0x010000    // Call ChangeProcs
#define E3dUF_CALL_OUTPUTS               0x020000    // Call Geometry Outputs
#define E3dUF_CALL_NODE_UPDATEPROCS      0x040000    // Call Update methods on Components of Nodes instantiating the Object
#define E3dUF_CALL_SCENE_CALLBACKS       0x080000    // Call Scene CallBacks

#define E3dUF_ALL                        0x0F7FF7

#define E3duf_CALLS      (E3dUF_CALL_CHANGEPROCS|E3dUF_CALL_OUTPUTS|E3dUF_CALL_NODE_UPDATEPROCS|E3dUF_CALL_SCENE_CALLBACKS)
#define E3duf_CALLBACKS  (E3dUF_CALL_CHANGEPROCS|E3dUF_CALL_NODE_UPDATEPROCS)
#define E3duf_NORMALS    (E3dUF_VERTEX_NORMAL|E3dUF_POLYGON_NORMAL)
#define E3duf_FULLGEO    (E3dUF_CONNECTIVITY|E3dUF_CV_POSITION|E3dUF_VERTEX_POSITION|E3dUF_VERTEX_NORMAL|E3dUF_POLYGON_NORMAL|E3dUF_REMOVE_STRIPS|E3dUF_EDGES|E3dUF_TRIANGULATE|E3dUF_VERTEX_ST)
#define E3duf_SHAPE      (E3dUF_VERTEX_POSITION|E3dUF_CV_POSITION|E3duf_NORMALS|E3dUF_VERTEX_ST)
#define E3duf_SPLINE     (E3dUF_CV_POSITION|E3dUF_CONNECTIVITY)






// E3dWindows_RedrawByChangeFlags() uses these flags that determine
// which windows should be redrawn and what the renderer should update.
// These should be set by:
// - SetFrameProcs of Animations
// - (undo/redo) operators
//

// Scene change flags
//
#define	E3dChgMATERIALTYPE          0x00000001
#define	E3dChgMATERIALPARAM         0x00000002

#define	E3dChgTEXTURE_IMAGE         0x00000004
#define	E3dChgTEXTURE_MAPPING       0x00000008
#define	E3dChgTEXTURE_FILTERING     0x00000010

#define E3dChgMATERIAL_ALL          (E3dChgMATERIALTYPE|E3dChgMATERIALPARAM|E3dChgTEXTURE_IMAGE|E3dChgTEXTURE_MAPPING|E3dChgTEXTURE_FILTERING)

#define	E3dChgLIGHT                 0x00000020
#define	E3dChgLIGHT_SAMPLING        0x00000040

#define	E3dChgCAMERA                0x00000080

#define	E3dChgPOINTSET              0x00000100

#define	E3dChgSPLINESHAPE           0x00000200
#define	E3dChgSPLINECONNECTIVITY    0x00000400
#define	E3dChgFACE                  0x00000800

#define	E3dChgMESHSHAPE             0x00001000
#define	E3dChgMESHTRANSFORM         0x00002000
#define	E3dChgVERTEX_ATTRIBUTE      0x00004000  // Color, Normal, TexCoords or Tangent - things that don't necessitate rebuilding KDTrees etc., but need in-place updates for RTriangles

#define	E3dChgHIDE_UNHIDE           0x00008000
#define	E3dChgTRANSFORM             0x00010000
#define	E3dChgHIERARCHY             0x00020000

#define	E3dChgREAD_TEXTURES         0x00040000
#define	E3dChgITEM_NAME             0x00080000
#define	E3dChgSCHEMATICS            0x00100000

#define	E3dChgANIMATION             0x00200000
#define	E3dChgANIMATION_TIME        0x00400000
#define	E3dChgANIMATION_SUBFRAME    0x00800000
#define	E3dChgANIMATION_SELECTION   0x01000000

#define	E3dChgSELECTION             0x02000000
#define	E3dChgLOCK_UNLOCK           0x04000000
#define	E3dChgWIREFRAME             0x08000000  // Wireframe symbol, e.g. bone width

#define	E3dChgATTRIBUTES            0x20000000

#define E3dSceneCHG_ALL             0x3FFFFFFF

// This is not a true "changed" flag, just used for E3dWindow_Redraw
#define	E3dChgPRE_VISUALIZE         0x40000000



// Flags specific to a Renderer
// These are not propagated by E3dRendererSet_MarkChanges to other Renderers within the RendererSet
//
#define	E3dRCHG_WINDOW_SIZE           0x00000001
#define	E3dRCHG_WINDOW_EXPOSE         0x00000002

#define	E3dRCHG_EXPOSURE              0x00000004
#define	E3dRCHG_RAY_SETTINGS          0x00000008
#define	E3dRCHG_ACCELERATOR_SETTINGS  0x00000010
#define	E3dRCHG_PHOTONMAP_SETTINGS    0x00000020

#define	E3dRCHG_ATTRIBUTES            0x00000040

#define	E3dRendererCHG_ALL            0x0000007F



#define	E3dChgGEOMETRY  (E3dChgPOINTSET|E3dChgSPLINESHAPE|E3dChgSPLINECONNECTIVITY|E3dChgFACE|E3dChgMESHSHAPE)



//--------------------------------
// Node / Object clone flags
//--------------------------------
// Continues after ECopyNAME
#define E3dCopyMODEL_UNIQUE_NAME   0x0002  // Generate unique names for Models
#define E3dCopyOBJECT_UNIQUE_NAME  0x0004  // Generate unique names for Objects
#define E3dCopyMATERIALS           0x0008
#define E3dCopyEDGES               0x0010
#define E3dCopyPOLYGONS            0x0020
#define E3dCopyCOMPONENTS          0x0040
#define E3dCopyLODS                0x0080
#define E3dCopyNODE_COMPONENTS     0x0100
#define E3dCopyANIMATIONS          0x0200



//----------------------------------------------------------------
// If these files are included, E3D will use OpenGL
//----------------------------------------------------------------
#ifdef USEOpenGL

#ifdef __APPLE__
#include <OpenGL/gl.h>
#else

#ifndef sgi
	#define E3dGLHAS_COLORINDEX_TEXTURE  1
	#define GL_GLEXT_PROTOTYPES          1
#endif

#include <GL/gl.h>
#endif	// __APPLE__


#ifndef sgi
	#define E3dGL_HAS_FBO                1
#endif


#ifdef sgi
#include <GL/glx.h>
#define GLXPbuffer  GLXPbufferSGIX
#else

#ifdef __APPLE__
 #include <OpenGL/glext.h>
#else
 #include <GL/glext.h>
#endif  // __APPLE__

#endif  // sgi

#endif // USEOpenGL



#ifdef E3d64BIT_INTS
typedef long long           E3dIndex;
typedef unsigned long long  E3dCount;
typedef long long           E3dInt;
typedef unsigned long long  E3dUInt;
#define E3dBitMaskBITS  64
#else
typedef int             E3dIndex;
typedef unsigned int    E3dCount;
typedef int             E3dInt;
typedef unsigned int    E3dUInt;
#define E3dBitMaskBITS  32
#endif  // E3d64BIT_INTS


//----------------------------------------------------------------
// If this symbol is defined, the 3D library uses GUI
// (Creates/manages pseudo models for lights, cameras etc.)
//----------------------------------------------------------------
#define _E3dGUI



#define	E3dMAX_HIERARCHY_DEPTH		256


//--------------------------------------------------------
// Macros
//--------------------------------------------------------

#define E3dM_MINMAX(mA, mB,  mMin, mMax)\
 if(mA < mB) { mMin = mA;mMax = mB; }\
 else { mMin = mB;mMax = mA; }

#define E3d_Min3(a,b,c) ( ((a) > (b)) ? ( ((b) > (c)) ? (c):(b)) : ( ((a) > (c)) ? (c) : (a)) )
#define E3d_Max3(a,b,c) ( ((a) < (b)) ? ( ((b) > (c)) ? (b):(c)) : ( ((a) > (c)) ? (a) : (c)) )


//--------------------------------------------------------
// Type definitions
//--------------------------------------------------------

#define	E3dNOMODEL	-1


enum { E3dGLSHORT = 0, E3dGLLONG, E3dGLSINGLE, E3dGLDOUBLE };


#define E3dMtxPos(r, c)	(c * 4 + r)


// What to include in bounding box calculations
//
#define E3dBB_NONE               0x00
#define E3dBB_SELECTED_GEOMETRY  0x01
#define E3dBB_SHADOW_CASTERS     0x02
#define E3dBB_SHADOW_RECEIVERS   0x04
#define E3dBB_TAGGED_POINTS      0x08
#define E3dBB_ALL                0x0F



enum
{
	E3dX = 0, E3dY, E3dZ, E3dW,  // Principal axes
	E3dXY, E3dXZ, E3dYZ,         // Principal planes
	E3dARBITRARY
};

#define E3dS	E3dX
#define E3dT	E3dY
#define E3dR	E3dZ
#define E3dQ	E3dW


// Rotation types
//
enum
{
	E3dXYZ = 1,
	E3dXZY,
	E3dYXZ,
	E3dYZX,
	E3dZXY,
	E3dZYX,
	E3dQUATERNION
};



#define E3dTRI_ERR		-1
#define E3dCOLINERR	3 * E3dMACHINE_ERROR	// One for each subtraction



#define E3dGLPIXELFORMAT	GL_ABGR_EXT
#define E3dGLPIXELTYPE		GL_UNSIGNED_BYTE

#define E3dALLOCQUICK	1
#define E3dALLOCFORGL	2

#define	E3dNO	0
#define	E3dYES	1


#define E3dFlagX	0x01
#define E3dFlagY	0x02
#define E3dFlagZ	0x04
#define E3dDoALL	0x07


#define E3dDoX		0x01
#define E3dDoY		0x02
#define E3dDoZ		0x04
#define E3dDoALL	0x07



#define E3dPOLY_DIRECTION	1  // 1: Exterior contours are defined counter-clockwise

typedef void*  E3dPointer;


typedef struct
{
	EBool  X, Y, Z;
} E3dBool3;


// Quaternion:
// X = AxisX * sin(angle * 0.5), Y = AxisY * sin(angle * 0.5), Z = AxisZ * sin(angle * 0.5), W = cos(angle*0.5)


#define _E3d_GenerateTangent_vp(mNormal, mDst) do\
{\
	const E3dCoord3  lNormal = mNormal;\
\
/* If there's no tangent given, generate one that:  */\
/* - is perpendicular to the normal                 */\
/* - lies in the X-Z plane                          */\
\
	if(E3d_Abs(lNormal.Y) > 0.75)\
	{\
		const E3dCoord3  lHelperTangent = { 0.0, 0.0, 1.0 };\
		E3d_Cross3_vvp(lNormal, lHelperTangent, mDst);\
	}\
	else\
	{\
		const E3dCoord3  lHelperTangent = { 0.0, 1.0, 0.0 };\
		E3d_Cross3_vvp(lNormal, lHelperTangent, mDst);\
	}\
	E3d_Normalize3_p(mDst);\
} while(0)


#define _E3d_GenerateTangent_pp(mNormal, mDst) do\
{\
	const E3dCoord3  lNormal = *mNormal;\
\
/* If there's no tangent given, generate one that:  */\
/* - is perpendicular to the normal                 */\
/* - lies in the X-Z plane                          */\
\
	if(E3d_Abs(lNormal.Y) > 0.75)\
	{\
		const E3dCoord3  lHelperTangent = { 0.0, 0.0, 1.0 };\
		E3d_Cross3_vvp(lNormal, lHelperTangent, mDst);\
	}\
	else\
	{\
		const E3dCoord3  lHelperTangent = { 0.0, 1.0, 0.0 };\
		E3d_Cross3_vvp(lNormal, lHelperTangent, mDst);\
	}\
	E3d_Normalize3_p(mDst);\
} while(0)


#define E3dCoord4_v(m,  mX, mY, mZ, mW)  E3dCoord3  m;m.X = mX;m.Y = mY;m.Z = mZ;m.W = mW


#ifdef E3dSIMD

#define E3dCoord2_v(m)                   E3dCoord2  m;m.unused2 = 0.0;m.unused3 = 0.0

#define E3dCoord2i_v(m,  mX, mY)         E3dCoord2  m = { .X= mX, .Y= mY, .unused0 = 0.0, .unused1 = 0.0 }

#define E3dCoord3_v(m)                   E3dCoord3  m;m.unused = 0.0
#define E3dCoord3i_v(m,  mX, mY, mZ)     E3dCoord3  m = { .X= mX, .Y= mY, .Z = mZ, .unused = 0.0 }

#define E3dFloat32_3_v(m)                E3dFloat32_3  m;m.unused = 0.0
#define E3dFloat32_3i_v(m,  mX, mY, mZ)  E3dFloat32_3  m = { .X = mX, .Y = mY, .Z = mZ, .unused= 0.0 }

#define E3dSIMD_Init_unused(m)           m.unused = 0.0

typedef struct
{
	float  X, Y, Z, unused;
} E3dFloat32_3 __attribute__((aligned(16)));

typedef struct
{
	union { E3dCoordinate	X;E3dCoordinate	U; };
	union { E3dCoordinate	Y;E3dCoordinate	V; };
	E3dCoordinate  unused0;
	E3dCoordinate  unused1;
} E3dCoord2;

typedef struct
{
	E3dCoordinate  X, Y, unused0, unused1;
} E3dCoord2 __attribute__((aligned(16)));

typedef struct
{
	E3dCoordinate  X, Y, Z, unused;
} E3dCoord3 __attribute__((aligned(16)));

typedef struct
{
	E3dCoordinate  X, Y, Z, W;
} E3dCoord4 __attribute__((aligned(16)));


// This type always has double-precision components
typedef struct
{
	double	X, Y, Z, unused;
} E3dDCoord3 __attribute__((aligned(32)));

#else	// E3dSIMD

#define E3dCoord2_v(m)                   E3dCoord2  m

#define E3dCoord2i_v(m,  mX, mY)         E3dCoord2  m = { .X= mX, .Y= mY }

#define E3dCoord3_v(m)                   E3dCoord3  m
#define E3dCoord3i_v(m,  mX, mY, mZ)     E3dCoord3  m = { .X= mX, .Y = mY, .Z = mZ }

#define E3dFloat32_3_v(m)                E3dFloat32_3  m
#define E3dFloat32_3i_v(m,  mX, mY, mZ)  E3dFloat32_3  m = { .X= mX, .Y= mY, .Z = mZ }

#define E3dSIMD_Init_unused(m)           m.unused = 0.0

typedef struct
{
	float  X, Y, Z;
} E3dFloat32_3;

typedef struct
{
//	E3dCoordinate	X, Y;
	union { E3dCoordinate	X;E3dCoordinate	U; };
	union { E3dCoordinate	Y;E3dCoordinate	V; };
} E3dCoord2;

typedef struct
{
	E3dCoordinate  X, Y, Z;
} E3dCoord3;

typedef struct
{
	E3dCoordinate  X, Y, Z, W;
} E3dCoord4;

typedef struct
{
	double  X, Y, Z;
} E3dDCoord3;
#endif	// E3dSIMD


#define E3dSTCore\
	E3dCoordinate  S, T;\
	E3dCoord3      Tangent;  /* Texture-space tangent (used for bump/normal mapping)	 */\
	uint16_t       Flags     /* For ST (UV) editing  E3dVtxSELECTED, E3dVtxTMPSELECTED */

// Texture coordinates
//
typedef struct
{
	E3dSTCore;
} E3dST;


static inline void E3dST_Init(E3dST* PST)
{
	PST->Flags = 0;
	E_SetXYZ_s(PST->Tangent,  0.0, 0.0, 0.0);
}


#define E3dCoord3SumCore\
	E3dCoord3      Vector;\
	int            Count

typedef struct
{
	E3dCoord3SumCore;
} E3dCoord3Sum;

#define E3dMAX_TEXTURE_ANISOTROPY	16


// Texture coordinates with MIP-mapping and anisotropic sampling support
//
typedef struct
{
	E3dSTCore;

	E3dST          STs[E3dMAX_TEXTURE_ANISOTROPY];int	NumOfSTs;		// Anisotropic filtering
	E3dCoordinate  dS, dT;
} E3dSTWithDiff;



#define	E3dMAX_NUM_OF_2DTEXTURES        16  // Max # of Textures per Material
#define	E3dMAX_NUM_OF_2DTEXTUREMAPPERS   8  // Max # of TextureMappers per Shadeable (e.g. PolyGroup)


#define E3dSurfacePointCore\
	E3dCoord3       Position,\
	                Normal;\
	E3dSTWithDiff   ST[E3dMAX_NUM_OF_2DTEXTUREMAPPERS];      /* Texture coordinates */\
	EcRGBAf32Color  TextureColors[E3dMAX_NUM_OF_2DTEXTURES]  /* Texture colors resulting from sampling the textures */

// Point used for shading surfaces and for TextureMappers
//
typedef struct
{
	E3dSurfacePointCore;
} E3dSurfacePointBase;


typedef struct
{
	E3dCoord4  AxisAngle;   // Exact axis (e.g. the X axis is (1.0, 0.0, 0.0)) and angle in degrees
	E3dCoord4  Quaternion;  // Quaternion equivalent for fast conversion to a matrix
} E3dRotation;

typedef struct
{
	E3dCoord3  Angles;    // Exact axis (e.g. the X axis is (1.0, 0.0, 0.0)) and angle in degrees
	uint8_t    Order;     // E.g. E3dZYX
} E3dEulerRotation;

typedef struct
{
	E3dCoordinate  A, B, C, D;    // Plane coefficients for the implicit formula (Ax+By+Cz+D=0)
} E3dPlane;


typedef union
{
	E3dPlane   Plane;
	E3dCoord4  Coord4;
} E3dCoordinate4;


#define E3dBBoxCore\
	union { E3dCoord3	Min;E3dCoordinate	Min4[4]; };\
	union { E3dCoord3	Max;E3dCoordinate	Max4[4]; }


typedef struct
{
	E3dBBoxCore;

	union { E3dCoord3  Center; E3dCoordinate  Center4[4];  };
	union { E3dCoord3  Extents;E3dCoordinate  Extents4[4]; };
} E3dBBoxCenterExtents;


typedef struct
{
	E3dBBoxCore;
} E3dBBox;

typedef struct
{
	union { E3dCoord3  Min; E3dCoordinate  Min4[4]; };
	union { E3dCoord3  Span;E3dCoordinate  Span4[4]; };
} E3dBBoxMinSpan;

#define MAX_HIERARCHYDEPTH	 32


#define MAX_MODELWINDOWNUM	4


enum
{
	E3dMDL_BINARY = 0,
	E3dMDL_ASCII
};



// Callback reasons
//
enum
{
	E3dCR_PICK = 1,
	E3dCR_ABORT    // E.g. Esc was pressed
};


// Interface call reasons
//
enum
{
	E3dICR_POINT_CHANGED = 1,
	E3dICR_POINTS_CHANGED
};


// EditProc return values
//
enum
{
	E3dEDITP_INFO_ADDED = 0,
	E3dEDITP_IN_USE,
	E3dEDITP_EDITING,
	E3dEDITP_NOTHING
};


// Orientations
//
#define E3dHORIZONTAL  0x01
#define E3dVERTICAL    0x02


// Schematics view connector types
//
enum
{
	E3dSHORTEST = 1,
	E3dPERPENDICULAR,
	E3dCURVED,
};


// Item (Model, Geometry etc.), action code
//
typedef int		(*E3dLockProc)(void*, short);

typedef struct
{
	char*        Name;
	E3dLockProc  LockProc;
} E3dLock;


enum
{
	E3dCENTER = 0,
	E3dMIN,
	E3dMAX,
	E3dBOTTOM,
	E3dVAL
};


#define	E3dGETRED        0x01
#define	E3dGETGREEN      0x02
#define	E3dGETBLUE       0x04
#define	E3dGETALPHA      0x08
#define	E3dGETINTENSITY  0x10


// Sub-camera indices for stereo and cube-map rendering
enum
{
	E3dCameraLEFT       = 0x01,
	E3dCameraRIGHT      = 0x02,
	E3dCameraPOSITIVE_X = 0x04,
	E3dCameraNEGATIVE_X = 0x08,
	E3dCameraPOSITIVE_Y = 0x10,
	E3dCameraNEGATIVE_Y = 0x20,
	E3dCameraPOSITIVE_Z = 0x40,
	E3dCameraNEGATIVE_Z = 0x80,
	E3dCameraALL        = 0xFF
};


#define E3dNORMALIZE	-999999.0


// Return flags for GLDraw and VkDraw methods / E3dNodeComponentClass  GLDrawFlags
//
typedef uint16_t  E3dHWDrawFlags;
#define E3dDRAW_DO_NOT_DRAW_GEOMETRY  0x0001  // Don't draw Mesh surface etc.
#define E3dDRAW_DO_NOT_DRAW_NODE      0x0002  // Don't draw Node Geometries
#define E3dDRAW_LAST                  0x0004  // GLDraw() etc. must be done after everything else had been drawn (e.g. requiers no Z-test overdraw, transparency etc.)
#define E3dDRAW_ALWAYS_ON_TOP         0x0008


extern char*       E3d_LaunchPath;
extern ECodePoint  E3d_CodePoint;

static inline E3dCoordinate  E3d_Min(const E3dCoordinate PA, const E3dCoordinate PB) { return((PA <= PB) ? PA : PB); }
static inline E3dCoordinate  E3d_Max(const E3dCoordinate PA, const E3dCoordinate PB) { return((PA >= PB) ? PA : PB); }

extern void   E3dShow_Init(E3dShow* PShow);
extern EBool  E3d_EnableBlankMessages;  // Enable warning message for "blank" functions

//--------------------------------
// App main.c
//--------------------------------
extern void  E3d_Log(const int PType, const char* PPrefix, const char* PMessage, ...);


#if defined (__cplusplus) && (!(defined (__CPLUSLIBS)))
}	// Close scope of 'extern "C"' declaration
#endif


#endif	// _E3D_Types_h
