//================================================================
// 3DPanel mode stack header file
//
// AUTHOR:  Gabor Nagy
// DATE:    2016-Oct-22 05:24:26
//
// Copyright (C) 1994-2021 By Gabor Nagy. All rights reserved.
//================================================================
#ifndef _E3D_ModeStack_h
#define _E3D_ModeStack_h

#include <Xe/Xe.h>


#if defined (__cplusplus) && (!(defined (__CPLUSLIBS)))
extern "C" {
#endif


// Pointer mode stacking
//
typedef struct E3dPointerModeClass_S  E3dPointerModeClass;
typedef struct E3dPointerMode_S       E3dPointerMode;

struct E3dPointerModeClass_S
{
	EBaseClassCore(E3dPointerMode);

	EBool  (*IsChildModeAllowed)(const E3dPointerMode* PMode, const E3dPointerMode* PChildMode);
};

struct E3dPointerMode_S
{
	EObjectCore(E3dPointerMode);

	XeString  Message;

	E3dID     PMID;                      // Unique ID returned by E3dp_PushPointerMode when this PointerMode was pushed onto the stack
	E3dID     MessageID;                 // Unique ID returned by E3dp_PushCompoundMessage when this PointerMode was pushed onto the stack
};


typedef struct E3dPointerModeEntry_S  E3dPointerModeEntry;

// PointerMode stack mode
//
struct E3dPointerModeEntry_S
{
	E3dID            PMID;
	E3dPointerMode*  Mode;    // Pointer to E3dPointerMode struct
};


extern E3dPointerMode*  E3dpPointerMode_ORBIT_STANDBY;
extern E3dPointerMode*  E3dpPointerMode_ORBIT_STANDBY_ORTHO;
extern E3dPointerMode*  E3dpPointerMode_ORBIT;

extern E3dPointerMode*  E3dpPointerMode_Z_SCROLL_ZOOM_STANDBY;
extern E3dPointerMode*  E3dpPointerMode_Z_SCROLL_ZOOM_SCROLL;
extern E3dPointerMode*  E3dpPointerMode_Z_SCROLL_ZOOM_ZOOM;
extern E3dPointerMode*  E3dpPointerMode_Z_SCROLL_ZOOM_SELECT_ZOOM_REGION;

extern E3dPointerMode*  E3dpPointerMode_Z_TRACK_ZOOM_STANDBY;
extern E3dPointerMode*  E3dpPointerMode_Z_TRACK_ZOOM_TRACK;
extern E3dPointerMode*  E3dpPointerMode_Z_TRACK_ZOOM_ZOOM;

extern E3dPointerMode*  E3dpPointerMode_DOLLY_STANDBY;
extern E3dPointerMode*  E3dpPointerMode_DOLLY_STANDBY_ORTHO;
extern E3dPointerMode*  E3dpPointerMode_DOLLY;

extern E3dPointerMode*  E3dpPointerMode_BANK_STANDBY;
extern E3dPointerMode*  E3dpPointerMode_BANK_STANDBY_ORTHO;
extern E3dPointerMode*  E3dpPointerMode_BANK;

extern E3dPointerMode*  E3dpPointerMode_CTRL_SCROLL_ZOOM_STANDBY;
extern E3dPointerMode*  E3dpPointerMode_CTRL_SCROLL_ZOOM_SCROLL;
extern E3dPointerMode*  E3dpPointerMode_CTRL_SCROLL_ZOOM_ZOOM;
extern E3dPointerMode*  E3dpPointerMode_CTRL_SCROLL_ZOOM_SELECT_ZOOM_REGION;

extern E3dPointerMode*  E3dpPointerMode_CTRL_ORBIT_DOLLY_TRACK_STANDBY;
extern E3dPointerMode*  E3dpPointerMode_CTRL_ORBIT_DOLLY_TRACK_ORBIT;
extern E3dPointerMode*  E3dpPointerMode_CTRL_ORBIT_DOLLY_TRACK_DOLLY;
extern E3dPointerMode*  E3dpPointerMode_CTRL_ORBIT_DOLLY_TRACK_TRACK;

extern E3dPointerMode*  E3dpPointerMode_ALT_ORBIT_TRACK_DOLLY_STANDBY;
extern E3dPointerMode*  E3dpPointerMode_ALT_ORBIT_TRACK_DOLLY_ORBIT;
extern E3dPointerMode*  E3dpPointerMode_ALT_ORBIT_TRACK_DOLLY_TRACK;
extern E3dPointerMode*  E3dpPointerMode_ALT_ORBIT_TRACK_DOLLY_DOLLY;

extern E3dPointerMode*  E3dpPointerMode_ALT_NONE_TRACK_ZOOM_STANDBY;
extern E3dPointerMode*  E3dpPointerMode_ALT_NONE_TRACK_ZOOM_TRACK;
extern E3dPointerMode*  E3dpPointerMode_ALT_NONE_SCROLL_ZOOM_SCROLL;
extern E3dPointerMode*  E3dpPointerMode_ALT_NONE_SCROLL_ZOOM_ZOOM;

extern E3dPointerMode*  E3dpPointerMode_DELETING;

extern E3dPointerMode*  E3dpPointerMode_SELECT_NODE;
extern E3dPointerMode*  E3dpPointerMode_SELECT_NODE_NODE;
extern E3dPointerMode*  E3dpPointerMode_SELECT_NODE_BRANCH;
extern E3dPointerMode*  E3dpPointerMode_SELECT_NODE_TREE;

extern E3dPointerMode*  E3dpPointerMode_SELECT_OBJECT;
extern E3dPointerMode*  E3dpPointerMode_SELECT_POLYGON;
extern E3dPointerMode*  E3dpPointerMode_SELECT_EDGE;
extern E3dPointerMode*  E3dpPointerMode_SELECT_OBJECT_COMPONENT_MULTI;

extern E3dPointerMode*  E3dpPointerMode_TAG_POINTS_STANDBY;
extern E3dPointerMode*  E3dpPointerMode_TAG_POINTS_TAG;
extern E3dPointerMode*  E3dpPointerMode_TAG_POINTS_UNTAG;
extern E3dPointerMode*  E3dpPointerMode_TAG_POINTS_INVERT;

extern E3dPointerMode*  E3dpPointerMode_PICK;

extern E3dPointerMode*  E3dpPointerMode_MOVE_POINT_STANDBY;
extern E3dPointerMode*  E3dpPointerMode_MOVE_POINT;

extern E3dPointerMode*  E3dpPointerMode_RECTANGLE_SELECT;


extern int                   E3dp_ModeChangeVerbosity;

extern E3dID                 E3dp_PushPointerMode(E3dPointerMode* PMode, XeString PMessage);
extern EBool                 E3dp_PopPointerMode(E3dPointerMode* PMode);
extern void                  E3dp_ResetPointerModeStack(void);
extern void                  E3dp_PointerModesFree(void);

extern E3dPointerModeEntry*  E3dp_GetCurrentPointerMode(void);
extern void                  E3dp_ResetWorkModes(void);


#if defined (__cplusplus) && (!(defined (__CPLUSLIBS)))
}	// Close scope of 'extern "C"' declaration
#endif

#endif	// _E3D_ModeStack_h
