//================================================================
// 3DPanel header file
//
// AUTHOR:  Gabor Nagy
// DATE:    2019-Jun-23 02:29:55
//
// Copyright (C) 1994-2021 By Gabor Nagy. All rights reserved.
//================================================================
#ifndef _E3D_Panel_h
#define _E3D_Panel_h

#include <Xe/Menu.h>

#ifndef _E3D_Node_h
#include <E3D/Node.h>
#endif

#ifndef _E3D_WindowAdv_h
#include <E3D/UI/WindowAdv.h>
#endif

#ifndef _E3D_Rendering_h
#include <E3D/Rendering.h>
#endif

#ifndef _E3D_Scene_h
#include <E3D/Scene.h>
#endif

#include <E3D/Languages.h>

#include <E3D/UI/Prefs.h>

extern XeApp            E3dp_XeApp;

extern E3dPolygonFlags  E3dp_PolygonSelectFlag;
extern uint16_t         E3dp_EdgeSelectFlag;

#define E3dDeleteSELECTED_NODES       0x0001
#define E3dDeleteSELECTED_OBJECTS     0x0002
#define E3dDeleteSELECTED_POLYGROUPS  0x0004
#define E3dDeleteSELECTED_POLYGONS    0x0008
#define E3dDeleteSELECTED_EDGES       0x0010
#define E3dDeleteSELECTED_POINTS      0x0020
#define E3dDisconnectSELECTED_POINTS  0x0040  // Remove corresponding VertexNodes of Polygons, but don't deleate the Mesh Vertex
#define E3dDeleteUNUSED_POINTS        0x0080


#if E3dFLOAT_BITS == 64
	#define XeScale_SetValue_Coord           XeScale_SetValue_f64
	#define Xe_ScaleGetValue_Coord           Xe_ScaleGetValue_f64
	#define XeScale_SetValueAndRange_Coord   XeScale_SetValueAndRange_f64
	#define XeScale_SetValueAndRanges_Coord  XeScale_SetValueAndRanges_f64
	#define Xe_ScaleSetIncrement_Coord       Xe_ScaleSetIncrement_f64
	#define m_Coordinate                    f64
	#define XeScaleNew_Coord             XeScale_Newf64
#else
	#if E3dFLOAT_BITS == 32
		#define XeScale_SetValue_Coord           XeScale_SetValue_f32
		#define Xe_ScaleGetValue_Coord           Xe_ScaleGetValue_f32
		#define XeScale_SetValueAndRange_Coord   XeScale_SetValueAndRange_f32
		#define XeScale_SetValueAndRanges_Coord  XeScale_SetValueAndRanges_f32
 		#define Xe_ScaleSetIncrement_Coord       Xe_ScaleSetIncrement_f32
		#define m_Coordinate                    f32
		#define XeScaleNew_Coord             XeScale_Newf32
	#else
ERROR
	#endif
#endif


#if defined (__cplusplus) && (!(defined (__CPLUSLIBS)))
extern "C" {
#endif

// Macros for Xe compound strings with Mode/Object names, mouse buttons etc.
//
#define E3dp_LMouseButtonS     XeFONT, XeFontFIXEDWIDTH,  XeCOLOR_RGBA,  LPrefs->ButtonNameRGBAp8Color, XeTEXT, "LMB",  XeCOLOR_RGBA, Xe_ForegroundRGBA8pColor,  XeFONT,XeFontDEFAULT
#define E3dp_MMouseButtonS     XeFONT, XeFontFIXEDWIDTH,  XeCOLOR_RGBA,  LPrefs->ButtonNameRGBAp8Color, XeTEXT, "MMB",  XeCOLOR_RGBA, Xe_ForegroundRGBA8pColor,  XeFONT,XeFontDEFAULT
#define E3dp_RMouseButtonS     XeFONT, XeFontFIXEDWIDTH,  XeCOLOR_RGBA,  LPrefs->ButtonNameRGBAp8Color, XeTEXT, "RMB",  XeCOLOR_RGBA, Xe_ForegroundRGBA8pColor,  XeFONT,XeFontDEFAULT
#define E3dp_RMouseButtonEscS  XeFONT, XeFontFIXEDWIDTH,  XeCOLOR_RGBA,  LPrefs->ButtonNameRGBAp8Color, XeTEXT, "RMB/Esc",  XeCOLOR_RGBA, Xe_ForegroundRGBA8pColor,  XeFONT,XeFontDEFAULT
#define E3dp_NodeS(mName)      XeCOLOR_RGBA, LPrefs->NodeBracketRGBAp8Color,  XeTEXT,  "[",  XeCOLOR_RGBA, LPrefs->NodeNameRGBAp8Color, XeTEXT, mName,  XeCOLOR_RGBA, LPrefs->NodeBracketRGBAp8Color,  XeTEXT,  "]",  XeCOLOR_RGBA, Xe_ForegroundRGBA8pColor
#define E3dp_ObjectS(mName)    XeCOLOR_RGBA, LPrefs->ObjectBracketRGBAp8Color,  XeTEXT,  "<",  XeCOLOR_RGBA, LPrefs->ObjectNameRGBAp8Color, XeTEXT, mName,  XeCOLOR_RGBA, LPrefs->ObjectBracketRGBAp8Color,  XeTEXT,  ">",  XeCOLOR_RGBA, Xe_ForegroundRGBA8pColor


typedef struct
{
	E3dNodeComponentCore;

//	uint32_t  SubCameras;
} E3dCameraDependenceNodeComponent;


#define	E3dpPOINTERMODE_STACKDEPTH  16


#define E3dSCHEMATICS_MIN_MAG    0.000001
#define E3dSCHEMATICS_MAX_MAG    0.136


#define E3dpNselectingEnabled  "selectingEnabled"
#define E3dpNdeletingEnabled   "deletingEnabled"

#define  E3dp3DPANEL_FRACTION_BASE  10000

// Maximum displayed grid-count for perspective mode
//
#define E3dpGRID_MAX  1000


enum { E3dp_MSG_ERROR };

// Transformation targets
//
enum
{
	E3dpTT_MODEL = 1,
	E3dpTT_MODEL_TMP_FROM_PTS,  // There were no selected points when the user tried to transform E3dpTT_SELECTED_POINTS, so we temporariy
	                            // switched to E3dpTT_MODEL. We'll have to switch back when the user releases the mouse button.
	E3dpTT_SELECTED_POINTS,
	E3dpTT_CENTER,
	E3dpTT_TEXTURE,
	E3dpTT_ANIMATION
};

#define caseE3dpTT_MODEL\
	case E3dpTT_MODEL:\
	case E3dpTT_MODEL_TMP_FROM_PTS


// Selecting modes
//
enum
{
	E3dpSELM_SINGLE,
	E3dpSELM_MULTI
};


#define E3dpSCALE_X      0x001
#define E3dpSCALE_Y      0x002
#define E3dpSCALE_Z      0x004
#define E3dpROTATE_X     0x010
#define E3dpROTATE_Y     0x020
#define E3dpROTATE_Z     0x040
#define E3dpTRANSLATE_X  0x100
#define E3dpTRANSLATE_Y  0x200
#define E3dpTRANSLATE_Z  0x400


//----------------------------------------
// Rendering hardware information
//----------------------------------------
// Vertex-array support
//
enum
{
	E3dVtxArrNONE = 0,
	E3dVtxArrNORMAL,
	E3dVtxArrFAST
};


// Shadow-mapping modes
//
enum
{
	E3dShadowNONE = 0,
	E3dShadowDUAL_TEXTURE,
	E3dShadowDEPTH_MAPPING
};



#ifndef _E3dPanel
#define _E3dPanel
typedef struct E3dPanel_S  E3dPanel;
#endif

struct E3dPanel_S
{
	EObjectCore(EObject);
	EList(, E3dWindow*, Windows);


	Widget       TopMatrixW, MiddleMatrixW;

	XeMenuRec*   VerticalMenuBar;
	Widget       MenuBarScrolledWindowW;

	Widget       M3DWindowLayoutW;

	Widget       RightPanelW,      // Panel to the right of the 3DWindows. Transform, Selection info, etc.
	             FoundObjectsListW,  // Names of selected Nodes
	             TransformTargetW, TransformModeW;
	Widget       ToolBarW, ToolBarBW, ToolBarPW,
	             StatusPanelW, StatusPanelBW, StatusPanelPW,
	             TimePanelW, TimePanelBW, TimePanelPW, TimeExtPanelMatrixW;

	int          GLFontFirstID;

	EBool        WindowRedrawEnabled, VerticalMenuBarOn, TransformPanelOn;
};





enum
{
	AM_ANIM,
	AM_MOVIE
};


#define E3dpICONSIZEBASE  26

#define E3dp_SetVMenuBArgs(mMode)\
do {\
	int  lSizeBase = 1;\
	switch(mMode)\
	{\
		case E3dpMBT_ICON:\
			EXeSetArg(XeNlabelType, XeIMAGE);\
			EXeSetArg(XeNmarginSizeX, 0);\
			EXeSetArg(XeNmarginLeft, 0);EXeSetArg(XeNmarginRight, 0);\
			EXeSetArg(XeNarrowSpacing, 0);\
			lSizeBase = E3dp_Prefs.VMenuIconSize;\
		break;\
\
		case E3dpMBT_TEXT:\
			EXeSetArg(XeNlabelType, XeSTRING);\
			EXeSetArg(XeNmarginSizeX, 0);\
			EXeSetArg(XeNmarginLeft, 0);EXeSetArg(XeNmarginRight, 0);\
			EXeSetArg(XeNarrowSpacing, 12);\
			lSizeBase = 32;\
		break;\
\
		case E3dpMBT_ICON_AND_TEXT:\
			EXeSetArg(XeNlabelType, XeIMAGE_STRING_VERTICAL);\
			EXeSetArg(XeNmarginSizeX, 0);\
			EXeSetArg(XeNmarginLeft, 0);EXeSetArg(XeNmarginRight, 0);\
			EXeSetArg(XeNarrowSpacing, 4);\
			lSizeBase = 32;\
		break;\
	}\
\
	EXeSetArg(XeNarrowWidth,   lSizeBase * 6 / E3dpICONSIZEBASE);\
	EXeSetArg(XeNarrowLength,  lSizeBase * 4 / E3dpICONSIZEBASE);\
	EXeSetArg(XeNmarginSizeY,  lSizeBase * 2 / E3dpICONSIZEBASE);\
} while(0)


typedef void  (*E3dCameraFlyToCallback)(const int PID, const float PSecondsLeft);


extern E3dLock  E3dObject_EditLock;
extern E3dLock  E3dNode_EditLock;


//--------------------------------
// From the app
//--------------------------------
extern void        E3dpCB_SaveImageAs(Widget PW, EPointer PClientData, EPointer PCallData);


extern XeCursor    E3dp_BusyCursor;
extern XeCursor    E3dp_OrbitCursor, E3dp_DollyCursor, E3dp_RollCursor, E3dp_TrackZoomCursor, E3dp_TrackCursor, E3dp_ZoomCursor;

extern void        E3dp_SetPushButtonArgs(EArg* PArgs, int* PArgCnt);



//--------------------------------
// GL/Scene.c
//--------------------------------
#define E3dGLM_PrintState() E3dGL_PrintState(__FUNCTION__, __FILE__, __LINE__)
extern void            E3dGL_PrintState(const char* PFunction, const char* PFile, const int PLine);
extern E3dHWDrawFlags  E3dGL_AimPointNodeGLDraw(void* P3DWin,  const E3dDrawContext* LDrawContext, const E3dDisplayFlags PDisplayFlags,  E3dNode* PNode, E3dNodeComponent* PInfo);
extern E3dHWDrawFlags  E3dGL_JointNodeGLDraw        (void* P3DWin,  const E3dDrawContext* LDrawContext, const E3dDisplayFlags PDisplayFlags,  E3dNode* PNode, E3dNodeComponent* PInfo);



typedef int  (*E3dShadowMapUpdateFunc)(E3dWindow* P3DWindow, E3dRenderer* PRenderer, E3dScene* PScene, E3dLightInstance* PLightInstance);
typedef int  (*E3dDrawShadowPassFunc)(E3dWindow* P3DWindow, E3dRenderer* PRenderer, const uint32_t PSubCameraID, const E3dDrawContext* PDrawContext,  E3dScene* PScene,  E3dLightInstance* PLightInstance,  const E3dMatrix PWorldToViewerMatrix, const EBool PDrawCamera);

extern E3dShadowMap*           (*E3dShadowMap_New)(void);
extern void (*E3dShadowMap_DrawFrustum)(E3dShadowMapT* PShadowMap);
extern E3dShadowMapUpdateFunc  E3dShadowMap_Update;
extern E3dDrawShadowPassFunc   E3d_DrawShadowPass;


//--------------------------------
// Panel/GL/FullWindowEffects.c
//--------------------------------
// Information used to generate a full-window pass fragment program
//
typedef struct
{
	int       AAMult;    // Anti-aliasing multiplier (e.g. 2 = 2x2 supersampling)
	EBool     UseGain, UseAdaptiveExposure, UseGamma,
	          UseGain1;  // Gain for 2nd texture (Texture1)
	EBool     Clamp;
	int       NumOfInputs;
	uint16_t  ReorientDstImage[2];
	float     LensCorrection, LensCorrectionR, LensCorrectionB;
	uint8_t   StereoMode;
} E3dFWCodeGenParams;

extern char*  E3dGL_FWGenerateFragmentProgram(const E3dFWCodeGenParams* PCodeGenParams, const int PLanguage);

extern int    E3dGL_Accumulate(E3dGPUContext* PGPUContext, E3dGPUFrameBuffer* PFBO0, E3dGPUFrameBuffer* PFBO1,  const EIndex PX, const EIndex PY,  const EIndex PSizeX, const EIndex PSizeY);

extern void   E3dGPUFrameBuffer_ToWindow(E3dGPUFrameBuffer* PFBO, E3dWindow* P3DWindow, E3dRenderer* PRenderer, const EpExposure* PExposure, const E3dBloom* PBloom, const int PAAMult, const uint16_t PSrcReorient);


#ifdef E3dGL_HAS_FBO
extern void   E3dGL_HDR_WindowFinishDrawing(E3dWindow* P3DWindow, E3dRenderer* PRenderer, E3dScene* PScene,  E3dGPUFrameBuffer* PFBO, E3dGPUFrameBuffer* PDstFBO, const int PAAMult, const uint16_t PSrcReorient);
#endif



//--------------------------------
// GL/Renderers.c
//--------------------------------
extern void   E3dGL_MainShadowMapInit(E3dScene* PScene, E3dWindow* P3DWindow, E3dRenderer* PRenderer, E3dLightInstance* PLightInstance);


//--------------------------------
// GL/Window.c
//--------------------------------
extern void   E3dWindow_DrawAreas(E3dpWindow* P3DWindow);



//--------------------------------
// Panel/Camera.c
//--------------------------------
extern E3dNodeComponentClass*  E3dCameraDependenceNodeComponent_class;

extern E3dModel*   E3d_SceneAddCameraModels(E3dScene* PScene,  E3dCamera* PCamera, E3dModel* PCameraModel,  E3dModel** PAimModelRet, const char* PCameraName);

extern EBool       E3dp_CameraControlButtonPress(E3dpWindow* P3DWindow,  const int PPtrX, const int PPtrY,  const int PButton, const unsigned int PModifiers);
extern EBool       E3dp_CameraControlButtonRelease(E3dpWindow* P3DWindow, const int PPtrX, const int PPtrY,  const int PButton, const unsigned int PModifiers);
extern int         E3dp_CameraControlPointerMove(E3dWindow* P3DWindow, const int PPtrX, const int PPtrY, const unsigned int PState);

extern int         E3dp_CameraFlyTo(E3dpWindow** P3DWindows, const EIndex PNumOf3DWindows, const float PDuration, const float PTimeStep, E3dCameraFlyToCallback PCallback);
extern float       E3dp_SceneFrameCameras(E3dScene* PScene, E3dpWindow* P3DWindow, E3dpWindow*** LFlyToWindowsP, EIndex* LNumOfFlyToWindowsP,  E3dBBox* PAllBBox,  unsigned int PWhatToInclude);

extern E3dCamera*  E3d_IsActiveCameraNodeOnWindow(const E3dNode* PNode, E3dpWindow* P3DWindow);


EList(extern, E3dShadowMap*, E3d_ShadowMaps);
extern EBool          E3dp_RedrawShadowMap;

extern void           E3dp_DisplayExit(void);
extern void           E3dGL_GLSLFree(void);


//--------------------------------
// Panel/MenuBar.c
//--------------------------------
extern void    E3dp_SetSelectionMode(int PMode);
extern void    E3dp_MenuBarOnOff(E3dPanel* P3DPanel, const EBool PSwitch);


//--------------------------------
// Panel/Panel.c
//--------------------------------
extern const EObjectClass  E3dPanel_class[];

extern XeFontSize*  E3dp_SchematicsFontSize;

extern EBool    E3dp_GUIFreeOnExit;
extern int      (*E3dp_TextureGUIUpdate)(E3dMaterial* PMaterial, E3d2DTexture* PTexture, EChangeFlags PChangedFlags);
extern int      (*E3dRendererSet_RenderFramesFunc)(E3dRendererSet* PRendererSet, E3dScene* PScene, E3dCamera* PCamera, E3dModel* PCameraModel);

extern int      E3dp_SelectionMode;

extern EBool    E3dp_Exiting;
extern E3dDrawContext  E3dp_NormalDrawContext;

extern E3dpWindow*     E3dp_Renderer3DWindow;

extern void            E3dp_GUIUpdateForNewScene(void);

extern void            E3dp_AddDrawStartCallback(EChangeFlags PEventMask, ECallbackFunc PCallbackProc, EPointer PClientData);
extern EIndex          E3dp_RemoveDrawStartCallback(ECallbackFunc PCallbackFunc, EPointer PClientData);

extern void            E3dp_AddDrawEndCallback(EChangeFlags PEventMask, ECallbackFunc PCallbackFunc, EPointer PClientData);
extern EIndex          E3dp_RemoveDrawEndCallback(ECallbackFunc PCallbackFunc, EPointer PClientData);


//--------------------------------
// Panel/PanelLayout.c
//--------------------------------
extern void    E3dWindows_Layout(E3dPanel* P3DPanel);
extern void    E3dWindows_PanelSetAttachments(E3dPanel* P3DPanel);
extern void    E3dp_MenuBarSetPanelAttachments(E3dPanel* P3DPanel);


//--------------------------------
// Panel/Renderer.c
//--------------------------------
extern const ETime      E3d_MaxInteractiveFrameMilliseconds;

extern void             E3dRenderer_GUIFromResources(ECoreObject* PObject, void* PPanelIn, void* PDrawerIn, int PAction);
extern int              E3dRendererSet_RenderFrame(E3dRendererSet* PRendererSet, E3dScene* PScene);

extern E3dRendererSet*  E3dp_RendererSet;
extern EIndex           E3d_BuiltInRendererClasses_Count;

extern E3dGPUContext*   E3dGL_Context;
extern E3dGPUContext*   E3dGL_ContextMSAA4;
extern E3dGPUContext*   E3dGL_ContextStereo;

#ifdef E3dGL_HAS_FBO
	extern E3dGPUFrameBuffer*  E3dGL_SharedFBO0;extern E3dWindow*  E3dGL_SharedFBO0Last3DWindow;
	extern E3dGPUFrameBuffer*  E3dGL_SharedFBO1;extern E3dWindow*  E3dGL_SharedFBO1Last3DWindow;
#endif

EList(extern, E3dPanel*, E3dp_Panels);

extern Widget      E3dp_RendererCanvasDialogW;
extern Widget      E3dp_RendererCanvasW;


// Default plugin icons, linked to ToolBar.a
//
extern EpImagePtr  _Image_Plugin, _Image_Plugin_Active, _Image_Plugin_Armed;

extern EpImagePtr  _Image_Key, _Image_Key_Active, _Image_Key_Armed;

extern EUndoStack  E3dp_UndoStack;

extern Widget      E3dp_TopShellW;

#define	E3dMAX_CLASS_RESOURCES  256

extern void          E3dScene_Read2DTextures(E3dScene* PScene, const EBool PForceReread, const EBool PAnimatedOnly);

extern void          E3dpCB_RenderAbort(EPointer PClientData);

#define E3dpOPENGL_CANVAS  1

extern int     (*E3dp_RenderFrameFunc)(E3dpWindow* P3DWindow, E3dScene* PScene,  E3dCamera* PCamera, E3dModel* PCameraModel, const EChangeFlags PSceneChangeFlags, const EChangeFlags PRendererChangeFlags);
extern int     (*E3dp_RenderWindowPopupProc)(E3dpWindow* P3DWindow);
extern int     E3dp_RenderWindowPopup(E3dpWindow* P3DWindow);

extern int     E3d_RayTrace(E3dWindow* P3DWindow, E3dRayTracer* PRenderer, E3dScene* PScene, E3dTraceRayFunc PTraceRayProc);


extern void    E3dPanel_RedrawWindows(E3dPanel* P3DPanel, const E3dViewModeFlags PViewModes, const E3dDisplayFlags PDisplayMode, const EChangeFlags PSceneChangedFlags, const EChangeFlags PRendererChangedFlags);

extern XeVisual*     E3dGL_Visual;
extern XeVisual*     E3dGL_VisualMSAA4;
extern XeVisual*     E3dGL_VisualStereo;

extern E3dPanel*     E3dPanel_New(const char* PName);
extern E3dPanel*     E3dPanel_Create(const char* PName, E3dScene* PScene, Widget PTopShellW, Widget PParentW,  const int PCols, const int PRows,  void (*PExitFunction)(void));

extern void          E3dp_Exit(void);


extern void          E3dp_UndoStackAddOps(EOps* POps, const char* PName);
extern void          E3dp_UndoStackAddOperator(EOperator* POperator, const char* PName);
extern EChangeFlags  E3dp_RunOperators(EOps* POps, const int PAction);
extern void          E3dp_Undo(void);
extern void          E3dp_Redo(void);


//--------------------------------
// Panel/Schematics.c
//--------------------------------
extern void    E3dNodeHrc_RefresSchemPositions(E3dModel* PRootModel);
extern int     E3dScene_LayoutOnSchematics(E3dScene* PScene, const EBool PAutoFrameSchematicsViews);


//--------------------------------
// Panel/ToolBar.c
//--------------------------------
extern Widget  E3dTool_FolderWs[];int  E3dTool_FolderWs_Count;

extern Widget  E3dTool_LightMatrixW,
               E3dTool_CreateMatrixW,
               E3dTool_SplineMatrixW,
               E3dTool_SurfaceMatrixW,
               E3dTool_GeometryMatrixW,
               E3dTool_CloneMatrixW,
               E3dTool_TransformMatrixW,
               E3dTool_HierarchyMatrixW,
               E3dTool_AnimationMatrixW;

extern void    E3dp_ToolBarOnOff(E3dPanel* P3DPanel, const EBool PSwitch);
extern void    E3dp_ToolBarExit(void);


//--------------------------------
// Panel/TransformPanel.c
//--------------------------------
extern int        E3dp_SelectionMode;

extern void       E3dp_TransformsInit(const char* PDataPath);
extern void       E3dp_TransformsExit(void);
extern void       E3dp_TransformPanelExit(void);

extern E3dMatrix  E3dp_WorldToLocalMatrix, E3dp_LocalToWorldMatrix;

extern E3dGeometry*  E3dp_ActiveGeometry;
extern E3dVertex*    E3dp_ActiveVertex;
extern int           E3dp_ActiveSubCV;
extern int           E3dp_OldPointFlags;

extern int           E3dp_BasePtrX, E3dp_BasePtrY;    // Pointer at time of button press
extern E3dCoord3     E3dp_TranslateBase;
extern E3dCoord3     E3dp_ScaleBase;

EList(extern, E3dShapeSnapshot*, E3dp_ShapeSnapshots);

extern EChangeFlags  E3dp_InteractiveTransformInit(E3dWindow* P3DWindow,  const int PPtrX, const int PPtrY);
extern EChangeFlags  E3dp_InteractiveTransformFree(void);

extern EOperator*    E3dp_MovePoint(E3dpWindow* P3DWindow,  const int PPtrX, const int PPtrY,  unsigned int PModifierState);
extern EChangeFlags  E3dp_StopMovePoint(void);



extern E3dLight*     E3dp_SceneAddDefaultLight(E3dScene* PScene);

extern EChangeFlags  E3dp_InteractiveTransform(E3dWindow* P3DWindow, const int PPtrX, const int PPtrY, const unsigned int PButtonState);

extern void          E3dp_PointMoveInit(E3dpWindow* P3DWindow,  const int PPtrX, const int PPtrY,  const E3dCoordinate PX, const E3dCoordinate PY, const E3dCoordinate PZ,  const EBool PStoreForUndo);

extern E3dCoord3     E3dp_ActivePointOriginalPos, E3dp_ActivePointMoveOffset;

extern EBool         E3dp_PointMoveInit2D(E3dpWindow* P3DWindow,  const int PPtrX, const int PPtrY,  const E3dCoordinate PX, const E3dCoordinate PY,  const EBool PStoreForUndo);
extern EBool         E3dp_PointMoveInitPlane(E3dpWindow* P3DWindow, int PPtrX, int PPtrY, E3dCoordinate PX, E3dCoordinate PY, E3dCoordinate PZ, EBool PStoreForUndo);

extern void          E3dp_InitSplineCVMove(E3dSpline* PSpline, E3dSplineCV* PCV, const EIndex PPointIndex, const int PItemType, E3dMatrix PLocalToWorldMatrix, E3dMatrix PWorldToLocalMatrix, E3dCoord3* LPosition);

#define E3dTrPanelTRANSFORM_VALUESTATES  0x01
#define E3dTrPanelOBJECT_LIST            0x02
#define E3dTrPanelALL                    0x03

typedef struct
{
	int       TransformTarget, TransformMode;
	uint16_t  ScaleFlags, RotateFlags, TranslateFlags;
	EBool     Transforming;                              // Scaling, Rotating, Translating etc. selected Models (keep buffer of rendered unselected models, grid, axis...)
	EBool     TransformObjectsSensitive;

	EListA(, EObject*, FoundObjects);

	Widget    ScaleXW, ScaleYW, ScaleZW,
	          RotateXW, RotateYW, RotateZW,
	          TranslateXW, TranslateYW, TranslateZW,
	          ScaleXBW, ScaleYBW, ScaleZBW,
	          RotateXBW, RotateYBW, RotateZBW, 
	          TranslateXBW, TranslateYBW, TranslateZBW,
	          FindW;
	Widget    TTB_ObjectW, TTB_TagW, TTB_CenterW, TTB_AnimW;
} E3dTransformPanel;

extern E3dTransformPanel  E3dp_TransformPanel;

extern void    E3dTransformPanel_Update(E3dTransformPanel* PPanel, const uint8_t PScaleFlags, const uint8_t PRotateFlags, const uint8_t PTranslateFlags,  const int PTransformTarget,  const uint16_t PUpdateFlags);
extern void    E3dTransformPanel_UpdateScalingUI(E3dTransformPanel* PPanel,  const E3dCoordinate PX, const E3dCoordinate PY, const E3dCoordinate PZ,  const uint16_t PFlags);
extern void    E3dTransformPanel_UpdateRotationUI(E3dTransformPanel* PPanel,  const E3dCoordinate PX, const E3dCoordinate PY, const E3dCoordinate PZ,  const uint16_t PFlags);
extern void    E3dTransformPanel_UpdateTranslationUI(E3dTransformPanel* PPanel,  const E3dCoordinate PX, const E3dCoordinate PY, const E3dCoordinate PZ,  const uint16_t PFlags);

extern void    E3dp_TransformPanelOnOff(E3dPanel* P3DPanel, const EBool PSwitch);

extern void    E3dp_ModelRemoveHrcFromScene(E3dModel* PRootModel);


//----------------------------------------
// Panel/GL/Display.c
//----------------------------------------
extern void    E3dGL_ViewDrawInit(E3dWindow* P3DWindow);
extern void    E3dGL_RenderWindowRedraw(E3dWindow* P3DWindow, E3dRenderer* PRenderer, E3dScene* PScene,  const EBool PSwapBuffers);


//----------------------------------------
// GL/Renderers.c
//----------------------------------------
extern void  E3dWindow_DrawExtras(E3dWindow* P3DWindow, E3dRenderer* PRenderer, const E3dDrawContext* PDrawContext,  E3dScene* PScene);



//----------------------------------------
// Application
//----------------------------------------
//--------------------------------
// gui.c
//--------------------------------
extern int     E3dp_ScaleWidth, E3dp_SliderLength, E3dp_SliderWidth;
extern void    E3dp_AddBuiltInFunctionsGUI(void);
extern void    E3dp_RemoveBuiltInFunctionsGUI(void);


//--------------------------------
// main.c
//--------------------------------
extern void    App_ReadSettings(const char* PName, EResource* PResources, void* PStructAddress);
extern void    App_SaveSettings(const char* PName, EResource* PResources, void* PStructAddress);


extern void    E3dp_IOModuleShowGUI(void* PIOModule, EIOModuleGUI* PIOModules, const EIndex PN, Widget PFormatsOptionMenuW);

EList(extern, ECallbackRec, _E3dp_DrawStartCallbacks);
EList(extern, ECallbackRec, _E3dp_DrawEndCallbacks);

#if defined (__cplusplus) && (!(defined (__CPLUSLIBS)))
}	// Close scope of 'extern "C"' declaration
#endif

#endif	// _E3D_Panel_h
