//================================================================
// 3DPanel header file
//
// AUTHOR:  Gabor Nagy
// DATE:    2019-Jun-23 22:23:32
//
// Copyright (C) 1994-2021 By Gabor Nagy. All rights reserved.
//================================================================
#ifndef _E3D_Panels_h
#define _E3D_Panels_h

#include <Xe/DropDownPanel.h>

#if defined (__cplusplus) && (!(defined (__CPLUSLIBS)))
extern "C" {
#endif

// Object panel
//
typedef struct
{
	XeInfoPanelCore;

	Widget      NameW, InstancesLabelW, OutputsLabelW, OutputsEnableW, OutputsShowW, OutputsReplaceInputW;
	E3dObject*  Object;       // Currently edited object
	EObject*    SubObject;    // PolyGroup, if Object is a Mesh, contour Spline of a Face

	EIndex      PointIndex;   // E.g. Current CV of a Spline

	EInterface  InputInterface;
} E3dObjectPanel;


typedef struct
{
	Widget  NameW, TypeW,  ValuesMatrixW, XW, YW, ZW, WW,  RemoveW;
} E3dTransformGUI;

// Node panel
//
typedef struct
{
	XeInfoPanelCore;

	Widget    NameW;
	E3dNode*  Node;

	Widget    TopMatrixW,
	          NodeOutputsMatrixW,
	          NodeOutputsLabelW, NodeOutputsEnabledW, NodeOutputsShowW,
	          BranchOutputsLabelW, BranchOutputsEnabledW, BranchOutputsShowW,
	          RotOrderMatrixW, RotOrderW,
	          TransformsScrolledWindowW, TransformsVScrollBarW, TransformsMatrixW,
	          StatsLW, StatsValLW,
	          BBoxValLW;
	EIndex    ObjectIdx;			// E.g. Current object index passed to E3dNodePanel_Create and E3dNodePanel_Update

	EBool     ConvertEulerAnglesWhenOrderChanges;  // When the user changes the rotation order on a Node, convert the rotation angles (to maintain the same orientation), on the selected Node or branch

	EListA(, E3dTransformGUI, Transforms);
} E3dNodePanel;


#define E3dNAME_FIELD                 0x01
#define E3dFOLLOW_SELECTION_CHECKBOX  0x02


//--------------------------------
// Panel/ObjectPanel.c
//--------------------------------
EList(extern, E3dObjectPanel*, E3dp_ObjectPanels);

extern void       (*E3dObject_EditFunc)(E3dModel* PModel, E3dObject* PObject, EIndex PLODIndex,  EObject* PSubObject, const EIndex PPointIndex,  E3dObjectComponentClass* POpenFolderForThisComponentClass);
extern EBool      (*E3dObject_IsEditedFunc)(E3dObject* PObject);

extern EBool      (*E3dPolyGroup_IsEditedFunc)(E3dPolyGroup* PPolyGroup);
extern void       (*E3dPolyGroup_UpdateDialogFunc)(const E3dPolyGroup* PPolyGroup);

extern E3dObjectPanel*  E3dObjectPanel_Create(Widget PParentW, const unsigned int PFlags, E3dObject* PObject, EObject* PSubObject, EIndex PSubSubObjIndex, E3dObjectComponentClass* POpenFolderForThisComponentClass);
extern void       E3dObjectPanel_UpdateSubSub(E3dObjectPanel* PObjectPanel, E3dObject* PObject, EObject* PSubObject, EIndex PPointIndex, E3dObjectComponentClass* POpenFolderForThisComponentClass);
extern void       E3dObjectPanel_Update(E3dObjectPanel* PObjectPanel, E3dObject* PObject, E3dObjectComponentClass* POpenFolderForThisComponentClass);
extern void       E3dObjectPanel_Free(E3dObjectPanel* PObjectPanel);

extern void       E3dObject_Edit(E3dObject* PObject, E3dObjectComponentClass* POpenFolderForThisClass);
extern EBool      E3dObject_IsEdited(E3dObject* PObject);

//--------------------------------
// Panel/NodePanel.c
//--------------------------------
extern void           (*E3dNode_EditFunc)(E3dNode* PNode, E3dNodeComponentClass* POpenFolderForThisClass);
extern EBool          (*E3dNode_IsEditedFunc)(E3dNode* PNode);

extern E3dNodePanel*  E3dNodePanel_Create(Widget PParentW, E3dNode* PNode, const EIndex PObjectIndex, E3dNodeComponentClass* POpenFolderForThisComponentClass);
extern void      E3dNodePanel_Update(E3dNodePanel* PNodePanel, E3dNode* PNode, const EIndex PObjectIndex, E3dNodeComponentClass* POpenFolderForThisComponentClass);
extern void      E3dNodePanel_UpdateValues(E3dNodePanel* PPanel, E3dNodeComponentClass* POpenFolderForThisComponentClass);
extern void      E3dNodePanel_Free(E3dNodePanel* PNodePanel);

extern Widget    E3dNode_DialogBoxNew(char* PName, E3dNode* PNode, E3dNodePanel** PNodePanelPtr, XeCallbackFunc PCB, EpImage* PIconImage);
extern Widget    E3dNode_DialogBoxNewVa(char* PName, E3dNode* PNode, E3dNodePanel** PNodePanelPtr, ...);
extern void      E3dp_JointNodeInfoGUIProc(E3dNode* PNode, E3dNodeComponent* PInfo, void* PPanelIn, void* PDrawerIn, int PAction);
extern void      E3dNode_Edit(E3dNode* PNode, E3dNodeComponentClass* POpenFolderForThisClass);
extern EBool     E3dNode_IsEdited(E3dNode* PNode);
extern void      E3dNode_Edit(E3dNode* PNode, E3dNodeComponentClass* POpenFolderForThisComponentClass);
extern void      E3dNodePanel_Exit(void);


#if defined (__cplusplus) && (!(defined (__CPLUSLIBS)))
}	// Close scope of 'extern "C"' declaration
#endif


#endif	// _E3D_Panels_h
