//================================================================
// Directory library header file
//
// AUTHOR:  Gabor Nagy
// DATE:    2018-Apr-01 06:55:17
//
// Copyright (C) 1994-2021 by Gabor Nagy. All rights reserved.
//================================================================

#ifndef _EDir_h
#define _EDir_h

#include <stdarg.h>		// For varargs
#include <sys/types.h>
#include <sys/stat.h>

#include <EParam.h>

#if defined (__cplusplus) && (!(defined (__CPLUSLIBS)))
extern "C" {
#endif

#ifdef WIN32
#include <dos.h>
#include <dir.h>
#endif

#ifdef UNIX
#include <sys/dir.h>
#endif

#include <time.h>
#include <stdlib.h>

#include <EStrings.h>


#ifdef UNIX
#ifdef sgi
#include <sys/statvfs.h>
#endif
#endif

#define EdirENTRY_ALLOC_STEP	16

#define	MAX_LINELEN	512

enum { EdirSORT_ALPHA=0, EdirSORT_TIME, EdirSORT_SIZE };
enum { EdirENTRY_NORMAL=0, EdirENTRY_CURRENT_DIR, EdirENTRY_PARENT_DIR };

// Flags to determine what to display in the directory list
//
#define EdirATTR_FLAGS    0x01
#define EdirUSER          0x02
#define EdirGROUP         0x04
#define EdirNUM_OF_LINKS  0x08
#define EdirSIZE          0x10
#define EdirDIR_SIZE      0x20
#define EdirTIME          0x40


#define EdirFILE_FILTER_CASE_SENSITIVE  0x01
#define EdirDIR_FILTER_CASE_SENSITIVE   0x02

typedef struct
{
	int                 IsDir, IsNFS;
	nlink_t             NumOfLinks;
	uid_t               UID;
	gid_t               GID;
	unsigned long long  Size;
	mode_t              Mode;
	time_t              ATime, CTime, MTime;
	char                Name[MAXPATHLEN + 1];
#ifdef UNIX
	unsigned long long  RealSize;                  // In case it's a symbolic link
	char                RealName[MAXPATHLEN + 1];
#endif
	EIOModule*          Format;
} EdirEntry;

extern EBool		EDir_ShowCurrentDir, EDir_ShowParentDir,
			EDir_DisplayDirSizes, EDir_ComputeDirSizes, EDir_ComputeCurrentDirSize, EDir_ComputeParentDirSize,
			EDir_DoNFS, EDir_ShowNFSDirs, EDir_FollowLinks;

extern EStrFilterTokens	EDir_UNIXFilterTokens;	// Filter token strings for UNIX "*" "[" "]"


extern int     EDir_SortOrder;
extern time_t  EDir_CurrentTime;
extern int     EDir_Max_EntryNum;

extern int     EDir_PathType(const char* PPath, const EBool PResolveSymLink);
extern int     EDir_GetDir(const char* PEntryName, const uint16_t PFlags, const void* PDirFilterIn, const void* PFileFilterIn, const EStrFilterTokens* PFilterTokens, EdirEntry** PEntriesPtr, EIndex* PNumOfEntriesPtr, EIndex* PNumOfDirsPtr, EIndex* PMaxNumOfEntriesPtr, long long* PTotalSizeRet);
extern int     Edir_RecurseDir(const char* PDirectory, const EBool PDoNFS, const EBool PDoSymLinks, EIndex* PNumOfDirsRet, int (*LCallback)(EdirEntry* PEntry, const char* PPath, const EIndex PNumOfDirsSoFar, const EIndex PNumOfFilesSoFar, EPointer PClientData), EPointer PClientData);
extern int     Edir_GetSizeOfDir(const char* PDirectory, const EBool PDoNFS, const EBool PDoSymLinks, long long* PTotalSizeRet, EIndex* PNumOfFilesRet, EIndex* PNumOfDirsRet);
extern void    EDir_EntryGetFormat(EdirEntry* PEntry);

#if defined (__cplusplus) && (!(defined (__CPLUSLIBS)))
}	// Close scope of 'extern "C"' declaration
#endif

#endif // _EDir_h
