//================================================================
// General plugin framework
//
// AUTHOR:  Gabor Nagy
// DATE:    2018-Apr-01 23:15:46
//
// Copyright (C) 1994-2021 by Gabor Nagy. All rights reserved.
//================================================================

#ifndef _EPlugins_h
#define _EPlugins_h

#include <EParam.h>
#include <EResource.h>
#include <EClass.h>
#include <ELog.h>

#if defined (__cplusplus) && (!(defined (__CPLUSLIBS)))
extern "C" {
#endif


typedef struct EPlugin_S      EPlugin;
typedef struct EPluginLock_S  EPluginLock;


typedef int	(*EPluginLockDelProc)(EPlugin* PPlugin, EPluginLock* PLock);
typedef int (*EPluginInitFunc)(EPlugin* PPlugin);      // Function to call at program start in the DSO
typedef int (*EPluginExitFunc)(const EBool PExiting);  // Name of the function to call when unloading the plugin

struct EPluginLock_S
{
	EPluginLockDelProc  LockDel;
};


typedef struct EPluginComponent_S       EPluginComponent;
typedef struct EPluginComponentClass_S  EPluginComponentClass;

struct EPluginComponentClass_S
{
	ECoreClassCore(EPluginComponent);
};



// Plugin structure
//
struct EPlugin_S
{
	EList(, EPluginLock*, Locks);           // The Plugin is in use (it has a dialog box displayed, it globally references Objects etc.). See E_PluginLockDel
	void*            DLHandle;              // DL handle associated with the DSO (necessary for unloading the DSO)
	char*            Name;                  // Set from "E_PluginName" or from the "Plugins.list" file
	char*            Info;                  // Summary of the plugin from "E_PluginInfo"
	EPluginInitFunc  InitFunction;          // Function to call at program start in the DSO
	EPluginExitFunc  ExitFunction;          // Name of the function to call when unloading the plugin
	char*            DSOName;               // Name of the Dynamic Shared Object (DSO)
	char*            PathName;              // Path where the DSO was found
	char*            ErrorMessage;          // If there was a problem loading this plugin...
	EResource*       Resources;             // List of resources (settings) set by E_PluginInit()
	char*            ResourceFileName;
	EBool            Load;                  // Load plugin on startup

	EList(, EPluginComponent, Components);  // Plugin global variables
};


extern int    EPlg_PluginInit(void* PItem, void* PTmpData);
extern int    EPlg_PluginGetValue(void* PP, void* PTmpData, const char* PCmdStr, const char* PValStr);
extern int    EPlg_PluginLoad(void* PItems, const int PNumOfItems, void* PItem, const char* PPathName, void* PTmpData, ELogFunc PLogFuntion);
extern void   EPlg_PluginUnload(EPlugin* PPlugin, const EBool PExiting);
extern void   EPlg_PluginFree(EPlugin* PPlugin, const EBool PFreeStruct);

extern EPluginLock*  E_PluginLockAdd(EPlugin* PPlugin, EPluginLockDelProc PLockDel);
extern int           E_PluginLockDel(EPlugin* PPlugin, EPluginLock** PLockP, const EBool PNotify);


#if defined (__cplusplus) && (!(defined (__CPLUSLIBS)))
}	// Close scope of 'extern "C"' declaration
#endif

#endif // _EPlugins_h
