//================================================================
// Threads header file
//
// AUTHOR:  Gabor Nagy
// DATE:    2019-Jun-23 01:36:54
//
// Copyright (C) 1994-2021 By Gabor Nagy. All rights reserved.
//================================================================
#ifndef _EThreads_h
#define _EThreads_h


#include <pthread.h>

#include <ETypes.h>



#ifdef __cplusplus
extern "C" {
#endif


typedef struct EThreadSet_S  EThreadSet;

#define EThreadCore\
	EThreadSet*  ThreadSet;\
	int          Number;\
	pthread_t    PThread;\
	EBool        PThreadCreated

typedef struct
{
	EThreadCore;
} EThread;


#define EThreadSetCore\
	pthread_mutex_t  ThreadStatus_mutex;\
	pthread_cond_t   Cond;\
\
	EBool            Initialized;  /* This helps avoid multiple initializations when an object with a ThreadSet (e.g. an E3dRenderer subclass) runs though its constructor chain */\
	EBool            RunThreads;   /* Set to TRUE by E_ThreadsStart and set to FALSE by E_ThreadsStop/E_ThreadsStopAndWait */\
\
	EThread*         Threads;EIndex  NumOfThreads, NumOfAllocatedThreads;\
	int              ThreadStructSize, NumOfThreadsStartingUp, NumOfRunningThreads

struct EThreadSet_S
{
	EThreadSetCore;
};


// Scanline-parallel process
//
#define EThreadsScanlinesInfoCore\
	pthread_mutex_t  GetWorkload_mutex;\
	EIndex_2         Size;\
	EIndex           NextScanline;\
	void*            CustomData

typedef struct
{
	EThreadsScanlinesInfoCore;
} EThreadsScanlinesInfo;

#define EScanlinesThreadCore\
	EThreadCore;\
	EThreadsScanlinesInfo*  SharedData  /* Data shared between threads */

typedef struct
{
	EScanlinesThreadCore;
} EScanlinesThread;


extern void   EThreadSet_Init(EThreadSet* PThreadSet, const int PThreadStructSize, const EBool PForceInit);
extern void   EThreadSet_Clear(EThreadSet* PThreadSet);
extern void   EThreadSet_FreeContents(EThreadSet* PThreadSet);

extern void   E_ThreadsInit(EThreadSet* PThreadSet, const int PNumOfThreads);
extern void   E_ThreadsStart(EThreadSet* PThreadSet, void *(*PStartRoutine)(void*));
extern EBool  E_ThreadsWaitForFinish(EThreadSet* PThreadSet);
extern EBool  E_ThreadsStop(EThreadSet* PThreadSet);
extern EBool  E_ThreadsStopAndWait(EThreadSet* PThreadSet);

extern void   EThreads_ScanlinesRun(EThreadSet* PThreadSet, EThreadsScanlinesInfo* PProc,  const size_t PThreadSize, void (PThreadInit)(EThread* PThread, void* PCustomData), void (PThreadFree)(EThread* PThread, void* PCustomData),
  void* (*PThreadMain)(void* PThread), void* PCustomData,
  void (*PProgressMessageFunc)(const char* PMessage, ...), const char* PProgressMessage, void (*PProgressFunc)(const float PCcomplete, void (*PAbortCallback)(EPointer PCustomData), EPointer PAbortFuncCustomData), void (*PAbortFunc)(EPointer PCustomData), EPointer PAbortFuncCustomData);

#ifdef __cplusplus
}	// Close scope of 'extern "C"' declaration
#endif


#endif // _EThreads_h
