//========================================================================
// E core library
//
// Undo/redo
//
// AUTHOR:  Gabor Nagy
// DATE:    2016-Feb-11 07:47:56
//
// Copyright (C) 1994-2021 by Gabor Nagy. All rights reserved.
//========================================================================

#ifndef _EUndo_h
#define _EUndo_h

#ifndef _EClass_h
#include <EClass.h>
#endif

#if defined (__cplusplus) && (!(defined (__CPLUSLIBS)))
extern "C" {
#endif

// Operator codes
//
enum
{
	EOpNONE = 0,
};


// Status Content codes
//
enum
{
	EActionNONE = 0
};


// Function codes for Proc() in EOperator
//
enum
{
	EOpDO = 1,
	EOpUNDO
};


typedef struct EOperator_S       EOperator;
typedef struct EOperatorClass_S  EOperatorClass;


// Operator list
//
typedef struct
{
	char*  Name;
	EList(, EOperator*, Operators);
	EListA(, EOpNode, OpNodes);     // Filled in by the Operator->Proc()s to indicate what changed
} EOps;


typedef EChangeFlags  (*EOperatorProc)(EOps* POps, EOperator* POperator, const int PAction);


#define EOperatorCore\
	EObjectCore(EOperator);\
	EChangeFlags  ChangeFlags  /* Instance-specific change flags. E.g. animation curve editing adds E3dChgANIMATION_SELECTION to E3dOperatorAddDelSplineCVs, to make EQX redraw animation views */

#define EOperatorClassCore\
	EBaseClassCore(EOperator);\
	EOperatorProc  Proc    /* The function that performs the operation  */


struct EOperator_S
{
	EOperatorCore;
};

struct EOperatorClass_S
{
	EOperatorClassCore;
};





typedef struct
{
	int           Ptr, NumOfSteps, NumOfStepsAllocated;
	int           MaxUndoSteps;
	EChangeFlags  Changed;    // Flags indicating what changed
	EOps**        Steps;
} EUndoStack;



extern int		EOp_AllocStep;


extern EOperator*       EOp_OperatorNew(EOperatorClass* PClass);

extern EOperatorClass*  EOp_OperatorClassFindByName(const char* PName);
extern EOperatorClass*  EOp_OperatorClassRegister(EOperatorClass* PClassDescriptor);
extern void             EOp_OperatorClassDeactivate(EOperatorClass* PClass);
extern void             EOp_OperatorClassesFree(void);


extern void      EOp_StackAddOps(EUndoStack* PStack, EOps* POps, const char* PName);

extern void      EOp_OpsInit(EOps* POps);
extern void      EOp_OpsFree(EOps* POps);
extern void      EOp_OpsFreeContents(EOps* POps);
extern void      EOp_OpsClear(EOps* POps);

extern uint32_t  EOp_OpsRun(EOps* POps, const int PAction);

extern void      EOp_UndoStackInit(EUndoStack* PUndoStack);
extern void      EOp_UndoStackFree(EUndoStack* PUndoStack);
extern void      EOp_OperatorOps1Append(EUndoStack* PUndoStack, EOperator* POperator, const char* PName);

extern EOps*     EOp_Undo(EUndoStack* PStack);
extern EOps*     EOp_Redo(EUndoStack* PStack);


#if defined (__cplusplus) && (!(defined (__CPLUSLIBS)))
}	// Close scope of 'extern "C"' declaration
#endif

#endif	// _EUndo_h
