//================================================================
// lib2D
//
// Camera, higher-level code. Image processing etc.
//
// AUTHOR:  Gabor Nagy
// DATE:    2019-Jun-23 17:29:14
//
// Copyright (C) 1994-2021 by Gabor Nagy. All rights reserved.
//================================================================
#ifndef _EpCameraRec_h
#define _EpCameraRec_h

#ifndef EClass_h
#include <EClass.h>
#endif  // EClass_h

#include <Image/Camera.h>

#if defined (__cplusplus) && (!(defined (__CPLUSLIBS)))
extern "C" {
#endif


typedef struct  EpCameraRec_S  EpCameraRec;

typedef int  (*EpCameraInitCallbackProc)(EpCameraRec* PCameraRec, EPointer PClientData);


struct EpCameraRec_S
{
	EpCamera*      Camera;

	EIndex         LastDrawnFrame;

	EPointer       ClientData;

	EpGLTexture*   DstOpenGLTexture;  // If not NULL, capture into this OpenGL texture

// Multi-threading
//
	pthread_t      CameraInit_pthread;
	XeIntervalId   CameraInit_CheckTimerID;

	EpPixelFormat  RecordedFrames_PixelFormat;
	EIndex_2       RecordedFrames_Size;
	EListA(, uint8_t*, RecordedFrames);  // Pixel buffers used when recording to memory

	EpImage*       TmpAccumulationImage;
	EpImage*       AccumulationImage;

	EpCameraInitCallbackProc  CameraInitCallbackProc;EPointer  CameraInitCallbackClientData;
};


typedef struct EpCamerasPanel_S  EpCamerasPanel;

#define EpCameraMAX_NRESOLUTIONS_FOR_RADIOBUTTONS  4

enum
{
	EMicrosecond = 1,
	EMillisecond,
	ESecond
};

typedef struct
{
	EpCamerasPanel*  CamerasPanel;

	EpCameraRec*     CameraRec;

	XeDrawer         Folder;
	Widget           CameraActiveW,  StatusMatrixW, StatusLabelW, ExposureProgressScaleW,  StatusLabel1W;
	Widget           ShutterTimeW, ShutterTimeUnitOptionMenuW, HWGainW, FramesPerSecondW;
	Widget           LumaGammaW;

	int              ShutterTimeUnit;  // EMicrosecond, EMillisecond, or ESecond

	Widget           FlipMatrixW,  FlipXW, FlipYW;

	Widget           TonemappingSeparatorW, TonemappingHMatrixW;
	Widget           PickGrayW;
	Widget           WhiteBalanceRW, WhiteBalanceGW, WhiteBalanceBW,  GainW,  GammaW,  ColorSaturationW,  AdaptiveExposureW;

	Widget           CapturePixelFormatWs[EpPixelFormatCOUNT];

	Widget           Demosaic_MatrixW,  Demosaic_RawW, Demosaic_DropInW, Demosaic_NearestW, Demosaic_BilinearW, Demosaic_MalvarW, Demosaic_VNGW;

	Widget           HDR_ButtonW, HDR_MatrixW, HDR_T1W, HDR_K1W,  HDR_T2W, HDR_K2W;

	Widget           ResolutionsMatrixW;
	EList(, Widget, ResolutionRadioButtonWs);
	Widget           ResolutionOptionMenu, ResolutionsOptionMenuPaneW;
	EList(, Widget, ResolutionOptionMenuButtonWs);

	Widget           ROI_ButtonW, ROI_MatrixW, ROI_PositionXW, ROI_PositionYW, ROI_SizeXW, ROI_SizeYW;

	Widget           BadPixelCorrectionW;

	Widget           CoolingMatrixW, Cooling_OnW, Cooling_PowerW, Cooling_TargetTemperatureW;

	Widget           FilterWheelMatrixW, FilterWheelRadioMatrixW;
	EList(, Widget, FilterWheelSlotRadioButtonWs);
	Widget           FilterWheelAutoCycleBW, FilterWheelAutoCycleTextW;

	Widget           GraticuleMatrixW;Widget  GraticuleNoneW, GraticuleCrosshairW, GraticuleCircleW;
	Widget           FocusAssistW;
} EpCameraPanel;


typedef void (*EpCamerasPanelCallbackFunc)(EpCamerasPanel* PCamerasPanel, EpCameraPanel* PCameraPanel, const int PReason, EPointer PClientData);

// Callback reasons
enum
{
	EpCR_CAMERA_STATUS_CHANGED = 1,
	EpCR_WHITEBALANCE_PICK_GRAY
};

struct EpCamerasPanel_S
{
	Widget                      ParentW;
	Widget                      ApplyChangesToAllCamerasW, StatusLabelW;
	EBool                       ApplyChangesToAllCameras;

	EpCamerasPanelCallbackFunc  CallbackFunc;
	EPointer                    CallbackClientData;

	EListA(, EpCameraPanel*, CameraPanels);

	EBool                       CreatePickGrayButtons;

	XeIntervalId  CameraGUIChangeCheckTimerID;
};

extern EpCamerasPanel*  EpCameras_Panel;
extern XeIntervalId     EpCamera_RedrawTimerID;

extern EpCameraRec*     EpCameraRec_New(const char* PName);
extern void             EpCameraRec_Free(EpCameraRec* PCameraRec);
extern int              EpCameraRec_Disconnect(EpCameraRec* PCameraRec);

extern EpCameraPanel*   EpCameraPanel_New(EpCamerasPanel* PCamerasPanel, Widget PParentW, EpCameraRec* PCameraRec);
extern void             EpCameraPanel_Free(EpCameraPanel* PPanel);
extern void             EpCameraPanel_Update(EpCameraPanel* PCameraPanel);

extern EpCamerasPanel*  EpCamerasPanel_New(Widget PParentW,  EpCamerasPanelCallbackFunc PCallbackFunc, EPointer PCallbackClientData,  EpImage* PIconImage);
extern void             EpCamerasPanel_Free(EpCamerasPanel* PCamerasPanel);
extern void             EpCamerasPanel_Update(EpCamerasPanel* PCamerasPanel, EpCameraRec** PCameraRecs, const EIndex PCameraRecs_Count);

extern int              EpCB_CameraRecConnectCamera(EpCameraRec* PCameraRec, EPointer PClientData);

extern void             EpCameraRec_ConnectCamera(EpCameraRec* PCameraRec, EpCameraInitCallbackProc PCallbackProc, EPointer PClientData);
extern void             EpTH_CameraGUIChangeCheck(EPointer PClientData, XeIntervalId* PTID, const int PCount);


extern EpCamerasPanelCallbackFunc  EpCamera_CallbackFunc;
extern EPointer                    EpCamera_CallbackClientData;


#if defined (__cplusplus) && (!(defined (__CPLUSLIBS)))
}  // Close scope of 'extern "C"' declaration
#endif

#endif // _EpCameraRec_h
