//================================================================================
// lib2D image library
//
// Image header file
//
// AUTHOR:  Gabor Nagy
// DATE:    2019-Jun-23 23:31:14
//
// Copyright (C) 1994-2021 by Gabor Nagy. All rights reserved.
//================================================================================
#ifndef _EImageOps_h
#define _EImageOps_h

#include <stdio.h>

#ifndef _EClass_h
#include <EClass.h>
#endif

#ifndef _EImage_h
#include <Image/Image.h>
#endif

#include <EUndo.h>


#if defined (__cplusplus) && (!(defined (__CPLUSLIBS)))
extern "C" {
#endif



#define Ep_ConvolveWithBorders(pTmpVals, pAddFunc, pShouldProcessPixel, pSetPixelFunc, pMax) do\
{\
	const EIndex  LKernelSizeXH = PKernelSizeX >> 1, LKernelSizeYH = PKernelSizeY >> 1;\
	const EIndex  LMaxX = LSizeX - LKernelSizeXH, LMaxY = LSizeY - LKernelSizeYH;\
\
/*--------------------------------------------------------*/\
/* Convolve Image 'interior' (no kernel mirroring)        */\
/*--------------------------------------------------------*/\
	{\
		EIndex  LY = LKernelSizeYH, LRowOffset = LY * LSizeX;for(;LY < LMaxY;LY++, LRowOffset += LSizeX)\
		{\
			EIndex  LX = LKernelSizeXH;for(;LX < LMaxX;LX++)\
			{\
				const EIndex  LPixelOffset = LX + LRowOffset;if(pShouldProcessPixel(LPixelOffset))\
				{\
					pTmpVals;\
\
					EIndex  LKY = -LKernelSizeYH, LKC = 0;for(;LKY <= LKernelSizeYH;LKY++)\
					{\
						EIndex  LKX = -LKernelSizeXH;for(;LKX <= LKernelSizeXH;LKX++, LKC++)\
						{\
							const float  LK = PKernel[LKC];pAddFunc(LPixelOffset + LKX + LKY * LSizeX, LK);\
						}\
					}\
					pSetPixelFunc(LPixelOffset, pMax);\
				}\
				else LDstPixels[LPixelOffset] = LSrcPixels[LPixelOffset];\
			}\
		}\
	}\
\
/*--------------------------------------------------------*/\
/* Convolve Image 'border' (kernel mirroring)             */\
/*--------------------------------------------------------*/\
/* Top bottom edge  */\
\
	{\
		EIndex  LY = 0, LRowOffset = LY * LSizeX;for(;LY < LKernelSizeYH;LY++, LRowOffset += LSizeX)\
		{\
			EIndex  LX = 0;for(;LX < LSizeX;LX++)\
			{\
				const EIndex  LPixelOffset = LX + LRowOffset;if(pShouldProcessPixel(LPixelOffset))\
				{\
					pTmpVals;\
\
					EIndex  LKY = -LKernelSizeYH, LKC = 0;for(;LKY <= LKernelSizeYH;LKY++)\
					{\
						EIndex  LKX = -LKernelSizeXH;for(;LKX <= LKernelSizeXH;LKX++, LKC++)\
						{\
							EIndex  LXN = LX + LKX, LYN = LY + LKY;\
							if(LXN < 0) LXN *= -1;else if(LXN >= LSizeX) LXN = LSizeX - LXN + LSizeX - 2;\
							if(LYN < 0) LYN *= -1;else if(LYN >= LSizeY) LYN = LSizeY - LYN + LSizeY - 2;\
							const float  LK = PKernel[LKC];pAddFunc(LXN + LYN * LSizeX, LK);\
						}\
					}\
					pSetPixelFunc(LPixelOffset, pMax);\
				}\
				else LDstPixels[LPixelOffset] = LSrcPixels[LPixelOffset];\
			}\
		}\
	}\
\
\
/* Bottom edge  */\
\
	{\
		EIndex  LY = LSizeY - LKernelSizeYH, LRowOffset = LY * LSizeX;for(;LY < LSizeY;LY++, LRowOffset += LSizeX)\
		{\
			EIndex  LX = 0;for(;LX < LSizeX;LX++)\
			{\
				const EIndex  LPixelOffset = LX + LRowOffset;if(pShouldProcessPixel(LPixelOffset))\
				{\
					pTmpVals;\
\
					EIndex  LKY = -LKernelSizeYH, LKC = 0;for(;LKY <= LKernelSizeYH;LKY++)\
					{\
						EIndex  LKX = -LKernelSizeXH;for(;LKX <= LKernelSizeXH;LKX++, LKC++)\
						{\
							EIndex  LXN = LX + LKX, LYN = LY + LKY;\
							if(LXN < 0) LXN *= -1;else if(LXN >= LSizeX) LXN = LSizeX - LXN + LSizeX - 2;\
							if(LYN < 0) LYN *= -1;else if(LYN >= LSizeY) LYN = LSizeY - LYN + LSizeY - 2;\
							const float  LK = PKernel[LKC];pAddFunc(LXN + LYN * LSizeX, LK);\
						}\
					}\
					pSetPixelFunc(LPixelOffset, pMax);\
				}\
				else LDstPixels[LPixelOffset] = LSrcPixels[LPixelOffset];\
			}\
		}\
	}\
\
\
/* Left edge  */\
\
	{\
		EIndex  LY = LKernelSizeYH, LRowOffset = LY * LSizeX;for(;LY < LMaxY;LY++, LRowOffset += LSizeX)\
		{\
			EIndex  LX = 0;for(;LX < LKernelSizeXH;LX++)\
			{\
				const EIndex  LPixelOffset = LX + LRowOffset;if(pShouldProcessPixel(LPixelOffset))\
				{\
					pTmpVals;\
\
					EIndex  LKY = -LKernelSizeYH, LKC = 0;for(;LKY <= LKernelSizeYH;LKY++)\
					{\
						EIndex  LKX = -LKernelSizeXH;for(;LKX <= LKernelSizeXH;LKX++, LKC++)\
						{\
							EIndex  LXN = LX + LKX, LYN = LY + LKY;\
							if(LXN < 0) LXN *= -1;else if(LXN >= LSizeX) LXN = LSizeX - LXN + LSizeX - 2;\
							if(LYN < 0) LYN *= -1;else if(LYN >= LSizeY) LYN = LSizeY - LYN + LSizeY - 2;\
							const float  LK = PKernel[LKC];pAddFunc(LXN + LYN * LSizeX, LK);\
						}\
					}\
					pSetPixelFunc(LPixelOffset, pMax);\
				}\
				else LDstPixels[LPixelOffset] = LSrcPixels[LPixelOffset];\
			}\
		}\
	}\
\
\
/* Right edge  */\
\
	{\
		EIndex  LY = LKernelSizeYH, LRowOffset = LY * LSizeX;for(;LY < LMaxY;LY++, LRowOffset += LSizeX)\
		{\
			EIndex  LX = LSizeX - LKernelSizeXH;for(;LX < LSizeX;LX++)\
			{\
				const EIndex  LPixelOffset = LX + LRowOffset;if(pShouldProcessPixel(LPixelOffset))\
				{\
					pTmpVals;\
\
					EIndex  LKY = -LKernelSizeYH, LKC = 0;for(;LKY <= LKernelSizeYH;LKY++)\
					{\
						EIndex  LKX = -LKernelSizeXH;for(;LKX <= LKernelSizeXH;LKX++, LKC++)\
						{\
							EIndex  LXN = LX + LKX, LYN = LY + LKY;\
							if(LXN < 0) LXN *= -1;else if(LXN >= LSizeX) LXN = LSizeX - LXN + LSizeX - 2;\
							if(LYN < 0) LYN *= -1;else if(LYN >= LSizeY) LYN = LSizeY - LYN + LSizeY - 2;\
							const float  LK = PKernel[LKC];pAddFunc(LXN + LYN * LSizeX, LK);\
						}\
					}\
					pSetPixelFunc(LPixelOffset, pMax);\
				}\
				else LDstPixels[LPixelOffset] = LSrcPixels[LPixelOffset];\
			}\
		}\
	}\
} while(0)


//================================================
// Function prototypes
//================================================

//--------------------------------
// ImageOps.c
//--------------------------------
extern EOperatorClass*  Ep_OperatorClassFindByName(const char* PName);
extern EOperatorClass*  Ep_OperatorClassRegister(EOperatorClass* PClassDescriptor);
extern void             Ep_OperatorClassDeactivate(EOperatorClass* PClass);
extern void             Ep_OperatorClassesFree(void);

extern void             Ep_OpsFree(void);

extern int      EpImage_FlipH(EpImage* PSImage, EpImage* PDImage);
extern int      EpImage_FlipV(EpImage* PSImage, EpImage* PDImage);

extern int      EpImage_Crop(const EpImage* PSImage, EpImage* PDImage, const EIndex PLeft, const EIndex PRight, const EIndex PTop, const EIndex PBottom,  const int PPixelType, const int PComponentsPerPixel, const int PPixelComponentType,  const EcRGBAiColor* PMarginColor);

typedef struct
{
	float  Mins[EcMAX_NUM_OF_SPECTRAL_SAMPLES], Maxes[EcMAX_NUM_OF_SPECTRAL_SAMPLES];
} EpRanges_f32;

typedef struct
{
	uint16_t  Mins[EcMAX_NUM_OF_SPECTRAL_SAMPLES], Maxes[EcMAX_NUM_OF_SPECTRAL_SAMPLES];
} EpRanges_u16;

typedef struct
{
	uint8_t  Mins[EcMAX_NUM_OF_SPECTRAL_SAMPLES], Maxes[EcMAX_NUM_OF_SPECTRAL_SAMPLES];
} EpRanges_u8;

typedef struct
{
	union
	{
		EpRanges_f32  Ranges_f32;
		EpRanges_u16  Ranges_u16;
		EpRanges_u8   Ranges_u8;
	};

	float  Means[EcMAX_NUM_OF_SPECTRAL_SAMPLES];
} EpPixelStats;


#define EpPixelStatRANGES               0x0001
#define EpPixelStatMEANS                0x0002
#define EpPixelStatMEANS_ALPHA_MASKED   0x0004

extern int      EpImage_PixelStatsAdv(const EpImage* PImage, const EpPixelBuffer* PPixelBuffer,  EpPixelStats* PStats, const uint32_t PFlags);
extern int      EpImage_Intensity(EpImage* PSImage, const short PMeteringMode);

extern int      EpImage_LumaRange(const EpImage* PImage, int* PLumaMinRet, int* PLumaMaxRet);
extern int      EpImage_LumaRangef(const EpImage* PImage, float* PLumaMin, float* PLumaMax);

extern void     EpImage_LumaHistogram(EpImage* PImage, int* PLumaHistogram);
extern void     EpImage_Histograms(const EpImage* PImage, const EpPixelBuffer* PPixelBuffer,  void* PMin, void* PMax,  int** PHistograms, const EIndex PNBins, const int PNChannels);


extern int      EpImage_HistogramEqualize(EpImage* PSImage, EpImage* PDImage, const float PStrength, const float PIntensity);

extern int      EpImage_ColorSaturation(const EpImage* PSImage, EpImage* PDImage, const float PSaturation);

extern int      EpImage_Invert(EpImage* PSImage, EpImage* PDImage, const int PChannels);
extern int      EpImage_RotateLeft90(EpImage* PSImage, EpImage* PDImage);
extern int      EpImage_RotateRight90(EpImage* PSImage, EpImage* PDImage);


extern int      EpImage_ColorDither(EpImage* PImage, EpImage* PImageD,  const int PRLevels, const int PGLevels, const int PBLevels,  const int PMethod, const EBool PInteractive);

extern EBool    EpImage_ChromaKey(EpImage* PSImage, EpImage* PDImage, const EcRGBAiColor* PTransparentColor, const float PThreshold, const float PThreshold2, const int PSamplingKernelRadius, void (*PShowProgressFunc)(float, void*), void* PShowProgressProcData);
extern EBool    EpImage_ChromaKeyRGBDist(EpImage* PSImage, EpImage* PDImage, const EcRGBAiColor* PTransparentColor, const float PThreshold, const float PThreshold2, const int PSamplingKernelRadius, void (*PShowProgressFunc)(float, void*), void* PShowProgressProcData);

extern int      EpImage_Paste(const EpImage* PImageDecal, EpImage* PImageBackground, EpImage* PDImage,  EIndex PSrcX, EIndex PSrcY,  const EIndex PDstX, const EIndex PDstY,  const EIndex PSizeX, EIndex PSizeY);
extern int      EpImage_Composite(const EpImage* PImageDecal, EpImage* PImageBackground, EpImage* PDImage,  EIndex PSrcX, EIndex PSrcY,  EIndex PDstX, EIndex PDstY, EIndex PSizeX, EIndex PSizeY,  const int PBlendMode, const uint8_t PAlphaSource);

extern int      EpImage_Gray8Paste(const EpImage* PImageDecal, EpImage* PImageBackground,  const int PPixelOperation,  const EIndex PSrcX, const EIndex PSrcY,  const EIndex PDstX, const EIndex PDstY,  const EIndex PSizeX, const EIndex PSizeY);
extern int      EpImage_Gray8PasteRGBAiColor(const EpImage* PImageDecal, EpImage* PImageBackground,  const EcRGBAiColor* PColor,  const EIndex PSrcX, const EIndex PSrcY,  const EIndex PDstX, const EIndex PDstY,  const EIndex PSizeX, const EIndex PSizeY,  const int PBlendMode);
extern int      EpImage_Gray8PasteClipRGBAiColor(const EpImage* PImageDecal, EpImage* PImageBackground,  const EcRGBAiColor* PColor,  const EIndex PSrcX, const EIndex PSrcY,  const EIndex PDstX, const EIndex PDstY,  const EIndex PSizeX, const EIndex PSizeY,
                 const EIndex PClipMinX, const EIndex PClipMinY,  const EIndex PClipSizeX, const EIndex PClipSizeY,  const int PBlendMode);

extern int      EpImage_BitmapPasteRGBAiColor(const EpImage* PImageDecal, const uint8_t* PBitmap, const int PSrcBytesPerLine, EpImage* PImageBackground,  const EcRGBAiColor* PColor,  const EIndex PSrcX, const EIndex PSrcY,  const EIndex PDstX, const EIndex PDstY,  const EIndex PSizeX, const EIndex PSizeY);
extern int      EpImage_BitmapPasteClipRGBAiColor(const EpImage* PImageDecal, const uint8_t* PBitmap, const int PSrcBytesPerLine, EpImage* PImageBackground,  const EcRGBAiColor* PColor,  const EIndex PSrcX, const EIndex PSrcY,  const EIndex PDstX, const EIndex PDstY,  const EIndex PSizeX, const EIndex PSizeY,  const EIndex PClipMinX, const EIndex PClipMinY,  const EIndex PClipSizeX, const EIndex PClipSizeY);

extern int      EpImage_sBlend(const EpImage* PImageDecal, EpImage* PImageBackground, EpImage* PDImage,  const float PWeight,  EIndex PSrcX, EIndex PSrcY,  EIndex PDstX, EIndex PDstY,  EIndex PSizeX, EIndex PSizeY);

extern int      EpImage_Add(EpImage* PSImage, EpImage* PDImage);
extern void     EpImage_Div(const EpImage* PSImage, EpImage* PDImage, const int PNFramesFinishedI);

extern int      EpImage_GainRGBAi(EpImage* PImage, const EcRGBAiColor* PColor);

extern int      EpImages_Stitch(EpImage** PSImages,  const EIndex PNumOfColumns, const EIndex PNumOfRows,  EpImage* PDImage, const EBool PScaleToFit, const int PVerbosity);


#if defined (__cplusplus) && (!(defined (__CPLUSLIBS)))
}	// Close scope of 'extern "C"' declaration
#endif


#endif // _EImageOps_h
