//================================================================
// lib2D
//
// - Tools
//
// Tools are functional units (objects) that can be connected
// into operation trees.
//
//
// AUTHOR:  Gabor Nagy
// DATE:    2019-Jun-23 23:36:42
//
// Copyright (C) 1994-2021 by Gabor Nagy. All rights reserved.
//================================================================

#ifndef _EPicTools_h
#define _EPicTools_h

#include <ETool.h>

#ifdef EpGUI
	#include <Image/2DWindow.h>
#else
	#include <Image/Image.h>
#endif  // EpGUI

#if defined (__cplusplus) && (!(defined (__CPLUSLIBS)))
extern "C" {
#endif


typedef struct EpToolClass_S  EpToolClass;

typedef struct EpTool_S       EpTool;


#ifdef EpGUI

typedef struct
{
	EBool       ShowResult,
	            Interactive, InteractivePropagate,
	            UseHardware;
	EBool       ResultDisplayed,
	            ReCenterWindow;        // Auto re-center in _Execute

// GUI
//
	EpCoord3    SchemPosition;         // Position of this Tool in the Schematics view
	Widget      DialogW;
	Widget      DialogTopMatrixW,
							OptionsMatrixW,  ShowResultBW, InteractiveBW, CacheResultBW, InteractivePropagateBW, UseHardwareBW, SeparatorW;
	EpWindow*   ResultWindow;
} EpToolGUI;

#endif  // EpGUI



#define EpToolCore\
	EToolCore;\
	EpImage*  Image;\
	uint16_t  InfoPanelComponents;\
	void*     GUI

struct EpTool_S
{
	EpToolCore;
};




struct EpToolClass_S
{
	EToolClassCore;

	int  (*GUI)(ETool* PTool, const char PAction);

	EpImage*  IconImages;

	uint16_t  HotKey;unsigned int  HotKeyModifiers;char  Mnemonic;  // Key shortcut
};


extern const EpToolClass  EpTool_class[];




#ifdef EpGUI
extern void    (*EpTool_Activate)(EToolClass* PToolClass);
extern void    (*EpTool_ActivateFunc)(Widget PW, EPointer PClientData, EPointer PCallData);
extern void    EpTool_ResultWindowOnOff(EpTool* PTool, Widget PShellW, Widget PParentW, const uint16_t PInfoPanelFlags);

extern Widget    EpTool_CreateStandardDialog(EpTool* PTool, const char* PTitle, const char* PIconName, XeCallbackFunc PApplyCallback);
extern Widget    EpTool_CreateFolders(Widget PParentW, Widget* PFolderTipsMatrixPtr);
extern Widget    EpTool_CreateAnimationTab(EpTool* PGUITool, Widget PTabPanelMatrixW);
#endif  // EpGUI

extern void    (*Ep_ExitFunction)(void);
extern int     (*EpToolClass_RegisterFunc)(EpToolClass* PClass);
extern int     (*EpToolClass_RemoveFunc)(EpToolClass* PClass);

extern int           Ep_BinMasks[];

EList(extern, EpToolClass*, Ep_ToolClasses);
EListA(extern, ETool*, Ep_Tools);

extern EpToolClass*  EpToolClass_Register(EpToolClass* PTemplate);
extern void          EpToolClass_Deactivate(EpToolClass* PClass);
extern void          EpToolClasses_Free(void);
extern void          EpTools_Free(void);



#if defined (__cplusplus) && (!(defined (__CPLUSLIBS)))
}	// Close scope of 'extern "C"' declaration
#endif

#endif // _EPicTools_h
