//================================================================================
// Xe Widget set
//
// Widget base classes
//
// AUTHOR:  Gabor Nagy
// DATE:    2019-Jun-23 21:56:51
//
// Copyright (C) 1994-2021 by Gabor Nagy. All rights reserved.
//===============================================================================
#ifndef _XeCoreP_h
#define _XeCoreP_h

#ifdef EX11
#include <Xe/XeX.h>
#endif

#include <Xe/XeP.h>

#ifdef XeTHREADS
#include <pthread.h>
#endif

#if defined (__cplusplus) && (!(defined (__CPLUSLIBS)))
extern "C" {
#endif


#ifdef XeTHREADS
	#define XeWidgetCoreThreads  pthread_t  RedrawThread;pthread_mutex_t  Redraw_mutex;EBool  RedrawThreadCreated;
#else
	#define XeWidgetCoreThreads
#endif


// X11-specific fields in the Core Widget
#ifdef EX11
#define XeWidgetCoreX11\
	Screen*         X_Screen;\
	Window          X_Window;\
	Colormap        X_Colormap;\
	EBool           X_ColormapAllocated;\
	GC              X_GC;                   /* Plain graphics context                                 */\
	EBool           X_ClearWindow;          /* Let X11 etc. clear the background on expose (we usually don't want this, as it causes flickering and wastes GPU time, but some Widget implementors may want a quick and dirty first prototype...) */\
	Pixmap          X_WindowMaskPixmap;     /* Non-rectangular Widgets                                */\
	Pixmap          X_CachePixmap;\
	int             X_CachePixmapSizeX, X_CachePixmapSizeY;
#else
#define XeWidgetCoreX11
#endif


#ifdef XeOpenGL
#define XeWidgetCoreGL\
	EBool           GL_ClearBackground
#else
#define XeWidgetCoreGL
#endif

#define XeWindowINPUT     0x01
#define XeWindowDRAWABLE  0x02

#define XeWidgetCore\
	XeRectObjCore;\
	WidgetList      Popups;Cardinal  NPopups;    /* List of popups                          */\
	XeEventTable    EventTable;                  /* Private to event dispatcher             */\
	XeTMData        TM;                          /* Translation management                  */\
	XeTranslations  Accelerators;                /* Accelerator (hotkey) translations       */\
	int             StackingPriority;            /* Higher number means Widget is in front of others */\
	EBool           Visible;                     /* Window is mapped and not occluded       */\
	EBool           MappedWhenManaged;           /* Map window when the widget is managed   */\
	uint8_t         WindowTypeFlags;             /* Flags of XeWindowINPUT and XeWindowDRAWABLE. If not 0, this Widget has its own X11 etc. Window. XeWindowINPUT means drawing is done on the Window of closest windowed ancestor, but event handling is done on this Widget's Window */\
                                               /* XeWindowDRAWABLE means the Window may be drawn into. */\
	EBool           Realized;                    /* Windowless widgets use this flag to tell that they have been realized (there's no window we can check, but we don't want to keep calling realize) */\
	EBool           Mapped;                      /* This is how the mapped status of windowless widgets is tracked                   */\
	EBool           FirstTimeMap;\
	EBool           RealResize;                  /* Windowless widgets are notified of position changes (as well as size changes).   */\
	                                             /* If the size did not change, Xe_SetGeometry() will set this field to FALSE before calling the resize method of a Widget */\
	uint8_t         ResizePolicyX, ResizePolicyY;          /* XeSHRINK|XeGROW */\
	uint8_t         SelfResizePolicyX, SelfResizePolicyY;  /* XeSHRINK|XeGROW */\
	uint16_t        RedrawPolicy;                /* Tells the subclass how to redraw itself. XeRedrawALL, etc.*/\
	XeCoord2        InternalSizeMax;             /* This is the max size Composite Widgets should grant themselves from descendant requests, or the max size Core Widgets should allow themseleves to be. -1 means no limit.  */\
	int16_t         CornerRadius, CornerTopLeftRadius, CornerTopRightRadius, CornerBottomLeftRadius, CornerBottomRightRadius;\
\
	XeColor         BackgroundColor;\
	EpImage*        BackgroundImage;\
	XeLayer*        BlendingLayers;\
	int             BlendLayersMax;              /* Only blend up to this many layers when compositing Images (transparent Widgets). The above array is pre-allocated to this count, to minimize malloc/realloc. */\
	uint16_t        BlendFunction;               /* If not 0 (EpSRCxALPHA_plus_DSTxONE_MINUS_ALPHA etc.), and either BackgroundColor or BackgroundImage is specified, XeX_RealizeWithDraw and _Xe_CoreRedraw will blend the Widget's Image on the background, otherwise it will simply paste the Image */\
\
	XeColor         BorderColor;\
	XeColor         ToolTipBackgroundColor;\
	XeColor         ToolTipForegroundColor;\
	XeFontList      ToolTipFontList;\
\
	Widget          ChildWithFocus;\
	XeGeneology     FocalPoint;\
	XeServerGrab*   KeyList;\
	XeServerGrab*   PtrList;\
	Widget          QueryEventDescendant;\
	unsigned int    MapHandlerAdded:1;\
	unsigned int    RealizeHandlerAdded:1;\
	unsigned int    ActiveHandlerAdded:1;\
	unsigned int    HaveFocus:1;\
	int8_t          KeyboardFocus;             /* -1: not set, 0: no focus 1: focus                      */\
\
	EBool           GeometryUpdateEnabled,     /* The application can enable / disable this, so if there are multiple changes, the Widget and its children will only get updated once */\
	                RedrawEnabled;             /* The application can enable / disable this, so if there are multiple changes, the Widget and its children will only get redrawn once */\
\
	EBool           ClearBackground;           /* Clear the background first, when redrawing this Widget */\
	EBool           BufferMotionEvents;        /* Collect all motion events from the queue when they are processed, and store their pointer positions in the XeEvent. This allows higher temporal resolution drawing (more sampled points on the curve) with less overhead than processing each, individual event. */\
	XeStyle*        Style;\
	float           GradientContrast;\
	EBool           GradientBackground;\
	EpImage*        CacheImage;\
	XeDisplay*      DisplayInfo;\
	XeVisual*       Visual;\
	int             D_DrawCount;               /* For debugging */\
	XeWidgetCoreThreads\
\
	XeWidgetCoreX11\
	XeWidgetCoreGL

/*
#define XeWidgetCore\
	EObjectCore(XeWidget);\
*/

typedef struct _XeWidget_S
{
	XeWidgetCore;
} XeWidgetRec;


#define XeCoreClassCore\
	XeRectObjectClassCore;\
	XeXRealizeFunc     X_Realize;                   /* Create X11, window or GL buffers etc. for the Widget */\
	XeActionList       Actions;Cardinal  NActions;  /* Action name to function table                        */\
	XeEnum             CompressExposure;            /* Compress Expose events for Widget                    */\
	EBool              CompressMotion;              /* Compress MotionNotify events for Widget              */\
	EBool              CompressEnterLeave;          /* Compress enter and leave events                      */\
	EBool              ReceiveVisibilityChange;     /* This class of Widgets should receive VisibilityNotify events  */\
	XeAcceptFocusFunc  AcceptFocusFunc;             /* Keyboard focus accept function                       */\
	String             TMTable                      /* Translation manager state machine                    */\


typedef struct _WidgetClassRec
{
	XeCoreClassCore;
} WidgetClassRec, CoreClassRec;

externalref WidgetClassRec  _XeCoreWidget_class;

extern void  XeInheritSetValuesAlmost(Widget POldW, Widget PW, XeWidgetGeometry* PRequestW, XeWidgetGeometry* PReply);
extern EBool XeInheritAcceptFocus(Widget PW, ETime* PTime);

extern void  XeInheritResize(Widget PWidget);
extern void  XeInheritRealize(Widget PWidget, XeXWValueMask* PMask, XSetWindowAttributes* PAttributes);
extern void  XeInheritRedraw(Widget PW, XeRedrawEvent* PEvent);

extern void  _Xe_CoreResize(Widget PW);
extern void  _Xe_CoreRedraw(Widget PW, XeRedrawEvent* PEvent, const EpImage* (*PRedrawProc)(Widget PW, EpImage* PImage), const EBool PRegenerateImageCache);
extern int   _XeCore_DrawBackgroundAndGetBlendMode(Widget PW, const int PBorder,  EpImage* PCacheImage, int* PBlendModeP);

#if defined (__cplusplus) && (!(defined (__CPLUSLIBS)))
} // Close scope of 'extern "C"' declaration
#endif

#endif // __XeCoreP_h
