//===============================================================================
// Xe UI toolkit
//
// Dialog box helper module
//
// AUTHOR:  Gabor Nagy
// DATE:    2019-Jun-23 00:30:24
//
// Copyright (C) 1994-2021 by Gabor Nagy. All rights reserved.
//===============================================================================
#ifndef _XeDialog_h
#define _XeDialog_h

#include <Xe/Manager.h>

#if defined (__cplusplus) && (!(defined (__CPLUSLIBS)))
extern "C" {
#endif

enum
{
	XeDIALOG_ERROR,
	XeDIALOG_INFORMATION,
	XeDIALOG_MESSAGE,
	XeDIALOG_QUESTION,
	XeDIALOG_TEMPLATE,
	XeDIALOG_NETWORK_CONNECTION,
	XeDIALOG_WORKING,
	XeDIALOG_WARNING,
	XeDIALOG_PROMPT,
	XeDIALOG_SELECTION,
	XeDIALOG_FILESELECTION
};


enum
{
	XePLACE_NONE = 0,
	XePLACE_TOP,
	XePLACE_ABOVE_FILENAME,
	XePLACE_BOTTOM
};


enum
{
	XeMOVE_TO_TRASH = 1,
	XeASK_AND_DELETE,
	XeDELETE
};


#define XeCURSOR_KEYS     0x0001
#define XePRIMARY_CLICK   0x0002
#define XePOINTER_MOTION  0x0004


#define XeDialogCore\
	int               Type;\
	Widget            MessageBoxW, WorkAreaW;\
	Widget            ImageLabelW, MessageLabelW, MessageLabelFrameW;\
	EList(, Widget, Buttons);\
	EguiCallbackList  OkCallback, CancelCallback

typedef struct
{
	XeDialogCore;
} XeDialog;


typedef struct
{
	Widget          ShowBitmapFontsW, AntiAliasingW, StatusLabelW, ListsMatrixW, FontFamilyListW, FontStyleListW, FontSizeListW, SampleLabelW;

	XeFontList      SampleFontList;  // Used for drawing the Font sample text
	XeCallbackFunc  CallbackFunc;EPointer CallbackClientData;

	uint8_t         Orientation;
	EBool           BitmapFontsEnabled, BitmapFontsShow,
	                ShowFontSizes, AntiAliasing;
	int             FontFamilyNameSizeY;

	uint8_t         UserSetAntiAliasing;  // EBoolUNSPECIFIED, TRUE, or FALSE. If TRUE, the user turned on anti-aliasing. If FALSE, he/she turned it off, so we should not try to use AA again when available, after we had to turn it off because a Font was selected that did not support AA.

	EList(, XeFontFamily*, DisplayedFontFamilies);

	const XeFontFamily*  CurrentFontFamily;
	XeFontStyleFlags     CurrentFontStyle;
	XeFont*              CurrentFont;
	float                CurrentFontSize;
} XeFontSelectionPanel;



enum
{
	XeCR_FONTFAMILY_SELECTED = 1,
	XeCR_FONTSTYLE_SELECTED,
	XeCR_FONTSIZE_SELECTED
};

typedef struct
{
	XeCallbackStructCore;
	int  ItemIndex;
} XeFontSelectionCallbackStruct;


#define MAX_MESSAGENUM        16  // Max number of separate message dialogs
#define	MAX_NUM_OF_DIALOGBUTTONS  8


extern void    XeDialogs_Init(XeApp PXeApp, Widget PParentW,
 const EcRGBAiColor* PBackgroundColor, const EcRGBAiColor* PListBackgroundColor, const EcRGBAiColor* PForegroundColor,
 const EcRGBAiColor* PTextFieldBackgroundColor, const EcRGBAiColor* PTextFieldForegroundColor, const EcRGBAiColor* PTextFieldFocusColor,
 const EcRGBAiColor* PHighlightColor);

extern void    XeDialogs_SetColors(
 const EcRGBAiColor* PBackgroundColor, const EcRGBAiColor* PListBackgroundColor, const EcRGBAiColor* PForegroundColor,
 const EcRGBAiColor* PTextFieldBackgroundColor, const EcRGBAiColor* PTextFieldForegroundColor, const EcRGBAiColor* PTextFieldFocusColor,
 const EcRGBAiColor* PHighlightColor);

extern void    XeDialogs_Free(void);


extern XeExtensionClass     XeDialog_class;


extern Widget  XeDialog_New(const int PType, const char* PName, Widget PParentW,  const EArg* PArgList, const int PArgCount);
extern Widget  XeDialog_NewVa(const int PType, const char* PName, Widget PParentW, ...);

extern Widget  Xe_TopDialogNew(XeApp PApp, const int PType, const char* PName,  const EArg* PArgList, const int PArgCount);
extern Widget  Xe_TopDialogNewVa(XeApp PApp, const int PType, const char* PName, Widget PParentW, ...);

extern Widget  XeDialog_GetChild(Widget PDialogW, const int PChild);
extern void    XeDialog_SetValuesVa(Widget PDialogW, ...);

extern Widget  XeDialog_AddButton(const char* PName, XeWidgetClass PClass, Widget PParentW, const EBool PDefault);

extern void    Xe_ErrorDisplayVa(Widget PParentW, const char* PTitle, const char* PMessage, XeCallbackFunc PCallback, ...);

extern void    Xe_QuestionDisplayVa(Widget PParentW, const char* PTitle, const char* PQuestion, XeCallbackFunc PCallback, ...);
extern void    Xe_QuestionUndisplay(void);

extern void    Xe_QuestionYesNo(Widget PParentW, const char* PTitle,  const char* PQuestion, const char* PYesText, const char* PNoText,  XeCallbackFunc PCallback, Widget PuCenterToWidget);

extern void    Xe_MessageDisplayVa(Widget PParentW, const char* PTitle, const char* PMessage, XeCallbackFunc PCallback, ...);
extern void    Xe_MessageDisplay(void);

extern void    Xe_ErrorDisplay(Widget PParentW, const char* PTitle, const XeString PXeStr);

extern Widget  Xe_SingleWorkingMessageDisplayVa(Widget PParentW, const char* PMessage, ...);

extern Widget  Xe_ProgressMessageDisplayVa(Widget PParentW, const int PWMessageID, const char* PMessage, ...);
extern void    XeUndisplayMessageWithTimer(unsigned long PMSecs, void (*PFunction)(void));

extern void    XeDialog_GetValuesVa(Widget PDialogW, ...);

extern void    XeStdSetPushButtonArgs(const int PMarginLeft, const int PMarginRight,  const int PMarginTop, const int PMarginBottom,  EArg* PArgs, int* PArgCnt);
extern void    XeStdSetToggleButtonArgs(const XeCoordinate PMarginLeft, const XeCoordinate PMarginRight,  const XeCoordinate PMarginTop, const XeCoordinate PMarginBottom,  EArg* PArgs, int* PArgCnt);
extern void    XeStdSetFolderButtonArgs(const XeCoordinate PMarginLeft, const XeCoordinate PMarginRight,  const XeCoordinate PMarginTop, const XeCoordinate PMarginBottom,  EArg* PArgs, int* PArgCnt);


// ConnectionDialog
//
extern void    Xe_HostListRefresh(Widget PW);
extern void    Xe_HostListSetFromString(char* PString);


// XeFontSelectionPanel
//
extern int          XeFont_SelectionPanelLoadFonts(XeDisplay* PDisplay, XeFontSelectionPanel* PPanel);
extern XeFontSize*  XeFont_SelectorPanelGetCurrentFontSize(XeFontSelectionPanel* PPanel);
extern void         XeFont_SelectionPanelSetFontFaimly(XeFontSelectionPanel* PPanel, const XeFontFamily* PFontFamily);
extern int          XeFont_SelectionPanelNew(XeFontSelectionPanel* PPanel, Widget PParentW,  XeCallbackFunc PCallbackFunc, EPointer PCallbackClientData,  const EArg* PArgs, const int PNArgs);
extern void         XeFont_SelectionPanelUpdate(XeFontSelectionPanel* PPanel, XeFont* PFont, XeFontSize* PFontSize);

extern void    XeFont_StyleListUpdate(XeFontSelectionPanel* PPanel);
extern int     XeFont_FaimlyListUpdate(XeFontSelectionPanel* PFontSelectionPanel, const EBool PShowBitmapFonts);
extern void    XeFont_SizeListUpdate(XeFontSelectionPanel* PPanel);

extern XeFont* XeFont_ByFamilyAndStyle(XeDisplay* PDisplay, const char* PFamilyName, const char* PStyleName, const EBool PBitmapFonts);

extern void	   _XeEH_FocusBack(Widget PW, XeEvent* PEvent, EPointer PClientData, EBool* PCDR);

#if defined (__cplusplus) && (!(defined (__CPLUSLIBS)))
}	// Close scope of 'extern "C"' declaration
#endif

#endif // _XeDialog_h
