//================================================================================
// Xe GUI API
//
// Drawers (pull-down panels, with an open/close arrow)
//
// AUTHOR:  Gabor Nagy
// DATE:    2019-Jun-23 16:17:25
//
// Copyright (C) 1994-2021 by Gabor Nagy. All rights reserved.
//================================================================================
#ifndef _XeDrawer_h
#define _XeDrawer_h

#ifndef _EClass_h
#include <EClass.h>
#endif

#include <Xe/PushButton.h>

#if defined (__cplusplus) && (!(defined (__CPLUSLIBS)))
extern "C" {
#endif

// For XeDrawer_NewVa
//
#define XeDrawerCREATE_TOP_MATRIX       0x0001
#define XeDrawerCREATE_SEPARATOR        0x0002
#define XeDrawerOPEN                    0x0004  // Open the drawer at creation
#define XeDrawerBUTTON_IN_H_MATRIX      0x0008  // Create a horizontal XeMatrix for the open/close button to allow for other Widgets next to it
#define XeDrawerACTIVE_INACTIVE_TOGGLE  0x0010
#define XeDrawerACTIVE                  0x0020
#define XeDrawerCREATE_IMAGE            0x0040
#define XeDrawerUNMANAGED               0x0080

//----------------------------------------
// For Folder panels
//----------------------------------------
typedef struct
{
	char    Name[64];
	Widget  ParentW, ButtonMatrixW, OpenCloseButtonW, ActiveButtonW, TopMatrixW, LabelW, DrawerW, SeparatorW, CloseButtonW;
	EBool   Open;
} XeDrawer;


typedef struct
{
	ECoreClass*  Class;
	XeDrawer     Drawer;
	void*        ResourcesGUI;
	EBool        Show;
} XeComponentClassDrawer;


#define XeInfoPanelCore\
	Widget                      NameMatrixW, FollowSelectionW, ApplyChangesToAllSelectedW;\
	Widget                      ClassFoldersParentW;\
	EList(, XeComponentClassDrawer*, ClassFolders);           /* One for each static Class   */\
	EList(, XeComponentClassDrawer*, ComponentClassFolders);  /* One for each dynamic Class  */\
	EBool                       FollowSelection,              /* Update panel when the selection changes  */\
	                            ApplyChangesToAllSelected     /* Apply changes made by the user to all selected, compatible Objects  */

typedef struct
{
	XeInfoPanelCore;
} XeInfoPanel;


extern Widget  XeDrawer_NewVa(Widget PParentW, XeDrawer* PFolder, const char* PName, const int PShadowThickness, const uint32_t PFlags, ...);
extern Widget  _XeDrawer_NewVa(char* PTitle, Widget PParentW,  EArg* PArgs, const int PArgCnt);
extern void    Xe_DrawerSetWorkArea(Widget PDrawerW, Widget PWAW);
extern void    Xe_DrawerFree(Widget PDrawerW);
extern void    Xe_DrawerOpenClose(XeDrawer* PFolder, const EBool POpen);
extern void    XeCB_DrawerOpenClose(Widget PW, EPointer PClientData, EPointer PCallData);


//----------------------------------------
// InfoPanel.c
//----------------------------------------
extern XeInfoPanel*             XeInfoPanel_New(const unsigned int PStructSize);
extern XeComponentClassDrawer*  XeInfoPanel_GetFolderByClass(const XeInfoPanel* PPanel, const void* PClass);


#if defined (__cplusplus) && (!(defined (__CPLUSLIBS)))
}	// Close scope of 'extern "C"' declaration
#endif


#endif	// _XeDrawer_h
