//================================================================
// Xe UI toolkit
//
// Main public header file
//
// AUTHOR:  Gabor Nagy
// DATE:    2019-Jun-23 03:43:51
//
// Copyright (C) 1994-2021 by Gabor Nagy. All rights reserved.
//================================================================
#ifndef _XeFSBox_h
#define _XeFSBox_h

#include <stdarg.h>

#include <EDir.h>

#include <Xe/XeTypes.h>


#if defined (__cplusplus) && (!(defined (__CPLUSLIBS)))
extern "C" {
#endif

typedef int  XeFSDialogGrpID;

#define XeFSDialogGrpNONE  -2
#define XeFSDialogGrpNEW   -1

#define XeCR_FILENAME_CHANGED  0x01

typedef struct XeFileSelectionPanel_S  XeFileSelectionPanel;

typedef struct
{
	XeCallbackStructCore;
	char*   Path;
	EList(, char*, Paths);  // Used if the XeList in the FSBox has multi-select enabled
} XeFileSelectionBoxCallbackStruct;

typedef void (*XeFSBoxCallbackFunc)(XeFileSelectionPanel* PPanel, EPointer PClientData, XeFileSelectionBoxCallbackStruct* PCallData);

struct XeFileSelectionPanel_S
{
	Widget               ParentShellW;
	Widget               FileNameDialogW, MatrixW, LabelW, TextW, BrowseBW, SaveBW;
	XeFSBoxCallbackFunc  CallbackFunc;EPointer  CallbackClientData;
	uint16_t             KeepFSBoxAfterOK:1, AbsolutePath:1, AutofillTextField:1, FSBoxMultiSelect:1;
};


#define XeFSPanelCREATE_HMATRIX      0x01
#define XeFSPanelCREATE_TEXTFIELD    0x02
#define XeFSPanelAUTOFILL_TEXTFIELD  0x04
#define XeFSPanelSAVE_BUTTON         0x08
#define XeFSPanelMULTISELECT         0x10

// FSPanel
//
extern XeFileSelectionPanel*  XeFileSelectionPanel_New(XeFileSelectionPanel* PPanel, const char* PName, Widget PParentW, Widget PParentShellW, const uint16_t PFlags,  const int PBlendLayersMax, const int PMarginWidth,  XeFSBoxCallbackFunc PCallbackFunc, EPointer PClientData);
extern void                   XeFileSelectionPanel_Update(XeFileSelectionPanel* PPanel, const char* PFileName);
extern void                   XeFileSelectionPanel_Free(XeFileSelectionPanel* PPanel, const EBool PFreeStruct);
extern Widget                 XeFileSelectionDialog_New(Widget PParentW,  const EArg* PArgs, const int PArgCnt);



// Procedure, provided by the application, to determine the type of a directory entry
//
typedef int  (*XeEntryIdFunc)(EdirEntry*, EdirEntry*);

extern void      XeFSBox_SetFilterList(Widget PDialogW, char* PFilterList);
extern char*     XeFSBox_GetFile(Widget PDialogW);
extern void      XeFSBox_SetFile(Widget PDialogW, char* PFileName);
extern EIndex    XeFSBox_GetSelectedFiles(Widget PDialogW, char*** PListRet, const EBool PFullPaths);

extern Widget    XeFSBox_GetChild(Widget PDialogW, const int PChild);
extern void      XeFSBox_DoSearch(Widget PDialogW, char* PPath);


#if defined (__cplusplus) && (!(defined (__CPLUSLIBS)))
}  // Close scope of 'extern "C"' declaration
#endif

#endif // _XeFSBox_h
