//================================================================
// Xe UI toolkit
//
// XeLabel
// Private  header file
//
// AUTHOR:  Gabor Nagy
// DATE:    2019-Jun-23 16:23:32
//
// Copyright (C) 1994-2021 by Gabor Nagy. All rights reserved.
//================================================================

#ifndef _XeLabelP_h
#define _XeLabelP_h

#include <Xe/PrimitiveP.h>
#include <Xe/Label.h>

#if defined (__cplusplus) && (!(defined (__CPLUSLIBS)))
extern "C" {
#endif


//================================
// Class definition
//================================

#define XeLabelClassCore\
	XePrimitiveClassCore

//================================
// Class record
//================================
typedef struct _XeLabelClassRec
{
	XeLabelClassCore;
} XeLabelClassRec;

extern XeLabelClassRec  _XeLabelWidget_class;


typedef struct
{
	XeString                String;
	XeColor                 ForegroundColor;

	EpImage*                NormalImage;EIndex  NormalImageChangeCount;
	EpImage*                InsensitiveImage;

	int                     ImageBlendMode;

	uint8_t                 ImageResizePolicy;

	int32_t*                ImageSizeXPtr;
	int32_t*                ImageSizeYPtr;

	XeCoord2                Size;                          // Total size of the Label, including the optional Image and String

	XeCoordinate            StringCenterOffsetY;
	XeCoord2                StringSize;    // Position and centered size of the String
	XeCoord2                StringPosition;
	XeCoord2                ImagePosition, ImageSize;      // Position and size of the Image

	XeStringBinarySegment*  StringBinary;int  StringNumOfSegments;

	XeCoord2                Spacing;
	XeString                AcceleratorString;             // The resource name is XeNacceleratorText
	XeCoord2                AcceleratorStringSize;
	XeCoordinate            AcceleratorStringX;
	XeStringBinarySegment*  AcceleratorStringBinary;int  AcceleratorStringNumOfSegments;

	int                     ShadowOffset, ShadowBlur;      // Offset in pixels, blur in 1/16th pixels
	uint8_t                 Type, StringAntiAliasing;

// Internal data
//
} XeLabel;


//================================
// Instance definition
//================================
// Instance part
//================================
typedef struct
{
// Resources
//
	char*          Accelerator;

	XeCoord2       LabelPosition;

	unsigned int   ShadowType;
	uint8_t        LabelAlignmentX, LabelAlignmentY;
	char           Mnemonic;

	XeFontList		 FontList;EIndex  FontListChangeCount;

	XeLabel        Label;


// Private variables
//
	EBool          FontListCreated;

#ifdef EX11
	GC             X_StringGC;
	Pixmap         X_BackgroundCachePixmap;  // Used for caching the background, so we can quickly change the label string, if there's no blending or anti-aliasing
	int            X_BackgroundCachePixmapSizeX, X_BackgroundCachePixmapSizeY;
	EBool          X_BackgroundCachePixmapValid;
#endif
} XeLabelPart;


//================================
// Instance record
//================================
typedef struct _XeLabelRec
{
	XeWidgetCore;
	XeBasePart       Base;
	XePrimitivePart  Primitive;
	XeLabelPart      Label;
} XeLabelRec;



//----------------------------------------
// Function prototypes
//----------------------------------------
extern void           _XeLabel_Default(XeLabel* PLabel);
extern void           _XeLabel_FreeContents(XeLabel* PLabel);
extern void           _XeLabel_LayoutStringAndImage(XeLabel* PLabel);
extern XeChangeFlags  _XeLabel_ValueChanged(XeVisual* PVisual, XeLabel* PLabelNew, XeLabel* PLabelCurrent, XeFontList PFontList);
extern XeChangeFlags  _XeLabel_CheckFontChange(XeLabelPart* POldLabel, XeLabelPart* PLabel);
extern void           _XeLabel_Initialize(Widget PW, XeLabelPart* PLabel, EcRGBAiColor** PForegroundColorP);


extern void	_XeLabelX_Draw(XeLabel* PLabel,  EpImage* PDstImage, const int PForegroundBlendMode, Drawable PXDrawable,  const XeVisual* PVisual,  GC PBackgroundXGC, GC PXGC,  const int PDrawableSizeX, const int PDrawableSizeY,
 const int PX, int PY,
 EpImage* PImage,
 const EcRGBAiColor* PBackgroundColor, const EcRGBAiColor* PBackgroundColorForShadows,
 const EcRGBAiColor* PStringColor,
 const EBool PSensitive, const EBool PUseColorsInString);


#ifdef EX11
extern void  XeX_LabelRealize(Widget PW, Mask* PValueMask, XSetWindowAttributes* PWinAttributes, const EpImage* (*PRedrawFunc)(Widget PW, EpImage* PCacheImage));
#endif

#if defined (__cplusplus) && (!(defined (__CPLUSLIBS)))
}	// Close scope of 'extern "C"' declaration
#endif

#endif // _XeLabelP_h
