//========================================================================
// Xe UI toolkit
//
// XeList
//  Public header file
//
// AUTHOR:  Gabor Nagy
// DATE:    2019-Jun-23 14:49:44
//
// Copyright (C) 1994-2021 by Gabor Nagy. All rights reserved.
//========================================================================
#ifndef _XeListWidget_h
#define _XeListWidget_h

#include <Xe/Xe.h>
#include <Xe/Label.h>

#if defined (__cplusplus) && (!(defined (__CPLUSLIBS)))
extern "C" {
#endif


enum
{
	XeSINGLE_SELECT = 1,
	XeMULTIPLE_SELECT
};


enum
{
	XeMultiSelectOFF = 1,
	XeMultiSelectON,
	XeMultiSelectTOGGLE
};


typedef struct
{
	XeCallbackStructCore;
	XeString 	 ItemStr;
	int        ItemLength;
	int        ItemPosition;
} XeListCallbackStruct;



typedef struct
{
	XeString  String;
	EpImage*  NormalImage;
	EpImage*  SelectedImage;
	EBool     Selected;
} XeListItem;


//========================================
// Initialize an XeListItem
//========================================
#define XeListItemInit(mItem)\
{\
	(mItem)->String = NULL;\
	(mItem)->NormalImage = NULL;\
	(mItem)->SelectedImage = NULL;\
	(mItem)->Selected = FALSE;\
}





extern XeWidgetClass             XeListWidget_class;
typedef struct _XeListClassRec*  XeListWidgetClass;
typedef struct _XeListRec*       XeListWidget;


extern void   XeList_AddItemUnselected(Widget PW, XeString PItem, const int PLocation);
extern void   XeList_DeleteAllItems(Widget PW);
extern void   XeList_DeselectAllItems(Widget PW);
extern void   XeList_SelectItemByIndex(Widget PW, const int PIndex);

extern void   XeList_ItemUpdate(Widget PW, XeListItem* PItem, const int PIndex);
extern void   XeList_SetItems(Widget PW, XeListItem* PItems, const int PNumOfItems, const EBool PResetScroll);
extern EIndex XeList_GetSelectedItemStrings(Widget PW, char*** PStrings);
extern EBool  XeList_ItemSelected(Widget PW, const EIndex PIndex);

extern void   XeList_ItemArrayFree(XeListItem* PItems, const int PNumOfItems);

extern void   XeList_UpdateAndRedraw(Widget PW);


extern EBool	XeList_GetSelectedPos(Widget PW,  int** PPosListRet, int* PPosCountRet);
extern EBool	XeList_ItemExists(Widget PW, XeString PItem);
extern Widget	XeScrolledList_New(Widget PParentW, const char* PName,  const EArg* PArgs, const int PNumOfArgs);


#if defined (__cplusplus) && (!(defined (__CPLUSLIBS)))
} // Close scope of 'extern "C"' declaration
#endif

#endif // _XeListWidget_h
