//===============================================================================
// Xe UI toolkit
//
// XeMenu
//  Public header file
//
// AUTHOR:  Gabor Nagy
// DATE:    2017-Aug-25 09:27:43
//
// Copyright (C) 1994-2021 by Gabor Nagy. All rights reserved.
//===============================================================================
#ifndef _XeMenuWidget_h
#define _XeMenuWidget_h

#include <Xe/Manager.h>

#if defined (__cplusplus) && (!(defined (__CPLUSLIBS)))
extern "C" {
#endif


#define XeMAX_MENULEVELS     16


// Menu widget types
//
enum
{
	XeMENU_BAR = 0,
	XeMENU_PANE,
	XeOPTION_MENU,
	XeOPTION_MENU_PANE
};

typedef struct
{
	char*           Name;
	EPointer        Value;
	XeCallbackFunc  Callback;
	EPointer        ClientData;
	EpCImage*       Icon;
} XeOptionMenuRec;

typedef struct XeMenuRec_S      XeMenuRec;
typedef struct XeMenuRootRec_S  XeMenuRootRec;

struct XeMenuRec_S
{
	XeMenuRootRec*  MenuBar;
	XeMenuRec*      Parent;
	XeMenuRec*      Sibling;
	XeMenuRec*      Child;
	int             CastShadowOffset;
	Widget          CascadeBW, MenuPaneW;
};



// XeMenuBar_New flags
//
#define	XeMenuBarSHADOWS                        0x01
#define	XeMenuBarVERTICAL                       0x02
#define XeMenuBarSHADED_MENUBAR_CASCADEBUTTONS  0x04

// XeMenu_New/XeMenu_NewVa flags
//
#define	EmNORMAL             0x00
#define	EmRADIO              0x01  // Menu has RadioBehavior
#define	EmCASCADE_WIDGET     0x01  // Force the menu's CascadeButtons to be CascadeButtonWidget (the default is xmCascadeButtonGadget)
#define	EmSHOW_ARROW         0x02  // Show menu posting direction with an arrow
#define	EmMENUPOST_ON_LEFT   0x08  // Where to post the menu relative to the CascadeButton
#define	EmMENUPOST_ON_RIGHT  0x10
#define	EmMENUPOST_ON_ABOVE  0x20
#define	EmMENUPOST_ON_BELOW  0x40

#define EmMENU_POST_DIRECTION  (EmMENUPOST_ON_LEFT | EmMENUPOST_ON_RIGHT | EmMENUPOST_ON_ABOVE | EmMENUPOST_ON_BELOW)


#define	EguiMENUBAR_LEVEL  0


//enum { EmCASTSHADOW_XSHAPE = 0, EmCASTSHADOW_SGITRUECOLOR };
//enum { Em_GADGET = 0, Em_WIDGET };
//enum { Em_PUSHBUTTON = 0, Em_TOGGLEBUTTON };


typedef struct
{
	XeCallbackStructCore;

// XeOptionMenuCallbackStruct part
//
	int       Index, NumOfItems;
	EPointer  Value;
} XeOptionMenuCallbackStruct;


//
// Xe/Menu.c
//
extern XeWidgetClass               XeMenuWidget_class;
typedef struct _XeMenuClassRec*  XeMenuWidgetClass;
typedef struct _XeMenuWRec*      XeMenuWidget;

extern Widget  XePulldownMenu_New(Widget PParentW, const char* PName,  EArg* Args, const int ArgCnt);

extern Widget  XeOptionMenu_New(Widget PParentW, const char* PName, Widget* PMenuPaneRet,  EArg* PArgs, int PArgCnt);

extern Widget	 XeMenu_PaneGetItemByNumber(Widget PW, const int PNumber);

extern Widget	 XeOptionMenu_GetItemByNumber(Widget PW, const int PNumber);

extern void	   XeOptionMenu_SetActiveItem(Widget PMenuW, Widget PItemW);

extern Widget	 XeOptionMenu_GetPane(Widget PW);
extern void    XeOptionMenu_SetOptionByNumber(Widget PMenuW, const int POptionNumber);

extern Widget	 XeMenu_GetChild(Widget PW, const int PChild);

//
// Xe/Menu.c
//
extern XeMenuRec*    XeMenuBar_New(const char* PName, Widget PParentW, const int PFlags, XeFontList PFontLst);

extern XeMenuRec*    XePopupMenu_RNew(Widget PParentW, const int PFlags, XeFontList PFontLst);
extern void          XeMenu_PopUp(XeMenuRec* PMenu,  const int PX, const int PY);

extern Widget        XePopupMenu_New(Widget PParentW, const char* PName,  EArg* PArgs, const int PArgCnt);

extern XeMenuRec*  XeMenu_New(XeMenuRec* PParentMenu, const char* PName, const char PMnemonic, const int PFlags, const int PCastShadowOffset);
extern XeMenuRec*  XeMenu_NewVa(XeMenuRec* PParentMenu, const char* PName, const int PFlags, ...);


extern Widget    XeMenu_SeparatorNew(XeMenuRec* PMenu);
extern Widget    XeMenu_PushButtonNew(XeMenuRec* PMenu, const char* PName, const char PMnemonic, const char* PAccelerator, const char* PAcceleratorText, const EBool PCreateOptionsButton, XeCallbackFunc PCallbackID, EPointer PClientData);
extern Widget    XeMenu_ToggleButtonNew(XeMenuRec* PMenu, const char* PName, const EBool PDefault, const char PMnemonic, const char* PAccelerator, const char* PAcceleratorText,  const uint8_t PSecondaryButtonType,  XeCallbackFunc PCallbackID, EPointer PClientData);
extern Widget    XeMenu_LabelNew(XeMenuRec* PMenu, const char* PName);

extern Widget    XeMenuBar_SeparatorNew(XeMenuRec* PMenu);
extern Widget    XeMenuBar_PushButtonNew(XeMenuRec* PMenuRoot, const char* PName, const char PMnemonic, const char* PAccelerator, const char* PAcceleratorText, XeCallbackFunc PCallbackID, EPointer PClientData);
extern Widget    Xe_BarPushButtonNewVa(XeMenuRec* PMenu, const char* PName, const char PMnemonic, const char* PAccelerator, const char* PAcceleratorText, XeCallbackFunc PCallbackID, EPointer PClientData, ...);
extern Widget    XeMenuBar_ToggleButtonNew(XeMenuRootRec* PMenuRoot, Widget PParentW, const char* PName, const EBool PDefault, const char PMnemonic, const char* PAccelerator, const char* PAcceleratorText, XeCallbackFunc PCallbackID, EPointer PClientData);

extern Widget    XeOptionMenu_CreateAndInit(Widget PParentW, Widget* PMenuWRet, XeOptionMenuRec* POptions, const int PNumOfItems, EPointer PValue, const char* PLabel, XeCallbackFunc PCallback, EPointer PClientData,  EArg* PArgs, int PArgCnt);
extern Widget    XeOptionMenu_CreateAndInitManaged(Widget PParentW, Widget* PMenuWRet,  XeOptionMenuRec* POptions, const int PNumOfItems,  EPointer PValue, const char* PLabel,  XeCallbackFunc PCallback, EPointer PClientData,  EArg* PArgs, int PArgCnt);

extern void      XeMenuBar_Free(XeMenuRec* PMenuBar);

extern Widget    Xe_RealizeCachedMenuItem(XeMenuRec* PMenu, EguiItem PItem, Widget PBeforeW);

extern void      XeMenus_Free(void);

extern EBool     XeShell_IsMenuPane(Widget PShellW);


#if defined (__cplusplus) && (!(defined (__CPLUSLIBS)))
}	// Close scope of 'extern "C"' declaration
#endif

#endif // _XeMenuWidget_h
