//===============================================================================
// Xe UI toolkit
//
// XeScale
//  Public header file
//
// AUTHOR:  Gabor Nagy
// DATE:    2019-Jun-23 21:42:09
//
// Copyright (C) 1994-2021 by Gabor Nagy. All rights reserved.
//===============================================================================
#ifndef _XeScaleWidget_h
#define _XeScaleWidget_h

#include <Xe/Manager.h>

#if defined (__cplusplus) && (!(defined (__CPLUSLIBS)))
extern "C" {
#endif

extern XeWidgetClass              XeScaleWidget_class;
typedef struct _XeScaleClassRec*  XeScaleWidgetClass;
typedef struct _XeScaleRec*       XeScaleWidget;

#define  XeRulerLEVEL1   0x0001
#define  XeRulerLEVEL2   0x0002
#define  XeRulerLEVEL3   0x0004

typedef struct
{
	int    TopValue;
	Pixel  Color;
} XeThermometerRange;


typedef struct
{
	XeCallbackStructCore;
	XeValue  Value;int  Index;
	char     Decimals;
	EBool    ValueChanged;               // Activate callbacks may be called when the value did not change (e.g. the user pressed enter). This flag indicates whether the value changed.

	EBool    CallValueChangedCallbacks;  // Set to TRUE by Xe. Should be set to FALSE by app if the app doesn't want the ValueChanged callbacks to be called
} XeScaleCallbackStruct;


typedef struct
{
	XeValue  Value, Minimum, Maximum, ScaleMinimum, ScaleMaximum, ResetValue;
} XeValueAndRanges;

extern EBool	XeTextF;

extern void		XeScale_SetValue(Widget PWidget, const int PValue, const EBool PNotify);
extern int		Xe_ScaleGetValue(Widget PWidget);
extern void		XeScale_SetValueAndRange(Widget PWidget, const int PValue,  const int PMinimum, const int PMaximum,  const EBool PNotify);
extern void		XeScale_SetValueAndRanges(Widget PWidget, const int PValue,  const int PMinimum, const int PMaximum,  const int PScaleMinimum, const int PScaleMaximum,  const EBool PNotify);

extern void		    XeScale_SetValue_i64(Widget PWidget, const long long PValue, const EBool PNotify);
extern long long	Xe_ScaleGetValue_i64(Widget PWidget);
extern void		    XeScale_SetValueAndRange_i64(Widget PWidget, const long long PValue,  const long long PMinimum, const long long PMaximum,  const EBool PNotify);
extern void		    XeScale_SetValueAndRanges_i64(Widget PWidget, const long long PValue,  const long long PMinimum, const long long PMaximum,  const long long PScaleMinimum, const long long PScaleMaximum,  const EBool PNotify);

extern float		XeScale_SetValue_f32(Widget PWidget, const float PValue, const EBool PNotify);
extern float		Xe_ScaleGetValue_f32(Widget PWidget);
extern void		  XeScale_SetValueAndRange_f32(Widget PWidget, const float PValue, const float PMinimum, const float PMaximum, const EBool PNotify);
extern void	  	XeScale_SetValueAndRanges_f32(Widget PWidget, const float PValue, const float PMinimum, const float PMaximum, const float PScaleMinimum, const float PScaleMaximum, const EBool PNotify);

extern double		XeScale_SetValue_f64(Widget PWidget, const double PValue, const EBool PNotify);
extern double		Xe_ScaleGetValue_f64(Widget PWidget);
extern void		  XeScale_SetValueAndRange_f64(Widget PWidget, const double PValue, const double PMinimum, const double PMaximum, const EBool PNotify);
extern void	  	XeScale_SetValueAndRanges_f64(Widget PWidget, const double PValue, const double PMinimum, const double PMaximum, const double PScaleMinimum, const double PScaleMaximum, const EBool PNotify);

extern void     Xe_ScaleSetIncrement_i32(Widget PW,  const int32_t PIncrement);
extern void     Xe_ScaleSetIncrement_i64(Widget PW,  const int64_t PIncrement);
extern void     Xe_ScaleSetIncrement_f32(Widget PW,  const float PIncrement);
extern void     Xe_ScaleSetIncrement_f64(Widget PW,  const double PIncrement);

extern Widget		XeScale_Newi32(Widget PParentW, const char* PName, EArg* Args, int ArgCnt);
extern Widget		XeScale_Newi64(Widget PParentW, const char* PName, EArg* Args, int ArgCnt);
extern Widget		XeScale_Newf32(Widget PParentW, const char* PName, EArg* Args, int ArgCnt);
extern Widget		XeScale_Newf64(Widget PParentW, const char* PName, EArg* Args, int ArgCnt);


#if defined (__cplusplus) && (!(defined (__CPLUSLIBS)))
}	// Close scope of 'extern "C"' declaration
#endif

#endif // _XeScaleWidget_h
