//================================================================================
// Xe UI toolkit
//
// XeScrolledWindow
//  Composite widget to hold the WorkWindow widget of XeScrolledWindow
//
// AUTHOR:  Gabor Nagy
// DATE:    2019-Jun-23 19:17:20
//
// Copyright (C) 1994-2021 by Gabor Nagy. All rights reserved.
//================================================================================
#ifndef _XeScrolledWindowP_h
#define _XeScrolledWindowP_h

#include <Xe/ManagerP.h>
#include <Xe/ScrolledWindow.h>

#if defined (__cplusplus) && (!(defined (__CPLUSLIBS)))
extern "C" {
#endif

//================================
// Class definition
//================================
// Class part
//================================
typedef struct _XeScrolledWindowClassPart
{
	EPointer  Extension;
} XeScrolledWindowClassPart;


//================================
// Class record
//================================
typedef struct _XeScrolledWindowClassRec
{
	XeCompositeClassCore;
	XeConstraintClassPart      Constraint;
	XeManagerClassPart         Manager;
	XeScrolledWindowClassPart  ScrolledWindow_class;
} XeScrolledWindowClassRec;

extern XeScrolledWindowClassRec  _XeScrolledWindow_class;


//================================
// Instance definition
//================================
// Instance part
//================================
typedef struct
{
// Resources
//
	XeColor       ButtonColor, SliderColor;

	XeColor       ScrollBarBackgroundColor, ScrollBarTroughColor;

	unsigned int  ShadowType, ScrollingPolicy;
	uint8_t       ScrollBarDisplayPolicyX, ScrollBarDisplayPolicyY;

	XeCoord2      MarginSize;                // Between the main shadows and the ScrollBars
	XeCoordinate  Spacing;                   // Between the ScrollBars and the WorkWindow

	XeCoordinate  ScrollBarShadowThickness;
	Widget        HScrollBarW, VScrollBarW,
	              WorkWindowW;               // The widget to be scrolled

// Scrolling state
//
	Widget        ClipWindowW;                       // The actual parent of Widget to be scrolled

	EList(, Widget, TopHeaderDrivingMatrixWs);

	XeCoord2      WorkWindowOffset,                  // Pixel of WorkWindow Widget in the top-left corner of the ScrollerWindow's area
	              WorkWindowSize,                    // Total size of the WorkWindow
	              WorkWindowVisibleSize;

	XeCoord2      WorkWindowMarginTotalSize;         // Total unusable portion of WorkWindowSize. Viewable area: WorkWindowSize - WorkWindowMarginTotalSize

// Private variables
//
	XeCoord2      ClipWPos, HScrollbarPos, VScrollbarPos;
	XeCoordinate  WorkAreaShadowThickness;
	XeCoord2      HScrollBarSize, VScrollBarSize;
} XeScrolledWindowPart;


//================================
// Instance record
//================================
typedef struct _XeScrolledWindowRec
{
	XeCompositeWidgetCore;

	XeBasePart            Base;
	XeManagerPart         Manager;
	XeScrolledWindowPart  ScrolledWindow;
} XeScrolledWindowRec;


//----------------------------------------
// Function prototypes
//----------------------------------------
extern void  _XeScrolledWindow_UpdateScrollbars(XeScrolledWindowWidget PScrolledWindowW,  const int PSWSizeX, const int PSWSizeY);
extern void  _XeScrolledWindow_SetScrollbars(Widget PW, int PWorkWindowSizeX,  const int PWorkWindowSizeY,  const int PWorkAreaSizeX, const int PWorkAreaSizeY);
extern void  _XeScrollWheelAction(Widget PW, XeEvent* PEvent,  String* PP, Cardinal* PNP);


#if defined (__cplusplus) && (!(defined (__CPLUSLIBS)))
}	// Close scope of 'extern "C"' declaration
#endif

#endif // _XeScrolledWindowP_h
