//===============================================================================
// Xe UI toolkit
//
// XeText
//  Public header file
//
// AUTHOR:  Gabor Nagy
// DATE:    2019-Jun-23 16:34:06
//
// Copyright (C) 1994-2021 by Gabor Nagy. All rights reserved.
//===============================================================================
#ifndef _XeTextWidget_h
#define _XeTextWidget_h

#include <Xe/TextField.h>

#if defined (__cplusplus) && (!(defined (__CPLUSLIBS)))
extern "C" {
#endif

extern XeWidgetClass               XeTextWidget_class;
typedef struct _XeTextClassRec*  XeTextWidgetClass;
typedef struct _XeTextRec*       XeTextWidget;


typedef struct
{
	XeTextPosition  Start, End;
	XeCoordinate    StartCol, StartRow, EndCol, EndRow;
#ifdef EX11
	Pixel           Color;
#endif
} XeTextHighlight;


extern void     XeText_Redraw(Widget PW);

extern EBool    XeText_Cut(Widget PW, Time PTime);
extern EBool    XeText_Copy(Widget PW, Time PTime);
extern EBool    XeText_Paste(Widget PW);

extern void     XeText_SetString(Widget PW, char* PString, const EBool PNotify, const EBool PRedraw);
extern char*		XeText_GetString(Widget PW);
extern void     XeText_Insert(Widget PW, XeTextPosition PHere, const char* PString, const EBool PRedraw);
extern void     XeText_Replace(Widget PW, XeTextPosition PStart, XeTextPosition PEnd, const char* PString, const EBool PRedraw);

extern void     XeText_ClearSelection(Widget PW, Time PTime);

extern EBool    XeText_GetSelectionPosition(Widget PW, XeTextPosition* PSelStart, XeTextPosition* PSelEnd);
extern void     XeText_SetSelection(Widget PW, XeTextPosition PStart, XeTextPosition PEnd, Time PTime, const EBool PRedraw, const EBool PNotify);
extern char*    XeText_GetSelection(Widget PW);

extern XeTextPosition  XeText_GetLastPosition(Widget PW);

extern XeTextPosition  XeText_GetInsertionPosition(Widget PW);
extern XeTextPosition  XeText_GetInsertionPositionXY(Widget PW,  XeCoordinate* PXP, XeCoordinate* PYP);

extern void     XeText_SetInsertionPosition(Widget PW, XeTextPosition PPos, const EBool PNotify, const EBool PRedraw);

extern XeTextPosition  XeText_GetTopPosition(Widget PW);
extern XeTextPosition  XeText_GetTopPositionXY(Widget PW,  XeCoordinate* PXP, XeCoordinate* PYP);
extern void     XeText_SetTopPosition(Widget PW, const XeTextPosition PTopPos, const EBool PRedrawIn);
extern void     XeText_SetTopPositionAndInsertionPosition(Widget PW, XeTextPosition PTopPos, XeTextPosition PCursorPos, const EBool PNotify, const EBool PRedraw);

extern int      XeText_GetVisibleRows(Widget PW);
extern int      XeText_GetVisibleCols(Widget PW);


extern void     XeText_SetHighlight(Widget PW, const XeTextPosition PLeft, const XeTextPosition PRight, const XeHighlightMode PMode, const Pixel PColor, const EBool PRedraw);
extern void     XeText_SetHighlights(Widget PW, XeTextHighlight* PHighlights, const int PNumOfHighlights, const EBool PRedraw);
extern void     XeText_ClearHighlights(Widget PW);


extern Widget   Xe_CreateScrolledText(Widget PParentW, const char* PName,  EArg* PArgs, const int PNumOfArgs);

extern EIndex   XeText_PixelXYToTextPos(Widget PW,  const XeCoordinate PX, const XeCoordinate PY,  XeCoordinate* PCharXP, XeCoordinate* PCharYP);
extern int      XeText_GetSubstring(Widget PW, const XeTextPosition PStart, const int PNChars, const int PBufSize, char* PBuffer);


#if defined (__cplusplus) && (!(defined (__CPLUSLIBS)))
} // Close scope of 'extern "C"' declaration
#endif

#endif // _XeTextWidget_h
