//===============================================================================
// Xe UI toolkit
//
// XeTextField
//  Public header file
//
// AUTHOR:  Gabor Nagy
// DATE:    2017-Aug-12 23:29:15
//
// Copyright (C) 1994-2021 by Gabor Nagy. All rights reserved.
//===============================================================================
#ifndef _XeTextField_h
#define _XeTextField_h

#include <Xe/Primitive.h>

#if defined (__cplusplus) && (!(defined (__CPLUSLIBS)))
extern "C" {
#endif

extern XeWidgetClass                  XeTextField_class;
typedef struct _XeTextFieldClassRec*  XeTextFieldClass;
typedef struct _XeTextFieldRec*       XeTextField;


enum { XeMULTI_LINE_EDIT, XeSINGLE_LINE_EDIT };

typedef struct
{
	XeCallbackStructCore;
	uint32_t  Modifiers;
} XeTextCallbackStruct;


// Text directions (for search)
//
typedef enum { XeTEXT_FORWARD, XeTEXT_BACKWARD } XeTextDirection;

typedef Atom	XeTextFormat;

#define XeFormat8_BIT   ((XeTextFormat)XA_STRING)   // 31
#define XeFormat16_BIT  ((XeTextFormat)2)  // They mean XA_SECONDARY ?
#define XeFormatWCHAR   ((XeTextFormat)3)

typedef enum
{
	XeSELECT_POSITION,
	XeSELECT_WHITESPACE,
	XeSELECT_WORD,
	XeSELECT_LINE,
	XeSELECT_ALL,
	XeSELECT_PARAGRAPH,
	XeSELECT_OUT_LINE
} XeTextScanType;


typedef struct
{
	const void*   Ptr;
	int           Length;
	XeTextFormat  Format;
} XeTextBlockRec;


typedef struct
{
	XeCallbackStructCore;
	EBool            DoIt;
	XeTextPosition   CurrInsert, NewInsert;
	XeTextPosition   StartPos, EndPos;
	XeTextBlockRec*  Text;
} XeTextVerifyCallbackStruct;



extern void      XeTextField_SetString(Widget PW, const char* PString, const EBool PNotify);
extern void      XeTextField_SetStringUTF8(Widget PW, const char* PString, const EBool PNotify);
extern void      XeTextField_SetStringW(Widget PW, const wchar_t* PString, const EBool PNotify);

extern char*     XeTextField_GetString(Widget PW);
extern char*     XeTextField_GetStringUTF8(Widget PW);
extern wchar_t*  XeTextField_GetStringW(Widget PW);

extern void      XeTextField_Insert(Widget PW, XeTextPosition PHere, char* PString);
extern void      XeTextField_Replace(Widget PW, XeTextPosition PStart, XeTextPosition PEnd, char* PString);

extern void      XeTextField_ClearSelection(Widget PW, Time PTime);

extern void      XeTextField_SetSelection(Widget PW, XeTextPosition PStart, XeTextPosition PEnd, Time PTime);

extern XeTextPosition  XeTextField_GetLastPosition(Widget PW);
extern void      XeTextField_SetInsertionPosition(Widget PW, XeTextPosition PPos);
extern void      XeTextField_SelectAll(Widget PW);


extern XeTranslations  XeStd_TextWidgetTransTable;


#if defined (__cplusplus) && (!(defined (__CPLUSLIBS)))
}	// Close scope of 'extern "C"' declaration
#endif

#endif // _XeTextField_h
