//===============================================================================
// Xe UI toolkit
//
// XeTextField
//  Private header file
//
// AUTHOR:  Gabor Nagy
// DATE:    2019-Jun-23 07:42:11
//
// Copyright (C) 1994-2021 by Gabor Nagy. All rights reserved.
//===============================================================================
#ifndef _XeTextFieldP_h
#define _XeTextFieldP_h

#include <Xe/PrimitiveP.h>
#include <Xe/TextField.h>

#if defined (__cplusplus) && (!(defined (__CPLUSLIBS)))
extern "C" {
#endif

#define XeSELECT	0x01
#define XeSCROLL	0x02
#define XeREDRAW	0x04


//================================
// Class definition
//================================

#define XeTextFieldClassCore\
	XePrimitiveClassCore

typedef struct _XeTextFieldClassRec
{
	XeTextFieldClassCore;
} XeTextFieldClassRec;

extern XeTextFieldClassRec  _XeTextField_class;


//================================
// Instance definition
//================================
// Instance part
//================================
typedef struct
{
// Resources
//
	XeColor                 ForegroundColor, TextSelectionColor, CursorColor;

	int                     PreferredNumOfRows, PreferredNumOfColumns;
	int                     NumOfRows, NumOfColumns;

	int                     NumOfTextColors;
	XeCoord2                StringAreaSize, TextMarginSize;

	XeCoord2                StringPosition;

	XeFontList              FontList;
	EBool                   FontListCreated;         // Default FontList was created in initialize()

	unsigned int            ShadowType;
	uint8_t                 StringAlignmentX, StringAlignmentY;

	char                    CursorType;              // XeIBEAM, XeSOLID

	XeCoordinate            StringCenterOffsetY;
	XeCoord2                StringSize;
	XeCoord2                StringOffset;            // Scroll-offset in pixels. These are synchronized with TopCharacter!

	XeTextPosition          TopCharacter,			       // Character position where displaying begins
	                        CursorPosition;
	XeCoordinate            CursorCol, CursorRow;
	XeCoord2                CursorPos;               // Cursor position in pixels

	int                     CursorBlinkRate;

	XeCallbackList          CursorMovedCallback, ValueChangedCallback, ModifyVerifyCallback, MotionVerifyCallback,
	                        ActivateCallback;

	EBool                   Editable, ActivateOnLeave, InsertMode,
	                        EditCompoundString,      // Edit the compound string / edit normal string and convert to compound for rendering
	                        CursorPositionVisible,
	                        ShadowColorsFromParentBackground;


	XeStringRenderSettings	RenderSettings;

// Private variables
//
	int                     CursorState;             // 0: off 255: on
	float                   ScrollSpeedX, ScrollSpeedY;
	XeCoord2                CursorWPos;              // Window-relative Cursor position in pixels
	XeCoord2                CursorSize;
	XeIntervalId            CursorTimerID, ScrollTimerID;
	EpImage                 AACacheImage;            // For supersampling (anti-aliasing)
	XeTextPosition          StringLength,
	                        SelectionStart, SelectionEnd,
	                        SelectionAnchor;
	char*                   String;                  // This could be UTF-8
	wchar_t*                WString;                 // X11 wide-character string (unicode)
	XeString                CompoundString;
	XeStringBinarySegment*  StringSegments;
	int                     StringNumOfSegments;
	XeIntervalId            DoubleClickTimer;
	char*                   Clipboard;
	EBool                   MotionSelectEnabled;

// X11
#ifdef EX11
	Pixmap                  X_UnderCursorPixmap;     // For blinking the cursor
	Pixel*                  X_TextColors;
	GC                      TextDrawGC, CursorDrawGC;
	Pixmap                  AACachePixmap;           // For supersampling (anti-aliasing). This Pixmap is the same size as AACacheImage. This is how we know when we need to resize it.
#endif
} XeTextFieldPart;


//================================
// Instance record
//================================
typedef struct _XeTextFieldRec
{
	XeWidgetCore;
	XeBasePart       Base;
	XePrimitivePart  Primitive;
	XeTextFieldPart  TextField;
} XeTextFieldRec;

#if defined (__cplusplus) && (!(defined (__CPLUSLIBS)))
}	// Close scope of 'extern "C"' declaration
#endif

#endif // _XeTextFieldP_h
