//================================================================
// Xe UI toolkit
//
// XeToggleButton
//  Private header file
//
// AUTHOR:  Gabor Nagy
// DATE:    2019-Jun-23 04:57:07
//
// Copyright (C) 1994-2021 by Gabor Nagy. All rights reserved.
//================================================================
#ifndef _XeToggleButtonP_h
#define _XeToggleButtonP_h

#include <Xe/LabelP.h>
#include <Xe/MenuButtonP.h>
#include <Xe/ToggleButton.h>

#if defined (__cplusplus) && (!(defined (__CPLUSLIBS)))
extern "C" {
#endif

//================================
// Class definition
//================================

#define XeToggleButtonClassCore\
	XeLabelClassCore

typedef struct _XeToggleButtonClassRec
{
	XeToggleButtonClassCore;
} XeToggleButtonClassRec;

extern XeToggleButtonClassRec  _XeToggleButton_class;


//================================
// Instance definition
//================================
// Instance part
//================================
typedef struct
{
// Resources
//
	XeColor         ArmColor, IndicatorColor, SelectColor;

	EpImage*        ArmImage;
	EpImage*        ActiveImage;

	XeCallbackList  ArmCallback, DisarmCallback, ActivateCallback, ValueChangedCallback;
	int             DefaultButtonShadowThickness;
	XeCoordinate    IndicatorSize, IndicatorGapWidth, IndicatorShadowThickness, SpacingX, Spacing;
	uint8_t         ButtonType, IndicatorType, SubButtonType;
	EBool           Editable, ActivateOnArm, Set,
	                IndicatorOn, CheckMarkOn,
	                FillOnSelect, VisibleWhenOff, ShowAsDefault,
	                SortDirectionArrowOn;
	Widget          FolderWidget;
	EBool           TurningOffEnabled;

// Private variables
//
	int             DefShadowSpacing;
	int             SubButtonSizeX;
	XeCoord2        IndicatorSizeP;
	XeCoord2        IndicatorPosition;
	EBool           IndicatorSizeXGiven, IndicatorSizeYGiven,
	                IndicatorSizeGiven;

	uint16_t        State;   // XeACTIVE | XeARMED | XeSUB_ARMED
	uint8_t         DragMode;
	uint8_t         ArrowDirection;
} XeToggleButtonPart;


//================================
// Instance record
//================================
typedef struct _XeToggleButtonRec
{
	XeWidgetCore;
	XeBasePart          Base;
	XePrimitivePart     Primitive;
	XeLabelPart         Label;
	XeMenuButtonPart    MenuButton;
	XeToggleButtonPart  ToggleButton;
} XeToggleButtonRec;




// Deferred Widgets

//-----------------------------
// Class
//-----------------------------
typedef struct
{
	EguiCoreClassPart  Core;
// ToggleButtonClassRec is empty
} EguiToggleButtonClassRec;

extern EguiToggleButtonClassRec  eguiToggleButtonClassRec;

//-----------------------------
// Instance
//-----------------------------
typedef struct
{
	XeCallbackFunc    Callback;
	EPointer          ClientData;
	char              Mnemonic;
	char*             Accel;
	char*             AccelText;
	char*             ToolTip;
	EguiCallbackList  ValueChangedCallback;
	EBool             State;
} EguiToggleButtonPart;

typedef struct _EguiToggleButtonRec
{
	EguiCorePart          Core;
	EguiToggleButtonPart  ToggleButton;
} EguiToggleButtonRec;

extern void  EguiCB_ToggleButton(Widget PW, EPointer PClientData, EPointer PCallData);




#if defined (__cplusplus) && (!(defined (__CPLUSLIBS)))
}	// Close scope of 'extern "C"' declaration
#endif

#endif // _XeToggleButtonP_h
