//================================================================
// Xe UI toolkit
//
// Main public header file
//
// AUTHOR:  Gabor Nagy
// DATE:    2019-Jun-23 23:33:42
//
// Copyright (C) 1994-2021 by Gabor Nagy. All rights reserved.
//================================================================
#ifndef _Xe_h
#define _Xe_h

#include <stdarg.h>

#include <Xe/XeTypes.h>

#include <EInput.h>


#include <Xe/Composite.h>


#ifdef XeOpenGL
extern EBool  XeGL_SupportsGL(Widget PW);
extern EBool  XeGL_IsInGLShell(Widget PW);
#else
#define XeGL_SupportsGL(pW)   FALSE
#define XeGL_IsInGLShell(pW)  FALSE
#endif

// Compress_exposure options
#define XeExposeCompressSERIES    ((XeEnum)TRUE)
#define XeExposeCompressMULTIPLE  2
#define XeExposeCompressMAXIMAL   3

// Modifiers
#define XeExposeGraphicsExpose        0x10
#define XeExposeGraphicsExposeMerged  0x20
#define XeExposeNoExpose              0x40
#define XeExposeNoRegion              0x80


#define XeUNSPECIFIED_INT     (-1)


#if defined (__cplusplus) && (!(defined (__CPLUSLIBS)))
extern "C" {
#endif

extern const int  _Xe_UnspecifiedInt;
#define XeUnspecifiedWindow       ((Window)2)
#define XeUnspecifiedWindowGroup  ((Window)3)

typedef struct
{
	EpImage*  Image;
	EIndex_2  Size;  // These fields allow us to tell if the Image has changed
	EIndex    ID;
	EIndex    ChangeCount;
} XeImage;

#define XeObjCore\
	char*  Name

#ifndef _XeP_h
struct _XeWidget_S
{
	XeObjCore;
};
#endif


typedef struct _XeObjectRec*    XeObject;
typedef struct _XeObjectClass*  XeObjectClass;


#ifndef VMS
externalref XeWidgetClass       XeObject_class;
#endif



// Built-in cursors
//
enum
{
	XeCursorLEFT_PTR = 1,
	XeCursorCENTER_PTR,
	XeCursorWATCH,
	XeCursorFOUR_ARROWS,
	XeCursorTOP_LEFT_CORNER,
	XeCursorTOP_RIGHT_CORNER,
	XeCursorBOTTOM_RIGHT_CORNER,
	XeCursorBOTTOM_LEFT_CORNER,
	XeCursorLEFT_SIDE,
	XeCursorRIGHT_SIDE,
	XeCursorTOP_SIDE,
	XeCursorBOTTOM_SIDE,
	XeCursorCROSSHAIR,
	XeCursorTARGET,
	XeCursorEXCHANGE,
	XeCursorSB_H_DOUBLE_ARROW
};



typedef struct XeStyle_S  XeStyle;

struct XeStyle_S
{
	void  (*Dummy)(void);
};

// Mouse pointer
//
typedef struct
{
	EpImage*  Image;

#ifdef EX11
	Display*  X_Display;
	Cursor    X_Cursor;
#endif
} XeCursor;



typedef struct
{
	EcRGBAiColor   RGBAi;        // Platform-independent color. HDR, fixed-point  0..65535 -> 0.0..1.0
	EcRGBAiColor*  RGBAiP;       // Temporary pointer, used for passing resource values (we can't pass structs, like EcRGBAiColor by value, as a resource).
	EcRGBAiColor*  RGBAiPtr;     // Pointer to Color. Using a pointer allows Xe for to support centralized settings and quick update of themes/colors.

	EBool          Set;          // This color was specified in a resource list (used in _WInit, etc.)
} XeColor;


//----------------------------------------------------------------
// Named colors For Xe_XPMToImageVa, XeX_XPMToPixmapVa etc.
//----------------------------------------------------------------
typedef struct
{
	char*         Name;
	EcRGBAiColor  RGBAi;
#ifdef EX11
	Pixel*        X_PixelPtr;
#endif
} XeNamedColor;



// XeExtension
// This object class is added to WMShell Widgets, to contain dialog box, file selection box, etc. data
//
typedef struct XeExtension_S           XeExtension;         // Forward-define XeExtension
typedef struct XeExtensionClassRec_S*  XeExtensionClass;    // Forward-define XeExtensionClass


#define XeCore(mType)\
	EEntityCore;\
	mType##Class  Class;\


#define XeExtensionCore\
	XeCore(XeExtension)

struct XeExtension_S
{
	XeExtensionCore;
};

/*
struct XeExtensionClass_S
{
	ECoreClassCore(XeExtension);
};
*/


extern XeExtension*  XeWMShellExtensionAdd(Widget PW, XeExtensionClass PClass);
extern XeExtension*  XeWMShellExtensionByClass(Widget PW, XeExtensionClass PClass);


typedef struct
{
	XeCallbackFunc  Callback;
	EPointer	      ClientData;
} XeCallbackRec, *XeCallbackList;

// Callbacks
//
typedef struct
{
	XeCallbackFunc  Proc;
	EPointer        ClientData;
} EguiCallbackRec;

typedef struct EguiCallbackList
{
	EList(, EguiCallbackRec, Callbacks);
} EguiCallbackList;



// For converting a Virtual keysym to a real keysym
//
typedef struct
{
	XeKeySym     keysym;
	XeModifiers  modifiers;
	XeKeySym     virtkey;
} XeVirtKeyBinding;



//#define XeUSE_XFT	1  // Use the new X FreeType interface library

#ifdef XeUSE_XFT
#include <X11/Xft/Xft.h>
#endif  // XeUSE_XFT

typedef XeGeometryResult  (*XeGeometryHandler)(Widget PW, XeWidgetGeometry* PRequest, XeWidgetGeometry* PReply);


#include <Xe/XeStrDefs.h>


#define EXeStart         ArgCnt = 0
#define EXeSetArg(n, v)  Args[ArgCnt].Name = (n), Args[ArgCnt].Value = (EArgVal)(v), ArgCnt++
#define EXeSetArgColor(m) do\
{\
	if(m.RGBAiPtr) EXeSetArg(XeNbackgroundColorPtr, m.RGBAiPtr);\
	else if(m.RGBAiP) EXeSetArg(XeNbackgroundColor, m.RGBAiP);\
	else EXeSetArg(XeNbackgroundColor, &m.RGBAi);\
} while(0)

#define XeToolTip8(m) do\
{\
	EXeSetArg(XeNtipString8, (EArgVal)(m));\
} while(0)

#define XeToolTip(m) do\
{\
	EXeSetArg(XeNtipString, (EArgVal)(m));\
} while(0)

#define XeSetToolTipXeString(mXeStr, mString) do\
{\
	mXeStr = XeString_NewVa(TRUE,  XeFONT, XeFontDEFAULT,  XeALIGNMENT, XeSTRING_ALIGN_LEFT,  XeTEXT, mString,  0);if(mXeStr)\
	{\
		EXeSetArg(XeNtipString, mXeStr);\
	}\
} while(0)



#define EXeSetImages2(m)  do { EXeSetArg(XeNlabelImage, m);EXeSetArg(XeNactiveImage, m##_Active);EXeSetArg(XeNarmImage, m); } while(0)
#define EXeSetImages3(m)  do { EXeSetArg(XeNlabelImage, m);EXeSetArg(XeNactiveImage, m##_Active);EXeSetArg(XeNarmImage, m##_Armed); } while(0)


typedef struct _EguiCoreRec*  EguiItem;
typedef struct _EguiCoreRec*  EguiCore;


typedef	int   XeIntervalId;

typedef void  (*XeTimerCallbackProc)(EPointer PClientData, XeIntervalId* PTID, const int PCount);


// XeNdeleteResponse values
//
enum
{
	XeDESTROY = 0, XeUNMAP_SHELL, XeDO_NOTHING
};


#define XeCallbackStructCore\
	int       Reason;\
	XeEvent*  Event


// Callback structures
//
typedef struct
{
	XeCallbackStructCore;
} XeCallbackStruct;


// Stretchable "themed" object (for scrollbar sliders, thermometers etc.)
//
typedef struct
{
	int       SizeX, SizeY,
	          StartCapLength,
	          MiddleTileLength,
	          EndCapLength;

	EpImage*  StartCap;
	EpImage*  MiddleTile;
	EpImage*  MiddleCache;
	EpImage*  EndCap;
} XeTile;



//------------------------------------------------------------------------
// Compound strings
//------------------------------------------------------------------------


//================================================================
// Defines for resources to be defaulted by vendors.
// String are initialized in Xeos.c
//================================================================
/*
#define XeDEFAULT_BACKGROUND            _XeSDEFAULT_BACKGROUND
#define XeDEFAULT_DARK_THRESHOLD        15
#define XeDEFAULT_LIGHT_THRESHOLD       77
#define XeDEFAULT_FOREGROUND_THRESHOLD  35
*/

typedef uint8_t  XeDirection;


#define XeDIRECTION_IGNORED            0x30

#define XeRIGHT_TO_LEFT_MASK           0x01  // 0x01 for bc
#define XeLEFT_TO_RIGHT_MASK           0x02
#define XeHORIZONTAL_MASK              0x03
#define XeTOP_TO_BOTTOM_MASK           0x04
#define XeBOTTOM_TO_TOP_MASK           0x08
#define XeVERTICAL_MASK                0x0c
#define XePRECEDENCE_HORIZ_MASK        0x40
#define XePRECEDENCE_VERT_MASK         0x80
#define XePRECEDENCE_MASK              0xc0

enum
{
	XeRIGHT_TO_LEFT_TOP_TO_BOTTOM = XeRIGHT_TO_LEFT_MASK | XeTOP_TO_BOTTOM_MASK | XePRECEDENCE_HORIZ_MASK,
	XeLEFT_TO_RIGHT_TOP_TO_BOTTOM = XeLEFT_TO_RIGHT_MASK | XeTOP_TO_BOTTOM_MASK | XePRECEDENCE_HORIZ_MASK,
	XeRIGHT_TO_LEFT_BOTTOM_TO_TOP = XeRIGHT_TO_LEFT_MASK | XeBOTTOM_TO_TOP_MASK | XePRECEDENCE_HORIZ_MASK,
	XeLEFT_TO_RIGHT_BOTTOM_TO_TOP = XeLEFT_TO_RIGHT_MASK | XeBOTTOM_TO_TOP_MASK | XePRECEDENCE_HORIZ_MASK,
	XeTOP_TO_BOTTOM_RIGHT_TO_LEFT = XeRIGHT_TO_LEFT_MASK | XeTOP_TO_BOTTOM_MASK | XePRECEDENCE_VERT_MASK,
	XeTOP_TO_BOTTOM_LEFT_TO_RIGHT = XeLEFT_TO_RIGHT_MASK | XeTOP_TO_BOTTOM_MASK | XePRECEDENCE_VERT_MASK,
	XeBOTTOM_TO_TOP_RIGHT_TO_LEFT = XeRIGHT_TO_LEFT_MASK | XeBOTTOM_TO_TOP_MASK | XePRECEDENCE_VERT_MASK,
	XeBOTTOM_TO_TOP_LEFT_TO_RIGHT = XeLEFT_TO_RIGHT_MASK | XeBOTTOM_TO_TOP_MASK | XePRECEDENCE_VERT_MASK,
	XeTOP_TO_BOTTOM = XeTOP_TO_BOTTOM_MASK | XeHORIZONTAL_MASK | XePRECEDENCE_MASK,
	XeBOTTOM_TO_TOP = XeBOTTOM_TO_TOP_MASK | XeHORIZONTAL_MASK | XePRECEDENCE_MASK,
	XeRIGHT_TO_LEFT = XeRIGHT_TO_LEFT_MASK | XeVERTICAL_MASK | XePRECEDENCE_MASK,
	XeLEFT_TO_RIGHT = XeLEFT_TO_RIGHT_MASK | XeVERTICAL_MASK | XePRECEDENCE_MASK,
	XeDEFAULT_DIRECTION = 0xff
};


extern EBool  XeDirectionMatch(XeDirection d1, XeDirection d2);

extern EBool  XeDirectionMatchPartial(XeDirection d1, XeDirection d2, XeDirection dmask);


typedef uint8_t          XeStringDirection;
typedef uint8_t          XeChar;                 // Opaque to outside
typedef unsigned short   XeWChar;                // Opaque to outside
typedef uint8_t*         XeString;               // Opaque to outside
typedef XeString*        XeStringTable;          // Opaque to outside
typedef uint8_t          XeStringComponentType;  // Component tags

// Tag codes:
//
#define XeStringCOMPONENT_MAGIC0  0xDF
#define XeStringCOMPONENT_MAGIC1  0x80

enum
{
	XeStringCOMPONENT_FONT = 1,
	XeStringCOMPONENT_TEXT,
	XeStringCOMPONENT_DIRECTION,
	XeStringCOMPONENT_SEPARATOR,
	XeStringCOMPONENT_LOCALE_TEXT,
	XeStringCOMPONENT_FULL_LENGTH,
	XeStringCOMPONENT_COLOR_NAME,
	XeStringCOMPONENT_COLOR_INDEX8,
	XeStringCOMPONENT_COLOR_INDEX16,
	XeStringCOMPONENT_COLOR_INDEX32,
	XeStringCOMPONENT_COLOR_RGB,
	XeStringCOMPONENT_COLOR_RGBA,
	XeStringCOMPONENT_ALIGN,
	XeStringCOMPONENT_FTEXT,         // Formatted text, just like in printf, followed by optional values
	XeStringCOMPONENT_ANTI_ALIASING  // 2, 3, 4 etc. For example : 2 means that the font is 2x larger, so rendered text is scaled down by 2 with filtering
};

// Shorthands for the above
//
#define	XeFONT           XeStringCOMPONENT_FONT
#define	XeTEXT           XeStringCOMPONENT_TEXT
#define	XeDIRECTION      XeStringCOMPONENT_DIRECTION
#define	XeSEPARATOR      XeStringCOMPONENT_SEPARATOR
#define	XeLOCALE_TEXT	   XeStringCOMPONENT_LOCALE_TEXT
#define	XeFULL_LENGTH	   XeStringCOMPONENT_FULL_LENGTH
#define	XeCOLOR_NAME	   XeStringCOMPONENT_COLOR_NAME
#define	XeCOLOR_INDEX8   XeStringCOMPONENT_COLOR_INDEX8
#define	XeCOLOR_INDEX16  XeStringCOMPONENT_COLOR_INDEX16
#define	XeCOLOR_INDEX32  XeStringCOMPONENT_COLOR_INDEX32
#define	XeCOLOR_RGB      XeStringCOMPONENT_COLOR_RGB
#define	XeCOLOR_RGBA     XeStringCOMPONENT_COLOR_RGBA
#define	XeALIGNMENT      XeStringCOMPONENT_ALIGN
#define	XeFTEXT          XeStringCOMPONENT_FTEXT
#define	XeAA             XeStringCOMPONENT_ANTI_ALIASING


// Codes 12-125 are reserved

#define XeStringCOMPONENT_END  ((XeStringComponentType)126)

// Codes 128-255 are user tags
#define XeStringCOMPONENT_USER_BEGIN  ((XeStringComponentType)128)
#define XeStringCOMPONENT_USER_END    ((XeStringComponentType)255)



#define XeSTRING_DEFAULT_TAG  NULL

enum { XeCOL_DEFAULT = 0, XeCOL_IDX8, XeCOL_IDX16, XeCOL_IDX32, XeCOL_RGB, XeCOL_RGBA };

#define XeCS_ESC  0x27


typedef struct
{
	EpImage*       Image;               // Each binary segment may have an Image

	XeFontSize*    FontSize;            // The Font for rendering this string segment
// String to identify the Font within a FontList. These are compared as pointers and not duplicated or freed by Xe,
// so it's the application's responsibility to maintain them (e.g. global const variables, such as const char  HeaderTag = "header1")
// This system allows applications to quickly and efficiently change the Fonts in existing compound strings, for example when the user changes the application preferences.
	const char*    FontTag;

	EcRGBAiColor   TextColor;           // Text foreground color

#ifdef EX11
	Pixel          X_TextColor;
	Pixmap         X_Pixmap;
#endif

	char*          String;              // ASCII or UTF-8 string
	wchar_t*       WString;             // Unicode string

	EInt32_2       Size,  Advance;
	int16_t        InkLeft, InkTop;     // Position for drawing glyph bitmaps, vectors, etc. (e.g. the first glyph may stic out to the left of the pen position)
	float          Ascent, Descent;
	uint8_t        Separator;           // New line after this segment. Number of new-line chars (1 on UNIX file: 0x0A, 2 on MS file: 0x0D,0x0A)
	uint8_t        Direction, Alignment,
	               AlignmentY;          // To the image...
	EBool          SetColor;            // A color was defined for this segment
} XeStringBinarySegment;


typedef struct
{
	XeStringBinarySegment*  Segments;
	unsigned int            NumOfSegments;
	int                     Length,        // Number of characters in line, excluding new line character(s)
	                        OrigLength;    // Number of characters in line, including new line character(s)
	int                     SizeX, SizeY, YStart;
} XeStringLine;


typedef struct
{
	EcRGBAiColor    BackgroundColor, CurrentColor;
	EpImage         GlyphImage;          // Internal buffer image for rendering glyphs with FreeType
	EpPixelBuffer*  GlyphPixelBuffer;
	EpImage         AACacheImage;        // Lazily-grown Image for down-scaling based anti-aliasing
	int             AACacheImageMargin;  // Used for blur-type operations, to avoid overflow
	int             ClipMinX, ClipMinY,  ClipSizeX, ClipSizeY;  // Clipping
#ifdef EX11
	Drawable  X_TargetDrawable;
	Pixmap    X_AACachePixmap;           // Lazily-grown Pixmap for down-scaling based anti-aliasing
	int       X_AACachePixmapSizeX, X_AACachePixmapSizeY;
	GC        X_BackgroundGC, X_GC;
	Pixel     X_BackgroundColor, X_CurrentColor;
#endif
} XeStringDrawContext;


typedef struct
{
	int   NumOfTabStops;
	int*  TabStops;
	int   UniformTabStops;
	char  TabUnits;    // XePIXELS, XeCHARACTERS
} XeStringRenderSettings;


enum { XeSTRING_DIRECTION_L_TO_R, XeSTRING_DIRECTION_R_TO_L, XeSTRING_DIRECTION_DEFAULT = 255 };
enum { XeSTRING_ALIGN_DEFAULT = 0, XeSTRING_ALIGN_LEFT, XeSTRING_ALIGN_RIGHT, XeSTRING_ALIGN_CENTER, XeSTRING_ALIGN_JUSTIFY };


typedef struct __XeStringContextRec*   XeStringContext;   // Opaque to outside
typedef struct __XeStringRec*          _XeString;         // Opaque to outside
typedef struct _XetStringContextRec*   _XeStringContext;  // Opaque to outside
typedef struct _XeFontListContextRec*  XeFontContext;     // Opaque to outside



// Keyboard focus policies for VendorShell
//
enum { XeEXPLICIT, XePOINTER };


///------------------------------------------------------------------------

// Text highlight modes
//
typedef enum
{
	XeHIGHLIGHT_NORMAL,
	XeHIGHLIGHT_SELECTED,
	XeHIGHLIGHT_SECONDARY_SELECTED
} XeHighlightMode;


typedef long  XeTextPosition;


#define XeCOPY_FAILED     0
#define XeCOPY_SUCCEEDED  1
#define XeCOPY_TRUNCATED  2

typedef struct
{
	Widget  W;
	int     SizeX;
} XeIndentRec;



enum
{
	XeSTATIC,
	XeDYNAMIC
};

// Cursor types for XeTextField and XeText
enum
{
	XeIBEAM,
	XeSOLID
};


typedef struct
{
	Widget  MatrixW;
	void*   IOModule;
} EIOModuleGUI;


// Used by XeWidgets_ChangeManage
//
typedef struct
{
	Widget  W;
	EBool   Manage;
} XeChangeManage;


// Used for blending Widgets with the background
//
typedef struct
{
	EpImage*       Image;int  X, Y;
	EcRGBAiColor*  Color;
} XeLayer;


typedef struct XeRectObj_S*       XeRectObj;
typedef struct _RectObjClassRec*  XeRectObjClass;
externalref XeWidgetClass         XeRectObj_class;

externalref XeWidgetClass         XeCoreWidget_class;


// Draw icon for menu PushButton or ToggleButton's sub-button
/*
#define _XeM_DrawSubButtonIcon(mWhat, mImage)\
{\
	int  lOptBX = LSizeX - LButton->SubButtonSizeX,\
	lOptBY = LEW->Base.MarginSizeY + LEW->Base.MarginTop,\
	lOptBSizeY = LSizeY - LEW->Base.MarginSizeY * 2 - LEW->Base.MarginTop - LEW->Base.MarginBottom;\
\
	EpImage*  lImage = _Xe_Theme->Image##mWhat;\
	EpImage*  lSImage = EpImage_New(0, 0,  EpRGBA, 0, EResUINT8,  FALSE);\
	EpImage_Scale(lImage, lSImage, LButton->SubButtonSizeX, lOptBSizeY, EpFilterGAUSSIAN, 0.75, NULL, NULL);\
	EpImage_Composite(lSImage, mImage, mImage,  0, 0,  lOptBX , lOptBY, lSImage->Size.X, lSImage->Size.Y,  EpSRCxALPHA_plus_DSTxONE_MINUS_ALPHA, EpADD);\
	E_Delete(lSImage);\
}
*/


#define _XeM_DrawSubButtonIcon(mWhat, mImage)\
{\
	const int  lOptBX = LSizeX - LButton->SubButtonSizeX,\
	           lOptBY = LEW->Base.MarginSize.Y + LEW->Base.MarginTop,\
	           lOptBSizeY = LSizeY - LEW->Base.MarginSize.Y * 2 - LEW->Base.MarginTop - LEW->Base.MarginBottom;\
\
	EpImage*        lImage = _Xe_Theme->Image_##mWhat;\
	const EpImage*  lScaledImage = Xe_ScaledImageFind(lImage, LButton->SubButtonSizeX, lOptBSizeY, TRUE);\
	EpImage_Composite(lScaledImage, mImage, mImage,  0, 0,  lOptBX , lOptBY, lScaledImage->Size.X, lScaledImage->Size.Y,  EpSRCxALPHA_plus_DSTxONE_MINUS_ALPHA, EpADD);\
/*{ char  s[256];sprintf(s, "tmp/%s()_%s:%d-%s.sgi", __FUNCTION__, __FILE__, __LINE__, PW->Name); EpImage_SaveToFile(LCacheImage, NULL,  s, NULL, 0,  NULL); }*/\
}



//--------------------------------
// Theme.c
//--------------------------------
extern void  Xe_BuildHImageFromTheme(XeTile* PTile, EpImage* PImage,  const EIndex PSizeX, const EIndex PSizeY,  const EpCImage* PImage_Left, const EpCImage* PImage_Middle, const EpCImage* PImage_Right);
extern void  Xe_BuildVImageFromTheme(XeTile* PTile, EpImage* PImage,  const EIndex PSizeX, const EIndex PSizeY,  const EpCImage* PImage_Top,  const EpCImage* PImage_Middle, const EpCImage* PImage_Bottom);


//--------------------------------
// Xe.c
//--------------------------------
extern int      Xe_Verbosity;
extern int      _XeX_RedrawVerbosity;

extern int      (*Xe_ItemAddFunction)(EguiItem PW);

extern EpImagePtr  Xe_ArrowDownImage, Xe_ArrowDownImage_Active;
extern EpImagePtr  Xe_ArrowLeftImage, Xe_ArrowLeftImage_Active;
extern EpImagePtr  Xe_ArrowRightImage, Xe_ArrowRightImage_Active;
extern EpImagePtr  _Image_Close, _Image_Close_Active, _Image_Close_Armed;
extern EpImagePtr  _Image_CloseButton, _Image_CloseButton_Active, _Image_CloseButton_Armed;

extern int      Xe_DrawerIndent;

extern int      XeMenu_ShadowOffset;
extern int      Xe_ScreenNumber;
extern int      Xe_DesktopIconSizeX, Xe_DesktopIconSizeY;
extern int		  Xe_ScreenSizeX, Xe_ScreenSizeY;

extern EBool    Xe_CommandKeyForCtrl;
extern EBool    Xe_ToolTips;
extern int      Xe_ToolTipShadowOffset;
extern int      Xe_ToolTipDelay;  // Milliseconds

extern int      XeDrawerIndent;


extern void   Xe_TileInit(XeTile* PTile);
extern void   Xe_TileFreeContents(XeTile* PTile);
extern EBool  Xe_TileGrow(XeTile* PTile,  const int PSizeX, const int PSizeY);

extern void   Xe_Popup(Widget PW, XeGrabMode PGrabKind);
extern void   Xe_Popdown(Widget PW);

extern void   XeAddCallbackN(Widget PW, const char* PCallbackName, XeCallbackFunc PCallback, EPointer PClientData);

extern void  XeAddCallback(Widget PW, const char* PCallbackName, XeCallbackFunc PProc, EPointer PClientData);
extern void  XtRemoveCallback(Widget PW, const char* PCallbackName, XeCallbackFunc PProc, EPointer PClientData);
extern void  XeAddCallbacks(Widget PW, const char* PCallbackName, XeCallbackList PCallbacks);
extern void  XtRemoveCallbacks(Widget PW, const char* PCallbackName, XeCallbackList PCallbacks);
extern void  XtRemoveAllCallbacks(Widget PW, const char* PCallbackName);
extern void  XtCallCallbacks(Widget PW, const char* PCallbackName, EPointer PCallData);
extern void  XtCallCallbackList(Widget PW, XeCallbackList PCallbacks, EPointer PCallData);


EList(extern, XeDisplay*, Xe_Displays);

//--------------------------------
// Function prototypes
//--------------------------------
extern XeApp        XeApp_New(const char* PName);
extern void         XeApp_Free(XeApp PApp);
extern XeDisplay*   XeApp_GetDisplay(XeApp PApp);
extern void         XeApp_CloseDisplays(XeApp PApp);

extern EObjectClass  XeVisual_class[];

extern XeDisplay*   XeDisplay_New(void);

extern void         _XeApp_CloseMarkedDisplays(XeApp PApp);

extern void         Xe_Init(void);
extern void         Xe_Free(void);

extern EArg*       Xe_MergeArgLists(const EArg* PArgs1, const int PNArgs1,  const EArg* PArgs2, const int PNArgs2);

extern XeDisplay*   Xe_OpenDisplay(XeApp PApp, const char* PDisplayName, const char* PAppName, const char* PClassName, int* PArgc, const char* const* PArgv);
extern XeVisual*    Xe_GetVisual(Widget PW);
extern XeDisplay*   Xe_GetDisplay(Widget PW);
extern XeVisual*    Xe_VisualGet(XeDisplay* PDisplay, const int PXScreenNumber, const uint16_t PTypeFlags, const uint16_t PDepth);

extern int          Xe_DisplaySizeX(XeDisplay* PDisplay);
extern int          Xe_DisplaySizeY(XeDisplay* PDisplay);

extern Widget    XeWidget_New(const char* PName, XeWidgetClass PClass, Widget PParent,  EArg* PArgs, const int PArgs_Count);
extern Widget    XeWidget_NewInsert(const char* PName, XeWidgetClass PClass,  Widget PParentW, const int PInsertAtIndex,  EArg* PArgs, const int PArgsN);
extern Widget    XeWidget_NewInsertAfter(const char* PName, XeWidgetClass PClass,  Widget PParentW, Widget PInsertAfterW,  EArg* PArgs, const int PArgsN);

extern Widget    XeWidget_WindowlessNew(const char* PName, XeWidgetClass PClass, Widget PParentW,  EArg* PArgs, const int PArgsN);
extern Widget    XeWidget_ManagedNew(const char* PName, XeWidgetClass PClass, Widget PParent, EArg* PArgs, const int PArgs_Count);
extern Widget    XeWidget_ManagedWindowlessNew(const char* PName, XeWidgetClass PClass, Widget PParentW,  EArg* PArgs, const int PArgsN);
extern Widget    XeWidget_NewVa(const char* PName, XeWidgetClass PClass, Widget PParentW, ...);
extern Widget    XeWidget_ManagedNewVa(const char* PName, XeWidgetClass PClass, Widget PParentW, ...);

extern void      XeWidget_Free(Widget PW);

extern int       XeWidget_SetCursor(Widget PW, const XeCursor* PCursor);


//
// Events
//

// This valid only for XeEventHandler_Remove and XeRawEventHandler_Remove. Don't use it to select events!

#define XeEventALL  ((XeEventMask) -1L)

extern void    XeEventHandler_Add(Widget PW, const XeEventMask PMask, const EBool PNonMaskable, XeEventHandler PFunc, EPointer PClosure);
extern void    XeEventHandler_Insert(Widget PW, const XeEventMask PEventMask, const EBool PNonMaskable,  XeEventHandler PFunc, EPointer PClientData,  XeListPosition PWhere);
extern void    XeEventHandler_Remove(Widget PW, const XeEventMask PEventMask, const EBool PNonMaskable,  XeEventHandler PFunc, EPointer PClientData);

extern void    XeRawEventHandler_Add(Widget PW, const XeEventMask PEventMask, const EBool PNonMaskable,  XeEventHandler PFunc, EPointer PClientData);
extern void    XeRawEventHandler_Insert(Widget PW, const XeEventMask PEventMask, const EBool PNonMaskable,  XeEventHandler PFunc, EPointer PClientData,  XeListPosition PWhere);
extern void    XeRawEventHandler_Remove(Widget PW, const XeEventMask PEventMask, const EBool PNonMaskable,  XeEventHandler PFunc, EPointer PClientData);

extern void    XeGrab_Add(Widget PW, const EBool PExclusive, const EBool PSpringLoaded);
extern void    XeGrab_Remove(Widget PW);
extern ETime   Xe_LastTimestampProcessed_ms(XeDisplay* PDisplay);

extern int     XeCursor_InitFromBuiltIn(XeDisplay* PDisplay, XeCursor* PCursor,  const int PID);
extern int     XeCursor_InitFromXPMs(XeDisplay* PDisplay, XeCursor* PCursor,  char** PXPMImage, const EcRGBAiColor* PColor0, const EcRGBAiColor* PColor1,  char** PMaskXPMImage);
extern void    XeCursor_FreeContents(XeCursor* PCursor);

extern float   XeActiveColorFactor, XeArmColorFactor, XeButtonColorFactor, XeSliderColorFactor, XeToggleButtonIndicatorColorFactor,
               XeListHeaderButtonColorFactor,
               XeTopShadowFactor, XeTopShadow2Factor, XeBottomShadowFactor, XeBottomShadow2Factor,
               XeOptionMenuPaneColorFactor,
               XeGradientBrightFactor, XeGradientDarkFactor,
               XeBottomShadow2Factor,
               XeGapColorFactor,
               XeTroughColorFactor;

extern void    Xe_SetMultiClickDelay_us(XeDisplay* PDisplay, const ETime PMicroseconds);
extern ETime   Xe_GetMultiClickDelay_us(XeDisplay* PDisplay);


//================================================================
// Event Gathering
//================================================================
typedef unsigned long  XeInputID;
typedef void      (*XeInputCallbackFunc)(EPointer PClientData, int* PSource, XeInputID* PId);
extern XeInputID  XeApp_AddInput(XeApp PApp, int PSource, EPointer PCondition, XeInputCallbackFunc PFunc, EPointer PClientData);
extern void       Xe_RemoveInput(XeInputID PId);




extern XeTheme*  _Xe_Theme;


extern XeStyle*  Xe_PushButtonStyle;
extern XeStyle*  Xe_ArrowButtonStyle;
extern XeStyle*  Xe_ToggleButtonStyle;
extern XeStyle*  Xe_ListHighlightStyle;
extern XeStyle*  Xe_ScrollBarStyle;

extern XeStyle*  XeStylePlain;
extern XeStyle*  XeStyleGradient;

extern int       Xe_ButtonGapWidth;
extern int       Xe_ButtonHighlightThickness;

extern int*      Xe_ClassDropDownPanelImageSizeXPtr;
extern int*      Xe_ClassDropDownPanelImageSizeYPtr;

#define XeCornerRadiusUNDEFINED  -9999

extern int       Xe_ButtonCornerRadius, Xe_DefeultButtonFrameSpacing;
extern int       Xe_IconSize;
extern int       Xe_FileSelectionBoxFolderIconSize, Xe_FileSelectionBoxIconSize;

extern void Xe_ShadowsCornerColorBands
(
	const int PShadowThickness, const uint16_t PType,  const EcRGBAiColor* PBackgroundColor, const EcRGBAiColor* PParentBackgroundColor,
	EpBand* PTopLftBands, const int PTopLftIndex, EpBand* PTopRgtBands, const int PTopRgtIndex,
	EpBand* PBtmLftBands, const int PBtmLftIndex, EpBand* PBtmRgtBands, const int PBtmRgtIndex
);

extern void  XeImage_Draw1PixelShadows(EpImage* PImage, const int PX, const int PY, const int PSizeX, const int PSizeY, const int PGapWidth, const EcRGBAp8Color PGapColor, const EcRGBAp8Color PTopShadowColor, const EcRGBAp8Color PBottomShadowColor);
extern void  XeImage_DrawShadows
(
	EpImage* PImage,  const int PX, const int PY,  const int PSizeX, const int PSizeY,  const int PGapWidth, const EcRGBAf32Color* PGapColor,
	const int16_t PTopLftRadius, EpBand* PTopLftBands,  const int16_t PTopRgtRadius, EpBand* PTopRgtBands,
	const int16_t PBtmLftRadius, EpBand* PBtmLftBands,  const int16_t PBtmRgtRadius, EpBand* PBtmRgtBands,
	const EBool PDrawCorners,
	const int PShadowThickness, const uint16_t PType, const uint8_t PGradient, const EcRGBAiColor* PBackgroundColor, const EcRGBAiColor* PParentBackgroundColor,
	const uint16_t PInsideBlendMode, const uint16_t POutsideBlendMode
);


#define _Xe_DrawFocusShadows(mEW, mImage, mOffset) do\
{\
	XeImage_DrawShadows(mImage, (mOffset), (mOffset),  Xe_SizeX(mEW) - (mOffset) * 2, Xe_SizeY(mEW) - (mOffset) * 2,  0, NULL,  0, NULL, 0, NULL,  0, NULL, 0, NULL,  FALSE,  mEW->Primitive.HighlightThickness, XeShadowIN|LShadowSpecFlags, 0,  LBackgroundColor, LBackgroundColor,  EpSRCxALPHA_plus_DSTxONE_MINUS_ALPHA, EpSRC);\
} while(0)


extern void  _XeImage_DrawArmedButtonShadows(EpImage* PImage,  const int PX, const int PY,  const int PSizeX, const int PSizeY,
              const int16_t PCornerTopLeftRadius, const int16_t PCornerTopRightRadius,
              const int16_t PCornerBottomLeftRadius, const int16_t PCornerBottomRightRadius,
              const EcRGBAiColor* PBackgroundColor, const EcRGBAiColor* PNormalBackgroundColor, const EcRGBAiColor* PParentBackgroundColor,
              const EcRGBAiColor* PGapColor, const int PGapWidth,
              const int PShadowThickness,
              const int16_t PInsideBlendMode, const int16_t POutsideBlendMode);

extern void		XeImage_DrawHighlight(EpImage* PImage, const XeStyle* PStyle, const int POrientation,
               const int PX, const int PY,  const int PSizeX, const int PSizeY,
               const int PHighlightBorder, const int PCornerRadius,
               EcRGBAf32Color* PColor, const float PGain, const float PContrast);

extern void   XeImage_DrawMetal(EpImage* PImage,  const int PX, const int PY,  const int PSizeX, const int PSizeY,
               const int POrientation,
               EcRGBAf32Color PColor, const float PGain, const float PContrast);

extern void   Xe_Warning(Widget PW, char* PMessage);
extern int    Xe_ConvertUnits(Widget PW, const int POrientation, const int PFromUnitType, const int PFromValue, const int PToUnitType);


extern int    _XeNumOfManagedChildren(Widget PW);

extern void   Xe_CallChangeManaged(Widget PW);
extern void   Xe_WidgetRealize(Widget PW);
extern void   Xe_WidgetUnrealize(Widget PW);
extern EBool  Xe_WidgetIsRealized(Widget PW);

extern void   Xe_WidgetSetSensitive(Widget PW, const EBool PSensitive);


extern void                 _XeColor_Init(XeColor* PXeColor);
extern void                 _XeColor_Set(XeColor* PXeColor, const EcRGBAiColor* PColor);
extern void                 _XeColor_SetPtr(XeColor* PXeColor, EcRGBAiColor* PColor);
extern const EcRGBAiColor*  _XeColor_Get(const XeColor* PColor);


//----------------------------------------
// FontList.c
//----------------------------------------
extern const float       XeFont_Sizes[];extern const int  XeFont_Sizes_Count;

extern float             XeFont_FontSizeMin;

extern char*             XeFont_LabelFontName;        extern XeFontSize*  Xe_LabelFontSize;
extern char*             XeFont_LabelBoldFontName;    extern XeFontSize*  Xe_LabelBoldFontSize;
extern char*             XeFont_FixedWidthFontName;   extern XeFontSize*  Xe_FixedWidthFontSize;
extern char*             XeFont_FinePrintFontName;    extern XeFontSize*  Xe_FinePrintFontSize;
extern char*             XeFont_Header1FontName;      extern XeFontSize*  Xe_Header1FontSize;
extern char*             XeFont_MenuFontName;         extern XeFontSize*  Xe_MenuFontSize;
extern char*             XeFont_DropDownPanelFontName;extern XeFontSize*  Xe_DropDownPanelFontSize;
extern char*             XeFont_ListFontName;         extern XeFontSize*  Xe_ListFontSize;
extern char*             XeFont_TextFieldFontName;    extern XeFontSize*  Xe_TextFieldFontSize;
extern char*             XeFont_NumbersFontName;      extern XeFontSize*  Xe_NumbersFontSize;

extern XeFontList        Xe_LabelFontList, Xe_Header1FontList, Xe_DropDownPanelFontList, XeMenu_FontList, Xe_ListFontList, Xe_TextFieldFontList, Xe_NumbersFontList, Xe_FixedWidthFontList;


extern XeFontListEntry*  XeFont_ListEntryNew(const char* PTag, XeFontSize* PFontSize);
extern void              XeFont_ListEntryFree(XeFontListEntry* PEntry);
extern XeFontList        XeFont_ListNew(const char* PName, XeFontSize* PFontSize);
extern void              XeFont_ListFree(XeFontList PFontList);
extern void              XeFont_ListAppendEntry(XeFontList PFontList, XeFontListEntry* PEntry);
extern XeFontListEntry*  XeFont_ListAddEntry(XeFontList PFontList, const char* PName, XeFontSize* PFontSize);
extern int               XeFont_ListSetEntry(XeFontList PFontList, const char* PFontTag, XeFontSize* PFontSize);
extern XeFontListEntry*  XeFont_ListGetDefaultEntry(XeFontList PFontList);
extern XeFontListEntry*  XeFont_ListGetEntryWithTagS(XeFontList PFontList, const char* PFontTag);

extern XeFontListEntry*  XeFont_ListGetEntryWithTag(XeFontList PFontList, const char* PFontTag);

extern XeFont*           XeFont_FontFamilyGetStyle(XeFontFamily* PFontFamily, const char* PStyleName);
extern XeFont*           XeFont_FontFamilyAddStyle(XeFontFamily* PFontFamily, const char* PFileName, const char* PStyleName, const EBool PReference);

extern XeFontSize*       XeFont_SizeAdd(XeFont* PFont, const float PSizeInPixels, const EBool PAntiAliasing);
extern XeFontSize*       XeFont_SizeGet(XeFont* PFont, const int PSizeInPixels, const EBool PAntiAliasing);
extern XeFontSize*       XeFont_FindClosestSize(const XeFont* PFont, const int PHeightInPixels, const int PMaxSizeDiff, const EBool PAntiAliasing, const EBool PSysBitmapOnly);
extern int               XeFont_Load(XeFont* PFont, XeFontSize* PFontSize);
extern int               XeFont_ParseFontName(const char* PXFontName,  char* PFoundryName, char* PFamilyName, char* PStyleStr, XeFontParams* PParams);
extern void              XeFont_GetNameBySize(const XeFont* PFont, const int PSizeInPixels, const EBool PAA, char* PName, const int PMaxLen);

extern char*             XeFontSize_GetName(XeFontSize* PFontSize);

extern EIndex            Xe_SysFontsCollect(XeDisplay* PDisplay, const EBool PDetails, const EBool PForceReload, const int PMinSizeInPixels, const uint8_t PSpacing, const char* PEncoding,const char* PCharset);

extern int               XeFont_CollectFTFonts(const char* PFontCacheFileName, const EBool PForceReload);
extern void              XeFont_SortFonts(XeFontFamilyPtrListA* PFontFamilies);
extern void              XeFont_SortSizes(XeFont* PFont);

extern int               XeFont_CollectFonts(XeDisplay* PDisplay, const EBool PCollectBitmapFonts, const EBool PForceReload);

extern XeFontFamily*    _XeFont_GetFontFamily(XeFontFamilyPtrListA* PList, const char* PFoundryName, const char* PFamilyName,  const EBool PAddIfDoesNotExist);


extern void              XeFont_GetXName(const XeFont* PFont, const int PFontSizeInPixels, char* PStr);


typedef struct
{
	Widget  W;
	int     SizeX;
} XeWidthAlignRec;


//---------------------------------------------------------------------------------------------------------

extern const char  XeTab[2];  // Special fly-weight strings

extern EBool       Xe_StringsEqual(const char* PString0, const char* PString1);

extern EBool       XeString_Dump(const XeString PXeStr, const EBool PRaw);
extern EBool       XeString_GetLtoR(XeString PXeStr, const char* PFontTag, char** PText);
extern EIndex      XeString_Length(XeString PXeStr);
extern XeString    XeString_Copy(const XeString PXeStr);
extern XeString    XeString_Concat(XeString PXeStr1, XeString PXeStr2);
extern XeString    XeString_SimpleNew(const char* PText);
extern XeString    XeString_New(const char* PString, const char* PTag);

extern XeString    XeString_NewV(const EBool PEntersToSeparators, va_list PVArgs);
extern XeString    XeString_NewVa(const int PEntersToSeparators, ...);
extern void        XeString_Free(XeString PXeStr);
extern void        XeString_Extents(const int PFormat, XeFontList PFontList, XeString PXeStr, int* PSizeX, int* PSizeY);

enum
{
	XeASCII = 1,
	XeUTF8
};

extern EIndex          XeString_ConvertToBinary(const XeVisual* PVisual,  const int PFormat, const XeFontList PFontList, const EcRGBAiColor* PForegroundColor, const XeString PXeStr,  XeStringBinarySegment** PSegmentsRet);

extern EIndex          XeString_8NumOfSegments(const char* PString);
extern EIndex          XeString_8ConvertToBinary(const XeVisual* PVisual,  XeFontList PFontList, const char* PString,  XeStringBinarySegment** PSegmentsRet);
extern EIndex          XeString_WConvertToBinary(const XeVisual* PVisual, XeFontList PFontList, const wchar_t* PString,  XeStringBinarySegment** PSegmentsRet);

extern void            XeString_BinaryDump(const XeStringBinarySegment* PSegments, const EIndex PNumOfSegments);

extern XeTextPosition  XeString_BinaryLength(const XeStringBinarySegment* PSegments, const EIndex PNumOfSegments);
extern char*           XeString_BinaryGetString8(const XeStringBinarySegment* PSegments, const EIndex PNumOfSegments);
extern char*           XeString_BinaryGetSubString8(const XeStringBinarySegment* PSegments, const EIndex PNumOfSegments, const XeTextPosition PStart, const XeTextPosition PEnd);
extern void            XeString_BinarySegmentSetFont(XeStringBinarySegment* PSegment, XeFontListEntry* PFontListEntry);


extern void	           XeString_BinaryFree(XeStringBinarySegment* PSegments, const EIndex PNumOfSegments);

extern XeStringBinarySegment*  XeString_BinaryCopy(const XeStringBinarySegment* PSegments, const EIndex PNumOfSegments);
extern XeStringBinarySegment*  XeString_BinaryInsertSegment(XeStringBinarySegment** PSegmentsP, int* PNumOfSegmentsP,  const EIndex PBefore);
extern XeStringBinarySegment*  XeString_BinaryInsertSegments(XeStringBinarySegment** PSegmentsP, int* PNumOfSegmentsP,  const EIndex PNumOfNewSegments, const EIndex PBefore);

extern void      XeString_BinaryUpdateFonts(XeStringBinarySegment* PSegments, const EIndex PNSegments,  XeFontList PFontList);

extern int       XeString_WCharPosToPixelX(XeFontSize* PFontSize,  const wchar_t* PWStr, const int PCharPos);
extern int       XeString_WGetClosestCharToPositionX(XeFontSize* PFontSize,  const wchar_t* PWStr, const int PStringLen,  const int PX,  int* PTextXP);
extern EBool     XeString_WGetExtents(XeFontSize* PFontSize,  const wchar_t* PString, const int PStartChar, const int PStringLen,  XeCoordinate* PBitmapLeftP, XeCoordinate* PBitmapTopP,  XeCoordinate* PSizeXP, XeCoordinate* PSizeYP,  XeCoordinate* PAdvanceXP, XeCoordinate* PAdvanceYP,  XeCoordinate* PMaxAscentP, XeCoordinate* PMaxDescentP, XeStringMetrics* PStringMetrics);
extern void      XeString_BinaryExtent(XeStringBinarySegment* PSegments, const EIndex PNumOfSegments, const XeStringRenderSettings* PRenderSettings, const EBool PComputeSegmentSizes,  XeStringMetrics* PResult);
extern int       XeString_BinaryPixelXToCharPos(XeStringBinarySegment* PSegment, const EIndex PNumOfSegments,  const EIndex PXIn);
extern EBool     XeString_BinaryEqual(const XeStringBinarySegment* PBinaryString0, const EIndex PSegmentsInString0, const XeStringBinarySegment* PBinaryString1, const EIndex PSegmentsInString1);

extern void      XeString_DrawContextInit(XeStringDrawContext* PContext);
extern void      XeString_DrawContextFreeContents(XeDisplay* PDisplay, XeStringDrawContext* PContext);

extern void      XeString_BinaryUpdateColors(const XeVisual* PVisual, const EcRGBAiColor* PForegroundColor,  XeStringBinarySegment* PSegments, const int PSegments_Count);
extern void      XeString_BinaryDraw(XeStringDrawContext* PContext, const XeVisual* PVisual, EpImage* PTargetImage, const int PBlendMode,  const int PDrawableSizeX, const int PDrawableSizeY, const float PBlurRadius,  const EcRGBAiColor* PBackgroundColor, const EcRGBAiColor* PForegroundXColor, const EBool PUseColorsInString,  const int PX, const int PY,  int PFullSizeX,  XeStringBinarySegment* PSegments, const int PSegments_Count, const XeStringRenderSettings* PRenderSettings, const int PAlignment);
extern void      Xe_DrawString(XeStringDrawContext* PContext,  const XeVisual* PVisual,  const char* PStr, const wchar_t* PWStr, const int PStartChar, const int PGlyphsToDraw,  const XeCoordinate PSizeX, const XeCoordinate PSizeY,
 XeFontSize* PFontSize, EpImage* PTargetImage, const int PBlendMode, const float PBlurRadius,  const XeCoordinate PDstX, const XeCoordinate PDstY,  const float PMaxAscent,  XeCoordinate* PAdvanceXP);

extern void      XeToolTipTimer(EPointer PClientData, XeIntervalId* PTID, const int PCount);
extern void      XeToolTipPopup(Widget PW);
extern void      XeToolTipPopdown(Widget PW);


extern EBool     Xe_SetFocus(Widget PW, const EBool PSendFocusLostEvent, const EBool PSendFocusGetEvent);
extern EBool     Xe_HasFocus(Widget PW);

extern void      Xe_TraversalRing(Widget* PTWs, const int PTrN);


typedef struct
{
	Widget*  Widgets;int  Widgets_Count;
	uint8_t  Direction;  // XeLEFT only, for now
	int      Indent;
} XeIndentSet;


extern XeIndentSet*  Xe_WidgetsIndentFromLeft(Widget* PWidgets, const int PNumOfWidgets, const int POffset);
extern void          Xe_WidgetIndentsUpdate(void);
extern void          Xe_WidgetIndentSetRemove(XeIndentSet* PSet);

extern int       Xe_GetPreferredSize(Widget PW,  const XeCoordinate PSuggestedSizeX, const XeCoordinate PSuggestedSizeY,  XeCoordinate* PSizeX, XeCoordinate* PSizeY);


extern void      Xe_GeometryUpdate(Widget PW, const uint8_t PUpdateFlags);
extern int       Xe_GeometryUpdateDisable(Widget PW);
extern void      Xe_GeometryUpdateEnable(Widget PW, const uint8_t PUpdateFlags);

extern void      Xe_BranchRedraw(Widget PW, const uint8_t PUpdateFlags, const EBool PUnmanaged, const EBool PClearDrawCaches);

extern void      Xe_CallExposeForWindowlessChildren(XeCompositeWidget PEW, const EBool PMarkedOnly, XeRedrawEvent* PEvent);

extern EBool     XeIsSubclass(const Widget PW, const XeWidgetClass PClass);
extern int       XeWidget_GetIndex(Widget PW);



// Resources
//
extern const char*  XeVaNestedList;
extern const char*  XeVaTypedArg;

extern void      Xe_SetValues(Widget PW, EArg* PArgs, Cardinal PNumOfArgs);
extern void      Xe_SetValuesVa(Widget PW, ...);
extern void      Xe_GetValues(Widget PW, EArg* PArgs, const int PNArgs);
extern void      Xe_GetValuesVa(Widget PW, ...);

extern void      Xe_GetResourceList(XeWidgetClass PWidgetClass, XeResource** PResourcesR, Cardinal* PNResourcesR);



extern void      XeAppMessage(XeApp PApp, const int PType, const char* PName, const char* PTypeName, const char* PFunction, const char* PFile, const int PLine, const char* PMessage,  const char** PParams, const int PNParams);

#define XeAppWarning(mApp, mName, mTypeName, mMessage,  mParams, mNParams)  XeAppMessage(mApp, ELogWARNING, mName, mTypeName, __FUNCTION__, __FILE__, __LINE__, mMessage,  mParams, mNParams)
#define XeAppError(mApp, mName, mTypeName, mMessage,  mParams, mNParams)  XeAppMessage(mApp, ELogERROR, mName, mTypeName, __FUNCTION__, __FILE__, __LINE__, mMessage,  mParams, mNParams)


//
// Manage.c
//
extern void    Xe_WidgetMap(Widget PW);
extern void    Xe_WidgetUnmap(Widget PW);

extern void    XeWidget_Manage(Widget PW);
extern void    XeWidgets_Manage(Widget* PWidgets, const int PNumOfChildren);
extern void    XeWidget_Unmanage(Widget PW);
extern void    XeWidgets_Unmanage(Widget* PWidgets, const int PNumOfChildren);
extern void    XeWidgets_ChangeManage(const XeChangeManage* PChildren, const int PNumOfChildren);

extern void    XeWidget_ManageNoNotify(Widget PW);
extern void    XeWidgets_ManageNoNotify(Widget* PWidgets, const int PNumOfChildren);
extern void    XeWidget_UnmanageNoNotify(Widget PW);
extern void    XeWidgets_UnmanageNoNotify(Widget* PWidgets, const int PNumOfChildren);



//
// Event.c
//
extern EBool   Xe_CallAcceptFocus(Widget PW, ETime* PTime);
extern void    XeApp_AddActions(XeApp PApp, XeActionList PActions, Cardinal PNActions);
extern void    Xe_MainLoop(XeApp PApp);
extern void    XeApp_SetExitFlag(XeApp PApp);

extern void    Xe_TranslateCoords(Widget PW,  const int PX, const int PY,  int* PRootXRet,  int* PRootYRet);

//
// X11/Event.c
//
extern XeIntervalId  XeTimeOut_Add(const int PMilliseconds, XeTimerCallbackProc PCallback, EPointer PClientData);
extern XeIntervalId  XeTimeOut_AddMicro(const ETime PMicroseconds, XeTimerCallbackProc PCallback, EPointer PClientData);
extern XeIntervalId  XeTimer_Add(const int PMilliseconds, XeTimerCallbackProc PCallback, EPointer PClientData);
extern XeIntervalId  XeTimer_AddMicro(const ETime PMicroseconds, XeTimerCallbackProc PCallback, EPointer PClientData);
extern void          XeTimer_Remove(XeIntervalId PID);
extern void          XeTimer_SetInterval(const XeIntervalId PID, const ETime PMilliseconds);
extern void          XeTimer_SetIntervalMicro(const XeIntervalId PID, const ETime PMicroseconds);

extern int           XeApp_Sync(XeApp PApp);


// Global variables
//
extern EBool         Xe_MultiThreaded;

extern EcRGBAiColor  Xe_BackgroundRGBAiColor,
                     Xe_ForegroundRGBAiColor,

                     Xe_BlackRGBAiColor,
                     Xe_TransparentBlackRGBAiColor,
                     Xe_WhiteRGBAiColor,
                     Xe_RedRGBAiColor,
                     Xe_GreenRGBAiColor,
                     Xe_BlueRGBAiColor,
                     Xe_CyanRGBAiColor,
                     Xe_MagentaRGBAiColor,
                     Xe_YellowRGBAiColor,

                     Xe_TroughRGBAiColor,

                     Xe_SliderRGBAiColor,

                     Xe_ListBackgroundRGBAiColor,
                     Xe_ListBackgroundAlternateRGBAiColor,
                     Xe_ListItemHighlightRGBAiColor,

                     XeMenu_BackgroundRGBAiColor,
                     XeMenu_ForegroundRGBAiColor,
                     XeMenu_ButtonArmRGBAiColor,

                     Xe_TextFieldBackgroundRGBAiColor,
                     Xe_TextFieldForegroundRGBAiColor,
                     Xe_TextFieldFocusRGBAiColor,
                     Xe_TextFieldTextSelectionRGBAiColor,

                     Xe_ButtonRGBAiColor,
                     Xe_ButtonActiveRGBAiColor,
                     Xe_ToolTipBackgroundRGBAiColor,
                     Xe_ScrollbarSliderRGBAiColor,

                     Xe_HighlightRGBAiColor,
                     Xe_ProgressBarRGBAiColor,
                     Xe_ToggleButtonLEDRGBAiColor;

extern EcRGBAiColor  Xe_TransparentRGBAiColor;

extern EcRGBAiColor  Xe_GapRGBAiColor,
                     Xe_ListHeaderButtonRGBAiColor,
                     XeOptionMenu_RGBAiColor;

extern uint32_t      Xe_ForegroundRGBA8pColor;

extern EcRGBAf32Color  Xe_TransparentBlackRGBAf32Color;


extern void          Xe_SetSensitive(Widget PW, const EBool PSensitive);

extern void          Xe_TranslateKeycode(XeDisplay* PDisplay, const XeKeyCode PKeyCode, XeModifiers PModifiers, XeModifiers* PModifiersR, XeKeySym* PKeysymR);
extern void          Xe_TranslateKey(XeDisplay* PDisplay, const XeKeyCode PKeyCode, XeModifiers PModifiers, XeModifiers* PModifiersRet, XeKeySym* PKeysymRet);

extern void          XeWidget_SetKeyboardFocus(Widget PSubtree, Widget PDescendant, const EBool PSendFocusLostEvent, const EBool PSendFocusGetEvent);
extern Widget        Xe_KeyboardFocusWidgetGet(Widget PW);

extern void          Xe_GrabKey(Widget PW, const XeKeyCode PKeyCode, const XeModifiers PModifiers, const EBool POwnerEents, int PPointerMode, int PKeyboardMode);
extern void          Xe_UngrabKey(Widget PW, const XeKeyCode PKeyCode, const XeModifiers PModifiers);
extern int           Xe_GrabKeyboard(Widget PWidget, const EBool POwnerEvents, const int PPointerMode, const int PKeyboardMode, const ETime PTime);
extern void          Xe_UngrabKeyboard(Widget PW, ETime PTime);

extern void          Xe_UngrabButton(Widget PW, const unsigned int PButton, const XeModifiers PModifiers);

extern void          Xe_UngrabPointer(Widget PW, const ETime PTime);


#define _Xe_GetRGBA8pColors()\
	const EBool     LDarkMode = Ec_LumaRGBAi(&Xe_BackgroundRGBAiColor) < 0x8000;\
	const uint32_t  LGreenRGBA8p = LDarkMode ? Ec_RGBA8p(0x00C000FF) : Ec_RGBA8p(0x008000FF);(void)LGreenRGBA8p;\
	const uint32_t  LOrangeRGBA8p = LDarkMode ? Ec_RGBA8p(0xB08000FF) : Ec_RGBA8p(0xA06000FF);(void)LOrangeRGBA8p;\
	const uint32_t  LYellowRGBA8p = LDarkMode ? Ec_RGBA8p(0xE0E000FF) : Ec_RGBA8p(0x906000FF);(void)LYellowRGBA8p;\
	const uint32_t  LRed1RGBA8p = LDarkMode ? Ec_RGBA8p(0x200000FF) : Ec_RGBA8p(0x400000FF);(void)LRed1RGBA8p;\
	const uint32_t  LRed2RGBA8p = LDarkMode ? Ec_RGBA8p(0x401000FF) : Ec_RGBA8p(0x600000FF);(void)LRed2RGBA8p;\
	const uint32_t  LRed3RGBA8p = LDarkMode ? Ec_RGBA8p(0x602000FF) : Ec_RGBA8p(0x800000FF);(void)LRed3RGBA8p;\
	const uint32_t  LRed4RGBA8p = LDarkMode ? Ec_RGBA8p(0xE05030FF) : Ec_RGBA8p(0xC00000FF);(void)LRed4RGBA8p;\
	const uint32_t  LBlue1RGBA8p = LDarkMode ? Ec_RGBA8p(0x000020FF) : Ec_RGBA8p(0x000040FF);(void)LBlue1RGBA8p;\
	const uint32_t  LBlue2RGBA8p = LDarkMode ? Ec_RGBA8p(0x000040FF) : Ec_RGBA8p(0x000060FF);(void)LBlue2RGBA8p;\
	const uint32_t  LBlue3RGBA8p = LDarkMode ? Ec_RGBA8p(0x204060FF) : Ec_RGBA8p(0x000080FF);(void)LBlue3RGBA8p;\
	const uint32_t  LBlue4RGBA8p = LDarkMode ? Ec_RGBA8p(0x40A0FFFF) : Ec_RGBA8p(0x0000C0FF);(void)LBlue4RGBA8p


extern const EObjectClass  XeFontFamily_class[];
extern const EObjectClass  XeFont_class[];
extern const EObjectClass  XeFontSize_class[];

extern const float         XeFont_AAMult;
EListA_s(extern, XeFontFamilyPtr, XeFont_Families);
EList(extern, char*, XeFont_Paths);


extern XeFontSize*     XeFontSize_New(XeDisplay* PDisplay, const char* PFontName, float* PAntiAliasing, const EBool PFallbackToDefault);
extern XeFontCharSet*  XeFontSize_CharSetAdd(XeFontSize* PFontSize, const char* PName);
extern void            XeFontSize_FreeContents(XeFontSize* PFontSize);
extern EBool           XeFontSize_HasLatin1(XeFontSize* PFontSize);

extern void            XeFont_FreeContents(XeFont* PFont);

extern XeFontSize*     Xe_DisplayGetDefaultFontSize(XeDisplay* PDisplay);

extern XeGlyph*        XeFontSize_GlyphLoadAndRender(XeFontSize* PFontSize, const int PGlyphIndex, const EBool PAA, const EBool PSetResolution);
extern int             XeFontSize_GlyphsLoadAndRender(XeFontSize* PFontSize, const int PEndCode, const EBool PAA);

extern void            XeFont_Exit(void);

extern void        XeImage_DrawButton(EpImage* PImage,
 const XeStyle* PStyle, const int POrientation,
 const int PX, const int PY,  const int PSizeX, const int PSizeY,
 const int PGapWidth, const EcRGBAiColor* PGapColor, const int PShadowThickness,
 const int PCornerTopLeftRadius, const int PCornerTopRightRadius,
 const int PCornerBottomLeftRadius, const int PCornerBottomRightRadius,
 const EcRGBAiColor* PBackgroundColor, const EcRGBAiColor* PParentBackgroundColor, const float PGain, const float PContrast);

extern void        XeImage_DrawButtonArmed(EpImage* PImage,
 const XeStyle* LStyle, const int POrientation,
 const int PX, const int PY, const int PSizeX, const int PSizeY,
 const int PGapWidth, const EcRGBAiColor* PGapXColor, const int PShadowThickness,
 const int PCornerTopLeftRadius, const int PCornerTopRightRadius,
 const int PCornerBottomLeftRadius, const int PCornerBottomRightRadius,
 const EcRGBAiColor* PBackgroundColor, const EcRGBAiColor* PNormalBackgroundColor, const EcRGBAiColor* PParentBackgroundColor);


extern void		     XeImage_DrawLeftRightVLines(EpImage* PImage,  const int PSizeX, const int PSizeY,  const EcRGBAiColor* PLeftLineColor,  const EcRGBAiColor* PRightLineColor);
extern void        XeImage_DrawGrip(EpImage* PImage,  const int PX, const int PY,  const int PSizeX, const int PSizeY,  const EcRGBAiColor* PHColor, const EcRGBAiColor* PDColor);

extern EBool       Xe_IsRealized(Widget PW);
extern EBool       Xe_ChainIsManaged(Widget PW);

#ifndef _XeP_h
extern Widget      Xe_Parent(Widget PW);
extern EBool       Xe_IsManaged(Widget PW);
extern EBool       Xe_IsSensitive(Widget PW);
extern int         Xe_SizeX(Widget PW);
extern int         Xe_SizeY(Widget PW);
#endif


typedef struct
{
	Widget        FrameW, DrawW;
	EpImage*      Image;
	EpImage*      BackgroundImage;  // To show alpha channel (e.g. a checker-board)
	EcRGBAiColor  BackgroundColor;
#ifdef EX11
	Pixmap        ImagePixmap;
	GC            DrawGC;
#endif
	EIndex_2      PreferredSize, Size;
	EBool         GCCreated;
} XeImageSwatch;


extern XeImageSwatch*  XeImageSwatch_New(Widget PParentW,  int PPreferredSizeX, int PPreferredSizeY,  XeImageSwatch* PSwatch);
extern void            XeImageSwatch_Free(XeImageSwatch* PSwatch, const EBool PFreeStruct);
extern void            XeImageSwatch_Redraw(XeImageSwatch* PSwatch);
extern void            XeImageSwatch_Update(XeImageSwatch* PSwatch,  const int PPreferredSizeX, const int PPreferredSizeY,  EpImage* PImage, EpToneMapper* PToneMapper);

typedef void  (*XeErrorHandler)(const uint8_t PType, const char* PMessage);

extern XeErrorHandler  XeApp_SetErrorHandler(XeApp PApp, XeErrorHandler PFunc);

extern const EcRGBAiColor*  Xe_InheritBackgroundColor(Widget PW);
extern int                  Xe_InheritBackgroundImage(Widget PW, EpImage** PImage);
extern int                  Xe_CollectBackgroundLayers(Widget PW, XeLayer* PLayers, const int PMaxLayers);


extern const char  _XePChars[];
extern EpImage*    Xe_XPMToImageVa(char** PXPM, ...);

extern XeApp       Xe_WidgetToApp(Widget PW);

extern Widget      Xe_ToolBarItemRealize(Widget PParentW, EguiItem PItem, Widget PBeforeW,  int32_t* PImageSizeXPtr, int32_t* PImageSizeYPtr);

extern const EpImage*  Xe_ScaledImageFind(EpImage* PImage,  const EIndex PSizeX, const EIndex PSizeY, const EBool PAddIfDoesNotExist);


extern XeApp      XeDisplay_ToApp(XeDisplay* PDisplay);

typedef void      (*XeKeyProc)(XeDisplay* PDisplay, const XeKeyCode PKeycode, XeModifiers PModifiers, XeModifiers* PModifiersR, XeKeySym* PKeysymR);
typedef void      (*XeCaseProc)(XeDisplay* PDisplay, XeKeySym PKeysym, XeKeySym* PLowerReturn, XeKeySym* PUpperReturn);
extern void       Xe_KeysymToKeycodeList(XeDisplay* PDisplay, XeKeySym PKeySym, XeKeyCode** PKeyCodesRet, Cardinal* PKeyCountRet);
extern XeKeySym*  Xe_GetKeysymTable(XeDisplay* PDisplay, XeKeyCode* PMinKeyCodeRet, int* PKeysymsPerKeycodeRet);
extern void       Xe_KeyTranslatorSet(XeDisplay* PDisplay, XeKeyProc PFunc);


//================================================================
// Translation Management
//================================================================

typedef struct _TranslationData*  XeTranslations;
typedef struct _TranslationData*  XeAccelerators;

extern XeTranslations  XeParseTranslationTable(const char* PStr);
extern XeAccelerators  XeParseAcceleratorTable(const char* PStr);
extern void            XeOverrideTranslations(Widget PW, XeTranslations PTrans);
extern void            XeAugmentTranslations(Widget PW, XeTranslations PTrans);

extern void   XtUninstallTranslations(Widget PW);

extern void   XeCallActionProc(Widget PW, const char* PAction, XeEvent* PEvent,  String* PParams, Cardinal PNParams);


#ifdef __APPLE__
extern int  Xe_CocoaAppStart(void);
extern int  Xe_CocoaDialog(void);
#endif


//
// Deferred Widgets
//

#define XeItemFree(m)  { if(m) EGUI_DestroyItem(m);m = NULL; }


typedef struct _EguiItemClassRec*       _EguiItemClass;
typedef void*                           EguiItemClass;

extern EguiItemClass                    eguiCoreClass;
typedef struct _EguiCoreClassRec*       EguiCoreClass;

// Public global variables and functions
//
extern int   XeApp_InitDisplay(XeApp PApp, const int PArgs_Count, const char* const* PArgs,  const EArg* PArgList, const int PArgCnt);

extern void  EGUI_DestroyItem(EguiItem PItem);

extern EguiItem  Xe_AddCustomWidget(const char* PLocation, const char* PName, XeCallbackFunc PCallback);

extern void  XeColorTheme_Dark(void);
extern void  XeColorTheme_Light(void);
extern void  Xe_UpdateColors(void);


#if defined (__cplusplus) && (!(defined (__CPLUSLIBS)))
}  // Close scope of 'extern "C"' declaration
#endif

#endif // _Xe_h
