//================================================================
// Xe UI toolkit
//
// OpenGL-specific functionality
//
// AUTHOR:  Gabor Nagy
// DATE:    2018-Apr-01 16:05:59
//
// Copyright (C) 1994-2021 by Gabor Nagy. All rights reserved.
//================================================================
#ifndef _XeGL_h
#define _XeGL_h

#include <Xe/XeTypes.h>

#ifdef __APPLE__
#include <OpenGL/gl.h>
#else

#ifndef sgi
#define GL_GLEXT_PROTOTYPES        1
#endif

#include <GL/gl.h>
#endif	// __APPLE__

#ifdef sgi
#include <GL/glx.h>
#define GLXPbuffer  GLXPbufferSGIX
#else

#ifdef __APPLE__
 #include <OpenGL/glext.h>
#else
 #include <GL/glext.h>
#endif  // __APPLE__

#endif  // sgi

//#endif // USEOpenGL


#if defined (__cplusplus) && (!(defined (__CPLUSLIBS)))
extern "C" {
#endif

// OpenGL-specific data for an XeFontSize (GPUData field)
typedef struct
{
	EObjectCore(EObject);

	int   FirstGlyphGLID;  // Used if UnicodeToGlyphGLCode is NULL
} XeGLFontSizeData;


#define XeMAX_PRINTF_BUFFER_SIZE  4096


extern const EObjectClass  XeGLFontSizeData_class[];

extern XeVisual*  XeGL_VisualNew(XeDisplay* PDisplay, const int PScreenNumber, const int* PGLVisualAttributes);
extern void       XeGL_FontSizeLoadGlyph(XeFontSize* PFontSize,  const int PGlyphIndex, const int PBase);
extern void       XeGL_UseXeFontSize(XeFontSize* PFontSize, const int PCodeMax,  const int PBase);
extern int        XeGL_LoadFontSize(XeFontSize* PFontSize);

extern void       XeGL_printff(const XeFontSize* PFontSize,  const float PX, const float PY, const float PZ,  const float PLineSpacing, const char* PText, ...);
extern void       XeGL_printfMLf(const XeFontSize* PFontSize,  const float PX, const float PY, const float PZ,  const float PStartX, const float PStartLine, const float PLineSpacing, const char* PText, ...);
extern void       XeGL_printfd(const XeFontSize* PFontSize,  const double PX, const double PY, const double PZ,  const double PLineSpacing, const char* PText, ...);
extern void       XeGL_printfMLd(const XeFontSize* PFontSize,  const double PX, const double PY, const double PZ,  const double PStartX, const double PStartLine, const double PLineSpacing, const char* PText, ...);
extern void       XeGL_wprintff(const XeFontSize* PFontSize,  const float PX, const float PY, const float PZ,  const float PLineSpacing, const wchar_t* PText);
extern void       XeGL_wprintfd(const XeFontSize* PFontSize,  const double PX, const double PY, const double PZ,  const double PLineSpacing, const wchar_t* PText);


extern GLenum                  XeGL_CheckStatus(const char* PFunction, const char* PFile, const int PLine, const char* PMessage, const EBool PShowNoError);
#define XeGLM_CheckStatus(mShowNoError)    XeGL_CheckStatus(__FUNCTION__, __FILE__, __LINE__, "", mShowNoError)
#define XeGLM_CheckStatusNL(mShowNoError)  XeGL_CheckStatus(__FUNCTION__, __FILE__, __LINE__, "\n", mShowNoError)

extern ELogFunc  XeGL_LogFunc;


#if defined (__cplusplus) && (!(defined (__CPLUSLIBS)))
}  // Close scope of 'extern "C"' declaration
#endif

#endif // _XeGL_h
