//================================================================================
// Xe UI toolkit
//
// Strictly internal stuff
//
// AUTHOR:  Gabor Nagy
// DATE:    2019-Jun-23 08:03:05
//
// Copyright (C) 1994-2021 by Gabor Nagy. All rights reserved.
//===============================================================================
#ifndef _XeI_h
#define _XeI_h

#include <Xe/XeP.h>
#include <XtXe/TranslateI.h>

#if defined (__cplusplus) && (!(defined (__CPLUSLIBS)))
extern "C" {
#endif

#ifndef X_NOT_POSIX
#ifdef _POSIX_SOURCE
#include <limits.h>
#else
#define _POSIX_SOURCE
#include <limits.h>
#undef _POSIX_SOURCE
#endif
#endif
#ifndef PATH_MAX
#ifdef WIN32
#define PATH_MAX 512
#else
#include <sys/param.h>
#endif
#ifndef PATH_MAX
#ifdef MAXPATHLEN
#define PATH_MAX MAXPATHLEN
#else
#define PATH_MAX 1024
#endif
#endif
#endif


#ifdef CRAY
#define WORD64
#endif

#if defined (_LP64) || \
    defined(__alpha) || defined(__alpha__) || \
    defined(__ia64__) || defined(ia64) || \
    defined(__sparc64__) || \
    defined(__s390x__) || \
    (defined(__hppa__) && defined(__LP64__)) || \
    defined(__amd64__) || defined(amd64) || \
    defined(__powerpc64__) || \
    (defined(sgi) && (_MIPS_SZLONG == 64))
#define LONG64
#endif


#if defined(XTHREADS) && defined(XUSE_MTSAFE_API)

#define _XeDisplay_ToApp(mD)  XeApp  LApp = (mD)->App
#define _Xe_WidgetToApp(w)    XeApp  LApp = (w)->App
#define _XeAppLock(app)       XeAppLock(app)
#define _XeAppUnlock(app)     XeAppUnlock(app)
#define _XeProcessLock()      XeProcessLock()
#define _XeProcessUnlock()    XeProcessUnlock()

// If Xe provides an API to query whether it's multi-threaded, neutralize _XeProcessLock()
//
extern void  (*_XeProcessLock)(void);

#else

#define _XeDisplay_ToApp(d)   do { } while(0)
#define _Xe_WidgetToApp(w)    do { } while(0)
#define _XeAppLock(app)       do { } while(0)
#define _XeAppUnlock(app)     do { } while(0)
#define _XeProcessLock()      do { } while(0)
#define _XeProcessUnlock()    do { } while(0)

#endif  // XTHREADS && XUSE_MTSAFE_API


// Window-to-Widget mapping
//
typedef struct _WWPair
{
	struct _WWPair*  Next;
#ifdef EX11
	Window           X_Window;
#endif
	Widget           widget;
} *WWPair;

typedef struct _XeW2WTable
{
	unsigned int  Mask;      // Size of hash table - 1
	unsigned int  Rehash;    // Mask - 2
	unsigned int  Occupied;  // Number of occupied entries
	unsigned int  Fakes;     // Number occupied by WWfake
	Widget*       Entries;   // The entries
	WWPair        Pairs;     // Bogus entries
} *XeW2WTable;


typedef enum
{
	XeNoServerGrab, 
	XePassiveServerGrab,
	XeActiveServerGrab,
	XePseudoPassiveServerGrab,
	XePseudoActiveServerGrab
} XeServerGrabType;


typedef struct
{
	XeServerGrab      Grab;  // Need a copy, in order to protect during grab.
	XeServerGrabType  GrabType;
} XeInputDevice;

typedef struct _XeGrabRec*  XeGrabList;

typedef struct
{
	XeGrabList     GrabList;
	XeInputDevice  Keyboard, Pointer;
	XeKeyCode      ActivatingKey;
	Widget*        Trace;
	int            TraceDepth, TraceMax;
	Widget         FocusWidget;
} XePerDisplayInput;


typedef struct _XeEventRec
{
	XeEventTable    next;
	XeEventMask     mask;      // Also select_data count for RecExt

	XeEventHandler	EventFunc;

	EPointer        ClientData;
	unsigned int    select:1;
	unsigned int    has_type_specifier:1;
	unsigned int    async:1;   // Not used, here for Digital extension? 
} XeEventRec;

typedef struct _XeGrabRec
{
	XeGrabList    next;
	Widget        widget;
	unsigned int  exclusive:1;
	unsigned int  SpringLoaded:1;
} XeGrabRec;


typedef struct
{
  Atom   Property;
  EBool  Available;
} XeSelectionPropRec;

typedef struct
{
  int                  PropCount;
  XeSelectionPropRec*  List;

#ifdef EX11
  Display*             X_Display;
  Atom                 incr_atom, indirect_atom, timestamp_atom;
#endif
} XePropListRec;


// Internal version of XeDisplay.
// The fields after XeDisplayCore are always there, but not accessible to application code.
//
typedef struct
{
	XeDisplayCore;

	XePerDisplayInput  InputState;               // State for modal grabs and keyboard focus
	ModToKeysymTable*  ModifiersToKeySyms;

	char*              Language;
	ETime              MultiClickDelay_us;
	XeExtSelectRec*    ExtSelect;int  ExtSelect_count;
	XeIntCallbackList  MappingCallbacks;         // For the translation manager
	unsigned long      KeySymsSerial;            // For tracking MappingNotify events
	int                KeySymsPerKeyCode;        // Number of KeySyms for each XeKeyCode
	uint8_t            IsModifier[32];           // 256-bit table that indicates for each key whether it's a modifier
	XeModifiers        ModeSwitch;	             // Keyboard group modifiers
	XeModifiers        NumLock;		               // Keyboard numlock modifiers
	XeTMKeyContext*    TMContext;                // For XeX_GetActionKeySym

#ifdef EX11
	XeKeySym           LockMeaning;              // Lock modifier meaning
	XeKeySym*          KeySyms;                  // XeKeyCode to XeKeySym table
	int                KeyCodeMin, KeyCodeMax;
	XeKeySym*          ModifierKeySyms;
	XeKeyProc          DefaultKeycodeTranslator;
	Region             X_Region;
	struct _XeXGC*     X_GCs;                    // For reusing X graphics contexts within an X Display
	Drawable**         X_Pixmaps;
	Time               X_LastTimestamp;
	XePropListRec*     X_SArray;                 // X context
	struct _XeW2WTable*  XeW2WTable;	           // X Window to Widget table
#endif
} XeDisplayI;


#define KEYBOARD TRUE
#define POINTER  FALSE

#define _XeIsServerGrab(g) ((g == XePassiveServerGrab) || (g == XeActiveServerGrab))

#define XeIsEitherPassiveGrab(g) ((g == XePassiveServerGrab) || (g == XePseudoPassiveServerGrab))

extern XeServerGrab* _Xe_CheckServerGrabsOnWidget(XEvent* PXEvent, Widget PW, const EBool PIsKeyboard);

#define _XeGetGrabList(pdi) (&(pdi)->GrabList)

extern Widget  _XeProcessKeyboardEvent(XKeyEvent* PXKeyEvent, Widget PW, XePerDisplayInput* PDI);

extern void    _Xe_ClearAncestorCache(Widget PW);


typedef enum
{
	XeCallbackNoList,
	XeCallbackHasNone,
	XeCallbackHasSome
} XeCallbackStatus;

typedef XeResource**  _XeCallbackTable;

extern XeIntCallbackList  _XtCompileCallbackList(XeCallbackList);



extern void _Xe_ShellGetPosition(Widget PW,  XeCoordinate* PX, XeCoordinate* PY);

#ifndef ERRORDB
#define ERRORDB "/usr/lib/X11/XtErrorDB"
#endif

#include <X11/Xalloca.h>  // For ALLOCATE_LOCAL


extern void     _XeAllocError(const char* PAllocType);
extern void     _Xe_ResourceListCompile(XeResource* PResources, const int PNResources);


typedef struct
{
	int                  RefCnt;                 // We need to reference-count these...
	XeIntervalId         ID;
	ETime                TriggerTime, Interval;  // Microseconds
	XeTimerCallbackProc  Callback;
	EPointer             ClientData;
	EBool                Active, Recurrent;
} XeTimeOut;


EListA(extern, XeTimeOut*, _XeTimeOut_s);  // Sorted by TriggerTime. Earliest first


#ifdef XTHREADS

/*
extern void (*_XeProcessLock)(void);
extern void (*_XeProcessUnlock)(void);
extern void (*_XeInitAppLock)(XeApp PApp);
*/


#define INIT_APP_LOCK(app)  if(_XeInitAppLock) (*_XeInitAppLock)(app)
#define FREE_APP_LOCK(app)  if(app && app->free_lock)(*app->free_lock)(app)

#define YIELD_APP_LOCK(app,push,pushed,level)  if(app && app->yield_lock) (*app->yield_lock)(app,push,pushed,level)
#define RESTORE_APP_LOCK(app,level,pushed)     if(app && app->restore_lock) (*app->restore_lock)(app,level,pushed)

#define XeXDPY_TO_APP(mD)        XeApp  LApp = (_XeProcessLock ? XeX_DisplayToApp(mD): NULL)
#define XeDPY_TO_APP(mD)         XeApp  LApp = (_XeProcessLock ? XeDisplay_ToApp(mD): NULL)

#else  // defined(XTHREADS)

#define INIT_APP_LOCK(app)
#define FREE_APP_LOCK(app)

#define XeXDPY_TO_APP(mD)
#define XeDPY_TO_APP(mD)

#endif // !defined(XTHREADS)


typedef struct
{
	char*  Start;
	char*  Current;
	int    BytesRemaining;
} XeHeap;


extern void  _XtCacheFlushTag(XeApp PApp, EPointer PTag);
extern void  _XtFreeActions(struct _ActionListRec* PActionTable);
extern void  _XtExtensionSelect(Widget PW);


extern void       _Xe_RemoveAllInputs(XeApp PApp);
extern int        _Xe_WaitForSomething(XeApp PApp, const EBool PIgnoreEvents, const EBool PIgnoreTimers, const EBool PIgnoreInputs, const EBool PIgnoreSignals, const EBool PBlock,
#ifdef XTHREADS
    const EBool PDropLock,
#endif
    unsigned long* PHowlong);

extern void       _XeWidgets_Manage(WidgetList PChildren, Cardinal PNChildren, Widget PParentW, const EBool PCallChangeManaged, const char* PCallerFunc);
extern void       Xe_SetMappedWhenManaged(Widget PW, const EBool PMappedWhenManaged);
extern EPointer  _XeInitializeActionData(struct _XeActionsRec* PActions, Cardinal PCount, const EBool PInPlace);

extern ETime       _Xe_HandleTimers(ETime PCurrentTime);


extern void        _XeCountVaList(va_list PVar, int* PTotalCount, int* PTypedCount);
extern void        _XeVaToTypedArgList(va_list PVar, const int PMaxCount, XeTypedArg** PArgsR, Cardinal* PNArgsR);

extern XeDisplay*  _XeX_AppAddDisplay(XeApp PApp, Display* PXDisplay);

extern void        _XeDoPhase2Destroy(XeApp PApp, int PDispatchLevel);
extern void        _Xe_DestroyServerGrabs(Widget PW);


extern void    _XeX_SelectionFree(XeDisplayI* PDisplay);


// EventUtil.c
extern void    _Xe_AncestorListFill(Widget** PListP, int* PMaxElemsP, int* PNumElemsP, Widget PStartW, Widget PEndW);

extern void    _XeEventsInit(void);
extern void    _Xe_EventTableFree(XeEventTable* PEventTable);
extern EBool   _Xe_InGrabList(Widget PW, XeGrabRec* PGrabList);
extern void    _Xe_SendFocusEvent(Widget PChild, const int PType);

extern void    _Xe_BuildKeysymTables(XeDisplayI* PDisplay);

#define _XeApp_IsSafeToDestroy(mApp)  ((mApp)->DispatchLevel == 0)

extern void  _Xe_AllocTMContext(XeDisplayI* PDisplay);

extern void     _XeGetLanguage(XeDisplayI* PDisplay);


extern XeCallbackStatus  XeHasCallbacks(Widget PW, const char* PCallbackName);


// Translation manager
extern void      _XeTranslateEvent(Widget PW, XeEvent* PEvent);
extern EBool     _XeComputeLateBindings(XeDisplay* PDisplay, LateBindingsPtr PLateModifiers, XeModifiers* PComputed, XeModifiers* PComputedMask);
extern void      _XeRemoveFromMappingCallbacks(Widget PW, EPointer PClientData /* target widget */, EPointer PCallData);



#ifdef EX11
extern void  _XeX_GCsFree(XeDisplayI* PDisplay);
#endif


#if defined (__cplusplus) && (!(defined (__CPLUSLIBS)))
} // Close scope of 'extern "C"' declaration
#endif


#endif // _XeI_h
