//================================================================================
// Xe UI toolkit
//
// Main private header file
//
// AUTHOR:  Gabor Nagy
// DATE:    2019-Jun-23 11:25:46
//
// Copyright (C) 1994-2021 by Gabor Nagy. All rights reserved.
//===============================================================================
#ifndef _XeP_h
#define _XeP_h

#include <Xe/XeStrDefs.h>

#include <Xe/Xe.h>

#if defined (__cplusplus) && (!(defined (__CPLUSLIBS)))
extern "C" {
#endif


#define XeObjectCore\
	XeObjCore;\
	Widget          Self;               /* Pointer to the Widget itself, for sanity checks */\
	XeWidgetClass   WClass;             /* Pointer to the Widget's ClassRec          */\
	Widget          Parent;             /* Parent Widget                             */\
	EBool           BeingDestroyed;     /* Marked for destroy                        */\
	XeCallbackList  DestroyCallbacks;   /* Who to call when the Widget is destroyed  */\
	EPointer        Constraints;        /* Constraint record                         */\
	XeApp           App

#define XeObjectClassCore\
	String             Name;                               /* Class name                          */\
	XeWidgetClass      Superclass;                         /* Superclass                          */\
	Cardinal           StructSize;                         /* Widget struct size in bytes         */\
	XeInitFunc         Init;                               /* Constructor                         */\
	XeWidgetFunc       Free;                               /* Destructor                          */\
	XeFunc             ClassInitialize;                    /* Class initialization proc           */\
	XeWidgetClassProc  ClassPartInitialize;                /* Dynamic Class initialization        */\
	XeClassFlags       ClassInitedFlags;                   /* 0, if class has not been initialized, XeShell_ClassFlag etc., when it has*/\
	XeResource*        WResources;Cardinal  WNResources;   /* Resources for subclass fields       */\
	XeSetValuesFunc    SetValues;                          /* Set subclass resource values        */\
	EArgsFuncB         SetValuesHook;                      /* Notify that SetValues called        */\
	EArgsFunc          GetValuesHook;                      /* Notify that get_values called       */\
	EPointer           CallbackPrivate                     /* List of callback offsets            */



typedef struct _XeEventRec*  XeEventTable;

typedef void    (*XeResourceDefaultFunc)(Widget PW, int POffset, XeSizeAddr* PValue);
typedef void    (*XeCreatePopupChildFunc)(Widget PW);

typedef struct
{
	String   Name;   // Resource name
	String   Type;   // Representation type name
	EArgVal  Value;  // Representation
	int      Size;   // Size of representation
} XeTypedArg;


struct _XtStateRec;  // Forward declare before use for C++

typedef struct
{
	XeTranslations       translations;      // Private to the Translation Manager
	XeActionFunc*        proc_table;        // Procedure bindings for actions
	struct _XtStateRec*  current_state;     // Translation Manager state
	ETime                LastEventTime_us;  // Microseconds
} XeTMData;


#ifdef DEBUG
#define Xe_CheckSubclass(mW, mWidgetClassPtr, mMessage)\
	if(!XeIsSubclass(((Widget)(mW)), (mWidgetClass)))\
	{\
		const char*  lParams[3];int  lNParams = 3;\
		lParams[0] = Xe_Class((Widget)(mW))->Name;\
		lParams[1] = (mWidgetClass)->Name;\
		lParams[2] = (mMessage);\
		XeAppError((mW)->App, "subclassMismatch", __FUNCTION__, "Widget class %s found when subclass of %s expected: %s",  lParams, lNParams);\
	}
#else
#define Xe_CheckSubclass(mW, mWidgetClass, mMessage)
#endif


extern void  _XeH_HandleFocus(Widget PW, XeEvent* PEvent, EPointer PClientData, EBool* PContRet);


typedef struct _XeObjectRec
{
	XeObjectCore;
} XeObjectRec;


typedef struct _XeObjectClass
{
	XeObjectClassCore;
} _XeObjectClass;


externalref _XeObjectClass  _XeOject_class;

externalref int  _XeInheritTranslations;

#define XeInheritTranslations     ((String)&_XeInheritTranslations)


// Keyboard focus
//
#define XeMY_ANCESTOR    0
#define XeMY_DESCENDANT  1
#define XeMY_COUSIN      2
#define XeMYSELF         3
#define XeUNRELATED      4
typedef char  XeGeneology;

typedef struct XeServerGrab_S  XeServerGrab;

struct XeServerGrab_S
{
	XeServerGrab*  Next;
	Widget         widget;
	uint32_t       OwnerEvents:1, PointerMode:1, KeyboardMode:1, HasExt:1, ConfineToIsWidgetWin:1;
	uint32_t       KeyBut;       // Key or button code
	uint32_t       Modifiers;    // Modifiers for key or button
	uint32_t       EventMask;
};



// OS-dependent file info for VirtKeys
//
#define XMBINDDIR "XMBINDDIR"
#ifndef XMBINDDIR_FALLBACK
#define XMBINDDIR_FALLBACK "/usr/lib/Xm/bindings"
#endif
#define XMBINDFILE "xmbind.alias"
#define MOTIFBIND ".motifbind"


typedef struct XeExtensionClassRec_S  XeExtensionClassRec;    // Forward-define XeExtensionClass

// Record for registering callbacks in the class record of an EguiItem
//
typedef struct
{
	char*         Name;
	unsigned int  Offset;  // Offset in the Object's instance record
} XeCallbackI;


#define XeBaseClassCore(mType)\
	mType##Class  SuperClass;\
	char*         Name;\
	unsigned int  StructSize

extern void  XeCallCallbacks(Widget PW, EguiCallbackList* PCBList, EPointer PCallData);

typedef EPointer  XeCacheRef;

extern void XeCallbackReleaseCacheRef(Widget PW,
    EPointer PClientData,  /* XeCacheRef */
    EPointer PCallData);

extern void XeCallbackReleaseCacheRefList(Widget PW, EPointer PClientData,  /* XeCacheRef* */
    EPointer PCallData);


struct XeExtensionClassRec_S
{
	XeBaseClassCore(XeExtension);
	EIndex  Instances_Count;
	void  (*Init)(Widget PW, XeExtension* PExtension);
	int   (*Free)(Widget PW, XeExtension* PExtension);
	EList(, XeResource, WResources);        // Resources of the XeExtension object
	EList(, XeCallbackI, CallbackOffsets);  // Offsets of Callback lists, in the XeExtension object
};



#define XeRectObjCore\
	XeObjectCore;\
	XeCoord2        Position;               /* Rectangle position                          */\
	XeCoord2        Size;                   /* Rectangle dimensions                        */\
	XeCoordinate    BorderWidth;            /* Rectangle border width                      */\
	XeCoord2        PreferredSize;\
	XeCoord2        GivenSize;              /* Size specified by the app, via XeNsizeX.XeNsizeY*/\
	EBool           Managed;                /* Is widget geometry managed?                 */\
	EBool           Sensitive;              /* Is widget sensitive to user events          */\
	EBool           AncestorSensitive;      /* are all ancestors sensitive?                */\
	EBool           IteratedQueryGeometry;  /* The QueryGeometry method of this Widget takes a suggested geometry into consideration, when computing its preferred geometry (e.g.: XeScrolledWindow). */\
	XeChangeFlags   ChangeFlags             /* Set by SetValues methods. XeChgREDRAW etc.  */


//
// RectObject
//
typedef struct XeRectObj_S
{
	XeRectObjCore;
} XeRectObjRec;


#ifdef XeOpenGL
#define XeRectObjectClassOpenGL\
  XeGL_RedrawFunc     GL_Redraw              /* Redraw the Widget with OpenGL         */
#else
#define XeRectObjectClassOpenGL
#endif



#define XeRectObjectClassCore\
	XeObjectClassCore;\
	XeWidgetFunc       Resize;                 /* Called when geometry management changed the Widget size  */\
	XeAlmostFunc       SetValuesAlmost;        /* Set values almost for geometry        */\
	XeGeometryHandler  QueryGeometry;          /* Return preferred geometry             */\
  XeRedrawFunc       Redraw                  /* Redraw method                         */


//
// RectObject class
//
typedef struct _RectObjClassRec
{
	XeRectObjectClassCore;
} RectObjClassRec;

externalref RectObjClassRec  _XeRectObj_class;



#include <Xe/CoreP.h>

typedef struct
{
	char*         Name;
	char*         Class;
	unsigned int  Size;
	void*         Offset;
	void*         Value;
} EguiResource;



//========================================
// Core	GUI item
//========================================
#ifndef _ECoreP_h
#define _ECoreP_h

#define E_IDC	0


// Class record
//
typedef struct
{
	EguiItemClass   SuperClass;
	char*           Name;
	unsigned int    InstanceSize;

	EguiResource*   Resources;EIndex  NumOfResources;

	XeCallbackI*    CallbackOffsets;EIndex  NumOfCallbacks;

	void            (*Destroy)(EguiItem);

	XeCallbackFunc  XtCBH;  // Callback handler for Xt
} EguiCoreClassPart;


typedef struct _EguiItemClassRec
{
	EguiCoreClassPart  Core;
} EguiItemClassRec, EguiCoreClassRec;

externalref EguiItemClassRec  eguiItemClassRec;
#define eguiCoreClassRec      eguiItemClassRec




// Core instance record
//
typedef struct
{
	char*              Name;
	EguiItem           Parent;
	char*              Location;  // This is for the application to know where to put an item (on a menu, a panel etc.)
	int                ClassID;   // For quick class identification
	EguiCoreClassRec*  Class;
	Widget             WidgetID;
	EPointer           UserData;
} EguiCorePart;

typedef struct _EguiCoreRec
{
	EguiCorePart  Core;
} EguiCoreRec;



//-----------------------------
// Class record
//-----------------------------
typedef struct
{
  void  (*AddChild)(EguiItem PItem, const char* PName,  EArg* PArgs, const int PArgCnt);
} EguiCompositeClassPart;


typedef struct _EguiCompositeClassRec
{
	EguiCoreClassPart       Core;
	EguiCompositeClassPart  Composite;
} EguiCompositeClassRec;

extern EguiCompositeClassRec  eguiCompositeClassRec;

//-----------------------------
// Instance
//-----------------------------
typedef struct
{
	EList(, EguiItem, Children);
} EguiCompositePart;

typedef struct _EguiCompositeRec
{
	EguiCoreClassPart  Core;
	EguiCompositePart  Composite;
} EguiCompositeRec;




#endif  // _ECoreP_h



//========================================
// List with optional buttons
//========================================
#ifndef _ListP_h
#define _ListP_h

typedef struct _EguiListClassRec
{
	EguiCoreClassPart       Core;
	EguiCompositeClassPart  Composite;
// ListClassRec is empty
} EguiListClassRec;

extern EguiListClassRec  eguiListClassRec;


typedef struct
{
	Widget            TopMatrixW, ListW, ButtonMatrixW;
	EguiCallbackList  SelectionCallback;
} EguiListPart;


typedef struct
{
 EguiCorePart		Core;
 EguiCompositePart	Composite;
 EguiListPart		List;
} EguiListRec;

#endif // _ListP_h



//----------------------------------------
// Fast class identification
//----------------------------------------
enum
{
	EguiSHELL = 1,
	EguiDIALOG,
	EguiFILESELECTIONBOX,
	EguiCOLOR_DIALOG,
	EguiLIST,
	EguiPUSHBUTTON,
	EguiSEPARATOR,
	EguiTOGGLEBUTTON,
	EguiCUSTOM
};

EList(extern, EguiItem, Xe_DeferredWidgets);

extern EguiItem   _EGUI_ItemNew(const char* PName, EguiItemClass PClass, EguiItem PParent, Widget PParentW);
extern void       _EGUI_AddItemToList(EguiItem PItem);

typedef struct
{
	const XeEvent*  Event;
} XeGlobalCallbackStruct;


typedef struct 
{
#ifndef USE_POLL
	fd_set  rmask;
	fd_set  wmask;
	fd_set  emask;
#endif
	int     nfds;
} FdStruct;


typedef struct _ConverterRec**  XeConverterTable;

typedef String  (*XeLanguageProc)(XeDisplay* PDisplay, String PXNL, EPointer PClientData);

typedef struct _XeLangFuncRec
{
	XeLanguageProc  Func;
	EPointer        ClientData;
} XeLangFunc;


typedef struct _ProcessContextRec
{
	XeApp             DefaultApp;
	EList(, XeApp, Apps);
	XeConverterTable  GlobalConverterTable;
	XeLangFunc        GlobalLangFunc;
} XeProcess;


typedef struct internalCallbackRec
{
	unsigned short  Count;
	char            is_padded;   // contains NULL padding for external form
	char            call_state;  // combination of _XtCB{FreeAfter}Calling
#ifdef LONG64
	unsigned int    align_pad;   // padding to align callback list
#endif
} InternalCallbackRec, *XeIntCallbackList;

typedef struct _XeSignalEvent
{
	struct _XeSignalEvent*  Next;
	XeApp                   App;
	EBool                   Notice;
} XeSignalEvent;

typedef unsigned long  XeInputMask;
#define XeInputNoneMask		0L
#define XeInputReadMask		(1L<<0)
#define XeInputWriteMask	(1L<<1)
#define XeInputExceptMask	(1L<<2)

typedef struct _InputEvent
{
	XeInputCallbackFunc  ie_proc;
	EPointer             ie_closure;
	struct _InputEvent*  ie_next;
	struct _InputEvent*  ie_oq;
	XeApp                App;
	int                  ie_source;
	XeInputMask          ie_condition;
} InputEvent;


typedef struct XeActionHookRec_S   XeActionHookRec;


struct _XeAppRec
{
	XeAppCore;

	EBool                     Alive, BeingDestroyed;

	XeProcess*                Process;

	EListA(, Widget, Shells);                               // All Shell Widgets created in this App (needed, for changing Fonts, Colors etc.)
	EListA(, XeDisplay*, Displays);
	EListA(, XeDisplay*, DisplayDestroyList);

	Widget                    WidgetUnderPhase2Free;
	int                       FreeCount;
	_XeFreeRec*               FreeList;int  FreeListSize;   // State data for 2-phase Widget freeing
	int                       DispatchLevel;

	unsigned long             SelectionTimeout;

	XeErrorMessageHandler     MessageFunc;
	XeErrorHandler            ErrorHandler;

	XeConverterTable          ConverterTable;
	struct _ActionListRec*    ActionTable;struct _ActionListRec*  ActionTableNext;
	XeActionHookRec*          ActionHookList;
	struct _TMBindCacheRec*   FreeBindings;
	XeLangFunc                LangFunc;

	XeCallbackFunc            GlobalCallback;
	uint32_t                  GlobalCallbackEventMask;
	InputEvent**              InputList;short  InputMax;
	InputEvent*               OutstandingQueue;
	char*                     DisplayNameTried;

	XeSignalEvent*            SignalQueue;

	ETime                     MultiClickDelay_us;  // In microseconds
#ifdef EX11
	FdStruct                  FDs;                 // X11 server connection file descriptors
	EBool                     RebuildFDList;
#endif

#ifdef XTHREADS
	LockPtr                   lock_info;
	ThreadAppProc             lock;
	ThreadAppProc             unlock;
	ThreadAppYieldLockProc    yield_lock;
	ThreadAppRestoreLockProc  restore_lock;
	ThreadAppProc             free_lock;
#endif
};


typedef void (*XeExtensionSelectProc)(Widget PW, int* PEventTypes, EPointer* PSelectData, int PCount, EPointer PClientData);

typedef struct _ExtensionSelectorRec
{
	XeExtensionSelectProc  Func;
	int                    Min, Max;
	EPointer              ClientData;
} XeExtSelectRec;


//================================================================================
// Reference-tracking correction for Images (Xe doesn't know about RefCnts etc.)
//================================================================================
#define _XeM_ImageChangeCheck(mStruct, mField)\
{\
	EpImage*  lOld = mStruct##Old->mField;\
	EpImage*  lNew = mStruct->mField;\
\
	if(lNew != lOld)\
	{\
/* A little trickery to make the reference-tracker work correctly	*/\
\
/* Restore pointer to old Object before Assign	*/\
		mStruct->mField = mStruct##Old->mField;\
		E_AssignToMember(lNew, mStruct,mField);\
\
		LChangeFlags |= XeChgREDRAW;\
		const EIndex	lOldSizeX = lOld ? lOld->Size.X : 0, lOldSizeY = lOld ? lOld->Size.Y : 0;\
		const EIndex	lNewSizeX = lNew ? lNew->Size.X : 0, lNewSizeY = lOld ? lNew->Size.Y : 0;\
		if(lOldSizeX != lNewSizeX || lNewSizeY != lOldSizeY) LChangeFlags |= XeChgSIZE;\
	}\
}


#define _Xe_InitColor(m0, m1, mColor) do\
{\
	Ec_SetRGBAi(&(m##RGBAiColor),  0, 0, 0, 0xFFFF);\
	if(m##RGBAiColorPtr)\
	{\
		m##RGBAiColor = *m##RGBAiColorPtr;\
	}\
	else\
	{\
		if(m##RGBAiColorP)\
		{\
			m##RGBAiColor = *m##RGBAiColorP;\
			m##RGBAiColorP = NULL;\
		}\
	}\
} while(0)


#define _DrawDefaultButtonShadow() do\
{\
	if(LDefBShadowThickness > 0)\
	{\
		if(!LClearBackground)  /* If ClearBackground is TRUE, we don't need to draw the spacing and the corners were already rounded by _XeCore_DrawBackgroundAndGetBlendMode(). */\
		{\
/* Draw spacing between the default button shadow and the button */\
/*			Ep_DrawRectangleRGBA8(LCacheImage, Ec_RGBAiToRGBA8(LParentBackgroundColor),  LDefBS1, LDefBS1,  LSizeX - LDefBS1 * 2, LSizeY - LDefBS1 * 2,  LDefBSS);*/\
/*			EpImage_RoundOffCorners(LCacheImage,  0, 0,  LSizeX, LSizeY,\
			 LDefaultButtonFrameCornerRadius, LDefaultButtonFrameCornerRadius,\
			 LDefaultButtonFrameCornerRadius, LDefaultButtonFrameCornerRadius,\
			 LParentBackgroundColor, LParentBackgroundColor, EpRegionALL, TRUE);\
*/\
		}\
\
		if(LButton->ShowAsDefault)\
		{\
			XeImage_DrawShadows(LCacheImage,  0, 0,  LSizeX, LSizeY,  0, NULL,\
			 LDefaultButtonFrameCornerRadius, NULL, LDefaultButtonFrameCornerRadius, NULL,  LDefaultButtonFrameCornerRadius, NULL, LDefaultButtonFrameCornerRadius, NULL,  TRUE,\
			 LDefBShadowThickness, XeShadowIN|LShadowSpecFlags, 0, LParentBackgroundColor, LParentBackgroundColor,  EpSRC, EpSRC);\
		}\
	}\
} while(0)


//typedef int  privateint;

/*
#ifdef XeDEVELOPMENT
#define XeDEBUG(p)	p
#else
#define XeDEBUG(p)
#endif
*/


#ifndef RectObjClassFlag
#define RectObjClassFlag     0x02
#define WidgetClassFlag      0x04
#define CompositeClassFlag   0x08
#define ConstraintClassFlag  0x10
#define XeShellClassFlag       0x20
#define WMShellClassFlag     0x40
#define TopLevelClassFlag    0x80
#endif // RectObjClassFlag


#define ASSIGN_MAX(x, y)	if ((y) > (x)) x = (y)


// XeScale and XeScrollbar pointer locations 
//
enum { XeOUTSIDE = 0, XeON_SCALE, XeON_SLIDER };

// NumObject and Scrollbar redraw flags
//
#define XeDO_VALUES           1   // Recompute pixel positions of slider/thermometer
#define XeDO_SCALE        (1<<1)  // Redraw scale background
#define XeDO_SLIDER       (1<<2)  // Redraw the slider
#define XeDO_THERMOMETER  (1<<3)  // Redraw the thermometer


// From Traversal.c
//
extern void	_Xe_NavigChangeManaged(Widget PW);


// Synthetic resource stuff
//

typedef enum
{ 
	XeSYNTHETIC_NONE, 
	XeSYNTHETIC_LOAD 
} XeImportOperator;

typedef void              (*XeExportProc)(Widget PW, int, EArgVal*);
typedef XeImportOperator  (*XeImportProc)(Widget PW, int, EArgVal*);


typedef struct _XeSyntheticResource
{
	String        Name;
	Cardinal      resource_size;
	Cardinal      resource_offset;
	XeExportProc  export_proc;
	XeImportProc  import_proc;
} XeSyntheticResource;


//------------------------------------------------------------------------
// Keyboard traversal stuff
//------------------------------------------------------------------------
typedef enum
{
	XeFOCUS_IN,
	XeFOCUS_OUT,
	XeENTER,
	XeLEAVE
} XeFocusChange;

typedef struct _XeFocusMovedCallbackStruct
{
	int       reason;
	XEvent*   event;
	EBool     cont;
	Widget    old_focus,new_focus;
	uint8_t   focus_policy;
} XeFocusMovedCallbackStruct, *XeFocusMovedCallback;


#define Xe_Parent(mW)           ((mW)->Parent)
#define Xe_Class(mW)            ((mW)->WClass)
#define Xe_Constraints(mW)      ((mW)->Constraints)
#define XeBeingDestroyed(mW)    ((mW)->BeingDestroyed)
#define XeDestroyCallbacks(mW)  ((mW)->DestroyCallbacks)

#define Xe_X(w)      (((Widget)(w))->Position.X)
#define Xe_Y(w)      (((Widget)(w))->Position.Y)
#define Xe_SizeX(w)  (((Widget)(w))->Size.X)
#define Xe_SizeY(w)  (((Widget)(w))->Size.Y)

#define Xe_GetDisplay(mW)  ((mW)->DisplayInfo)
#define Xe_GetVisual(mW)   ((mW)->Visual)

#define XeFOCUS_RESET   1
#define XeFOCUS_IGNORE  2


//------------------------------------------------------------------------


extern char  _XeStrNormalColor[];
extern char  _XeStrRed[];
extern char  _XeStrGreen[];
extern char  _XeStrYellow[];
extern char  _XeStrPurple[];
extern char  _XeStrCyan[];

extern char*  _XeGetNormalColorCode(char* PPtr);
extern char*  _XeGetErrorColorCode(char* PPtr);
extern char*  _XeGetWarningColorCode(char* PPtr);
extern char*  _XeGetDirColorCode(char* PPtr);

extern XeTranslations  _XeScaleTextFieldTransTable;

extern void  _Xe_ImageGetScaledSize(const EIndex POrigSizeX, const EIndex POrigSizeY,  int32_t* PSizeXPtr, int32_t* PSizeYPtr,  XeCoord2* PSizeP);

extern void  XeImageDrawArrowRGBA8(EpImage* PCacheImage, const int PX, const int PY, const int PSizeX, const int PSizeY,  const int PDirection,  EcRGBA8Color PColor);

extern void  _XeCastShadow(Widget PW,  const int PRXPos, const int PRYPos,  const int PMSizeX, const int PMSizeY,  Window* PShadowBXWindowPtr, Window* PShadowRXWindowPtr, Pixmap* PShadowBXMaskPtr, Pixmap* PShadowXRMaskPtr, const int PShadowOffset);

extern void  _XeTraverse(Widget PW, XeEvent* PEvent, const char PActionCode, const EBool PSendFicysLostEvent, const EBool PSendFocusGetEvent);
extern void  _XeATraverse(Widget PW, XeEvent* PEvent,  String* PP, Cardinal* PNP);

extern void  _XeInstallTranslations(Widget PW);

extern EBool  XeToolkitThreadInitialize(void);
extern void   XeAppLock(XeApp PApp);
extern void   XeAppUnlock(XeApp PApp);

extern void  (*_XeProcessLock)(void);
extern void  (*_XeProcessUnlock)(void);
extern void  (*_XeInitAppLock)(XeApp PApp);


#ifdef XTHREADS

#define XeLOCK_PROCESS     if(_XeProcessLock) (*_XeProcessLock)()
#define XeUNLOCK_PROCESS   if(_XeProcessUnlock) (*_XeProcessUnlock)()
#define XeWIDGET_TO_APP(w) XeApp  LApp = (w && _XeProcessLock ? (w)->App : NULL)
#define XeLOCK_APP(mApp)   if(mApp && mApp->lock)(*mApp->lock)(mApp)
#define XeUNLOCK_APP(mApp) if(mApp && mApp->unlock)(*mApp->unlock)(mApp)

#else  // ifdef XTHREADS

#define XeLOCK_PROCESS
#define XeUNLOCK_PROCESS
#define XeWIDGET_TO_APP(w)
#define XeLOCK_APP(mApp)
#define XeUNLOCK_APP(mApp)

#endif

extern void  XeProcessLock(void);
extern void  XeProcessUnlock(void);



EListA(extern, EpScaledImage*, _Xe_ScaledImages);

extern const float  _Xe_GradientContrastDefault;

extern XeProcess  _Xe_Process;

extern int   _Xe_AppsToFree_Count;
extern void  _XeApps_Free(void);

extern void  _XePopupInitialize(XeApp PApp);
extern void  _XeDoFreeBindings(XeApp PApp);


//--------------------------------
// Menu.c
//--------------------------------
extern EBool  _XeIsInMenuPane(Widget PW);


//--------------------------------
// Geometry.c
//--------------------------------
extern XeGeometryResult  XeMakeGeometryRequest(Widget PW, XeWidgetGeometry* PDesired);
extern Widget            Xe_WindowedAncestorAndRelativePosition(Widget PW, const uint8_t PWindowTypeFlags,  int* PXP, int* PYP);

extern EBool             Xe_IsUpdateEnabled(Widget PW);

extern void              _Xe_CallResize(Widget PW,  const uint8_t PResizeFuncActions);
extern void              _Xe_SetGeometry(Widget PW,  const int PX, const int PY,  const int PSizeX, const int PSizeY,  const int PBorderWidth,  const uint8_t PResizeFuncActions);
extern void              Xe_SetGeometry(Widget PW,  const int PX, const int PY,  const int PSizeX, const int PSizeY,  const int PBorderWidth);
extern void              Xe_SetSize(Widget PW, const int PSizeX, const int PSizeY,  const int PBorderWidth);
extern void              Xe_SetPosition(Widget PW,  const int PX, const int PY);
extern XeGeometryResult  _Xe_RequestGeometry(Widget PW, XeWidgetGeometry* PRequest, XeWidgetGeometry* PReply, EBool* PClearRectObj);
extern char*             XeDBGeometryReplyToString(XeGeometryResult PA);


//--------------------------------
// Events.c
//--------------------------------
extern int   XeUnbufferButtonEvents(Widget PW, const int PButton);
extern void  _XeTimersFree(void);


extern void  XeTranslationsFree(XeTranslations PT);


typedef struct _XeEventRecExt
{
	int        type;
	EPointer  select_data[1];  // Actual dimension is [mask]
} XeEventRecExt;


extern char*  XeHomeDir(void);

extern EBool  _XeCheckSubclassFlag(Widget PW, const XeClassFlags PFlag);
extern EBool  _XeIsSubclassOf(Widget PW, XeWidgetClass PWidgetClass, XeWidgetClass PFlagClass, const XeClassFlags PFlag);
extern EBool  XeIsSubclass(Widget PW, XeWidgetClass PWidgetClass);
extern Widget _XeWindowedAncestor(Widget PW);


#define Xe_SupportsWindow(m)              (XeIsWidget(m) && (m)->WindowTypeFlags != 0)
#define Xe_SupportsWindowType(m, mFlags)  (XeIsWidget(m) && ((m)->WindowTypeFlags & (mFlags)) == mFlags)
#define Xe_IsManaged(m)                   (XeIsRectObj(m) ? (m)->Managed : FALSE)
#define Xe_IsSensitive(m)                 (XeIsRectObj(m) ? ((m)->Sensitive && (m)->AncestorSensitive) : FALSE)


extern void           Xe_WidgetClassInit(XeWidgetClass PWidgetClass);

extern void           _Xe_ImageInit(XeImage* PImageRec);
extern void           _Xe_ImageFree(XeImage* PImageRec);
extern XeChangeFlags  _Xe_ImageChangeCheck(XeImage* PImageRecOld, XeImage* PImageRec);


extern void           _Xe_SetValues(Widget PW,  const EArg* PArgs, const int PNArgs,  const EBool PGeoRequests, const EBool PRedraw);

// We were included from XeI.h - expose internal stuff
#ifdef _XeI_h

// These 4 macros a very handy internally, we can't expose them to applications, as they violate the rule that arguments are to be evaluated exactly once.

#define XeIsRectObj(object)      (_XeCheckSubclassFlag(object, XeRectObj_ClassFlag))
#define XeIsWidget(object)       (_XeCheckSubclassFlag(object, XeWidget_ClassFlag))
#define XeIsComposite(mW)        (_XeCheckSubclassFlag(mW, XeCompositeWidget_ClassFlag))
#define XeIsConstraint(mW)       (_XeCheckSubclassFlag(mW, XeConstraintWidget_ClassFlag))
#define XeIsShell(mW)            (_XeCheckSubclassFlag(mW, XeShell_ClassFlag))
#define XeIsOverrideShell(mW)    (_XeIsSubclassOf(mW, (XeWidgetClass)XeOverrideShellWidget_class,    (XeWidgetClass)XeShellWidget_class, XeShell_ClassFlag))
#define XeIsTopLevelShell(mW)    (_XeCheckSubclassFlag(mW, TopLevelClassFlag))
#define XeIsWMShell(mW)	         (_XeCheckSubclassFlag(mW, WMShellClassFlag))
#define XeIsVendorShell(mW)      (_XeIsSubclassOf(mW, (XeWidgetClass)XeVendorShellWidget_class,      (XeWidgetClass)XeWMShellWidget_class,       XeWMShell_ClassFlag))
#define XeIsTransientShell(mW)   (_XeIsSubclassOf(mW, (XeWidgetClass)XeTransientShellWidget_class,   (XeWidgetClass)XeWMShellWidget_class,       XeWMShell_ClassFlag))
#define XeIsApplicationShell(mW) (_XeIsSubclassOf(mW, (XeWidgetClass)XeApplicationShellWidget_class, (XeWidgetClass)XeTopLevelShellWidget_class, XeTopLevel_ClassFlag))


#define StringToQuark(string)  XrmStringToQuark(string)
#define StringToName(string)   XrmStringToName(string)

extern void             _Xe_ResourcesFlatten(XeWidgetClass PWC);
extern void             _XeConstraintResourcesFlatten(XeConstraintWidgetClass PWC);
extern XeCacheRef*      _XeGetResources(Widget PW,  EArg* PArgs, const int PNArgs);
extern void             _Xe_ResourceAddressFromParent(Widget PW, int POffset, XeSizeAddr* PValue);
extern void             _Xe_CopyToArg(void* PSrc, EArgVal* PDst, const unsigned int PVSize);
#else

// Non Xe-internal code should access internals through functions
//
extern EBool  XeIsRectObj(Widget PW);
extern EBool  XeIsWidget(Widget PW);
extern EBool  XeIsComposite(Widget PW);
extern EBool  XeIsConstraint(Widget PW);
extern EBool  XeIsShell(Widget PW);
extern EBool  XeIsOverrideShell(Widget PW);
extern EBool  XeIsTopLevelShell(Widget PW);
extern EBool  XeIsWMShell(Widget PW);
extern EBool  XeIsVendorShell(Widget PW);
extern EBool  XeIsTransientShell(Widget PW);
extern EBool  XeIsApplicationShell(Widget PW);

#endif

typedef void (*XeDestructor)(XeApp PApp, XeSizeAddr* PTo, EPointer PConverterData, XeSizeAddr* PArgs, Cardinal* PNArgs);

typedef int            XeCacheType;
#define XeCacheNone       0x001
#define XeCacheAll        0x002
#define XeCacheByDisplay  0x003
#define XeCacheRefCount   0x100


typedef enum
{
// Address mode       Meaning
//----------------------------------------
	XeAddress,          // Address
	XeBaseOffset,       // Offset
	XeImmediate,        // Constant
	XeResourceString,   // Resource name string
	XeResourceQuark,    // Resource name quark
	XeWidgetBaseOffset, // Offset from ancestor
	XeProcedureArg      // Procedure to invoke
} XeAddressMode;

typedef struct
{
	XeAddressMode  address_mode;
	EPointer       address_id;
	Cardinal       size;
} XeConvertArgRec, *XeConvertArgList;


typedef EBool (*XeTypeConverter)(XeDisplay* PDisplay, XeSizeAddr* PArgs, Cardinal* PNArgs, XeSizeAddr* PFrom, XeSizeAddr* PTo, EPointer* PConverterData);

extern void   XeSetTypeConverter(const char* PFromType, const char* PToType, XeTypeConverter PConverter, XeConvertArgList PConvertArgs, Cardinal PNumArgs, XeCacheType PCacheType, XeDestructor PDestructor);
extern EBool  XeConvertAndStore(Widget PW, const char* PFromType, XeSizeAddr* PFrom, const char* PToType, XeSizeAddr* PToInOut);
extern EBool  XeCallConverter(XeDisplay* PDisplay, XeTypeConverter PConverter, XeSizeAddr* PArgs, Cardinal PNArgs, XeSizeAddr* PFrom, XeSizeAddr* PToInOut, XeCacheRef* PCacheRefRet);
extern void   XeDisplayStringConversionWarning(XeDisplay* PDisplay, const char* PFromValue, const char* PToType);

const EcRGBAiColor* Xe_GetBackgroundColor(const EcRGBAiColor* PColor, EcRGBAiColor* PDstColor, const uint16_t PBlendFunction);

extern void XeAppSetTypeConverter(XeApp PApp, const char* PFromType, const char* PToType, XeTypeConverter PConverter,
    XeConvertArgList PConvertArgs, Cardinal PNArgs,  XeCacheType PCacheType, XeDestructor PDestructor);

extern void  _XeStringMetrics_ToSizeAndOffset(const XeStringMetrics* PMetrics, XeCoord2* PTotalSizeP, XeCoordinate* POffsetYP);

#if defined (__cplusplus) && (!(defined (__CPLUSLIBS)))
} // Close scope of 'extern "C"' declaration
#endif


#endif // _XeP_h
