//================================================================
// Xe UI toolkit
//
// X11-specific header file
//
// AUTHOR:  Gabor Nagy
// DATE:    2019-Jun-23 04:34:56
//
// Copyright (C) 1994-2021 by Gabor Nagy. All rights reserved.
//================================================================
#ifndef _XeX_h
#define _XeX_h

#include <Xe/Xe.h>

#include <X11/Xatom.h>

#if defined (__cplusplus) && (!(defined (__CPLUSLIBS)))
extern "C" {
#endif


#define XeM_AddXPoint(mPoints, mCounter,  mX, mY)\
do {\
	mPoints[mCounter].x = mX, mPoints[mCounter].y = mY, mCounter++;\
} while(0)


typedef unsigned long  XeXGCMask;      // Mask of values that are used by a Widget
typedef unsigned long  XeXWValueMask;

extern Pixel         XeX_Black, XeX_White;

extern XeDisplay*    _XeX_OpenDisplay(XeApp PApp, const char* PDisplayName, const char* PAppName, const char* PAppClassName,  int* PArgC, const char* const* PArgV);
extern XeApp         XeX_DisplayToApp(Display* PXDisplay);

extern char*         XeX_KeyToAccel(const unsigned short PKey, const unsigned int PModifiers, char* PStr);
extern char*         XeX_KeyToAccelText(const unsigned short PKey, const unsigned int PModifiers, char* PStr);
extern void          XeX_ConvertCase(XeDisplay* PDisplay, XeKeySym PKeysym, XeKeySym* PLowerRet, XeKeySym* PUpperRet);
extern unsigned int  XeX_ModifiersFromEvent(const XeEvent* PEvent);

#ifdef _XeP_h

#define XeX_Window(mW)    ((mW)->X_Window)
#define XeX_Display(mW)   ((mW)->DisplayInfo->X_Display)
#define XeX_Colormap(mW)  ((mW)->X_Colormap)
#define XeX_Screen(mW)    ((mW)->X_Screen)
#define XeX_GetGC(mW)     ((mW)->X_GC)

extern XeDisplay*  XeX_XDisplayToDisplay(Display* PXDisplay);
extern void        XeX_Realize(Widget PW, Mask* PValueMask, XSetWindowAttributes* PWinAttributes);
extern void        XeX_RealizeWithDraw(Widget PW, Mask* PValueMask, XSetWindowAttributes* PWinAttributes, const EpImage* (*PRedrawFunc)(Widget PW, EpImage* PCacheImage));
extern void        XeX_CreateWindow(Widget PW, XeXWValueMask PValueMask, XSetWindowAttributes* PAttributes);

extern void        _XeX_CoreRedraw(Widget PW, XeRedrawEvent* PEvent, void (*PRedrawProc)(Widget, Drawable));
extern void        _XeX_WidgetDrawFromImage(Widget PW, const EpImage* PImage);

#else

extern Display*    XeX_Display(Widget PW);
extern Screen*     XeX_Screen(Widget PW);
extern Window      XeX_Window(Widget PW);
extern Colormap    XeX_Colormap(Widget PW);
extern Pixmap      XeX_GetCachePixmap(Widget PW);
extern GC          XeX_GetGC(Widget PW);

#endif


extern EBool     _XeX_HandleTruecolorIcons(XeDisplay* PDisplay, XEvent* PXEvent);


extern void      XeX_Redraw(Widget PW);

extern Window    XeX_ClosestDrawableAncestorWindow(Widget PW);
extern Window    XeX_DrawableWindowAndPosition(Widget PW,  int* PXP, int* PYP);

extern Pixel     _XeX_GetXColor(const XeVisual* PVisual, const XeColor* PXeColor);

extern Pixel     XeX_GetNamedColor(const XeVisual* PVisual, const char* PColorName);
extern Pixel     XeX_GetRGBColor(const XeVisual* PVisual,  const unsigned int PR, const unsigned int PG, const unsigned int PB);


// X11 Drawable <-> EpImage
//
extern Pixmap    XeX_ImageToPixmapVa(const XeVisual* PVisual, const EpImage* PImage, ...);
extern int       XeXImage_PasteToDrawable(const EpImage* PImage, const XeVisual* PVisual, GC PGC, Drawable PDrawable,
                  const int PSrcX, const int PSrcY,  const int PDstX, const int PDstY,  const int PSizeX, const int PSizeY);
extern int       XeX_ImageCompositeToDrawable(const EpImage* PImage,  const XeLayer* PBackgroundLayers, const int PNBackgroundLayers,  const XeVisual* PVisual,
                  const int PSrcX, const int PSrcY,  const int PDstX, const int PDstY,  const int PSizeX, const int PSizeY, const uint16_t PBlendFunction,  GC PXGC, Drawable PXDrawable);

extern int       XeX_DrawableToImage(const XeVisual* PVisual, const Drawable PDrawable, Pixel PBackgroundXColor, EpImage* PImage);

extern int       XeX_DrawablePasteToImage(const XeVisual* PVisual, Drawable PXDrawable, const Pixel PBackgroundXPixel, EpImage* PImage,
                  const int PSrcX, const int PSrcY,  const int PDstX, const int PDstY,  const int PSizeX, const int PSizeY);

// Drawing
//
extern void      XeX_FillGradientRectangle(const XeVisual* PVisual, Drawable PPixmap, GC PGC, const int POrientation, const int PX, const int PY, const int PSizeX, const int PSizeY, const EcRGBAf32Color* PColor0, const EcRGBAf32Color* PColor1);

extern void      XeX_DrawShadows(const XeVisual* PVisual, Drawable PXDrawable, GC PXGC,  const int PX, const int PY,  const int PSizeX, const int PSizeY,  const int PGapWidth, const int PGapCornerRadius, const EcRGBAf32Color* PGapColor,
 const int PShadowThickness, const uint16_t PType, const uint8_t PGradient, const EcRGBAiColor* PBackgroundColor, const EcRGBAiColor* PParentBackgroundColor);

extern XeFontSize*  _XeX_FontSizeCreate(XeDisplay* PDisplay, const char* PXName,  const float PSizeInPixels, const float PAntiAliasing);
extern EIndex       _XeX_SysFontsCollect(XeDisplay* PDisplay, const EBool PDetails, const EBool PForceReload, const int PMinSizeInPixels, const uint8_t PSpacing, const char* PCharset, const char* PEncoding);

extern int          XeFontX_Load(XeFont* PFont, XeFontSize* PFontSize);

extern void         XeX_WTextExtents(const XeFontSize* PFontSize, const wchar_t* PString, const int PLength, int* PDir, int* PAsc, int* PDesc, XeCharStruct* POverall);


extern int          XeXDrawable_SaveToFile(XeVisual* PVisual, Drawable PXDrawable, const char* PFileName);


// Conversion operations with XeNconversion (for XeX_CImageToPixmapVa)
//
#define XeX_ALPHA_TO_1BIT_PIXMAP  0x0001

extern Pixmap      XeX_XPMToPixmapVa(const XeVisual* PVisual, char** PXPM, ...);
extern Pixmap      XeX_CImageToPixmapVa(const XeVisual* PVisual, EpCImage* PImage, ...);


#define _XeX_DrawFocusShadows(mEW, mDrawable, mOffset) do\
{\
	XeX_DrawShadows(LXDisplay, mDrawable, LGC, (mOffset), (mOffset), Xe_SizeX(mEW) - (mOffset) * 2, Xe_SizeY(mEW) - (mOffset) * 2, 0, 0, LBottomShadowXColor, mEW->Primitive.HighlightThickness, XeShadowIN|LShadowSpecFlags, 0,  LBackgroundXColor, LTopShadowXColor, LBottomShadowXColor);\
} while(0)


extern XeEventMask  _XeX_EventTypeToMask(const int PEventType);

extern Widget     XeX_WindowToWidget(XeDisplay* PDisplay, Window PXWindow);
extern Widget     XeX_WindowToWidgetXD(Display* PXDisplay, Window PXWindow);

extern void       XeX_DrawableRegister(Display* PXDisplay, Drawable PXDrawable, Widget PW);
extern void       XeX_DrawableUnregister(Display* PXDisplay, Drawable PXDrawable);


//
// Selection
//
typedef EPointer  XeRequestId;

typedef EBool  (*XeXConvertSelectionFunc)(Widget PW, Atom* PSelection, Atom* PTarget, Atom* PTypeReturn, EPointer* PValueRet, unsigned long* PLengthRet, int* PFormatRet);
typedef void   (*XeXLoseSelectionFunc)(Widget PW, Atom* PSelection);
typedef void   (*XeXSelectionDoneFunc)(Widget PW, Atom* PSelection, Atom* PTarget);
typedef void   (*XeXSelectionCallbackProc)(Widget PW, EPointer PClientData, Atom* PSelection, Atom* PType, EPointer PValue, unsigned long* PLength, int* PFormat);
typedef void   (*XeXLoseIncrementalSelectionProc)(Widget PW, Atom* PSelection, EPointer PClientData);
typedef void   (*XeXIncrementalSelectionDoneProc)(Widget PW, Atom* PSelection, Atom* PTarget, XeRequestId* PReceiverID, EPointer PClientData);
typedef EBool  (*XeXConvertIncrementalSelectionProc)(Widget PW, Atom* PSelection, Atom* PTarget, Atom* PType, EPointer* PValue, unsigned long* PLength, int* PFormat, unsigned long* PMaxLength, EPointer PClientData, XeRequestId* PReceiverID);

typedef void   (*XeXCancelConvertSelectionFunc)(Widget PW, Atom* PSelection, Atom* PTarget, XeRequestId* PReceiverID, EPointer PClientData);


extern EBool   XeX_OwnSelection(Widget PW, const Atom PSelection, const Time PTime, XeXConvertSelectionFunc PConvert, XeXLoseSelectionFunc PLose, XeXSelectionDoneFunc PDone);
extern void    XeX_DisownSelection(Widget PW, Atom PSelection, const Time PTime);
extern void    XeX_GetSelectionValue(Widget PW, Atom PSelection, Atom PTarget, XeXSelectionCallbackProc PCallback, EPointer PClientData, const Time PTime);
extern void    XeX_GetSelectionValues(Widget PW, Atom PSelection, Atom* PTargets, const int PCount, XeXSelectionCallbackProc PCallback, EPointer* PClientData, const Time PTime);

extern XeEventMask  _XeX_BuildEventMask(Widget PW);
extern void         _XeX_ComputeWindowAttributes(Widget PW, XeXWValueMask* PValueMask, XSetWindowAttributes* PValues);

extern XEvent*    XeX_NextEvent(Display* PXDisplay, XEvent* PXEvent);
extern EBool      XeX_DispatchEvent(XEvent* PXEvent);
extern int        XeX_UnbufferButtonEvents(Display* PDisplay, Window PWindow, const int PButton);

extern void      XeStringX_BinaryLinesDraw(const XeVisual* PVisual,  Drawable PXDrawable, const int PDrawableSizeX, const int PDrawableSizeY, GC PBackgroundGC, GC PGC,  const EcRGBAiColor* PBackgroundColor, const EcRGBAiColor* PForegroundColor, EBool PUseColorsInString,  const int PX, const int PY,  int PFullSizeX,  XeStringLine* PLines, const int PLines_Count, XeStringBinarySegment* PSegments, const int PSegments_Count, const XeStringRenderSettings* PRenderSettings, const int PAlignment);

extern GC        XeX_GCNew(Widget PW, const int PDepth, XeXGCMask PXValueMask, XGCValues* PXValues, XeXGCMask PXDynamicMask, XeXGCMask PUnusedMask);
extern void      XeX_GCFree(Widget PW, GC PXGC);

typedef void     (*XeXRealizeFunc)(Widget PWidget, XeXWValueMask* PMask, XSetWindowAttributes* PAttributes);

extern XeKeySym   XeX_GetActionKeySym(XEvent* PXEvent, XeModifiers* PModifiersR);


extern void          XeX_GrabButton(Widget PW, const int PButton, const XeModifiers PModifiers, const EBool  POwnerEvents, const unsigned int PEventMask, const int PPointerMode, const int PKeyboardMode, Window PConfineToXWindow, Cursor PXCursor);
extern int           XeX_GrabPointer(Widget PW, const EBool POwnerEvents, const unsigned int PEventMask, const int PPointerMode, const int PKeyboardMode, Window PConfineToXWindow, Cursor PXCursor, const ETime PTime);


extern XrmDatabase   XeX_GetDatabase(Display* PXDisplay);



#if defined (__cplusplus) && (!(defined (__CPLUSLIBS)))
}  // Close scope of 'extern "C"' declaration
#endif

#endif // _XeX_h
