#ifndef _Xt_TranslateI_h
#define _Xt_TranslateI_h

#define CACHE_TRANSLATIONS

#define TM_NO_MATCH (-2)

#define _XtRStateTablePair "_XtStateTablePair"

typedef unsigned char   TMByteCard;
typedef unsigned short  TMShortCard;
typedef unsigned long   TMLongCard;
typedef short           TMShortInt;


typedef struct _CompiledAction
{
	XrmQuark      signature;
	XeActionFunc  proc;
} CompiledAction, *CompiledActionTable;


typedef struct _ActionListRec*  ActionList;
typedef struct _ActionListRec
{
	ActionList           next;
	CompiledActionTable  table;
	TMShortCard          count;
} ActionListRec;


typedef struct XeTMTypeMatchRec_S       XeTMTypeMatchRec;
typedef struct XeTMTypeMatchRec_S*      XeTMTypeMatch;
typedef struct XeTMModifierMatchRec_S   XeTMModifierMatchRec;
typedef struct XeTMModifierMatchRec_S*  XeTMModifierMatch;
typedef struct XeTMEventRec_S           XeTMEvent;
typedef struct XeTMEventRec_S*          XeTMEventPtr;

typedef EBool (*MatchProc)(XeTMTypeMatch typeMatch, XeTMModifierMatch modMatch, XeTMEventPtr eventSeq);

typedef struct _ModToKeysymTable
{
	XeModifiers  mask;
	int        count;
	int        idx;
} ModToKeysymTable;

typedef struct _LateBindings
{
	unsigned int    knot:1;
	unsigned int    pair:1;
	unsigned short  ref_count;  // Garbage collection
	XeKeySym        keysym;
} LateBindings, *LateBindingsPtr;

typedef short ModifierMask;

typedef struct _ActionsRec*  ActionPtr;

typedef struct _ActionsRec
{
	int        idx;         // Index into quarkTable to find proc
	String*    params;      // Pointer to array of params
	Cardinal   num_params;  // Number of params
	ActionPtr  next;        // Next action to perform
} ActionRec;

typedef struct _XtStateRec*  StatePtr;
typedef struct _XtStateRec
{
	unsigned int  isCycleStart:1;
	unsigned int  isCycleEnd:1;
	TMShortCard   typeIndex;
	TMShortCard   modIndex;
	ActionPtr     actions;    // rhs list of actions to perform
	StatePtr      nextLevel;
} StateRec;


#define XtTableReplace   0
#define XtTableAugment   1
#define XtTableOverride  2
#define XtTableUnmerge   3

typedef unsigned int  _XtTranslateOp;


// New Definitions
struct XeTMModifierMatchRec_S
{
	TMLongCard       modifiers;
	TMLongCard       modifierMask;
	LateBindingsPtr  lateModifiers;
	EBool            standard;
};

struct XeTMTypeMatchRec_S
{
	TMLongCard  EventType;
	TMLongCard  eventCode;
	TMLongCard  eventCodeMask;
	MatchProc   matchEvent;
};

typedef struct _TMBranchHeadRec
{
	unsigned int  isSimple:1;
	unsigned int  hasActions:1;
	unsigned int  hasCycles:1;
	unsigned int  more:13;
	TMShortCard   typeIndex;
	TMShortCard   modIndex;
} TMBranchHeadRec, *TMBranchHead;

// NOTE: Elements of this structure must match those of TMComplexStateTreeRec and TMParseStateTreeRec.
//
typedef struct _TMSimpleStateTreeRec
{
	unsigned int      isSimple:1;
	unsigned int      isAccelerator:1;
	unsigned int      mappingNotifyInterest:1;
	unsigned int      refCount:13;
	TMShortCard       numBranchHeads;
	TMShortCard       numQuarks;   // # of entries in quarkTbl

	TMShortCard       unused;      // to ensure same alignment
	TMBranchHeadRec*  branchHeadTbl;
	XrmQuark*         quarkTbl;    // table of quarkified rhs
} TMSimpleStateTreeRec, *TMSimpleStateTree;

// NOTE: elements of this structure must match those of TMSimpleStateTreeRec and TMParseStateTreeRec.
//
typedef struct _TMComplexStateTreeRec
{
	unsigned int      isSimple:1;
	unsigned int      isAccelerator:1;
	unsigned int      mappingNotifyInterest:1;
	unsigned int      refCount:13;
	TMShortCard       numBranchHeads;
	TMShortCard       numQuarks;      // # of entries in quarkTbl

	TMShortCard       numComplexBranchHeads;
	TMBranchHeadRec*  branchHeadTbl;
	XrmQuark*         quarkTbl;       // table of quarkified rhs
	StatePtr*         complexBranchHeadTbl;
} TMComplexStateTreeRec, *TMComplexStateTree;


// NOTE: elements of this structure must match those of TMSimpleStateTreeRec and TMComplexStateTreeRec.
//
typedef struct _TMParseStateTreeRec
{
	unsigned int       isSimple:1;
	unsigned int       isAccelerator:1;
	unsigned int       mappingNotifyInterest:1;
	unsigned int       isStackQuarks:1;
	unsigned int       isStackBranchHeads:1;
	unsigned int       isStackComplexBranchHeads:1;
	unsigned int       unused:10;        // to ensure correct alignment
	TMShortCard        numBranchHeads;
	TMShortCard        numQuarks;        // # of entries in quarkTbl

	TMShortCard        numComplexBranchHeads;
	TMBranchHeadRec*   branchHeadTbl;
	XrmQuark*          quarkTbl;         // table of quarkified rhs
	StatePtr*          complexBranchHeadTbl;
	TMShortCard        branchHeadTblSize;
	TMShortCard        quarkTblSize;     // total size of quarkTbl
	TMShortCard        complexBranchHeadTblSize;
	StatePtr           head;
} TMParseStateTreeRec, *TMParseStateTree;


typedef union _TMStateTreeRec
{
	TMSimpleStateTreeRec   simple;
	TMParseStateTreeRec    parse;
	TMComplexStateTreeRec  complex;
} *TMStateTree, **TMStateTreePtr, **TMStateTreeList;

typedef struct _TMSimpleBindProcsRec
{
	XeActionFunc*  Funcs;
}TMSimpleBindProcsRec, *TMSimpleBindProcs;

typedef struct _TMComplexBindProcsRec
{
	Widget          widget;    // widgetID to pass to action Proc
	XeTranslations  aXlations;
	XeActionFunc*   Funcs;
} TMComplexBindProcsRec, *TMComplexBindProcs;

typedef struct _TMSimpleBindDataRec
{
	unsigned int          isComplex:1;    // must be first
	TMSimpleBindProcsRec  bindTbl[1];     // variable length
} TMSimpleBindDataRec, *TMSimpleBindData;

typedef struct _TMComplexBindDataRec
{
	unsigned int               isComplex:1;    // must be first
	struct _ATranslationData*  accel_context;  // for GetValues
	TMComplexBindProcsRec      bindTbl[1];     // variable length
} TMComplexBindDataRec, *TMComplexBindData;

typedef union _TMBindDataRec
{
	TMSimpleBindDataRec   simple;
	TMComplexBindDataRec  complex;
} *TMBindData;

typedef struct _TranslationData
{
	unsigned char             hasBindings;      // Must be first
	unsigned char             operation;        // Replace, augment, override
	TMShortCard               numStateTrees;
	struct _TranslationData*  composers[2];
	XeEventMask               eventMask;
	TMStateTree               stateTreeTbl[1];  // Variable length
} TranslationData;


// ATranslations is returned by GetValues for translations that contain accelerators. The TM can differentiate between this
// and TranslationData (that don't have a bindTbl) by looking at the first field (hasBindings) of either structure.
// All ATranslationData structures associated with a widget are chained off the BindData record of the widget.
//
typedef struct _ATranslationData
{
	unsigned char              hasBindings;   // Must be first
	unsigned char              operation;
	struct _TranslationData*   Translations;  // Atual translations
	struct _ATranslationData*  next;          // Chain the contexts together
	TMComplexBindProcsRec      bindTbl[1];    // Accelerator bindings
} ATranslationData, *ATranslations;


typedef struct _TMGlobalRec
{
	XeTMTypeMatchRec**        typeMatchSegmentTbl;
	TMShortCard               numTypeMatches;
	TMShortCard               numTypeMatchSegments;
	TMShortCard               typeMatchSegmentTblSize;
	XeTMModifierMatchRec**    modMatchSegmentTbl;
	TMShortCard               numModMatches;
	TMShortCard               numModMatchSegments;
	TMShortCard               modMatchSegmentTblSize;
	EBool                     newMatchSemantics;
#ifdef TRACE_TM
	XeTranslations*           tmTbl;
	TMShortCard               numTms;
	TMShortCard               tmTblSize;
	struct _TMBindCacheRec**  bindCacheTbl;
	TMShortCard               numBindCache;
	TMShortCard               bindCacheTblSize;
	TMShortCard               numLateBindings;
	TMShortCard               numBranchHeads;
	TMShortCard               numComplexStates;
	TMShortCard               numComplexActions;
#endif
} TMGlobalRec;

extern TMGlobalRec  _XtGlobalTM;



typedef struct _TMConvertRec
{
	XeTranslations	old;  // Table to merge into
	XeTranslations	new;  // Table to merge from
} TMConvertRec;

#define _XeEventTimerEventType ((TMLongCard)~0L)
#define KeysymModMask		(1L<<27) // private to TM
#define AnyButtonMask		(1L<<28) // private to TM

typedef struct _EventRec
{
	TMLongCard       modifiers;
	TMLongCard       modifierMask;
	LateBindingsPtr  lateModifiers;
	TMLongCard       EventType;
	TMLongCard       eventCode;
	TMLongCard       eventCodeMask;
	MatchProc        matchEvent;
	EBool            standard;
} XeTMTEvent;

typedef struct _EventSeqRec*  EventSeqPtr;
typedef struct _EventSeqRec
{
	XeTMTEvent   TMTEvent; // Event description
	StatePtr     state;    // Private to state table builder
	EventSeqPtr  next;     // Next event on line
	ActionPtr    actions;  // r.h.s.
} EventSeqRec;

typedef EventSeqRec  EventRec;
typedef EventSeqPtr  EventPtr;

struct XeTMEventRec_S
{
	XeEvent*    Event;
	XeTMTEvent  TMTEvent;
};

typedef struct XeActionHookRec_S   XeActionHookRec;
typedef struct XeActionHookRec_S*  XeActionHook;

typedef void  (*XeActionHookFunc)(Widget PW, EPointer PClientData, String PActionName, XeEvent* PEvent,  String* PParams, Cardinal* PNParams);

struct XeActionHookRec_S
{
	XeActionHookRec*  Next;  // Must remain first
	XeApp             App;
	XeActionHookFunc  Func;
	EPointer          ClientData;
};

// Choose a number between 2 and 8
#define TMKEYCACHELOG2  6
#define TMKEYCACHESIZE  (1<<TMKEYCACHELOG2)

typedef struct
{
	unsigned char  modifiers_return[256];  // Constant per XeKeyCode, key proc
	XeKeyCode      keycode[TMKEYCACHESIZE];
	unsigned char  modifiers[TMKEYCACHESIZE];
	XeKeySym       keysym[TMKEYCACHESIZE];
} XeTMKeyCache;

typedef struct
{
	XeEvent*       Event;
	unsigned long  serial;
	XeKeySym       keysym;
	XeModifiers    modifiers;
	XeTMKeyCache   keycache;  // Keep this last, to keep offsets to others small
} XeTMKeyContext;



#define TM_MOD_SEGMENT_SIZE 	16
#define TM_TYPE_SEGMENT_SIZE 	16

#define TMGetTypeMatch(idx)      ((XeTMTypeMatch) &((_XtGlobalTM.typeMatchSegmentTbl[((idx) >> 4)])[(idx) & 15]))
#define TMGetModifierMatch(idx)  ((XeTMModifierMatch) &((_XtGlobalTM.modMatchSegmentTbl[(idx) >> 4])[(idx) & 15]))

// Useful Access Macros
#define TMNewMatchSemantics()  (_XtGlobalTM.newMatchSemantics)
#define TMBranchMore(branch)   (branch->more)
#define TMComplexBranchHead(tree, br)  (((TMComplexStateTree)tree)->complexBranchHeadTbl[TMBranchMore(br)])

#define TMGetComplexBindEntry(bindData, idx)  ((TMComplexBindProcs)&(((TMComplexBindData)bindData)->bindTbl[idx]))

#define TMGetSimpleBindEntry(bindData, idx)   ((TMSimpleBindProcs)&(((TMSimpleBindData)bindData)->bindTbl[idx]))


#define _Xe_KeySymTablesInit(mDisplay)  { if((mDisplay)->KeySyms == NULL) _Xe_BuildKeysymTables(mDisplay); }


//
// Internal Functions
//

extern String    _XtPrintXlations(Widget PW, XeTranslations PTranslations, Widget PAccelWidget, const EBool PIncludeRHS);
extern void      _XtRegisterGrabs(Widget PW);
extern void      _XtAddEventSeqToStateTree(EventSeqPtr PEventSeq, TMParseStateTree PStateTree);
extern EBool     _XtMatchUsingStandardMods(XeTMTypeMatch PTypeMatch, XeTMModifierMatch PModMatch, XeTMEventPtr PEventSeq);
extern EBool     _XtMatchUsingDontCareMods(XeTMTypeMatch PTypeMatch, XeTMModifierMatch PModMatch, XeTMEventPtr PEventSeq);
extern EBool     _XtRegularMatch(XeTMTypeMatch PTypeMatch, XeTMModifierMatch PModMatch, XeTMEventPtr PEventSeq);
extern EBool     _XtMatchAtom(XeTMTypeMatch pTypeMatch, XeTMModifierMatch PModMatch, XeTMEventPtr PEventSeq);


#define _XtCBCalling           1
#define _XtCBFreeAfterCalling  2

typedef EBool         (*_XtConditionProc)(EPointer);

extern void             _XeAddCallback(XeIntCallbackList* PCBL, XeCallbackFunc, EPointer);
extern void             _XeAddCallbackOnce(XeIntCallbackList* PCBL, XeCallbackFunc, EPointer);
extern XeCallbackList   _XtGetCallbackList(XeIntCallbackList* PCBL);
extern void             _XtRemoveAllCallbacks(XeIntCallbackList* PCBL);
extern void             _XtRemoveCallback(XeIntCallbackList* PCBL, XeCallbackFunc, EPointer);
extern void             _XtPeekCallback(Widget, XeCallbackList, XeCallbackFunc*, EPointer*);
extern void             _XtCallConditionalCallbackList(Widget, XeCallbackList, EPointer, _XtConditionProc);

#ifndef NO_MIT_HACKS
extern void           _XtDisplayTranslations(Widget PW, XEvent* PXEvent, String* PParams, Cardinal* PNarams);

extern void           _XtDisplayAccelerators(Widget PWidget, XEvent* PXEvent, String* PParams, Cardinal* PNParams);
extern void           _XtDisplayInstalledAccelerators(Widget PW, XEvent* PXEvent, String* PParams, Cardinal* PNParams);
#endif  // ifndef NO_MIT_HACKS

extern XeTranslations _XtCreateXlations(TMStateTree* PStateTrees, TMShortCard PNStateTrees, XeTranslations PFirst, XeTranslations PSecond);

extern EBool  _XtCvtMergeTranslations(XeDisplay* PDisplay, XeSizeAddr* PArgs, Cardinal* PNArgs, XeSizeAddr* PFrom, XeSizeAddr* PTo, EPointer* PClosureRet);

void _XtRemoveStateTreeByIndex(XeTranslations PTranslations, TMShortCard PI);

void _XtFreeTranslations(XeApp PApp, XeSizeAddr* PToVal, EPointer PClosure,  XeSizeAddr* PArgs, Cardinal* PNArgs);

extern TMShortCard     _XtGetModifierIndex(XeTMTEvent* PTMTEvent);
extern TMShortCard     _XtGetQuarkIndex(TMParseStateTree PStateTreePtr, XrmQuark PQuark);
extern XeTranslations  _XtGetTranslationValue(Widget PW);
extern TMShortCard     _XtGetTypeIndex(XeTMTEvent* PTMTEvent);

extern void    _XtRemoveTranslations(Widget PW);
extern void    _XtDestroyTMData(Widget PW);
extern void    _XtMergeTranslations(Widget PW, XeTranslations PNewTranslations, _XtTranslateOp POperation);

extern TMStateTree  _XtParseTreeToStateTree(TMParseStateTree PParseTree);

extern String  _XtPrintActions(ActionRec* PActions, XrmQuark* PQuarkTbl);
extern String  _XtPrintState(TMStateTree PStateTree, TMBranchHead PBranchHead);
extern String  _XtPrintEventSeq(EventSeqPtr PEventSeq, Display* PXDisplay);

typedef EBool  (*_XtTraversalProc)(StatePtr PState, EPointer PData);

extern void  _XtTraverseStateTree(TMStateTree PTree, _XtTraversalProc PFunc, EPointer PData);
extern void  _XtTranslateInitialize(void);
extern void  _XtAddTMConverters(XeConverterTable PTable);
extern void  _XtUnbindActions(Widget PW, XeTranslations PTranslations, TMBindData PBindData);
extern void  _XtUnmergeTranslations(Widget PW, XeTranslations PTranslations);

extern void  _XtFreeActionRecs(ActionPtr PActions);

#endif // _Xt_TranslateI_h
